/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_FlxMtx_H
#define fesslix_FlxMtx_H

#include "flxmath.h"

#include <algorithm>
#include <map>
#include <vector>

class FlxMtx;
class FlxMtxSym;

// -------------------------- VECTOR ----------------------------------------------------------------------

/**
* @brief Scalarproduct of two vector - sizes have to match (or at least size of v2 has to be greather than size of v1)
*/
inline const tdouble dot(const tVec& v1, const tVec& v2) {
  #if FLX_KAHAN_DOT
    register pdouble tm=0.0;
  #else
    register tdouble tm=0.0;
  #endif
  const size_t v1s = v1.size();
  const tdouble* v1p = &v1[0];
  const tdouble* v2p = &v2[0];
  for (register size_t i=0;i<v1s;++i) {
    tm+=v1p[i]*v2p[i];
  }
  #if FLX_KAHAN_DOT
    return tm.cast2double();
  #else
    return tm;
  #endif
}
inline const pdouble dot(const pVec& v1, const pVec& v2) {
  register pdouble tm=0.0;
  const size_t v1s = v1.size();
  const pdouble* v1p = &v1[0];
  const pdouble* v2p = &v2[0];
  for (register size_t i=0;i<v1s;i++) {
    tm+=v1p[i]*v2p[i];
  }
  return tm;
}

/**
* @brief Maximum-Norm of a vector (vector has to contain at least one element - this is not checked)
*/
inline const tdouble NormMax(const tVec& v) {
  tdouble tm=fabs(v[0]);
  const size_t vs = v.size();
  const tdouble* vp = &v[0];
  for (register size_t i=1;i<vs;++i) {
    if (tm < fabs(vp[i])) {
      tm=fabs(vp[i]);
    }
  }
  return tm;
}

/**
* @brief 2-Norm of a vector (vector has to contain at least one element - this is not checked)
*/
inline const tdouble Norm2(const tVec& v) {
  #if FLX_KAHAN_2NORM
    pdouble tm=0.0;
  #else
    tdouble tm=0.0;
  #endif
  const size_t vs = v.size();
  const tdouble* vp = &v[0];
  for (register size_t i=0;i<vs;++i) {
    tm+=vp[i]*vp[i];
  }
  #if FLX_KAHAN_2NORM
    return std::sqrt(tm.cast2double());
  #else
    return std::sqrt(tm);
  #endif
}
inline const tdouble Norm2(const tdouble* v, const tuint N) {
  #if FLX_KAHAN_2NORM
    pdouble tm=0.0;
  #else
    tdouble tm=0.0;
  #endif
  for (register size_t i=0;i<N;++i) {
    tm+=v[i]*v[i];
  }
  #if FLX_KAHAN_2NORM
    return std::sqrt(tm.cast2double());
  #else
    return std::sqrt(tm);
  #endif
}

/**
* @brief square of the 2-Norm of a vector (vector has to contain at least one element - this is not checked)
*/
inline const tdouble Norm2_NOroot(const tVec& v) {
  #if FLX_KAHAN_2NORM
    pdouble tm=0.0;
  #else
    tdouble tm=0.0;
  #endif
  const size_t vs = v.size();
  const tdouble* vp = &v[0];
  for (register size_t i=0;i<vs;++i) {
    tm+=vp[i]*vp[i];
  }
  #if FLX_KAHAN_2NORM
    return tm.cast2double();
  #else
    return tm;
  #endif
}

/**
* @brief check if there are values less than TOL in a vector
*/
inline void checkTOL(tVec& v) {
  const size_t vs = v.size();
  tdouble* vp = &v[0];
  const tdouble GT = GlobalVar.TOL();
  for (register size_t i=0;i<vs;++i) {
    if (std::fabs(vp[i]) <= GT) vp[i] = 0.0;
  }
}

inline void pVec_assemble(pVec& w, const tVec& v) {
  const size_t vs = v.size();
  const tdouble* vp = &v[0];
  pdouble* wp = &w[0];
  for (register size_t i=0;i<vs;++i) {
    wp[i] = vp[i];
  }
}
inline void pVec_add(pVec& w, const tVec& v) {
  const size_t vs = v.size();
  const tdouble* vp = &v[0];
  pdouble* wp = &w[0];
  for (register size_t i=0;i<vs;++i) {
    wp[i] += vp[i];
  }
}
inline void pVec_mult(pVec& w, const tdouble& v) {
  const size_t ws = w.size();
  pdouble* wp = &w[0];
  for (register size_t i=0;i<ws;++i) {
    wp[i] *= v;
  }
}
inline const pVec operator*(const pVec& left, const tdouble& right) {
  pVec result(left);
  const size_t rs = left.size();
  pdouble* resultp = &result[0];
  for (register size_t i=0;i<rs;++i) {
    resultp[i]*=right;
  }
  return result;
}
inline void tVec_assemble(tVec& w, const pVec& v) {
  const size_t ws = w.size();
  const pdouble* vp = &v[0];
  tdouble* wp = &w[0];
  for (register size_t i=0;i<ws;++i) {
    wp[i] = vp[i].cast2double();
  }
}

void tVec_simple_plot(std::ostream& os, const tVec& V);
std::ostream& operator<<(std::ostream& os, const tVec& V);
std::ostream& operator<<(std::ostream& os, const bVec& V);
std::ostream& operator<<(std::ostream&os, const std::vector< tuint >& V);

void tVec_cast2tuintNo0(const tVec& V, iVec& R);

// -------------------------- MATRIX ----------------------------------------------------------------------

/* Matrix Types defined in this library
 * ====================================
 * 
 * FlxMtx_base (*)
 * 
 *   FlxMtx			full nxm-Matrix
 *   FlxMtxTransformation	a transformating matrix consisting of (nxn)-matrices of type FlxMtx
 *   FlxMtxLTri			full nxn lower triangular matrix
 *   FlxMtxSparsLTri		sparse nxn lower triangular matrix
 * 
 *   FlxMtx_baseS (*)
 *     FlxMtxSparsSymILU	incomplete LU factorization of a sparse symmetric matrix
 *     FlxMtxSparsSymLU		Cholesky decomposition of a (sparse) symmetric matrix
 *     FlxMtxSym		full symmetric nxn matrix
 *     FlxMtxDiag		diagonal matrix
 *     FlxMtxIdentity		identity matrix
 *     FlxMtxSparsSym		sparse symmetric nxn matrix
 *     FlxMtxSparsSFEMSym	sparse symmetric matrix needed for SFEM problems (coefficients are of type FlxMtxSparsSym)
 *     FlxMtxPrecnInvSFEMSym		-> preconding
 *     FlxMtxPrecnILUSFEMSym		-> preconding
 *     
 */


/**
* @brief The base class of matrices in Fesslix
*/
class FlxMtx_base {
  public:
    virtual ~FlxMtx_base() {};
    virtual FlxMtx_base* copy();
    virtual const tVec operator*(const tVec& v) const;
    virtual const pVec operator*(const pVec& v) const;
    virtual FlxMtx operator*(const FlxMtx_base& rhs) const;
    virtual FlxMtx_base& operator*=(const tdouble& s) = 0;
    virtual const tnlong nrows() const = 0;
    virtual const tnlong ncols() const = 0;
    /**
    * @brief Matrix-vector product: w=M*v;
    */
    virtual void MultMv(const tVec& v, tVec& w) const = 0;
    /**
    * @brief Matrix-vector product: w=M*v;
    */
    virtual void MultMv(const pVec& v, pVec& w) const;
    /**
    * @brief add value v to coefficient (i,j)  -  has to be within range of matrix (not checked!!!)
    */
    virtual void add_value(const tnlong& i, const tnlong& j, const tdouble& v)=0;
    /**
    * @brief get coefficient (i,j)  -  has to be within range of matrix (not checked!!!)
    */
    virtual const tdouble operator()(const tnlong& i, const tnlong& j) const = 0;
    /**
    * @brief sets all coefficients in the matrix to zero
    */
    void set_zeroMtx() {operator*=(0.0);};
    
    /**
    * @brief output-function of the matrix
    */
    virtual void output_Mtx(std::ostream& sout) const;
    /**
    * @brief output matrix in Octave format
    */
    virtual void output_OctaveMtx(std::ostream& sout, bool checkTOLv=false, bool doendl=true) const;
    /**
    * @brief returns the largest coefficient
    */
    virtual const tdouble max() const;
    /**
    * @brief returns the largest coefficient
    */
    virtual const tdouble min() const;
};

std::ostream& operator<<(std::ostream& os, const FlxMtx_base& M);

class FlxMtxDiag;

/**
* @brief the base class for symmetric matrices in Fesslix
*/
class FlxMtx_baseS : public FlxMtx_base {
  protected:
    FlxMtx_baseS* Minv;
    virtual void assembleMinv(int i);
    virtual void preconding(const tVec& r, tVec& z, const int i=0) const;
    virtual void preconding(const pVec& r, pVec& z, const int i=0) const;
  public:
    FlxMtx_baseS() : Minv(NULL) {}
    virtual ~FlxMtx_baseS();
    /**
    * @brief Conjugate Gradient algorithm
    */
    // TODO change to MultMV
    virtual const bool solve_CG(tVec& x, tVec& f, tdouble &eps, tnlong& iter, tuint pcn, bool startZero);
    /**
    * @brief returns the diagnol matrix of this matrix
    */
    virtual FlxMtxDiag* get_diag() const;
    /**
    * @brief returns the inverse of this matrix
    */
    virtual FlxMtx_baseS* get_Inverse();
    /**
    * @brief checks if the matrix is positive definite
    */
    virtual const bool isPositiveDefinite(tuint& r, const bool fixIt=false) = 0;
    
    virtual void output_Mtx(std::ostream& sout) const;
};

class FlxMtxTransformation;
class FlxMtxSpars;

/**
* @brief stores a full matrix
*/
class FlxMtx : public FlxMtx_base {
  private:
    tnlong rsize;
    tnlong csize;
    tVec mtx;
  public:
    const tnlong nrows() const {return rsize;};
    const tnlong ncols() const {return csize;};

    FlxMtx(const tnlong& rsize,const tnlong& csize) :rsize(rsize),csize(csize),mtx(rsize*csize) {};
    FlxMtx* copy();
    /**
    * @brief Matrix-vector product: w=M*v;
    */
    virtual void MultMv(const tVec& v, tVec& w) const;
    FlxMtx& operator+=(const FlxMtx& m);
    FlxMtx& operator*=(const tdouble& s) {mtx*=s; return *this; };
//     FlxMtx operator*(const FlxMtx_base& m) const;
    tdouble& operator()(const tnlong& i, const tnlong& j);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    /**
    * @brief res = Transpose(M).M
    */
    void TransposeMmultM(FlxMtxSym& res) const;
    /**
    * @brief w = Transpose(M).v
    */
    void TransposeMmultVec(const tVec& v, tVec& w) const;
    /**
    * @brief returns the matrix-vector (constant!)
    */
    const tVec& get_mtx_tVec() const { return mtx; }
    
    friend class FlxMtxSpars;
    friend void MtxProd_BTKB(const FlxMtx&, const FlxMtxSym&, FlxMtxSym&);
    friend void MtxProd_BTKB(const FlxMtxTransformation&, const FlxMtxSym&, FlxMtxSym&);
};

/**
* @brief stores a transformation matrix
*/
class FlxMtxTransformation : public FlxMtx_base {
  private:
    tnlong rsize;
    std::vector<FlxMtx*> Ttm;
  public:
    const tnlong nrows() const {return rsize;};
    const tnlong ncols() const {return rsize;};

    FlxMtxTransformation(const std::vector<FlxMtx*> &Ttm);
    ~FlxMtxTransformation();
//     FlxMtx* copy();
    FlxMtxTransformation& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    /**
    * @brief Matrix-vector product: w=M*v;
    */
    virtual void MultMv(const tVec& v, tVec& w) const;
    
    friend void MtxProd_BTKB(const FlxMtxTransformation&, const FlxMtxSym&, FlxMtxSym&);
};


class FlxMtxSparsSym;

/**
* @brief stores a lower triangular matrix
*/
class FlxMtxLTri : public FlxMtx_base {
  protected:
    tnlong msize;
    tVec mtx;
  public:
    const tnlong nrows() const {return msize;};
    const tnlong ncols() const {return msize;};

    FlxMtxLTri(const tnlong& msize) :msize(msize),mtx((msize*msize+msize)/2) {};
    FlxMtxLTri* copy();
    /**
    * @brief Matrix-vector product: w=M*v;
    * NOTE: for this special case, w and v can refer to the same vector
    */
    virtual void MultMv(const tVec& v, tVec& w) const;
//     FlxMtx& operator+=(const FlxMtxLTri& m);
    FlxMtxLTri& operator*=(const tdouble& s) {mtx*=s; return *this; };
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    /**
    * @brief Assembles this matrix by doing a CholeskyDecomposition of Sm
    */
    FlxMtxLTri& CholeskyDec(FlxMtxSym& Sm);
    FlxMtxLTri& CholeskyDec(FlxMtxSparsSym& Sm);
    /**
    * @brief Invert the Matrix
    */
    FlxMtxLTri& Invert();
    /**
    * @brief solves the following problem: v=Transpose(*this)*v;
    */
    void TransMultVec(tVec& v);
    
    friend void MtxProd_BTKB(const FlxMtxLTri& B, const FlxMtxSym& E, FlxMtxSym& K);
    friend class FlxMtxSym;
    friend class FlxMtxSparsLTri;
};


// Matrix - Example:
//    0: 1: 2: 3:
// 0: 1
// 1:    2
// 2: 3     4
// 3:    8     6
// 
//      0  1  2  3  4  5  6  7  8  9  10
// ija  5  6  7  9  11 0  1  0  2  1  3
// sa   X  X  X  X  X  1  2  3  4  8  6

// length of the arrays: ija[ija[0]-1]
// dimension: ija[0]-1
/**
* @brief stores a sparse matrix (nxm)
*/
class FlxMtxSpars : public FlxMtx_base {
  protected:
    tdouble* sa;
    tnlong* ija;
    FlxMtxSpars() : sa(NULL),ija(NULL) {}
    /**
    * @brief allocates just the memory ...
    */
    FlxMtxSpars(const tnlong nmax);
  public:
    const tnlong nrows() const {return ija[0]-1;};
    const tnlong ncols() const {return ija[0]-1;};

    /**
    * @brief creates a sparse matrix out of a full matrix
    */
    FlxMtxSpars(const FlxMtx& mtx);
    /**
    * @brief creates a copy of the matrix
    */
    FlxMtxSpars(const FlxMtxSpars& mtx);
    virtual ~FlxMtxSpars();
    FlxMtxSpars* copy();
    /**
    * @brief Matrix-vector product: w=M*v;
    * NOTE: does not check dimensions of v and w
    */
    virtual void MultMv(const tVec& v, tVec& w) const;
    /**
    * @brief multiplies a certain row with a vector
    * @param rownumber number of the corresponding row (please make sure index is valid)
    */
    virtual const tdouble MultRowV(const tVec& v, const tnlong rownumber) const;
    /**
    * @brief solves the following problem: w=Transpose(*this)*v;
    */
    virtual void TransMultVec(const tVec& v, tVec& w);
    FlxMtxSpars& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);

};


// Matrix - Example:
//    0: 1: 2: 3:
// 0: 1
// 1:    2
// 2: 3     4
// 3:    7     6
// 
//      0  1  2  3  4  5  6 
// ija  5  5  5  6  7  0  1
// sa   1  2  4  6     3  7
// length of the arrays: ija[ija[0]-1]
// dimension: ija[0]-1
/**
* @brief stores a sparse lower triangular matrix
*/
class FlxMtxSparsLTri : public FlxMtxSpars {
  public:
    /**
    * @brief allocates just the memory ...
    */
    FlxMtxSparsLTri(const tnlong nmax) : FlxMtxSpars(nmax) {}
    /**
    * @brief creates a sparse lower triangular matrix out of the lower triangular matrix
    */
    FlxMtxSparsLTri(const FlxMtxLTri& mtx);
    /**
    * @brief creates a copy of the matrix
    */
    FlxMtxSparsLTri(const FlxMtxSparsLTri& mtx);
    /**
    * @brief Assembles this matrix by doing a CholeskyDecomposition of Sm
    */
    FlxMtxSparsLTri(FlxMtxSparsSym& Sm);
    /**
    * @brief ... creates a diagonal matrix
    */
    FlxMtxSparsLTri(FlxMtxDiag& mtx);
    FlxMtxSparsLTri* copy();
    /**
    * @brief Matrix-vector product: w=M*v;
    * NOTE: for this special case, w and v can refer to the same vector
    */
    virtual void MultMv(const tVec& v, tVec& w) const;
    virtual const tdouble MultRowV(const tVec& v, const tnlong rownumber) const;
    /**
    * @brief Calculates v=L^-1*v -> inverse matrix is not explicitly computed !!!
    */
    void MultInv(const tVec& v,tVec& w);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    /**
    * @brief solves the following problem: v=Transpose(*this)*v;
    */
    void TransMultVec(tVec& v);
    virtual void TransMultVec(const tVec& v, tVec& w);
    /**
    * @brief solves the following problem: return=D*(*this);
    */
    FlxMtxSparsLTri* MultM_Mdiag(const FlxMtxDiag& D) const;
    
    /**
    * @brief Doing a sparse Cholesky decomposition into saP and ijaP
    */
    static void CholeskyDec(tdoublePtr& saP, tnlong*& ijaP, FlxMtxSparsSym& mtx);
};

/**
* @brief sparse symmetric matrix - incomplete LU factorization of a sparse symmetric matrix
*/
class FlxMtxSparsSymILU : public FlxMtx_baseS {
  protected:
    tdouble* sa;
    tnlong* ija;
    FlxMtxSparsSymILU() {};
  public:
    const tnlong nrows() const {return ija[0]-1;};
    const tnlong ncols() const {return ija[0]-1;};
    
    FlxMtxSparsSymILU(const FlxMtxSparsSym& mtx, bool mod0diagentry=false);
    virtual ~FlxMtxSparsSymILU();
    /**
    * @brief NOTE: this is not a simple Matrix-Vector multiplication ... this solves a linear system of equations (THIS)*w=v
    */
    virtual void MultMv(const tVec& v, tVec& w) const;
    virtual void MultMv(const pVec& v, pVec& w) const;
    FlxMtx_base& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false) { return true;};
};

/**
* @brief sparse symmetric matrix - Cholesky decomposition of a (sparse) symmetric matrix
*/
class FlxMtxSparsSymLU : public FlxMtxSparsSymILU {
  public:
    FlxMtxSparsSymLU(FlxMtxSparsSym& mtx);
    const tdouble Percentage_of_NonZero() const { return (ija[ija[0]-1]+ija[0])/(pow2(ija[0]-1.)-ija[0]+1)*2;}
};

/**
* @brief stores a full symmetric matrix
*/
class FlxMtxSym : public FlxMtx_baseS {
  private:
    tnlong msize;
    tVec mtx;
    void preconding(const tVec& r, tVec& z, const int precn = 0) const;
  public:
    const tnlong nrows() const {return msize;};
    const tnlong ncols() const {return msize;};
    FlxMtxSym(const tnlong& msizeV) :msize(msizeV),mtx((msizeV*msizeV+msizeV)/2) {};
    FlxMtxSym(const FlxMtxLTri& LTL);
    FlxMtxSym(const FlxMtx_baseS& S);
    FlxMtxSym* copy();
    virtual void MultMv(const tVec& v, tVec& w) const;
    /**
    * @brief Multiplies a slice vector v (a lot of zeros in the beginning and the end) with this matrix
    * @param li lower index to start
    * @param ui upper index to stop
    * @param v the slice vector (has to be defined between li and ui)
    * @param w the result vector (has to be of the same size as the dimension of the matrix)
    */
    void MultMv_slice(const tVec& v, tVec& w, const size_t& li, const size_t& ui) const;
    void MultMv(const pVec& v, pVec& w) const;
    FlxMtxSym& operator+=(const FlxMtxSym& m);
    FlxMtxSym& operator*=(const tdouble& s) {mtx*=s; return *this; };
    tdouble& operator()(const tnlong& i, const tnlong& j);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false);
    tdouble* get_VecPointer() { return &mtx[0];}
    /**
    * @brief assembles vec (dim=N*N) with the elements of this matrix (returns a full matrix)
    * @param vec NxN-vector (size is NOT checked)
    */
    void get_VecPointer_full(tdouble* vec) const;
    /**
    * @brief returns the matrix-vector (constant!)
    */
    const tVec& get_mtx_tVec() const { return mtx; }
    
    static const bool isPositiveDefinite_ext(tVec& mtxe, tuint rows, tuint& r, const bool fixIt = false);
    friend void MtxProd_BTKB(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K);
    friend void MtxProd_BTKB(const FlxMtxLTri& B, const FlxMtxSym& E, FlxMtxSym& K);
    friend void MtxProd_BTKB(const FlxMtx_base& B, const FlxMtx_baseS& E, FlxMtxSym& K);
    friend void MtxProd_BTKB(const FlxMtxTransformation&, const FlxMtxSym&, FlxMtxSym&);
};

/**
* @brief stores a diagonal matrix
*/
class FlxMtxDiag : public FlxMtx_baseS {
  private:
    tnlong msize;
    tVec mtx;
  public:
    const tnlong nrows() const {return msize;};
    const tnlong ncols() const {return msize;};

    FlxMtxDiag(const tnlong& msizeV,const tdouble defVal = 0.0) :msize(msizeV),mtx(defVal,msizeV) {};
    FlxMtxDiag(const FlxMtx_baseS& mb);
    virtual void MultMv(const tVec& v, tVec& w) const;
    void MultMv(const pVec& v, pVec& w) const;
    FlxMtxDiag& operator+=(const FlxMtxDiag& m) { mtx+=m.mtx; return *this; };
    FlxMtxDiag& operator*=(const tdouble& s) {mtx*=s; return *this; };
    const tdouble operator()(const tnlong& i, const tnlong& j) const { if (i==j) return mtx[i]; else return 0.0; }
    tdouble& operator()(const tnlong&i) { return mtx[i]; }
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false);
    /**
    * @brief returns the matrix-vector (constant!)
    */
    const tVec& get_mtx_tVec() const { return mtx; }
    
    virtual FlxMtx_baseS* get_Inverse();
};

/**
* @brief the identity matrix
*/
class FlxMtxIdentity : public FlxMtx_baseS {
  private:
    tnlong msize;
  public:
    const tnlong nrows() const {return msize;};
    const tnlong ncols() const {return msize;};

    FlxMtxIdentity(const tnlong& msizeV) :msize(msizeV) {};
    virtual void MultMv(const tVec& v, tVec& w) const { w = v; }
    void MultMv(const pVec& v, pVec& w) const { w = v; }
    FlxMtxIdentity& operator+=(const FlxMtx_base& m);
    FlxMtxIdentity& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const { if (i==j) return 1.0; else return 0.0; }
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false) { return true; }
    
    virtual FlxMtx_baseS* get_Inverse();
};

/**
* @brief stores a sparse symmetric matrix
*/
class FlxMtxSparsSym : public FlxMtx_baseS {
  private:
    tdouble* sa;
    tnlong* ija;
    void preconding(const tVec& r, tVec& z, const int precn = 0) const;
  public:
    const tnlong nrows() const {return ija[0]-1;};
    const tnlong ncols() const {return ija[0]-1;};

    FlxMtxSparsSym(const FlxMtxSym& Mtx);
    virtual ~FlxMtxSparsSym();
    virtual void MultMv(const tVec& v, tVec& w) const;
    void MultMv(const pVec& v, pVec& w) const;
    FlxMtx_base& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    void set_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false);
    
    virtual void output_Mtx(std::ostream& sout) const;
    const tdouble Percentage_of_NonZero() const { return (ija[ija[0]-1]+ija[0])/(pow2(ija[0]-1.)-ija[0]+1)*2;}
    
    friend class FlxMtxLTri;
    friend class FlxMtxSparsSymILU;
    friend class FlxMtxSparsLTri;
};

class FlxMtxSparsSFEMSym : public FlxMtx_baseS {
  private:
    /**
    * @brief Dimension of the stiffness matrix
    */
    tnlong Kdim;
    tdouble* sa;
    FlxMtx_baseS** sb;
    tnlong* ija;
    std::map<FlxMtx_baseS*, tuint> box;
    virtual void assembleMinv(int i);
  public:
    const tnlong nrows() const {return (ija[0]-1)*Kdim;};
    const tnlong ncols() const {return nrows();};

    /**
    * @brief Assembles this matrix
    * fM: the multiply coefficients
    * KM: the matrix coefficients
    * StfMtxV: the indexed stiffness matrices as pointers
    * P: the size of the homogeneous chaos
    */
    FlxMtxSparsSFEMSym(std::valarray< tdouble >& fM, std::valarray< int >& KM, FlxMtx_baseS** StfMtxV, tulong P);
    virtual ~FlxMtxSparsSFEMSym();
    virtual void MultMv(const tVec& v, tVec& w) const;
    void MultMv(const pVec& v, pVec& w) const;
    FlxMtx_base& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false);
    
    virtual const bool solve_CG(tVec& x, tVec& f, tdouble &eps, tnlong& iter, tuint pcn, bool startZero);
    
    virtual void output_Mtx(std::ostream& sout) const;
};

class FlxMtxPrecnInvSFEMSym : public FlxMtx_baseS {
  private:
    tVec c0iiInv;
    tnlong Kdim;
    FlxMtxSym* Kinv;
  public:
    const tnlong nrows() const {return c0iiInv.size()*Kdim; }
    const tnlong ncols() const {return nrows();};

    FlxMtxPrecnInvSFEMSym(FlxMtxSparsSym& K, const tVec& c0ii);
    ~FlxMtxPrecnInvSFEMSym();
    virtual void MultMv(const tVec& v, tVec& w) const;
    FlxMtx_base& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false) { return true; }
};

class FlxMtxPrecnILUSFEMSym : public FlxMtx_baseS {
  private:
    tVec c0iiInv;
    tnlong Kdim;
    FlxMtxSparsSymILU* ILU;
  public:
    const tnlong nrows() const {return c0iiInv.size()*Kdim; }
    const tnlong ncols() const {return nrows();};

    FlxMtxPrecnILUSFEMSym(FlxMtxSparsSym& K, const tVec& c0ii, bool FullDecomp=false, bool mod0diagentry=false);
    ~FlxMtxPrecnILUSFEMSym();
    virtual void MultMv(const tVec& v, tVec& w) const;
    void MultMv(const pVec& v, pVec& w) const;
    FlxMtx_base& operator*=(const tdouble& s);
    const tdouble operator()(const tnlong& i, const tnlong& j) const;
    void add_value(const tnlong& i, const tnlong& j, const tdouble& v);
    const bool isPositiveDefinite(tuint& r, const bool fixIt=false) { return true; }
};


/**
* @brief Matrix multiplication of type K=Trans(B)*E*B
*/
void MtxProd_BTKB(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K);
FlxMtxSym MtxProd_BTKB(const FlxMtx& B, const FlxMtxSym& E);
void MtxProd_BTKB(const FlxMtxLTri& B, const FlxMtxSym& E, FlxMtxSym& K);
FlxMtxSym MtxProd_BTKB(const FlxMtxLTri& B, const FlxMtxSym& E);
void MtxProd_BTKB(const FlxMtx_base& B, const FlxMtx_baseS& E, FlxMtxSym& K);
FlxMtxSym MtxProd_BTKB(const FlxMtx_base& B, const FlxMtx_baseS& E);
void MtxProd_BTKB(const FlxMtxTransformation& B, const FlxMtxSym& E, FlxMtxSym& K);
FlxMtxSym MtxProd_BTKB(const FlxMtxTransformation& B, const FlxMtxSym& E);
void MtxProd_BTKB_1D(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K);
/**
* @brief Matrix multiplication (for a transformation where just a part of K has to be transformed)
* @param B the transformation matrix (3x3-matrix)
* @param E the element stiffness matrix in local coordinates (2x2-matrix)
* @param K the element stiffenss matrix in global coordinates (4x4-matrix)
* @param tranformFirst if true: the first DOF should be transfomed; if false: the second DOF should be transfomed
*/
void MtxProd_BTKB_1D_part(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K, bool tranformFirst);

/**
* @brief Calculates v*Transpose(v)
*/
FlxMtxSym VecDyadProd(const tVec& v);


#endif // fesslix_FlxMtx_H


