/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxdata.h"

#include "flxfem_element_ptruss.h"

#include <fstream>

using namespace std;

FlxData* FlxDataBase::data = NULL;
FlxMatrixReader* FlxReaderBase3::mtxReader = NULL;
#if FLX_DEBUG
  int FlxData::Cinst = 0;
  int FlxReadManager::Cinst = 0;
  int FlxTimerBox::Cinst = 0;
  int FlxIgnoreBox::Cinst = 0;
  int FlxSubBox::Cinst = 0;
  int FlxOstreamBox::Cinst = 0;
  int FlxIstreamBox::Cinst = 0;
#endif

FlxReadManager::FlxReadManager()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxReadManager' created ...";
      throw FlxException("FlxReadManager::FlxReadManager", ssV.str() );
    }
  #endif
}

void FlxReadManager::push(ReadStream* readerV)
{

  if (readerV == NULL) {
    ostringstream ssV_2;
    ssV_2 << "Empty Reader.";
    throw FlxException("FlxReadManager::push_1", ssV_2.str() ); 
  }

  s.push(readerV);
  reader = readerV;
}

void FlxReadManager::pop()
{
  if (s.empty()) {
    ostringstream ssV_2;
    ssV_2 << "No reader on the stack.";
    throw FlxException("FlxReadManager::pop_1", ssV_2.str() ); 
  }
  s.pop();
  if ( s.empty() ) {
//     ostringstream ssV_2;
//     ssV_2 << "No reader on the stack.";
//     throw FlxException("FlxReadManager::pop_2", ssV_2.str() ); 
    reader = NULL;
  } else {
    reader = s.top();
  }
}

FlxFunction* FlxReadManager::parse_function(string funStr)
{
  ReadStream* rs = new ReadStream(std::string(funStr));
  push(rs);
  FlxFunction* value = NULL;
  try {
    value = new FlxFunction(funReader);	
  } catch (FlxException &e) {
    pop();
    delete rs;
    throw;
  }
  pop();
  delete rs;
  return value;
}

FlxData::FlxData()
#if FLX_USE_PYTHON
:PythBox(this)
#endif
{ 
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxData' created ...";
      throw FlxException("FlxData::FlxData", ssV.str() );
    }
  #endif
  FlxDataBase::set_data(this);
  FEM = new flxFEM(&ConstantBox);
  FlxRndBox::set_RFbox(&RFbox);
  FlxString::set_StrFunBox(&StrFunBox);
}

FlxData::~FlxData()
{
  delete FEM;
}

std::ostream& FlxOstreamBox::get ( string name ) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  map<std::string, std::ostream*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return (*pos->second);
  } else {
    ostringstream ssV_2;
    ssV_2 << "The output-stream '" << name << "' does not exist.";
    throw FlxException("FlxOstreamBox::get_1", ssV_2.str(), "In oder to use an output-stream, you have to define it first."); 
  }
}

bool FlxOstreamBox::delete_stream(ostream* strm)
{
  // if stream is the logStream
    if ( strm == &(GlobalVar.get_log()) ) return false;
  // if stream is a FileStream
      ofstream* thestream = dynamic_cast<ofstream*> (strm);
      if ( thestream != NULL ) {
	thestream->close();
	delete thestream;
	return true;
      }
    // if stream is a DistributorStream
      flxStreamAlloc* thestreamD = dynamic_cast<flxStreamAlloc*> (strm);
      if ( thestreamD != NULL ) {
	delete thestreamD;
	return true;
      }
   return false;
}

FlxOstreamBox::FlxOstreamBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxOstreamBox' created ...";
      throw FlxException("FlxOstreamBox::FlxOstreamBox", ssV.str() );
    }
  #endif
}

FlxOstreamBox::~FlxOstreamBox() {
  for (map<std::string, std::ostream*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    if (pos->first!="cout") delete_stream(pos->second);
  }
}

void FlxOstreamBox::insert( std::string name, std::ostream* value) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pair<std::string, std::ostream*> Element(name, value);
  map<std::string, std::ostream*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    ostringstream ssV;
    ssV << "A stream with that name '" << name << "' does already exist.";
    throw FlxException("FlxOstreamBox::insert_1", ssV.str() ); 
  }
  box.insert(Element);
}

void FlxIstreamBox::insert(string name, FlxIstream* value, bool errSerious)
{
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pair<std::string, FlxIstream*> Element(name, value);
  map<std::string, FlxIstream*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    pos->second->copyStream(value,errSerious);
  }
  box.insert(Element);
}

FlxIstream& FlxIstreamBox::get(string name)
{
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  map<std::string, FlxIstream*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return (*pos->second);
  } else {
    ostringstream ssV_2;
    ssV_2 << "The input-stream '" << name << "' does not exist.";
    throw FlxException("FlxIstreamBox::get_1", ssV_2.str(), "In oder to use an input-stream, you have to define it first."); 
  }
}

FlxIstreamBox::FlxIstreamBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxIstreamBox' created ...";
      throw FlxException("FlxIstreamBox::FlxIstreamBox", ssV.str() );
    }
  #endif
}

FlxIstreamBox::~FlxIstreamBox()
{
  for (map<std::string, FlxIstream*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    delete pos->second;
  }
}


FlxSubBox::FlxSubBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxSubBox' created ...";
      throw FlxException("FlxSubBox::FlxSubBox", ssV.str() );
    }
  #endif
}

FlxSubBox::~FlxSubBox() {
  for (map<std::string, FlxObjBase*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxSubBox::insert(std::string name, FlxObjBase* value) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pair<std::string, FlxObjBase*>Element(name, value);
  if ( ! box.insert(Element).second ) {
    delete value;
    ostringstream ssV;
    ssV << "Procedure '" << name << "' is already defined.";
    throw FlxException("FlxSubBox::insert_1", ssV.str() );
  }
}

FlxObjBase* FlxSubBox::get(std::string name) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  map<std::string, FlxObjBase*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}


FlxIgnoreBox::FlxIgnoreBox(): iL_recur_active(0) 
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxIgnoreBox' created ...";
      throw FlxException("FlxIgnoreBox::FlxIgnoreBox", ssV.str() );
    }
  #endif
}

bool FlxIgnoreBox::isOnIgnoreList_recur( std::string objName ) {
  transform(objName.begin(), objName.end(), objName.begin(), (int(*)(int)) std::tolower);
  vector<string>::const_iterator iter = find ( ignoreList_recur.begin(), ignoreList_recur.end(), objName);
  return (iter != ignoreList_recur.end() )?true:false; 
}



void FlxTimerBox::insert(string name, FlxTimer* value) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pair<std::string, FlxTimer*>Element(name, value);
  if ( ! box.insert(Element).second ) {;
    delete value;
    ostringstream ssV;
    ssV << "Timer '" << name << "' is already defined.";
    throw FlxException("FlxTimer::insert_1", ssV.str() );
  }
}


FlxTimer* FlxTimerBox::get(string name) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  map<std::string, FlxTimer*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    ostringstream ssV;
    ssV << "Timer '" << name << "' does not exist.";
    throw FlxException("FlxTimer::get_1", ssV.str() );
  }
}


void FlxTimerBox::deleteEl(std::string name)
{
  FlxTimer* tt = get(name);
  delete tt;
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  box.erase(name);
}

FlxTimerBox::FlxTimerBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxTimerBox' created ...";
      throw FlxException("FlxTimerBox::FlxTimerBox", ssV.str() );
    }
  #endif
}

FlxTimerBox::~FlxTimerBox() {
  for (map<std::string, FlxTimer*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxProgress::start(const tuint N_v) {
  if (!active) return;
  N = N_v;
  last_t = clock();
  op << "0%"; op.flush(); op << "\b\b";
  last_p = 0;
  is_running = true;
}

void FlxProgress::tick_t(const tuint idx) {
  const tuint p = int((idx*100)/N);
  if (p==last_p) return;
  op << p << "%"; op.flush();
  op << "\b\b";
  if (p>=10) {
    op << '\b';
    if (p>=100) {
      op << '\b';
    }
  }
  last_p = p;
  last_t = clock();
}

void FlxProgress::stop() {
  if (!active) return;
  op << "    "; op.flush(); op << "\b\b\b\b";
//   op.flush();
  is_running = false;
}

void FlxObjBase::exec() {
  if (NOTdolog) NOTdolog_start();
  task();
  if (NOTdolog) NOTdolog_stop();
  #if FLX_DEBUG
    GlobalVar.slog_flush();
  #endif
  if ( Next != NULL ) {
    Next->exec();
  }
}

void FlxObjBase::attach_obj(FlxObjBase* NextV) {
  if ( Next == NULL ) {
    Next = NextV;
  } else {
    Next->attach_obj(NextV);
  }
}


FlxObjBase::~FlxObjBase() {
  if (Next != NULL) {
    delete Next;
  }
}