/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxdata_H
#define fesslix_flxdata_H

#include "flxrandomFEM.h"

class FlxObjBase;

class FlxReadManager : public FlxReaderBase3 {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::stack<ReadStream*> s;
  public:
    FlxReadManager();
    /**
    * @brief Set a new reader on top of the stack
    */
    void push(ReadStream* readerV);
    /**
    * @brief Removes the reader from the stack -> memory is NOT deallocated
    */
    void pop();
    
    FlxFunction* parse_function(std::string funStr);
    
    static void set_funReader( FlxFunctionReader* funReaderV) {funReader = funReaderV;};
    static void set_mtxReader( FlxMatrixReader* mtxReaderV) { mtxReader = mtxReaderV;};
};


class flxBufAlloc : public std::streambuf
{
  private:
    typedef std::streambuf base_type;
    typedef base_type::traits_type traits_type;

    flxBufAlloc( const flxBufAlloc& );
    flxBufAlloc& operator=( const flxBufAlloc& ); // Kopieren verhindern

    std::streambuf* m_sb1;
    std::streambuf* m_sb2;

  protected:
    virtual int_type overflow( int_type c = traits_type::eof() )  {   
        if( m_sb1 && m_sb2 
            && !traits_type::eq_int_type( m_sb1->sputc( c ), traits_type::eof() )
            && !traits_type::eq_int_type( m_sb2->sputc( c ), traits_type::eof() )
            )
            return traits_type::not_eof( c );
        return traits_type::eof();  // einer der Streambufs tut nicht
    }

  public:
    flxBufAlloc ( std::ostream& os1, std::ostream& os2 ) : m_sb1( os1.rdbuf() ) , m_sb2( os2.rdbuf() ) { }

};

class flxDummyAlloc : public std::streambuf
{
  private:
    typedef std::streambuf base_type;
    typedef base_type::traits_type traits_type;

    flxDummyAlloc( const flxDummyAlloc& );
    flxDummyAlloc& operator=( const flxDummyAlloc& ); // Kopieren verhindern

  protected:
    virtual int_type overflow( int_type c = traits_type::eof() )  {   
       return c;
    }

  public:
    flxDummyAlloc ( ) { }

};

class flxStreamAlloc : public std::ostream {
  private:
    flxBufAlloc theBuf;

  public:
    flxStreamAlloc ( std::ostream& os1, std::ostream& os2 ) : std::ostream(&theBuf), theBuf(os1, os2) { }
};

class flxDummyOstream : public std::ostream {
  private:
    flxDummyAlloc theBuf;

  public:
	  flxDummyOstream ( ) :std::ostream(&theBuf) {  }
};





/**
* @brief A class for storing output-streams (ostream&)
*/
class FlxOstreamBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, std::ostream*> box;
    bool delete_stream(std::ostream* strm);
  public:
    FlxOstreamBox();
    ~FlxOstreamBox();
    /**
    * @brief Insert a ostream.
    * @param name An unique name
    * @param value The ostream to store (a pointer to that ostream is stored)
    */
    void insert ( std::string name, std::ostream* value);
    /**
    * @brief Get a ostream by name.
    * @param name The name of the ostream to return.
    * @return std::ostream&
    */
    std::ostream& get ( std::string name);
};


/**
* @brief A class for storing input-streams (ostream&)
*/
class FlxIstreamBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, FlxIstream*> box;
    bool delete_stream(FlxIstream* strm);
  public:
    FlxIstreamBox();
    ~FlxIstreamBox();
    /**
    * @brief Insert a ostream.
    * @param name An unique name
    * @param value The ostream to store (a pointer to that ostream is stored)
    */
    void insert ( std::string name, FlxIstream* value, bool errSerious=true);
    /**
    * @brief Get a FlxIstream by name.
    * @param name The name of the ostream to return.
    * @return FlxIstream&
    */
    FlxIstream& get ( std::string name );
};

class FlxSubBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, FlxObjBase*> box;
  public:
    FlxSubBox();
    ~FlxSubBox();
    void insert (std::string name, FlxObjBase* value);
    FlxObjBase* get(std::string name);
};


#if FLX_USE_PYTHON
class FlxData;
class FlxPythonBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    bool initialized;
    std::map<std::string, PyObject*> module_box;
    
    
    void init_python() { if (initialized) return; else init_python_actual(); }
    void init_python_actual();
    PyObject* get_module(const std::string& name);
    PyObject* get_variable(const std::string& mod_name, const std::string& var_name);
  public:
    FlxPythonBox(FlxData* dref);
    ~FlxPythonBox();
    void insert_module (const std::string& name, const bool throw_err = true);
    PyObject* get_function(const std::string& mod_name, const std::string& fun_name);
    const tdouble get_variable_double(const std::string& mod_name, const std::string& var_name);
    void set_variable(const std::string& mod_name, const std::string& var_name, const tdouble value);
    void register_path(const std::string& path);
    void execute_cmd(const std::string& cmdStr);
};
#endif


class FlxIgnoreBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::vector<std::string> ignoreList_recur;
    int iL_recur_active;
  public:
    FlxIgnoreBox ();
    void set_iL_recur(std::string objName ) { ignoreList_recur.push_back(objName); }
    void activate_iL_recur() { iL_recur_active++; }
    void deactivate_iL_recur() { iL_recur_active--; }
    bool isActive_iL_recur() { return (iL_recur_active > 0)?true:false; }
    bool isOnIgnoreList_recur( std::string objName );
};


class FlxTimer {
  private:
    bool is_running;
    clock_t startV, delta;
  
  public:
    FlxTimer() : is_running(false), delta(0) {}
    void start() { if (!is_running) { is_running = true; startV = clock(); } }
    void stop() { if (is_running) { is_running = false; delta += clock()-startV; } }
    clock_t get_ticks () { return delta; }
    bool IsRunning () { return is_running; }
};

class FlxTimerBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, FlxTimer*> box;
  public:
    FlxTimerBox();
    ~FlxTimerBox();
    void insert (std::string name, FlxTimer* value);
    FlxTimer* get(std::string name);
    void deleteEl ( std::string name );  
};

class FlxProgress {
  private:
    const bool active;
    std::ostream& op;
    tuint N;	// total size
    clock_t last_t;	// last time of action
    tuint last_p;
    bool is_running;
    tuint size_e;	// number of chars written during last step
    
    void tick_t(const tuint idx);
  public:
    FlxProgress(std::ostream& op) : active(GlobalVar.prgBar),op(op),last_p(0),is_running(false),size_e(0) { }
    
    void start(const tuint N_v);
    inline void tick(const tuint idx) { if (!active) return;
      if (tdouble(clock()-last_t)<0.2*CLOCKS_PER_SEC) return;
      tick_t(idx);
    };
    void stop();
};

/**
* @brief A class for storing data
*/
class FlxData {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
  public:
    FlxOstreamBox OstreamBox;
    FlxIstreamBox IstreamBox;
    FlxConstantBox ConstantBox;
    FlxVarBox VarBox;
    FlxFunctionBox FunBox;
    FlxConstMtxBox ConstMtxBox;
    FlxMtxFunBox MtxFunBox;
    FlxSubBox SubBox;
    FlxIgnoreBox IgnoreBox;
    FlxTimerBox TimerBox;
    GaussIntegration GaussInt;
    flxFEM *FEM;
    FlxRndBox RndBox;
    FlxRFBox RFbox;
    FlxReadManager ReadManager;
    FlxStringFunBox StrFunBox;
    #if FLX_USE_PYTHON
      FlxPythonBox PythBox;
    #endif
    
    FlxData();
    ~FlxData();
};

class FlxDataBase {
  protected:
    static FlxData* data;
  public:
    ~FlxDataBase() {};
     /**
    * @brief Set the DataBox 
    * @param dataV a pointer to the DataBox to use
    */
    static void set_data ( FlxData *dataV ) { data = dataV; };
};


// ****************** flxOBJECTS *************************************************

/**
* @brief The base class for all other 'object classes'
*/
class FlxObjBase : public FlxDataBase {
  private:
    /**
    * @brief The next 'object class'
    *
    * If the command is NOT inside of a loop or an if-block, 'Next' will be NULL.
    */
    FlxObjBase* Next;
    /**
    * @brief true if this object should not log
    */
    const bool NOTdolog;
    void NOTdolog_start() { if (GlobalVar.logLevel>2) GlobalVar.logLevel=2; }
    void NOTdolog_stop() { GlobalVar.logLevel_strong_reset(); }
  protected:
    /**
    * @brief This function has to be overriden by all 'object classes' (that is the hart)
    */
    virtual void task() = 0;
    //virtual void task() {};
  public:
    FlxObjBase (bool dolog) : Next(NULL),NOTdolog(!dolog) {}
    virtual ~FlxObjBase();
    /**
    * @brief Execute this and all following 'object classes'
    */
    virtual void exec();
    /**
    * @brief Attach an 'object class' to the end of the list of 'object classes'
    */
    void attach_obj ( FlxObjBase* NextV );
};

/**
* @brief object class: dummy class - does nothing
*/
class FlxObjDummy : public FlxObjBase {
  private:
    void task() {};
  public:
    FlxObjDummy ( ) : FlxObjBase(true) {}
};



#endif // fesslix_flxdata_H
