/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxfunction_ext_H
#define fesslix_flxfunction_ext_H

#include "flxfunction.h"



/**
* @brief creates a FlxPoint out of FlxFunctions
*/
class FlxFunPoint {
  public:
    /**
    * @brief frame of reference (coordinate system)
    */
    enum FoR { 
      cartesian, cylindrical, spherical
    };
    static FoR get_FoR(char strlFoR,bool errSerious=true);
    static char get_FoR(FoR elFoR);
  private:
    FoR FrORef;
    FlxFunction *d1, *d2, *d3;
  public:
    FlxFunPoint(const FlxFunPoint& ref);
    FlxFunPoint(ReadStream *reader, FlxFunctionReader *funReader,bool errSerious=true);
    FlxFunPoint(const FoR FrORef, FlxFunction *d1, FlxFunction  *d2, FlxFunction *d3);
    ~FlxFunPoint();
    FlxFunPoint& operator=(const FlxFunPoint& ref);
    /**
    * @brief returns a pointer to a point
    * note: memory has to be deallocated
    */
    const flxPoint calc() const;
    /**
    * @brief frame of reference (coordinate system)
    * @param dir x, y or z
    */
    const bool is_direction_zero(const char dir);
    
    friend std::ostream& operator<<(std::ostream& os, const FlxFunPoint& val);
};

std::ostream& operator<<(std::ostream& os, const FlxFunPoint& val);


/**
* @brief integrates a function numerically using Gaussian quadrature
* @return the result of the integration
* @param funI the function to integrate
* @param theconst the integration variable (const-variable)
* @param start start of the integral
* @param end end of the integral
* @param GPN the number of Gauss-points to use
*/
const tdouble FlxFun_GaussIntegration(FunBase* funI, tdouble* theconst, const tdouble& start, const tdouble& end, const tuint& GPN, GaussIntegration& GI);

/**
* @brief estimates the number of Gauss-points required in order to integrate funI with an error less than err
* @return number of Gauss-points 
*   returns 0 if funI is constant w.r.t. theconst
*   returns GPmax+1 if the iteration did not converge within the specified bounds
* @param funI the function to integrate
* @param theconst the integration variable (const-variable)
* @param start start of the integral
* @param end end of the integral
* @param GPmax the maximum number of Gauss-points to check the integration-error
* @param GPtest the number of Gauss-points to use as reference solution
* @param err the error (relative error) required to stop the iteration
*/
const tuint FlxFun_EstimateGaussPoint(FunBase* funI, tdouble* theconst, const tdouble& start, const tdouble& end, const tuint& GPmax, const tuint& GPtest, const tdouble& err, GaussIntegration& GI);

/**
* @brief finds a root of a function using the bisection algorithm
* @return the result of the root-search
* @param funR the function to analyze
* @param theconst the root-search variable (const-variable)
* @param start start of the search-interval
* @param end end of the search-interval
* @param dx epsilon for interval size
* @param dy epsilon for function-value
*/
const tdouble FlxFun_RootSearch_Bisec(FunBase* funR, tdouble* theconst, tdouble start, tdouble end, const tdouble dx, const tdouble dy);
const tdouble FlxFun_RootSearch_RegulaFalsi(FunBase* funR, tdouble* theconst, tdouble start, tdouble end, const tdouble dx, const tdouble dy);

/**
* @brief finds the minimum of a function (1D) using the Golden section search
* @return the result of the minimum-search
* @param funR the function to analyze
* @param theconst the optimization-search variable (const-variable)
* @param start start of the search-interval
* @param end end of the search-interval
* @param tau stop-criteria of the Golden section search
* @param nex number of the maximum number of interations in the exponential search
*/
const tdouble FlxFun_Optimize1D(FunBase* funR, tdouble* theconst, tdouble start, tdouble end, const tdouble tau, const tuint nex);


#endif // fesslix_flxfunction_ext_H
