/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxfunction_ope_calc_H
#define fesslix_flxfunction_ope_calc_H

#include "flxfunction_ext.h"


/**
* @brief arithmetic class: stores a constant from the ConstantBox
*/
class FunConst : public FunBase, public FlxBoxBase {
  private:
    tdouble* thenumber;
  public:
    FunConst (tdouble* TheNumber) : thenumber(TheNumber) {};
    const tdouble calc() { return *thenumber; }
    const bool search_circref(FlxFunction* fcr) { return false; }
    const tuint precedence() { return 0; }
    const std::string write();
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const theconst) { return thenumber==theconst; }
    const tdouble expectation() { return *thenumber; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: stores a arithmetic expression from the VarBox
*/
class FunVar : public FunBase, public FlxBoxBase {
  private:
    FlxFunction* thefun;
  public:
    FunVar (FlxFunction* TheFun) : thefun(TheFun) {};
    const tdouble calc() { return thefun->calc(); }
    const bool search_circref(FlxFunction* fcr);
    const tuint precedence() { return 0; }
    const std::string write();
    void dependOn_RV(bVec& bv) { thefun->dependOn_RV(bv); }
    const bool dependOn_Const(const tdouble* const thenumber) { return thefun->dependOn_Const(thenumber); }
    const tdouble expectation() { return thefun->expectation(); }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: add two values
*/
class FunAdd : public FunBaseOperat2 {
  public:
    FunAdd (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() { return child_1->calc() + child_2->calc(); }
    const std::string write();
    const tuint precedence() { return 12; }
    const tdouble expectation() { return child_1->expectation() + child_2->expectation(); }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: subtract two values
*/
class FunSub : public FunBaseOperat2 {
  public:
    FunSub (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() { return child_1->calc() - child_2->calc(); }
    const std::string write();
    const tuint precedence() { return 12; }
    const tdouble expectation() { return child_1->expectation() - child_2->expectation(); }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: Negation
*/
class FunNot : public FunBaseOperat1 {
  public:
    FunNot (FunBase *child1) : FunBaseOperat1(child1) {};
    const tdouble calc() { 
      if ( fabs(child_1->calc())<=GlobalVar.TOL() ) { return 1.0; }
      else { return 0.0; }
    }
    const std::string write();
    const tuint precedence() { return 0; }
};

/**
* @brief arithmetic class: multiply two values
*/
class FunMult : public FunBaseOperat2 {
  public:
    FunMult (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() { return child_1->calc()*child_2->calc(); }
    const std::string write();
    const tuint precedence() { return 11; }
    const tdouble expectation();
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: divide two values
*/
class FunMult_Div : public FunBaseOperat2 {
  public:
    FunMult_Div (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() {  return child_1->calc() / child_2->calc(); }
    const std::string write();
    const tuint precedence() { return 11; }
    const tdouble expectation();
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: raise a value to a given power
*/
class FunPower : public FunBaseOperat2 {
  public:
    FunPower (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() { return std::pow (child_1->calc(), child_2->calc()); }
    const std::string write();
    const tuint precedence() { return 10; }
    const tdouble expectation();
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: compares two values ((not) equal to)
*/
class FunEqual : public FunBaseOperat2 {
  private:
    const bool isEqual;
  public:
    FunEqual (FunBase *child1, FunBase *child2, const bool IsEqual) : FunBaseOperat2(child1, child2), isEqual(IsEqual) {};
    const tdouble calc() { 
      if (fabs(child_1->calc()-child_2->calc()) <= GlobalVar.TOL() ) { return isEqual; }
      else { return !isEqual; }
    }
    const std::string write();
    const tuint precedence() { return 14; }
};

/**
* @brief arithmetic class: compares two values (<, <=, >=, >)
*/
class FunLessThan : public FunBaseOperat2 {
  private:
    bool isEqual;
    bool isLess;
  public:
    FunLessThan (FunBase *child1, FunBase *child2, bool IsEqual, bool IsLess) : FunBaseOperat2(child1, child2), isEqual(IsEqual), isLess(IsLess) {};
    const tdouble calc();
    const std::string write();
    const tuint precedence() { return 13; }
};

/**
* @brief arithmetic class: Logical AND
*/
class FunAnd : public FunBaseOperat2 {
  public:
    FunAnd (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() { 
      if (!(fabs(child_1->calc()) > GlobalVar.TOL())) return false;
      return (fabs(child_2->calc()) > GlobalVar.TOL());
    }
    const std::string write();
    const tuint precedence() { return 15; }
};

/**
* @brief arithmetic class: Logical AND
*/
class FunOr : public FunBaseOperat2 {
  public:
    FunOr (FunBase *child1, FunBase *child2) : FunBaseOperat2(child1, child2) {};
    const tdouble calc() { 
      if (fabs(child_1->calc()) > GlobalVar.TOL()) return true;
      return (fabs(child_2->calc()) > GlobalVar.TOL());
    }
    const std::string write();
    const tuint precedence() { return 16; }
};

/**
* @brief arithmetic class: ternary operator
*/
class FunTernary : public FunBaseOperat3 {
  public:
    FunTernary (FunBase *child1, FunBase *child2, FunBase *child3) : FunBaseOperat3(child1, child2, child3) {};
    const tdouble calc() { 
      if ( fabs(child_1->calc()) > GlobalVar.TOL() ) {
	return child_2->calc();
      } else {
	return child_3->calc();
      }
    }
    const std::string write();
    const tuint precedence() { return 17; }
};


#endif // fesslix_flxfunction_ope_calc_H
