/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxglobal_H
#define fesslix_flxglobal_H

#include "flxplatform.h"
#include "pdouble.h"
#include "flxexception.h"

const tdouble PI = 3.14159265358979323846264338327950;
const tdouble GAMMA = 0.5772156649015328606065120900824;   // Euler's constant

void print_FLX_GNU(std::ostream& sout);
void print_GNU(std::ostream& sout);

#include <valarray>
typedef std::valarray<tdouble> tVec;
typedef std::valarray<pdouble> pVec;
typedef std::valarray<bool> bVec;
typedef std::valarray<tuint> iVec;
typedef std::valarray<tnlong> lVec;

size_t bVec_NumbTrue(const bVec& bv);

void FlxError(bool errSerious, std::string errnumber, std::string Titel = "", std::string Msg = "");

class strGlobalVar {
  private:
    std::ostream* sdummy;
    std::ostream* slogP;
    /**
    * @brief stores a pointer to the standard output stream
    */
    std::ostream* stdcout;
    
    tuint D2S_Prec;
    tuint D2S_Type;
    tdouble D2S_BValU;
    tdouble D2S_BValL; 
    bool D2S_Del0;
    bool D2S_DelP;
    tdouble TOLval;
    /**
    * @brief stores if the pre-log is active
    */
    bool prelog_active;
    /**
    * @brief stores if the pre-log is active
    */
    std::string prelog_stream;
    /**
    * @brief this is the default loglevel
    */
    int logLevel_strong;
    /**
    * @brief stores the directory of the binary (executable)
    */
    std::string exe_dir;
    /**
    * @brief truncates new_line at the end of prelog
    */
    void prelog_prepare_write();
    /**
    * @brief true, if fesslix is run in leak-check-mode
    */
    bool leak_check;
  public:
    /**
    * @brief this is the temporary loglevel - can be reset by objects
    */
    int logLevel;
    tuint MT19937_init_calls;
    bool MT19937_init_RAND;
    bool MT19937_init_seed;
    tuint MT19937_init_seedvalue;
    tnlong Exp_Num_Intervals_1D;
    tnlong Exp_Num_Samples_1D;
    tdouble Exp_Num_Diff_Redistribute_1D;
    tdouble Exp_Num_Lbound_1D;
    tdouble Exp_Num_Ubound_1D;
    tdouble Exp_Num_Bound_MCI;
    tnlong Exp_Num_Intervals_MCI;
    tnlong Exp_Num_SmpPerInterval_MCI;
    tnlong Exp_Num_Smp_Redistribute_MCI;
    tdouble Exp_Num_Diff_Redistribute_MCI;
    tuint Exp_Num_Rnd_Moments;
    tuint Exp_Num_Rnd_Moments_Central;
    size_t LegendrePolyH_init_numb;
    bool allowInteractiveMode;
    bool prgBar;		// true if progress-bar is activated for timeconsuming jobs
    std::string configDir;
    std::string configDir_local;
    std::string pwd;
    FlxAlert alert;
    const tdouble sqrtEps;		// square-root of precision of epsilon
    void set_TOL(tdouble tolV);
    /**
    * @brief if a value is smaller or equal than TOL, it can be considered to be zero
    */
    const tdouble& TOL() const { return TOLval; }
    
    strGlobalVar();
    ~strGlobalVar();

    void set_sdummy(std::ostream* sdummyV) { sdummy = sdummyV; }
    void set_slog(std::ostream* slogV) { slogP = slogV; }
    void set_stdcout(std::ostream* stdcoutV) { stdcout = stdcoutV; }
    std::ostream& get_log() { return *slogP; }
    /**
    * @brief returns the logging-stream
    * (1) alerts and errors are logged (ALERT)
    * (2) warnings are logged (WARNING)
    * (3) normal but significant information is logged (NOTICE)
    * (4) informational (INFO)
    * (5) debug-level messages (DEBUG)
    */
    std::ostream& slog(int logLevel_);
    void slog_flush() { slogP->flush(); }
    bool check_logNOTcout() { return (slogP != stdcout); };
    std::string get_flxPrompt() { return "fesslix:> "; }
    
    std::string Double2String(tdouble dv, bool checkTOL = false);
    void Double2String_setPrec(tuint prec);
    void Double2String_setType(tuint type);
    void Double2String_setBValU(tdouble bv);
    void Double2String_setBValL(tdouble bv);
    void Double2String_setDel0(bool d0);
    void Double2String_setDelP(bool dP);
    void Double2String_log();
    /**
    * @brief set the strong loglevel
    */
    void logLevel_strong_set(int logL);
    /**
    * @brief set the temporary loglevel to the strong loglevel
    */
    void logLevel_strong_reset() { logLevel = logLevel_strong; };
    /**
    * @brief activates/deactivates the prelog
    */
    void prelog_activated(bool b);
    /**
    * @brief returns true if the prelog is active
    */
    bool prelog_isActive() { return prelog_active; }
    /**
    * @brief write the pre-log to the log - and clear it afterwards
    */
    void prelog_write();
    /**
    * @brief clear the pre-log
    */
    void prelog_clear();
    /**
    * @brief force writing the pre-log to the log - do not clear it afterwards
    */
    std::string prelog_force_write();
    /**
    * @brief check if the pre-log is not empty
    */
    bool prelog_isNOTempty() { if(prelog_stream!="") return true; else return false; }
    /**
    * @brief append a character to the pre-log
    */
    void prelog_append(char c);
    /**
    * @brief append a string to the pre-log
    */
    void prelog_append(std::string str);
    /**
    * @brief set exe_dir
    * @param str expects argv[0] from the main-function
    */
    void exe_dir_set(std::string str);
    /**
    * @brief returns the exe dir
    */
    std::string exe_dir_get() { return exe_dir; }
    /**
    * @brief sets leak_check to true
    */
    void set_leak_check_mode() { leak_check = true; }
    /**
    * @brief returns whether fesslix is run in leak-check mode
    */
    const bool is_leak_check() const { return leak_check; }
};
#ifdef fesslix_flxglobal_CPP
  strGlobalVar GlobalVar;
#else
  extern strGlobalVar GlobalVar;
#endif

#endif // fesslix_flxglobal_H