/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxio_H
#define fesslix_flxio_H

/*#include <fstream>*/
#include <algorithm>
#include <stack>
#include <vector>

#include "flxmath.h"


std::ofstream* open_stream(std::string filename);

//------------------------------------------------------------

class istream_warper {
  private:
    /**
    * @brief The file stream
    */
    std::istream *theStream;
    /**
    * The name of the file which is streamed.
    */
    const std::string filename;
    std::stack<int> mystack;
  public:
    istream_warper(std::istream *theStream, const std::string filename);
    ~istream_warper();
    const std::string& get_FileName() const { return filename; }
    const int get();
    const int peek();
    const bool eof();
    void putback(const int ch);
    const std::string get_line(const char delim=10);
};


//------------------------------------------------------------


/**
* @brief A class for extracting information out of a file.
*/
class ReadStream { 
  public:
    /**
    * @brief Labeles the type of a character.
    *
    * STRING: 'A'-'Z', 'a'-'z', '_'
    * NUMERAL: '0'-'9'
    * BRAKET: '{', '}', '[', ']', '(', ')'
    * QUOTE: '"', '''
    * COMMENT: '#'
    * NONE: tab, newline, blank
    * ENDOFFILE: eof
    * OTHER: any other character between '!'-'~'
    * UNKNOWN: any other character
    */
    enum InpType { STRING, NUMERAL, OTHER, BRAKET, QUOTE, NONE, COMMENT, ENDOFFILE, UNKNOWN };
    static InpType getType(char ch);
  protected:
    /**
    * @brief The file stream (warper)
    */
    istream_warper *theStream;
    /**
    * @brief How many blanks fit into a tab?
    */
    const int TabWidth;	// tab width (for calculation of column number)
    /**
    * @brief The current line number.
    */
    int lineNumb;	// current line number
    /**
    * @brief The current column number.
    */
    int charNumb;	// current column number
    /**
    * @brief if true: output will be written to pre-log
    */
    bool do_log;
    /**
    * @brief gets the next character in the stream --> and write to pre-log
    */
    void log_char(char c) { if (do_log) GlobalVar.prelog_append(c); }
    /**
    * @brief gets the next character in the stream --> and write to pre-log
    */
    const char stream_getnext_log();
    /**
    * @brief Go to the next character in the stream which is ( !( NONE || COMMENT) )
    */
    void setNext ( const bool DOlog=true );
  public:
    /**
    * @brief Create a ReadStream class.
    * @param FileName The name of the file to open.
    * @param tabWidth = 8; How many blanks fit into a tab?
    * @return ReadStream
    */
    ReadStream ( const char* FileName, bool do_log=false, int tabWidth = 8 );
    /**
    * @brief Create a ReadStream class.
    * @param strV The string to read.
    * @param tabWidth = 8; How many blanks fit into a tab?
    * @return ReadStream
    */
    ReadStream ( std::string strV, bool do_log=false, int tabWidth = 8 );
    
//     ReadStream ( std::istream& istr, int tabWidth = 8 );  // TODO: does not work correctly
    
    virtual ~ReadStream ( );
    /**
    * @brief Indicates the position where an error occured.
    * @return String describing the current position in text file (line number, column number, name of the file).
    */
    const std::string getCurrentPos ();
    /**
    * @brief Returns the type of the next character in the stream.
    * @return InpType
    */
    const InpType getNextType ();

    /**
    * @brief Returns the next character in the string.
    * @param DOsetNext = false; If this parameter is true, 'setNext()' is executed afterwards.
    * @return char
    * @see setNext()
    */
    const char getChar ( const bool DOsetNext = true, const bool DOlog=true );
    
    /**
    * @brief Returns the next character in the string - and throws an error if it is not the expected one.
    * @param theChar The expected character.
    * @return char
    * @throw FlxException
    */
    const char getChar ( const char theChar, const bool errSerious = true, const bool DOsetNext = true );

    /**
    * @brief Returns the next character in the stream without moving the current position in the stream.
    * @return char
    */
    const char whatIsNextChar ( );
    /**
    * @brief Returns the next string of length 'length' in the stream without moving the current position in the stream.
    * allowed are a-z A-z 0-1 _
    * @return char
    */
    const std::string whatIsNextString ( const int length, const bool lowercase = false );

    /**
    * @brief Returns the next word in the string.
    * 
    * The fist character of the word has to be of the type STRING.
    * For all other character the types STRING and NUMERAL are valid.
    * A character of any other type indicates the end of the word.
    * The function 'setNext()' is executed afterwards.
    * @return A string containing the extracted word.
    * @see setNext()
    * @throw FlxException
    */
    const std::string getWord (bool lowercase = false, bool errSerious = true, const bool numbeg=false);
    /**
    * @brief Checks if the next word in the stream is 'strWord'
    */
    const std::string getWord(const char* strWordC, bool errSerious = true);
    /**
    * @brief Checks if the next expression in the stream is 'strExpr'
    */
    void getExpr(const char* strWordC, bool errSerious = true);

    /**
    * @brief Returns the next bool-value in the string.
    * 
    * Allowed is:
    *  'true' or 'false'
    *  '1' or '0'
    * @return true or false
    * @see setNext()
    * @throw FlxException
    */
    const bool getBool(const bool errSerious = true);

    /**
    * @brief Returns the next word in the string.
    * 
    * The text TEXT must have the following format: "TEXT"
    * @return A string containing the extracted text.
    * @see setNext()
    * @throw FlxException
    */
    const std::string getText (const bool MultiLine = false, const bool errSerious = true);
  
    /**
    * @brief Checks if the next character can be the starting character of a number.
    * @return true or false
    */
    const bool nextCanBeNumber();

    /**
    * @brief Returns the next number in the string.
    * 
    * The format of the string describing the number must be of this format:
    * ['+','-']'0'-'9'...'0'-'9'[('e', 'E')['+','-']'0'-'9'...'0'-'9']
    * Note: the use of 'e' or 'E' to extract an integer value can cause problems.
    * The function 'setNext()' is executed afterwards.
    * @return The extracted numeric value.
    * @see setNext()
    * @throw FlxException
    */
    template <class T> const T getNumber ( const bool errSerious = true, bool signallow=true, const bool floating=true) {
      std::ostringstream ssV;
      char ch = theStream->peek();
      bool dezsep = !floating;
      bool dezexp = !floating; 
      
      // check if the first character is of type NUMERAL or if it is a leading sign
      if ( ! nextCanBeNumber() ) {
	std::ostringstream ssV_2;
	ssV_2 << "Character (" << ch << ") [" << theStream->peek() << "] not expected. (Only '0'-'9' and leading sign is valid.)";
	throw FlxException("ReadStream::getNumber_1", ssV_2.str(), getCurrentPos());
      } 
      
      // read the number into a stringstream
      while ( getNextType() == NUMERAL || ch == '.' || ch == 'E' || ch == 'e' || ch == '+' || ch == '-' ) {
	if ( (ch == '+'  || ch == '-') && signallow == false ) break;
	ssV << getChar(false);
	signallow = false;
	if ( ch == '.'  || ch == 'e' || ch == 'E' ) {	// check if '.', 'e' or 'E' have been used multiple times
	  if ( (ch == '.' && dezsep == false) || ( (ch == 'e' || ch == 'E') && dezexp == false ) ) {
	    if ( ch == '.' ) dezsep = true;
	    else {
	      dezexp = true;
	      signallow = true;
	    }
	  } else {
	    std::ostringstream ssV_2;
	    ssV_2 << "Multiple use of '" << ch << "'.";
	    FlxError(errSerious,"ReadStream::getNumber_2", ssV_2.str(), getCurrentPos());
	  }
	}
	ch = theStream->peek();
      }
    
      // convert the sting to an int (and ::check success)
      std::istringstream ssI( ssV.str() );
      T T1;
      bool errV = false;
      if ( ! dezexp ) {
	if ( ! (ssI >> T1) ) errV = true;
      } else {					// if 'e' or 'E' have been used (problematic for NOT double values)
	std::istringstream ssI2( ssV.str() );
	double d1;
	if ( ! ( ssI >> T1 && ssI2 >> d1 ) ) errV = true;
	if ( ! errV ) {
	  double d2 = d1 - T1;
	  if ( d2 > 1.0 || d2 < -1.0 ) {
	    T1 = T(d1);
	    d2 = d1 - T1;
	    if ( d2 > 1.0 || d2 < -1.0 ) errV = true;
	  }
	}
      }
      if ( errV ) {
	std::ostringstream ssV_2;
	ssV_2 << "The number (" << ssV.str() << ") could not be converted into a numeric value.";
	FlxError(errSerious,"ReadStream::getNumber_3", ssV_2.str(), getCurrentPos());
      }
      setNext();

      if ( whatIsNextChar() == '%' ) {			// check if there is a '%' after the number
	getChar('%');
        T1 = T1 / 100;
      } else if ( whatIsNextString(2) == "°" ) {	// check if there is a '°' after the number
	getExpr("°");
        T1 = T(T1 * (PI / 180));
      }
      
      setNext();
      return T1;
     }

};



//------------------------------------------------------------


class FlxReaderBase {
  protected:
    static ReadStream *reader;
  public:
    virtual ~FlxReaderBase() {};
};


//------------------------------------------------------------



class FlxIstream {
  protected:
    /**
    * @brief the name of the inputReader
    */
    std::string name;
    /**
    * @brief generate an error if end of file is not respected (Warning by default)
    */
    bool errEOF;
    
    
    /**
    * @brief this function is executed if the end of the file is reached
    */
    void reachedEOF();
  public:
    FlxIstream(std::string nameV, bool errEOFv);
    virtual ~FlxIstream() {}
    
    virtual void copyStream(FlxIstream* rhsB, bool errSerious=true) = 0;
    /**
    * @brief returns the next value
    * @return false if EOF is reached
    */
    virtual bool get_value(tdouble& v, bool suppressErr = false) = 0;
    /**
    * @brief returns the next value
    * @param v the vector to assemble
    * @param lastIndex if EOF occured, it contains the index where the error occured within the vector
    * @return false if EOF is reached
    */
    bool get_vec(tVec& v, size_t &lastIndex, bool suppressErr = false);
};

class FlxIstream_file : public FlxIstream {
  private:
    /**
    * @brief random sequence input: reader
    */
    ReadStream* iReader;
    /**
    * @brief random sequence input: blocksize
    */
    tuint blocksize;
    /**
    * @brief random sequence input: sequence vector
    */
    tVec* SeqVec;
    /**
    * @brief current index in SeqVec
    */
    tuint index;
    /**
    * @brief last valid index in SeqVec
    */
    tuint lastindex;
    /**
    * @brief sets the input to the next number (ignoring ',' and ';')
    */
    void set_next();
    /**
    * @brief read another block
    */
    void read_block();
    /**
    * @brief check for end of file
    */
    bool check_eof();
    
    // reading specific columns
      /**
      * @brief the number of columns in the file
      */
      int Cnumb;
      /**
      * @brief the column numbers to read - starts counting with 1 !!!
      */
      std::vector<int> Cvec;
      /**
      * @brief the current column-number of the input (the next to read)
      */
      int curCol;
      /**
      * @brief the current index in the vector Cvec; (the next to check)
      */
      int curVecIdx;
      
    /**
    * @brief set the input stream to a new stream
    */
    void resetStream(ReadStream* iReaderV, bool errEOFv, tuint blocksizeV);
    
  public:
    FlxIstream_file(std::string nameV, ReadStream* iReaderV, bool errEOFv, tuint blocksizeV, const int Cnumb, const std::vector<int> &Cvec);
    ~FlxIstream_file();
   
    void copyStream(FlxIstream* rhsB, bool errSerious=true);
    bool get_value(tdouble& v, bool suppressErr = false);
};


class FlxIstream_vector : public FlxIstream {
  private:
    std::vector<tdouble> iVec;
    /**
    * @brief current index in iVec
    */
    tnlong index;
    /**
    * @brief number of entries in iVec
    */
    tnlong numbEl;
    
  public:
    FlxIstream_vector(std::string nameV, FlxIstream* is, bool errEOFv, const tnlong Nreserve);
    
    void copyStream(FlxIstream* rhsB, bool errSerious=true);
    bool get_value(tdouble& v, bool suppressErr = false);
    
    /**
    * @brief append a number to the end of the stream
    */
    void appendNumber(const tdouble &d);
    
    /**
    * @brief sorts the vector and resets the stream
    */
    void sortStream();
    /**
    * @brief resets the stream
    */
    void reset_stream() { index = 0; }
    /**
    * @brief sorts the vector and resets the stream
    */
    tnlong get_total_size() { return numbEl; }
    /**
    * @brief returns the first value in the set
    */
    tdouble get_first_value() { return iVec.empty()?0:iVec[0]; }
    /**
    * @brief returns the last value in the set
    */
    tdouble get_last_value() { return iVec.empty()?0:iVec[iVec.size()-1]; }
};




//------------------------------------------------------------ inline functions ------------------------------------




#endif // fesslix_flxio_H
