/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */


#ifndef fesslix_flxmath_H
#define fesslix_flxmath_H

#include <cmath>

#include "flxglobal.h"



/**
* @brief throw away all numbers before the dot
*/
inline const tdouble frac(const tdouble &x) {  // digits after the dot
  if ( x >= 0.0 ) return x - std::floor(x);
  return std::ceil(x) - x;
}
inline const tdouble arccot(const tdouble &x) {
  return PI * 0.5 - std::atan(x);
}
inline const tdouble arctanh(const tdouble &x) {
  return 0.5 * std::log((1.0 + x) / (1.0 - x));
}

inline const tdouble sign(const tdouble &x) {
  return (x>0)?1.0:((x<0)?-1.0:0.0);
}

template <typename T>
inline T pow_int(T x, const tuint n)
{
  switch(n) {
    case 0:
      return 1;
    case 1:
      return x;
    case 2:
      return x*x;
    case 3:
      return x*x*x;
    case 4:
    {
      T v = x*x;
      return v*v;
    }
    case 5:
    {
      T v = x*x;
      return v*v*x;
    }
    default:
    {
      T v = T(1);
      for (tuint i=0;i<n;++i) v *= x;
      return v;
    }
  };
}

template <typename T>
inline T pow2(T x)
{
  return x*x;
}

template <typename T>
inline T pow_b2(T n)
{
  return 1<<n;
}

/**
* @brief calculate the Logarithm of the Gamma FunctionList
* @brief returns the value ln(GAMMA(xx)] for xx > 0
*/
const tdouble GammaLn(const tdouble &xx);

inline const tdouble BetaFun(const tdouble &z, const tdouble &w) {
  return std::exp(GammaLn(z)+GammaLn(w)-GammaLn(z+w));
}

const tdouble flxgamma(const tdouble x);
const tdouble flxgamma(const tdouble a, const tdouble z);
/**
* @brief regularized incomplete Beta function
*/
const tdouble iBeta_reg(const tdouble alpha, const tdouble beta, const tdouble x);
/**
* @brief inverse of the regularized incomplete Beta function
*/
const tdouble iBeta_reg_inv(const tdouble alpha, const tdouble beta, const tdouble p);

/**
* @brief Factorial Function
*/
const tdouble Factorial(const int n);
const tdouble DoubleFactorial(const int n);
const tdouble FactorialLn(const int n);

/**
* @brief calculate the Binomial Coefficient (n, m)
*/
inline const tdouble BinomCoeff(const int n, const int m)
{
  return std::floor(0.5+std::exp(FactorialLn(n)-FactorialLn(m)-FactorialLn(n-m)));
}


inline const double round_flx(const double r) {
  return (r > 0.0) ? std::floor(r + 0.5) : std::ceil(r - 0.5);
}



#endif 	// fesslix_flxmath_H
