/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxmtxfun_H
#define fesslix_flxmtxfun_H

#include "flxmtxfun_data.h"

//=================== General Definitions ===========================

class FlxMtxFunBox;

class FlxMtxBoxBase {
  protected:
    static FlxConstMtxBox* ConstMtxBox;
    static FlxMtxFunBox* MtxFunBox;
  public:
    virtual ~FlxMtxBoxBase() {};
    static void set_ConstMtxBox(FlxConstMtxBox* ConstMtxBoxV) { ConstMtxBox = ConstMtxBoxV; }
    static void set_MtxFunBox(FlxMtxFunBox* MtxFunBoxV) { MtxFunBox = MtxFunBoxV; }
};

//=================== Calc-Definitions ==========================

class FunMtxBase {
  public:
    virtual ~FunMtxBase() {};
    virtual SMtxBase* calc() = 0;
    virtual const bool search_circref(FlxFunction* fcr) = 0;
    virtual const std::string write() = 0;
    virtual const tuint precedence() = 0;
    virtual void dependOn_RV(bVec& bv) = 0;
    virtual const bool dependOn_Const(const tdouble* const thenumber) = 0;
};

typedef FunMtxBase* FunMtxBasePtr;

class FunMtxDummy : public FunMtxBase {
  public:
    SMtxBase* calc() { return new SMtxScalar(0.0); }
    const bool search_circref(FlxFunction* fcr) {return false;};
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
};

/**
* @brief arithmetic class: stores a matrix with functions in it ... like Number in FlxFunction
*/
class SMtxValue : public FunMtxBase {
  private:
    const tuint nrows;
    const tuint ncols;
    std::vector<FunBase*>* vecV;
  public:
    SMtxValue(std::vector<FunBase*>* vecV, const tuint nrows, const tuint ncols) : nrows(nrows), ncols(ncols), vecV(vecV) {};
    virtual ~SMtxValue ();
    SMtxBase* calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
};

/**
* @brief The base class for all arithmetic classes with two operands.
*/
class FunMtxBaseOperat2 : public FunMtxBase {
  protected:
    FunMtxBase *child_1;
    FunMtxBase *child_2;
  public:
    FunMtxBaseOperat2 (FunMtxBase *child1, FunMtxBase *child2) : child_1(child1), child_2(child2) {};
    virtual ~FunMtxBaseOperat2 () {
      delete child_1;
      delete child_2;
    }
    const bool search_circref(FlxFunction* fcr) { return child_1->search_circref(fcr) || child_2->search_circref(fcr) ; }
    void dependOn_RV(bVec& bv) { child_1->dependOn_RV(bv); child_2->dependOn_RV(bv); }
    const bool dependOn_Const(const tdouble* const thenumber) { return child_1->dependOn_Const(thenumber) || child_2->dependOn_Const(thenumber); }
};


//=================== Read-Definitions ==========================

/**
* @brief Each arithmetic class needs a reader-class - this is the base class for all other arithmetic read classes.
*/
class FunMtxReadBase : public FlxReaderBase2, public FunReadFunBase_0, public FlxBoxBase {
  protected:
    FunMtxReadBase *Next;
    const int priority;
    FunMtxReadBase* insert ( FunMtxReadBase *TheFunRead );
    static FunMtxReadBase *startLink;
  public: 
    FunMtxReadBase (const int Priority, bool isEndNode = false);
    virtual ~FunMtxReadBase () { delete Next; };
    virtual FunMtxBase* read ( bool errSerious ) = 0;
};

/**
* @brief if there is no suitable arithmetic read class found, this 'read'-function will be used (end of the list of arithmetic read classes)
*/
class FunMtxReadEND : public FunMtxReadBase {
  protected:
    FunMtxBase* read ( bool errSerious );
  public:
    FunMtxReadEND () : FunMtxReadBase(-1, true) {}
};

/**
* @brief head of the list of arithmetic read classes
*/
class FunMtxReadSTART : public FunMtxReadBase {
  public:
    FunMtxReadSTART (  );
    void insert ( FunMtxReadBase *TheFunRead ) { FunMtxReadBase::insert(TheFunRead); }
    FunMtxBase* read ( bool errSerious );
};


/**
* @brief This class builds an arithmetic function out of a stream
*/
class FlxMatrixReader {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    /**
    * @brief A list containing all arithmetic read classes.
    */
    FunMtxReadSTART* FunctionList;
  public:
    /**
    * @brief Create a FlxFunctionReader class.
    *
    * An instance of each arithmetic read class has to be created !!!
    *
    * @param StrmReader The Stream-Reader class
    */
    FlxMatrixReader ( ); 		// FunMtxReadXXXXXX Objects have to be listed there
    virtual ~FlxMatrixReader () {
      delete FunctionList;
    }

    /**
    * @brief builds an arithmetic function (a ReadStream class (see constructor) is used for input)
    */
    FunMtxBase* read ( bool errSerious = true);
};

//=================== Dealing with FlxFunctions and matrices as parameter ==========================

class FunBaseFun_Mtx : public FunBaseFun_multPara {
  protected:
    std::list< FunMtxBase* > *ParaList_Mtx;
  public:
    FunBaseFun_Mtx (std::vector<FunBase*> *ParaListV, std::list< FunMtxBase* > *ParaList_Mtx) : FunBaseFun_multPara(ParaListV), ParaList_Mtx(ParaList_Mtx) {};
    FunBaseFun_Mtx (std::list< FunMtxBase* > *ParaList_Mtx) : FunBaseFun_multPara(new std::vector<FunBase*>), ParaList_Mtx(ParaList_Mtx) {};
    virtual ~FunBaseFun_Mtx ();
    virtual const bool search_circref(FlxFunction* fcr);
    const std::string write();
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunBaseFun_MtxConst : public FunBaseFun_multPara {
  protected:
    static FlxConstantBox* scalarConsts;
    static FlxVarBox* varConsts;
    static FlxConstMtxBox* mtxConsts;
    std::list< FlxString* >* ParaList_MtxName;
  public:
    FunBaseFun_MtxConst (std::vector<FunBase*> *ParaListV, std::list< FlxString* >* ParaList_MtxName) : FunBaseFun_multPara(ParaListV), ParaList_MtxName(ParaList_MtxName) {};
    FunBaseFun_MtxConst (std::list< FlxString* >* ParaList_MtxName) : FunBaseFun_multPara(new std::vector<FunBase*>), ParaList_MtxName(ParaList_MtxName) {};
    virtual ~FunBaseFun_MtxConst();
    virtual const bool search_circref(FlxFunction* fcr);
    const std::string write();
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; };
    
    static void set_dataMtxConsts ( FlxConstMtxBox* mtxConstsV ) { mtxConsts = mtxConstsV; }
    static void set_dataScalarConsts ( FlxConstantBox* scalarConstsV ) { scalarConsts=scalarConstsV; }
    static void set_dataVarConsts ( FlxVarBox* varConstsV ) { varConsts=varConstsV; }
};

class FunReadFunBase_Mtx : public FunReadFunBase {
  protected:
    static FunMtxReadSTART* MtxFunList;
    std::list< FunMtxBase* >* read_para( tuint Pnmb, bool errSerious);
  public:
    virtual ~FunReadFunBase_Mtx() {};
    static void set_dataMtx ( FunMtxReadSTART* MtxFunListV) { MtxFunList = MtxFunListV; }
};

class FunReadFunBase_MtxConst : public FunReadFunBase {
  protected:
    std::list< FlxString* >* read_para( tuint Pnmb, bool errSerious);
  public:
    virtual ~FunReadFunBase_MtxConst() {};
};

//=================== Dealing with Matrix-Functions ==========================


class FunReadMtxFunBase : public FlxReaderBase2, public FunReadFunBase_0, public FlxBoxBase {
  protected:
    static FunMtxReadSTART* MtxFunList;
  public:
    virtual FunMtxBase* read ( bool errSerious ) = 0;
    virtual ~FunReadMtxFunBase() {};
    static void set_dataMtx ( FunMtxReadSTART* MtxFunListV) { MtxFunList = MtxFunListV; FunReadFunBase_Mtx::set_dataMtx(MtxFunListV); }
};

class FlxMtxFunBox : public FlxBoxBase {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, FunReadMtxFunBase*> box;
  public:
    FlxMtxFunBox ();
    ~FlxMtxFunBox ();
    void insert ( std::string name, FunReadMtxFunBase* fR);
    FunReadMtxFunBase* get ( std::string name);
};

class FlxReaderBase3 : public FlxReaderBase2 {
  protected:
    static FlxMatrixReader *mtxReader;
  public:
    virtual ~FlxReaderBase3() {};
};

//=================== FlxMtxFun-Definition ==========================

/**
* @brief Using (not developing) this library, this is the class you will have to work with.
*
* FlxMtxFun *fun1 = new FlxMtxFun(mtxReader);
* 
*/
class FlxMtxFun {
  private:
    /**
    * @brief the arithmetic function
    */
    FunMtxBase *fun;
    /**
    * @brief instances of this function (used to copy this function)
    */
    tuint* instances;
  public:
    FlxMtxFun (FlxMatrixReader *mtxReader, bool errSerious=true) : fun(mtxReader->read(errSerious)),instances(new tuint(0)) {}
    FlxMtxFun (FunMtxBase* funV) : fun(funV),instances(new tuint(0)) {}
    FlxMtxFun (const FlxMtxFun& FlxF);
    FlxMtxFun& operator=( const FlxMtxFun& FlxF);
    ~FlxMtxFun ();
    /**
    * @brief calculates the arithmetic function
    */
    SMtxBase* calc() { return fun->calc(); }
    /**
    * @brief calculates the arithmetic function - and returns an error if result isn't a vector
    */
    const tVec calcVec();
    const iVec cast2tuintNo0();
    /**
    * @brief set arithmetic expression of funA to this class; and deletes funA;
    */
    void assign ( FlxMtxFun *funA );
    /**
    * @brief returns a string of the function
    */
    std::string write() { return fun->write() ; }
};







#endif // fesslix_flxmtxfun_H
