/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxmtxfun_data.h"
#include "flxmtxfun.h"

using namespace std;

#if FLX_DEBUG
  int FlxConstMtxBox::Cinst = 0;
#endif

void SMtxBase::check_1(const SMtxBase& rhs) const
{
  if (rhs.get_ncols() != get_ncols() || rhs.get_nrows() != get_nrows() ) {
    std::ostringstream ssV;
    ssV << "Rows and columns of both matrices do not match.";
    throw FlxException("SMtxBase::check_1", ssV.str() );
  }
}

SMtxBase* SMtxBase::Transpose()
{
  throw FlxException_NotImplemented("SMtxBase::Transpose");
}

ostream& operator<<(ostream& os, const SMtxBase& M)
{
  os << "{";
  tuint n = M.get_nrows();
  tuint m = M.get_ncols();
  tuint i,j;
  for (i = 0; i < n; ++i) {
    for (j = 0; j < m; ++j) {
      os << " " << GlobalVar.Double2String( M(i,j) );
      if (j+1<m) os << ',';
    }
    if ( i+1 < n ) os << ';' << std::endl;
  }
  os << " }";
  return os;
}

SMtxScalar& SMtxScalar::operator+=(const SMtxBase& rhs)
{
  check_1(rhs);
  val+=rhs(0,0);
  return *this;
}

SMtxScalar& SMtxScalar::operator*=(const tdouble& rhs)
{
  val*=rhs;
  return *this;
}

const tVec SMtxScalar::get_tVec() const
{
  tVec t(1);
  t[0] = val;
  return t;
}


FlxMtx_base* SMtxVector::get_Matrix()
{
  tuint n = get_nrows();
  tuint m = get_ncols();
  FlxMtx* mtxT = new FlxMtx(n, m);
  tuint i,j;
  for (i = 0; i < n; ++i) {
    for (j = 0; j < m; ++j) {
      (*mtxT)(i,j) = (*val)[i+j];
    }
  }
  return mtxT;
}

SMtxVector& SMtxVector::operator+=(const SMtxBase& rhs)
{
  check_1(rhs);
  const SMtxVector* t = dynamic_cast<const SMtxVector*> (&rhs);
  if (t!=NULL) {
    *val+=*(t->val);
  }
  return *this;
}

SMtxVector& SMtxVector::operator*=(const tdouble& rhs)
{
  *val*=rhs;
  return *this;
}

const tVec SMtxVector::get_tVec() const
{
  return *val;
}


SMtxMatrix& SMtxMatrix::operator+=(const SMtxBase& rhs)
{
  check_1(rhs);
  const SMtxMatrix* t = dynamic_cast<const SMtxMatrix*> (&rhs);
  if (t!=NULL) {
    const FlxMtx* u = dynamic_cast<const FlxMtx*> (t->val);
    FlxMtx* v = dynamic_cast<FlxMtx*> (val);
    if (u!=NULL && v !=NULL) {
      *v+=*u;
    } else {
      throw FlxException_NotImplemented("SMtxMatrix::operator+=");
    }
  }
  return *this;
}

SMtxMatrix& SMtxMatrix::operator*=(const tdouble& rhs)
{
  *val*=rhs;
  return *this;
}

const tVec SMtxMatrix::get_tVec() const
{
  const tuint r = get_nrows();
  const tuint c = get_ncols();
  tVec t(r*c);
  for (tuint i=0;i<r;++i) {
    for (tuint j=0;j<c;++j) {
      t[i*c+j] = operator()(i,j);
    }
  }
  return t;
}


FlxConstMtxBox::FlxConstMtxBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxConstMtxBox' created ...";
      throw FlxException("FlxConstMtxBox::FlxConstMtxBox", ssV.str() );
    }
  #endif
  FlxMtxBoxBase::set_ConstMtxBox(this);
  FunBaseFun_MtxConst::set_dataMtxConsts(this);
}

FlxConstMtxBox::~FlxConstMtxBox()
{
  for (map<string, SMtxBase*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxConstMtxBox::insert(string name, SMtxBase* value)
{
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pair<std::string, SMtxBase*>Element(name, value);
  if ( ! box.insert(Element).second ) {  // ... and if the entry already exists ...
    // find it 
      map<std::string, SMtxBase*>::iterator pos;
      pos = box.find(name);
      if ( !(pos != box.end()) ) {
	return; // that's not okay ... but should never happen
      }
    // delete the current value
      delete pos->second;
    // insert the new entry
      pos->second = value;    
  }
}

SMtxBase* FlxConstMtxBox::get(string name, const bool err_if_unknown)
{
  map<std::string, SMtxBase*>::iterator pos;
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    if (err_if_unknown) {
      std::ostringstream ssV;
      ssV << "Matrix-constant '" << name << "' has not yet been defined.";
      throw FlxException("FlxConstMtxBox::get", ssV.str() );
    } else {
      return NULL;
    }
  }
}

void FlxConstMtxBox::declareC(const string name)
{
  if ( get(name) == NULL ) {
    insert(name, new SMtxScalar(0.0));
  }
}

void FlxConstMtxBox::freeC(const string name)
{
  delete get(name,true);
  box.erase(name);
}



