/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxmtxfun_data_H
#define fesslix_flxmtxfun_data_H

#include "flxstring.h"

class SMtxBase {
  protected:
    /**
    * @brief check if rows and columns of matrices match.
    */
    void check_1(const SMtxBase& rhs) const;
  public:
    virtual ~SMtxBase () {};
    virtual SMtxBase* copy() = 0;
    virtual const tuint get_nrows() const = 0;
    virtual const tuint get_ncols() const = 0;
    /**
    * @brief get coefficient (i,j)  -  has to be within range of matrix (not checked!!!)
    */
    virtual const tdouble operator()(const tuint& i, const tuint& j) const = 0;
    /**
    * @brief change a coefficient (i,j)  -  has to be within range of matrix (not checked!!!)
    */
    virtual void insert(const tuint& i, const tuint& j, const tdouble& d) = 0;
    virtual SMtxBase& operator+=(const SMtxBase& rhs) = 0;
    virtual SMtxBase& operator*=(const tdouble& rhs) = 0;
    virtual SMtxBase* Transpose();
    virtual const tdouble max() const = 0;
    virtual const tdouble min() const = 0;
    virtual const tVec get_tVec() const = 0;
};

std::ostream& operator<<(std::ostream& os, const SMtxBase& M);

class SMtxScalar : public SMtxBase {
  private:
    tdouble val;
  public:
    SMtxScalar(tdouble value) : val(value) {}
    virtual ~SMtxScalar () {};
    SMtxScalar* copy() { return new SMtxScalar(*this); }
    const tuint get_nrows() const { return 1; }
    const tuint get_ncols() const { return 1; }
    const tdouble operator()(const tuint& i, const tuint& j) const { return val;}
    void insert(const tuint& i, const tuint& j, const tdouble& d) { val = d; }
    SMtxScalar& operator+=(const SMtxBase& rhs);
    SMtxScalar& operator*=(const tdouble& rhs);
    SMtxBase* Transpose() { return this; }
    const tdouble max() const { return val; }
    const tdouble min() const { return val; }
    const tVec get_tVec() const;
};

class SMtxVector : public SMtxBase {
  private:
    tVec* val;
    tuint nrows;
    bool transpose;
  public:
    SMtxVector(const tVec& vt) : val(new tVec(vt)), nrows(vt.size()), transpose(false) {}
    SMtxVector(tVec* vec, tuint nrows, bool transposeV = false) : val(vec), nrows(nrows), transpose(transposeV) {};
    SMtxVector(const SMtxVector& ref) : val(new tVec(*(ref.val))), nrows(ref.nrows), transpose(ref.transpose) {}
    virtual ~SMtxVector () {delete val;}
    SMtxVector* copy() { return new SMtxVector(*this); }
    tVec& get_Vector() { return *val;}
    FlxMtx_base* get_Matrix();
    const tuint get_nrows() const { return (transpose?1:nrows); }
    const tuint get_ncols() const { return (transpose?nrows:1); }  
    const tdouble operator()(const tuint& i, const tuint& j) const { return (*val)[transpose?j:i];}
    void insert(const tuint& i, const tuint& j, const tdouble& d) { (*val)[transpose?j:i] = d; }
    SMtxVector& operator+=(const SMtxBase& rhs);
    SMtxVector& operator*=(const tdouble& rhs);
    SMtxBase* Transpose() { transpose = !transpose; return this; }
    const tdouble max() const { return val->max(); }
    const tdouble min() const { return val->min(); }
    const tVec get_tVec() const;
};

class SMtxMatrix : public SMtxBase {
  private:
    FlxMtx_base* val;
  public:
    SMtxMatrix(FlxMtx_base* mtx) : val(mtx) {};
    SMtxMatrix(const SMtxMatrix& ref) : val(ref.val->copy()) {};
    virtual ~SMtxMatrix () {delete val;}
    SMtxMatrix* copy() { return new SMtxMatrix(*this); }
    FlxMtx_base& get_Matrix() { return *val; }
    const tuint get_nrows() const { return val->nrows(); }
    const tuint get_ncols() const { return val->ncols(); }  
    const tdouble operator()(const tuint& i, const tuint& j) const { return val->operator()(i,j);}
    void insert(const tuint& i, const tuint& j, const tdouble& d) { val->add_value(i, j, -1.0*val->operator()(i,j)); val->add_value(i, j, d); }
    SMtxMatrix& operator+=(const SMtxBase& rhs);
    SMtxMatrix& operator*=(const tdouble& rhs);
    const tdouble max() const { return val->max(); }
    const tdouble min() const { return val->min(); }
    const tVec get_tVec() const;
};


//=================== Constant-Box ==========================

/**
* @brief A class for storing constant values (the result of the evaluated function is stored).
*/
class FlxConstMtxBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, SMtxBase*> box;
  public:
    FlxConstMtxBox();
    ~FlxConstMtxBox ();
    /**
    * @brief insert: the reference of a variable is changed by 'insert' (difference to 'FlxConstantBox')
    */
    void insert ( std::string name, SMtxBase* value);
    /**
    * @brief get: NOTE: you have to COPY the returned reference
    */
    SMtxBase* get ( std::string name, const bool err_if_unknown = false );
    void declareC ( const std::string name );
    /**
    * @brief removes matrix 'name' from the list
    */
    void freeC ( const std::string name );
};




#endif // fesslix_flxmtxfun_data_H
