/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxmtxfun_fun_read.h"
#include "flxmtxfun_fun.h"

using namespace std;

void flxmtxfun_fun_insert(FlxMtxFunBox& MtxFunBox, FlxFunctionBox& FunBox)
{
  MtxFunBox.insert("unimtx", new FunReadMtxFunNewMtx() );
  MtxFunBox.insert("pow", new FunReadMtxFunPow() );
  MtxFunBox.insert("transpose", new FunReadMtxFunTranspose() );
  MtxFunBox.insert("filter", new FunReadMtxFunFilter() );
  MtxFunBox.insert("seq", new FunReadMtxFunSeq() );
  MtxFunBox.insert("mulmtxcoeff", new FunReadMultVecCoeff() );
  
  FunBox.insert("max", new FunReadFunMax() );
  FunBox.insert("min", new FunReadFunMin() );
  FunBox.insert("mtxcoeff", new FunReadFunMtxCoeff() );
  FunBox.insert("mtxrows", new FunReadFunMtxRows() );
  FunBox.insert("mtxcols", new FunReadFunMtxCols() );
  FunBox.insert("mtxsum", new FunReadFunMtxSum() );
}

FunMtxBase* FunReadMtxFunNewMtx::read(bool errSerious)
{
  FunBase* rows=NULL; FunBase* cols=NULL; FunBase* uval=NULL;
  try {
    rows = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    cols = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    uval = FunctionList->read(errSerious);
  } catch (FlxException &e) {
    if (rows !=NULL) { delete rows; }
    if (cols !=NULL) { delete cols; }
    if (uval !=NULL) { delete uval; }
    throw;
  }
  return new FunNewMtx(rows,cols,uval);
}

FunMtxBase* FunReadMtxFunPow::read(bool errSerious)
{
  FunMtxBase* child_1=NULL; FunBase* child_2=NULL;
  try {
    child_1 = MtxFunList->read(errSerious);
    reader->getChar(',',errSerious);
    child_2 = FunctionList->read(errSerious);
  } catch (FlxException &e) {
    if (child_1) { delete child_1; }
    if (child_2) { delete child_2; }
    throw;
  }
  return new FunMtxPow(child_1,child_2);
}

FunMtxBase* FunReadMtxFunTranspose::read(bool errSerious)
{
  FunMtxBase* mtx = MtxFunList->read(errSerious);
  return new FunTranspose(mtx);
}

FunMtxBase* FunReadMultVecCoeff::read(bool errSerious)
{
  FunMtxBase* mtx1 = MtxFunList->read(errSerious);
  reader->getChar(',');
  FunMtxBase* mtx2 = MtxFunList->read(errSerious);
  return new FunMultVecCoeff(mtx1,mtx2);
}

FunMtxBase* FunReadMtxFunSeq::read(bool errSerious)
{
  if (reader->getNextType() != ReadStream::STRING) {
    ostringstream ssV_2;
    ssV_2 << "Name of the 'seq' variable to use expected.";
    FlxError(errSerious,"FunReadMtxFunFilter_seq::read_1", ssV_2.str(), reader->getCurrentPos());
  }
  tdouble* cv = NULL; FunBase* startF=NULL; FunBase* funCond=NULL; FunBase* funConst=NULL;
  try {
    cv = ConstantBox->get(reader->getWord(true,errSerious),true);
    reader->getChar('=',errSerious);
    startF = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    funCond = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    funConst = FunctionList->read(errSerious);
  } catch (FlxException &e) {
    if (cv!=NULL) delete cv;
    if (startF!=NULL) delete startF;
    if (funCond!=NULL) delete funCond;
    if (funConst!=NULL) delete funConst;
    throw;
  }
  return new FunMtxSeq(cv,startF,funCond,funConst);
}

FunMtxBase* FunReadMtxFunFilter::read(bool errSerious)
{
  if (reader->getNextType() != ReadStream::STRING) {
    ostringstream ssV_2;
    ssV_2 << "Name of the 'const' variable to use expected.";
    FlxError(errSerious,"FunReadMtxFunFilter::read_1", ssV_2.str(), reader->getCurrentPos());
  }
  tdouble* cv = NULL; FunMtxBase* seqMtx = NULL; FunBase* fltfun = NULL;
  try {
    cv = ConstantBox->get(reader->getWord(true),true);
    reader->getChar(',',errSerious);
    seqMtx = MtxFunList->read(errSerious);
    reader->getChar(',',errSerious);
    fltfun = FunctionList->read(errSerious);
  } catch (FlxException &e) {
    if (cv!=NULL) delete cv;
    if (seqMtx!=NULL) delete seqMtx;
    if (fltfun!=NULL) delete fltfun;
    throw;
  }
  return new FunMtxFilter(cv,seqMtx,fltfun);
}

const std::string FunMtxFilter::write() {
  return "filter(" + ConstantBox->get(cv) + "," + mtx->write() + "," + fltfun->write() + ")";
}

SMtxBase* FunMtxFilter::calc()
{
  SMtxBase* tmtx = mtx->calc();
  const tuint nr = tmtx->get_nrows();
  const tuint nc = tmtx->get_ncols();
  tuint i,j;
  for ( i = 0; i < nr; ++i) {
    for (j = 0; j < nc; ++j) {
      *cv = tmtx->operator()(i,j);
      tmtx->insert(i,j,fltfun->calc());
    }
  }
  return tmtx;
}

const bool FunMtxFilter::dependOn_Const(const tdouble* const thenumber) {
  if (mtx->dependOn_Const(thenumber)) return true;
  if (cv==thenumber) return false;
  return fltfun->dependOn_Const(thenumber);
}

const std::string FunTranspose::write() {
  return "transpose(" + mtx->write() + ")";
}

const std::string FunNewMtx::write() {
  return "unimtx(" + rows->write() + "," + cols->write() + "," + uval->write() + ")";
}

void FunNewMtx::dependOn_RV(bVec& bv) {
  rows->dependOn_RV(bv); cols->dependOn_RV(bv); uval->dependOn_RV(bv);
}

const bool FunNewMtx::dependOn_Const(const tdouble* const thenumber) {
  if (rows->dependOn_Const(thenumber)) return true;
  if (cols->dependOn_Const(thenumber)) return true;
  if (uval->dependOn_Const(thenumber)) return true;
  return false;
}

const bool FunNewMtx::search_circref(FlxFunction* fcr)
{
  return rows->search_circref(fcr) || cols->search_circref(fcr) || uval->search_circref(fcr);
}

SMtxBase* FunNewMtx::calc()
{
  const tuint r = tuint_from(rows->calc(),"Number of rows",true,false);
  const tuint c = tuint_from(cols->calc(),"Number of columns",true,false);
  // uniform value
    const tdouble v = uval->calc();
    
  if (r == 1 && c == 1) {  	// Scalar
    return new SMtxScalar(v);
  } else if (c == 1) {		// Vector
    tVec *vec = new tVec(r);
    *vec = v;
    return new SMtxVector(vec,r,false);
  } else if (r == 1) {		// Vector (Transposed)
    tVec *vec = new tVec(c);
    *vec = v;
    return new SMtxVector(vec,c,true);
  } else {			// Matrix
    FlxMtx* mtx = new FlxMtx(r,c);
    if (v==0.0) {
      mtx->set_zeroMtx();
    } else {
      tuint i,j;
      for (i = 0; i < r; ++i) {
	for (j = 0; j < c; ++j) {
	  mtx->add_value(i,j,v);
	}
      }
    }
    return new SMtxMatrix(mtx);
  }
}

SMtxBase* FunMtxPow::calc()
{
  SMtxBase* v1 = child_1->calc();
  const tuint potv = tuint_from(child_2->calc(),"expected integer value for power of a matrix");
  const tuint nrows = v1->get_nrows();
  const tuint ncols = v1->get_ncols();
  if (nrows!=ncols) {
    std::ostringstream ssV;
    ssV << "Matrix must be a square matrix!";
    throw FlxException("FunMtxPow::calc", ssV.str() );
  }
  if (ncols == 1 && nrows == 1 ) {
    v1->insert(0,0,pow(v1->operator()(0,0),int(potv)));
    return v1;
  } else {
    if (potv==0) {
      for (tuint i=0;i<nrows;++i) {
	for (tuint j=0;j<ncols;++j) {
	  if (i==j) v1->insert(i,j,1.0);
	  else v1->insert(i,j,0.0);
	}
      }
      return v1;
    } else if (potv==1) {
      return v1;
    } else {
      SMtxMatrix* m1 = dynamic_cast<SMtxMatrix*> (v1);
      const FlxMtx_base& mtx1 = m1->get_Matrix();
      FlxMtx m2 = mtx1*mtx1;
      for (tuint i=2;i<potv;++i) {
	m2 = m2*mtx1;
      }
      delete v1;
      FlxMtx_base* m2c = m2.copy();
      return new SMtxMatrix(m2c);
    }
  } 
}

const std::string FunMtxPow::write()
{ 
  return "pow(" + child_1->write() + "," + child_2->write() + ")";
}

const std::string FunMultVecCoeff::write()
{
  return "multveccoeff(" + mtx1->write() + "," + mtx2->write() + ")";
}

SMtxBase* FunMultVecCoeff::calc()
{
  SMtxBase* sm1 = mtx1->calc();
  SMtxBase* sm2 = mtx2->calc();
  const tuint R=sm1->get_nrows();
  const tuint C=sm1->get_ncols();
  if (C==sm2->get_ncols() && R == sm2->get_nrows() ) {
    SMtxBase* sr;
    if (C==1 && R==1) {
      sr = new SMtxScalar(0.0);
    } else if (C == 1) {
      tVec *t = new tVec(R);
      sr = new SMtxVector(t,R,false);
    } else if (R == 1) {
      tVec *t = new tVec(C);
      sr = new SMtxVector(t,C,true);
    } else {
      sr = new SMtxMatrix(new FlxMtx(R,C));
    }
    for (tuint i=0;i<R;++i) {
      for (tuint j=0;j<C;++j) {
	sr->insert(i,j,sm1->operator()(i,j)*sm2->operator()(i,j));
      }
    }
    delete sm1;
    delete sm2;
    return sr;
  } else {
    delete sm1;
    delete sm2;
    std::ostringstream ssV;
    ssV << "Size of matrices does not match.";
    throw FlxException("FunMultVecCoeff::calc", ssV.str() );
  }
}

const std::string FunMtxSeq::write()
{
  return "seq(" + ConstantBox->get(cv) + "=" + startF->write() + "," + funCond->write() + "," + funConst->write() + ")";
}

SMtxBase* FunMtxSeq::calc()
{
  *cv = startF->calc();
  stack<tdouble,list<tdouble> > q;
  while ( funCond->calc() > 0.0 ) {
    q.push(*cv);
    *cv = funConst->calc();
  }
  tVec* t = new tVec(q.size());
  for (tnlong i = q.size(); i > 0; --i) {
    (*t)[i-1] = q.top();
    q.pop();
  }
  return new SMtxVector(t, t->size());
}

void FunMtxSeq::dependOn_RV(bVec& bv) 
{
  startF->dependOn_RV(bv);
  funCond->dependOn_RV(bv);
  funConst->dependOn_RV(bv);
}

const bool FunMtxSeq::dependOn_Const(const tdouble* const thenumber) {
  if (startF->dependOn_Const(thenumber)) return true;
  if (cv==thenumber) return false;
  if (funCond->dependOn_Const(thenumber)) return true;
  return funConst->dependOn_Const(thenumber);
}

const bool FunMtxSeq::search_circref(FlxFunction* fcr)
{
  return startF->search_circref(fcr) || funCond->search_circref(fcr) || funConst->search_circref(fcr);
}

const tdouble FunMax::calc()
{
  tdouble cv = 0.0;
  bool b1 = true;
  for (list< FlxString* >::const_iterator i = ParaList_MtxName->begin(); i!=ParaList_MtxName->end(); ++i) {
    const std::string mtxname = (*i)->eval_word(true);
    // optian the local max-value
      SMtxBase* mtx = mtxConsts->get(mtxname,false);
      tdouble t;
      if (mtx==NULL) {
	tdouble* tp = scalarConsts->get(mtxname,false);
	if (tp==NULL) {
	  FlxFunction* fp = varConsts->get(mtxname);
	  if (fp == NULL) {
	    std::ostringstream ssV;
	    ssV << "A variable (const,var,mtxconst) with the name '" << mtxname << "' does not exist.";
	    throw FlxException_NeglectInInteractive("FunMax::calc", ssV.str() );
	  } else {
	    t = fp->calc();
	  }
	} else {
	  t = *tp;
	}
      } else {
	t = mtx->max();
      }
    if (b1) {
      cv = t;
      b1 = false;
    } else {
      if (t > cv) cv = t;
    }
  } 
  return cv;
}

const tdouble FunMin::calc()
{
  tdouble cv = 0.0;
  bool b1 = true;
  for (list< FlxString* >::const_iterator i = ParaList_MtxName->begin(); i!=ParaList_MtxName->end(); ++i) {
    const std::string mtxname = (*i)->eval_word(true);
    // optian the local min-value
      SMtxBase* mtx = mtxConsts->get(mtxname,false);
      tdouble t;
      if (mtx==NULL) {
	tdouble* tp = scalarConsts->get(mtxname,false);
	if (tp==NULL) {
	  FlxFunction* fp = varConsts->get(mtxname);
	  if (fp == NULL) {
	    std::ostringstream ssV;
	    ssV << "A variable (const,var,mtxconst) with the name '" << mtxname << "' does not exist.";
	    throw FlxException_NeglectInInteractive("FunMin::calc", ssV.str() );
	  } else {
	    t = fp->calc();
	  }
	} else {
	  t = *tp;
	}
      } else {
	t = mtx->min();
      }
    if (b1) {
      cv = t;
      b1 = false;
    } else {
      if (t < cv) cv = t;
    }
  } 
  return cv;
}

const tdouble FunMtxRows::calc()
{
  const std::string mtxname = (*ParaList_MtxName->begin())->eval_word(true);
  SMtxBase* mtx = mtxConsts->get(mtxname,true);
  return mtx->get_nrows();
}

const tdouble FunMtxCols::calc()
{
  const std::string mtxname = (*ParaList_MtxName->begin())->eval_word(true);
  SMtxBase* mtx = mtxConsts->get(mtxname,true);
  return mtx->get_ncols();
}

const tdouble FunMtxSum::calc()
{
  const std::string mtxname = (*ParaList_MtxName->begin())->eval_word(true);
  SMtxBase* mtx = mtxConsts->get(mtxname,true);
  const tVec tv = mtx->get_tVec();
  const tdouble* tvp = &tv[0];
  tdouble rv = 0.0;
  const tuint nrows = mtx->get_nrows();
  const tuint ncols = mtx->get_ncols();
  tuint c = 0;
  for (tuint i=0;i<nrows;++i) {
    for (tuint j=0;j<ncols;++j) {
      rv += tvp[c];
      ++c;
    }
  }
  return rv;
}

FunBase* FunReadFunMtxCoeff::read ( bool errSerious ) 
{
  std::list< FlxString* >* plst = new std::list< FlxString* >();
  FunBase* rid = NULL;
  FunBase* cid = NULL;
  try {
    plst->push_back(new FlxString(false,false));
    reader->getChar(',');
    rid = FunctionList->read(errSerious);
    reader->getChar(',');
    cid = FunctionList->read(errSerious);
  } catch (FlxException &e) {
    for (list< FlxString* >::const_iterator i = plst->begin(); i!=plst->end(); ++i) {
      delete (*i);
    }
    delete plst;
    if (rid) delete rid;
    if (cid) delete cid;
    throw;
  }
  return new FunMtxCoeff(plst, rid, cid );
}

const tdouble FunMtxCoeff::calc()
{
  const std::string mtxname = (*ParaList_MtxName->begin())->eval_word(true);
  SMtxBase* mtx = mtxConsts->get(mtxname,true);
  const tuint i = tuint_from(rid->calc(),"row index",true,true);
  const tuint j = tuint_from(cid->calc(),"column index",true,true);
  if (i>=mtx->get_nrows() || j>=mtx->get_ncols()) {
    std::ostringstream ssV;
    ssV << "Index not within valid bounds: (" << i << "," << j << ") [" << mtx->get_nrows() << "," << mtx->get_ncols() << "]";
    throw FlxException("FunMtxCoeff::calc", ssV.str() );
  }
  return mtx->operator()(i,j);
}

const bool FunMtxCoeff::search_circref(FlxFunction* fcr) {
  if (rid->search_circref(fcr)) return true;
  if (cid->search_circref(fcr)) return true;
  return FunBaseFun_MtxConst::search_circref(fcr);
}

const std::string FunMtxCoeff::write() {
  return write_v() + "(" + (*ParaList_MtxName->begin())->write() + "," + rid->write() + "," + cid->write() + ")";
}

void FunMtxCoeff::dependOn_RV(bVec& bv) {
  rid->dependOn_RV(bv);
  cid->dependOn_RV(bv);
  FunBaseFun_MtxConst::dependOn_RV(bv);
}

const bool FunMtxCoeff::dependOn_Const(const tdouble* const thenumber) {
  if (rid->dependOn_Const(thenumber)) return true;
  if (cid->dependOn_Const(thenumber)) return true;
  return FunBaseFun_MtxConst::dependOn_Const(thenumber);
}
