/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxmtxfun_fun_calc_H
#define fesslix_flxmtxfun_fun_calc_H

#include "flxmtxfun.h"


/**
* @brief arithmetic class: creates a matrix where all coefficients are assigned with the same value
*/
class FunNewMtx : public FunMtxBase {
  private:
    FunBase* rows;
    FunBase* cols;
    FunBase* uval;
  public:
    FunNewMtx (FunBase* rows, FunBase* cols, FunBase* uval) : rows(rows), cols(cols), uval(uval) {};
    ~FunNewMtx() { delete rows; delete cols; delete uval;}
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
};

class FunMtxPow : public FunMtxBase {
  private:
    FunMtxBase* child_1;
    FunBase* child_2;
  public:
    FunMtxPow (FunMtxBase* child_1, FunBase* child_2) : child_1(child_1), child_2(child_2) {};
    ~FunMtxPow() { delete child_1; delete child_2; }
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr) { return child_1->search_circref(fcr) || child_2->search_circref(fcr) ; }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { child_1->dependOn_RV(bv); child_2->dependOn_RV(bv); }
    const bool dependOn_Const(const tdouble* const thenumber) { return child_1->dependOn_Const(thenumber) || child_2->dependOn_Const(thenumber); }
};

/**
* @brief arithmetic class: multiply (divide) two values
*/
class FunTranspose : public FunMtxBase {
  private:
    FunMtxBase* mtx;
  public:
    FunTranspose (FunMtxBase* mtx) : mtx(mtx) {};
    ~FunTranspose() { delete mtx;}
    SMtxBase* calc () { return (mtx->calc())->Transpose(); };
    const bool search_circref(FlxFunction* fcr) { return mtx->search_circref(fcr); }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { mtx->dependOn_RV(bv); }
    const bool dependOn_Const(const tdouble* const thenumber) { return mtx->dependOn_Const(thenumber); }
};

/**
* @brief arithmetic class: Filter
*/
class FunMtxFilter : public FunMtxBase, public FlxBoxBase {
  private:
    tdouble* cv;
    FunMtxBase* mtx;
    FunBase* fltfun;
  public:
    FunMtxFilter (tdouble* cv, FunMtxBase* mtx, FunBase* fltfun) : cv(cv), mtx(mtx), fltfun(fltfun) {};
    ~FunMtxFilter() { delete mtx; delete fltfun; }
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr) { return mtx->search_circref(fcr) || fltfun->search_circref(fcr); }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { mtx->dependOn_RV(bv); fltfun->dependOn_RV(bv); } 
   const bool dependOn_Const(const tdouble* const thenumber);
};

/**
* @brief arithmetic class: generates a filter-sequence
*/
class FunMtxSeq : public FunMtxBase, public FlxBoxBase {
  private:
    tdouble* cv;
    FunBase* startF;
    FunBase* funCond;
    FunBase* funConst;
  public:
    FunMtxSeq (tdouble* cv, FunBase* startF, FunBase* funCond, FunBase* funConst) : cv(cv), startF(startF), funCond(funCond), funConst(funConst) {};
    ~FunMtxSeq() { delete startF; delete funCond; delete funConst; }
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
};

/**
* @brief arithmetic class: multiply (divide) two values
*/
class FunMultVecCoeff : public FunMtxBase {
  private:
    FunMtxBase* mtx1;
    FunMtxBase* mtx2;
  public:
    FunMultVecCoeff (FunMtxBase* mtx1, FunMtxBase* mtx2) : mtx1(mtx1), mtx2(mtx2) {};
    ~FunMultVecCoeff() { delete mtx1; delete mtx2; };
    SMtxBase* calc();
    const bool search_circref(FlxFunction* fcr) { return mtx1->search_circref(fcr) || mtx2->search_circref(fcr); }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { mtx1->dependOn_RV(bv); mtx2->dependOn_RV(bv); }
    const bool dependOn_Const(const tdouble* const thenumber) { return mtx1->dependOn_Const(thenumber) || mtx2->dependOn_Const(thenumber); }
};

//=================== FlxFunction - Part ==========================

/**
* @brief arithmetic class: maximum coefficient in a matrix
*/
class FunMax : public FunBaseFun_MtxConst {
  public:
    FunMax (std::list< FlxString* > *ParaList_Mtx) : FunBaseFun_MtxConst(ParaList_Mtx) {};
    const tdouble calc();
    const std::string write_v() { return "max"; }
};

/**
* @brief arithmetic class: maximum coefficient in a matrix
*/
class FunMin : public FunBaseFun_MtxConst {
  public:
    FunMin (std::list< FlxString* > *ParaList_Mtx) : FunBaseFun_MtxConst(ParaList_Mtx) {};
    const tdouble calc();
    const std::string write_v() { return "min"; }
};

/**
* @brief arithmetic class: coefficient of a matrix
*/
class FunMtxCoeff : public FunBaseFun_MtxConst {
  private:
    FunBase* rid;
    FunBase* cid;
  public:
    FunMtxCoeff (std::list< FlxString* > *ParaList_Mtx, FunBase* rid, FunBase* cid) : FunBaseFun_MtxConst(ParaList_Mtx), rid(rid), cid(cid) {};
    ~FunMtxCoeff() { delete rid; delete cid;} 
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const std::string write_v() { return "mtxcoeff"; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
};

/**
* @brief arithmetic class: number of rows of a matrix
*/
class FunMtxRows : public FunBaseFun_MtxConst {
  public:
    FunMtxRows (std::list< FlxString* > *ParaList_Mtx) : FunBaseFun_MtxConst(ParaList_Mtx) {};
    const tdouble calc();
    const std::string write_v() { return "mtxrows"; }
};

/**
* @brief arithmetic class: number of columns of a matrix
*/
class FunMtxCols : public FunBaseFun_MtxConst {
  public:
    FunMtxCols (std::list< FlxString* > *ParaList_Mtx) : FunBaseFun_MtxConst(ParaList_Mtx) {};
    const tdouble calc();
    const std::string write_v() { return "mtxcols"; }
};

/**
* @brief arithmetic class: sum over all coefficients of a matrix
*/
class FunMtxSum : public FunBaseFun_MtxConst {
  public:
    FunMtxSum (std::list< FlxString* > *ParaList_Mtx) : FunBaseFun_MtxConst(ParaList_Mtx) {};
    const tdouble calc();
    const std::string write_v() { return "mtxsum"; }
};


#endif // fesslix_flxmtxfun_fun_calc_H

