/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxobjcommon_H
#define fesslix_flxobjcommon_H

#include "flxobjects.h"

class FlxCreateObjReaders_Common : public FlxCreateObjReaders {
  public:
    void createObjReaders (FlxObjectReadBox* objReadBox );
    void createFunReaders (FlxData* dataBox );
};








// -------------------------- OBJECTS ----------------------------------------------------------------------



/**
* @brief object class: define a variable expression
*
* var varName = FlxFunction
*/
class FlxObjVar : public FlxObjBase {
  private:
    std::string cname;
    FlxFunction *fun;
    void task() { data->VarBox.insert(cname, fun); fun=NULL; }
  public:
    FlxObjVar ( bool dolog, std::string cnameV, FlxFunction* funV ) : FlxObjBase(dolog), cname(cnameV), fun(funV) {}
    ~FlxObjVar() { if (fun!=NULL) delete fun; }
};

/**
* @brief object class: define a Matrix-constant
*
* mtxconst constName = FlxMtxFun
*/
class FlxObjMtxConst : public FlxObjBase {
  private:
    const std::string cname;
    FlxMtxFun *fun;
    void task() { data->ConstMtxBox.insert(cname, fun->calc()); }
  public:
    FlxObjMtxConst ( const bool dolog, const std::string cnameV, FlxMtxFun* funV ) : FlxObjBase(dolog), cname(cnameV), fun(funV) {}
    ~FlxObjMtxConst() { delete fun; }
};


/**
* @brief object class: define a Matrix-constant
*
* mtxcoeff constName(i,j) = FlxFunction;
*/
class FlxObjMtxCoeff : public FlxObjBase {
  private:
    std::string cname;
    FlxFunction *i_fun;
    FlxFunction *j_fun;
    FlxFunction *c_fun;
    void task();
  public:
    FlxObjMtxCoeff ( bool dolog, std::string cnameV, FlxFunction *i_fun, FlxFunction *j_fun, FlxFunction *c_fun ) : FlxObjBase(dolog), cname(cnameV), i_fun(i_fun), j_fun(j_fun), c_fun(c_fun) {}
    ~FlxObjMtxCoeff() { delete i_fun; delete j_fun; delete c_fun; }
};

/**
* @brief object class: define a function
*
* fun funName = FlxFunction
*/
class FlxObjFun : public FlxObjBase {
  private:
    std::string cname;
    FunReadFunBase *fun;
    void task();
  public:
    FlxObjFun ( bool dolog, std::string cnameV, FunReadFunBase* funV ) : FlxObjBase(dolog), cname(cnameV), fun(funV) {}
    ~FlxObjFun();
};

/**
* @brief object class: define a function
*/
class FlxObjEnd : public FlxObjBase {
  private:
    void task() { throw FlxEndE(); }
  public:
    FlxObjEnd(bool dolog) : FlxObjBase(dolog) {};
};

/**
* @brief object class: define a function
*/
class FlxObjExit : public FlxObjBase {
  private:
    void task() { throw FlxExitE(); }
  public:
    FlxObjExit(bool dolog) : FlxObjBase(dolog) {};
};

/**
* @brief object class: start interactive mode
*/
class FlxObjInteractive : public FlxObjOutputBase, public FlxEvaluateCmdBase {
  private:
    bool resetInputLog;
    void do_resetInputLog();
    void print_prompt();
    void task();
  public:
    FlxObjInteractive ( bool dolog, std::string ostreamV ) : FlxObjOutputBase(dolog, ostreamV), resetInputLog(false) {}
};


/**
* @brief object class: calculate an arithmetic expression and output the result
*
* calc FlxFunction
*/
class FlxObjCalc : public FlxObjOutputBase {
  private:
    FlxFunction *fun;
    tdouble *ansptr;
    void task() { *ansptr=fun->calc(); sout() << fun->write() << " = " << GlobalVar.Double2String(*ansptr,checkTOL) << std::endl; }
  public:
    FlxObjCalc ( const bool dolog, FlxFunction* funV, std::string ostreamV, const bool checkTOL );
    ~FlxObjCalc() { delete fun; }
};

/**
* @brief object class: 
*
* print_hlegendrepoly(nth)
*/
class FlxObjPrintHLegendrePoly : public FlxObjOutputBase {
  private:
    FlxFunction *fun;
    const bool deriv;
    void task();
  public:
    FlxObjPrintHLegendrePoly ( bool dolog, FlxFunction* funV, std::string ostreamV, bool verboseV, const bool deriv ) : FlxObjOutputBase(dolog,ostreamV,verboseV), fun(funV), deriv(deriv) {}
    ~FlxObjPrintHLegendrePoly() { delete fun; }
};

/**
* @brief object class: 
*
* print_hgegenbauerpoly(nth)
*/
class FlxObjPrintHGegenbauerPoly : public FlxObjOutputBase {
  private:
    FlxFunction *fun;
    void task();
  public:
    FlxObjPrintHGegenbauerPoly ( bool dolog, FlxFunction* funV, std::string ostreamV, bool verboseV ) : FlxObjOutputBase(dolog,ostreamV,verboseV), fun(funV) {}
    ~FlxObjPrintHGegenbauerPoly() { delete fun; }
};

/**
* @brief object class: output a string
*
* echo "text"
*/
class FlxObjEcho : public FlxObjOutputBase {
  private:
    FlxString* strV;
    void task() { strV->eval(sout()); sout() << std::endl; }
  public:
    FlxObjEcho ( bool dolog, FlxString* strV, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), strV(strV) {}
    ~FlxObjEcho() { delete strV; }
};

/**
* @brief object class: output a string
*
* warranty;
*/
class FlxObjWarranty : public FlxObjOutputBase {
  private:
    void task();
  public:
    FlxObjWarranty ( bool dolog, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV) {}
    ~FlxObjWarranty() {}
};

/**
* @brief object class: calculate a matrix expression and output the result
*
* mtxcalc FlxMtxFun
*/
class FlxObjMtxCalc : public FlxObjOutputBase {
  private:
    FlxMtxFun *fun;
    void task();
  public:
    FlxObjMtxCalc ( bool dolog, FlxMtxFun* funV, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), fun(funV) {}
    ~FlxObjMtxCalc() { delete fun; }
};

/**
* @brief object class: calculate a matrix expression and output the result
*
* transform_mtx2octave FlxMtxFun
*/
class FlxObjTransformMtx2Octave: public FlxObjOutputBase {
  private:
    FlxString *funStr;
    void task();
  public:
    FlxObjTransformMtx2Octave ( bool dolog, FlxString *funStr, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), funStr(funStr) {}
    ~FlxObjTransformMtx2Octave() { delete funStr; }
};

/**
* @brief object class: deletes a defined matrix constant
*
* mtxconst_free FlxMtxFun
*/
class FlxObjMtxConst_free: public FlxObjOutputBase {
  private:
    FlxString *funStr;
    void task();
  public:
    FlxObjMtxConst_free ( bool dolog, FlxString *funStr, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), funStr(funStr) {}
    ~FlxObjMtxConst_free() { delete funStr; }
};

/**
* @brief object class: calculate an arithmetic expression and output the result
*
* calc FlxFunction
*/
class FlxObjStreamSet_Prec : public FlxObjOutputBase {
  private:
    FlxFunction *fun;
    void task();
  public:
    FlxObjStreamSet_Prec ( bool dolog, FlxFunction* funV, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), fun(funV) {}
    ~FlxObjStreamSet_Prec() { delete fun; }
};

/**
* @brief object class: creates a new output-stream (write into a file)
*
* filestream streamname filename
*/
class FlxObjFileStream : public FlxObjBase {
  private:
    const std::string streamname;
    FlxString* filename;
    const bool trunc;
    void task();
  public:
    FlxObjFileStream ( bool dolog, std::string streamnameV, FlxString* filename, bool truncV = true ) 
      : FlxObjBase(dolog), streamname(streamnameV), filename(filename), trunc(truncV) {}
    ~FlxObjFileStream() { delete filename; }
};

/**
* @brief object class: input file stream (read numbers from a file)
*/
class FlxObjInputFileStream : public FlxObjBase {
  private:
    const std::string streamname;
    FlxString* filename;
    FlxFunction* bs;
    const bool erreof;
    int Cnumb;
    const std::vector<int> Cvec;
    void task();
  public:
    FlxObjInputFileStream ( bool dolog, std::string streamnameV, FlxString* filename, FlxFunction* bs, int Cnumb, const std::vector<int> Cvec, bool erreofV = false ) 
      : FlxObjBase(dolog), streamname(streamnameV), filename(filename), bs(bs), erreof(erreofV), Cnumb(Cnumb), Cvec(Cvec) {}
    ~FlxObjInputFileStream() { delete filename; delete bs; }
};

/**
* @brief object class: input vector stream (read numbers from a vector)
*/
class FlxObjInputVectorStream : public FlxObjBase {
  private:
    const std::string streamname;
    FlxString* inputStreamName;
    FlxFunction* Nreserve;
    const bool erreof;
    void task();
  public:
    FlxObjInputVectorStream ( bool dolog, std::string streamname, FlxString* inputStreamName, FlxFunction* Nreserve, bool erreofV = false ) 
      : FlxObjBase(dolog), streamname(streamname), inputStreamName(inputStreamName), Nreserve(Nreserve), erreof(erreofV) {}
    ~FlxObjInputVectorStream();
};

/**
* @brief object class: append a number to an input vector stream
*/
class FlxObjivstream_append : public FlxObjBase {
  private:
    FlxIstream_vector &istream;
    FlxFunction* fun;
    void task() { istream.appendNumber(fun->calc()); }
  public:
    FlxObjivstream_append ( bool dolog,FlxIstream_vector &istream, FlxFunction* fun )
      : FlxObjBase(dolog), istream(istream), fun(fun) {}
    ~FlxObjivstream_append() { delete fun; }
};

/**
* @brief object class: writes an input stream to an output stream
*/
class FlxObjistream_write : public FlxObjOutputBase {
  private:
    FlxString* isname;
    void task();
  public:
    FlxObjistream_write ( bool dolog, FlxString* isname, std::string ostreamV )
      : FlxObjOutputBase(dolog,ostreamV), isname(isname) {}
    ~FlxObjistream_write() { delete isname; }
};

/**
* @brief object class: reads input from a file
*
* read "PATHNAME";
*/
class FlxObjReadFile : public FlxObjBase, public FlxEvaluateCmdBase {
  private:
    FlxString* filename;
    void task();
  public:
    FlxObjReadFile ( bool dolog, FlxString* filenameV ) : FlxObjBase(dolog), filename(filenameV) {}
    ~FlxObjReadFile() { delete filename; }
};

/**
* @brief object class: creates a new Distributor-Stream
*
* filestream streamname filename
*/
class FlxObjDistributorStream : public FlxObjBase {
  private:
    const std::string streamname;
    const std::string stream1;
    const std::string stream2;
    void task();
  public:
    FlxObjDistributorStream ( bool dolog, std::string streamnameV, std::string stream1V, std::string stream2V ) 
      : FlxObjBase(dolog), streamname(streamnameV), stream1(stream1V), stream2(stream2V) {}
};

/**
* @brief object class: set default values of optional parameters
*
* default DefSetName = value;
*/
class FlxObjDefault : public FlxObjBase {
  private:
    void* value;
    void task();
    FlxOptionalParaBase* para;
  public:
    FlxObjDefault( bool dolog, void* valueV, FlxOptionalParaBase* paraV) : FlxObjBase(dolog), value(valueV), para(paraV) {};
    ~FlxObjDefault();
};


/**
* @brief object class: control structure - if
*
* if (FlxFunction) { ... } [ else { ... }; ]
*/
class FlxObjIf : public FlxObjBase {
  private:
    FlxFunction* funIf;
    FlxObjBase* InternListThen;
    FlxObjBase* InternListElse;
    void task() {
      if ( funIf->calc() > 0.0 ) { if (InternListThen != NULL) InternListThen->exec(); }
      else { if (InternListElse != NULL) InternListElse->exec(); }
    }
  public:
    FlxObjIf ( bool dolog, FlxFunction* funIfV, FlxObjBase* InternListThenV, FlxObjBase* InternListElseV ) : FlxObjBase(dolog), funIf(funIfV), InternListThen(InternListThenV), InternListElse(InternListElseV) {}
    ~FlxObjIf();
};

/**
* @brief object class: control structure - while-loop
*
* while ( FlxFunction ) { ... };
*/
class FlxObjWhile : public FlxObjBase {
  private:
    unsigned int maxCycles;
    FlxFunction* funWhile;
    FlxObjBase* InternListLoop;
    void task() { 
      if ( maxCycles <= 0 ) {
	while ( funWhile->calc() > 0.0 ) { InternListLoop->exec(); }
      } else {
	unsigned int curCycles = 0;
	while ( funWhile->calc() > 0.0 && maxCycles > curCycles) { InternListLoop->exec(); curCycles++; } 
	if ( funWhile->calc() > 0.0 ) {
	  std::ostringstream ssV;
	  ssV << "While-Loop: maximum number of loop-passes exceeded (" << maxCycles << ").";
	  throw FlxException("FlxObjWhile::task_1", ssV.str(), "This might be an infinite loop ..." );
	}
      }
    }
  public:
    FlxObjWhile ( bool dolog, FlxFunction* funWhileV, FlxObjBase* InternListLoopV, unsigned int MaxCycles=0) 
      : FlxObjBase(dolog), maxCycles(MaxCycles), funWhile(funWhileV), InternListLoop(InternListLoopV) {}
    ~FlxObjWhile();
};

/**
* @brief object class: control structure - for-loop
*
* for ( const-name, flxfunction, flxfunction ) { ... };
*/
class FlxObjFor : public FlxObjBase {
  private:
    unsigned int maxCycles;
    tdouble* theConst;
    FlxFunction* funCond;
    FlxFunction* funConst;
    FlxObjBase* InternListLoop;
    FlxObjConst* ConstDef;
    void task() { 
      ConstDef->exec();
      if ( maxCycles <= 0 ) {
	while ( funCond->calc() > 0.0 ) {
	  InternListLoop->exec();
	  *theConst = funConst->calc();
	}
      } else {
	unsigned int curCycles = 0;
	while ( funCond->calc() > 0.0 && maxCycles > curCycles ) {
	  InternListLoop->exec();
	  *theConst = funConst->calc();
	  curCycles++;
	}
	if ( funCond->calc() > 0.0 ) {
	  std::ostringstream ssV;
	  ssV << "For-Loop: maximum number of loop-passes exceeded (" << maxCycles << ").";
	  throw FlxException("FlxObjFor::task_1", ssV.str(), "This might be an infinite loop ..." );
	}
      }
    }
  public:
    FlxObjFor ( bool dolog, tdouble* theConstV, FlxObjConst* ConstDefV, FlxFunction* funCondV, FlxFunction* funConstV, FlxObjBase* InternListLoopV, unsigned int MaxCycles=0 ) 
      : FlxObjBase(dolog), maxCycles(MaxCycles), theConst(theConstV), funCond(funCondV), funConst(funConstV), InternListLoop(InternListLoopV), ConstDef(ConstDefV) {}
    ~FlxObjFor();
};

/**
* @brief object class: control structure - for-loop
*
* sfor ( const-name; to; start0 ) { ... };
*/
class FlxObjSFor : public FlxObjBase {
  private:
    tdouble* theConst;
    FlxFunction* funTo;
    bool start0;
    FlxObjBase* InternListLoop;
    void task() { 
		int n = funTo->cast2tuintW0(false);
      for (int i = (start0?0:1); i <= n; ++i) {
	*theConst = i;
	InternListLoop->exec();
      }
    }
  public:
    FlxObjSFor ( bool dolog, tdouble* theConst, FlxFunction* funTo, bool start0, FlxObjBase* InternListLoop ) 
      : FlxObjBase(dolog), theConst(theConst), funTo(funTo), start0(start0), InternListLoop(InternListLoop) {}
    ~FlxObjSFor() { delete funTo; delete InternListLoop;}
};

/**
* @brief object class: timer
*
* time { CODE_TO_STOP; };
*/
class FlxObjTime : public FlxObjOutputBase {
  private:
    FlxObjBase* InternListTime;
    void task();
  public:
    FlxObjTime ( bool dolog, FlxObjBase* InternListTime, std::string ostreamV ) : FlxObjOutputBase(dolog, ostreamV), InternListTime(InternListTime) {}
    ~FlxObjTime();
};

/**
* @brief object class: define a timer
*
* timer define TIMER_NAME
*/
class FlxObjTimerDefine : public FlxObjBase {
  private:
    std::string tname;
    void task() { data->TimerBox.insert(tname, new FlxTimer() ); 
      GlobalVar.slog(4) << "timer: timer '" << tname << "' defined." << std::endl; }
  public:
    FlxObjTimerDefine ( bool dolog, std::string tnameV ) : FlxObjBase(dolog), tname(tnameV) {}
};

/**
* @brief object class: start a timer
*
* timer start TIMER_NAME
*/
class FlxObjTimerStart : public FlxObjBase {
  private:
    std::string tname;
    void task() { GlobalVar.slog(4) << "timer: timer '" << tname << "' started. (t=" << GlobalVar.Double2String(tdouble(data->TimerBox.get(tname)->get_ticks())/CLOCKS_PER_SEC) << ")" << std::endl;
      data->TimerBox.get(tname)->start(); }
  public:
    FlxObjTimerStart ( bool dolog, std::string tnameV ) : FlxObjBase(dolog), tname(tnameV) {}
};

/**
* @brief object class: stop a timer
*
* timer stop TIMER_NAME
*/
class FlxObjTimerStop : public FlxObjBase {
  private:
    std::string tname;
    void task() { data->TimerBox.get(tname)->stop();
      GlobalVar.slog(4) << "timer: timer '" << tname << "' stopped. (t=" << GlobalVar.Double2String(tdouble(data->TimerBox.get(tname)->get_ticks())/CLOCKS_PER_SEC) << ")" << std::endl; }
  public:
    FlxObjTimerStop ( bool dolog, std::string tnameV ) : FlxObjBase(dolog), tname(tnameV) {}
};

/**
* @brief object class: delete a timer
*
* timer delete TIMER_NAME
*/
class FlxObjTimerDelete : public FlxObjBase {
  private:
    std::string tname;
    void task() { data->TimerBox.deleteEl(tname); 
      GlobalVar.slog(4) << "timer: timer '" << tname << "' deleted." << std::endl; }
  public:
    FlxObjTimerDelete ( bool dolog, std::string tnameV ) : FlxObjBase(dolog), tname(tnameV) {}
};

/**
* @brief object class: print time of a timer
*
* timer print TIMER_NAME
*/
class FlxObjTimerPrint : public FlxObjOutputBase {
  private:
    std::string tname;
    void task();
  public:
    FlxObjTimerPrint ( bool dolog, std::string tnameV, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV), tname(tnameV)  {}
};

/**
* @brief object class: plot some functions (for GNUplot)
*
* funplot f1,f2,...;
*/
class FlxObjFunPlot : public FlxObjOutputBase {
  protected:
    std::vector<bool> B;
    std::vector<FlxFunction*> V;
    std::vector<FlxMtxFun*> M;
    virtual void task();
  public:
    FlxObjFunPlot(bool dolog,const std::vector<bool>& B, const std::vector<FlxFunction*>& V, const std::vector<FlxMtxFun*>& M, std::string ostreamV, const bool check_TOL): FlxObjOutputBase(dolog,ostreamV,false,check_TOL), B(B), V(V), M(M) {}
    ~FlxObjFunPlot();
};

/**
* @brief object class: set a correlation between random variables
*
* rnd corr set rv1 RV_1 rv2 RV_2 rho CORR;
*/
class FlxObjRndCorr_Set : public FlxObjBase {
  protected:
    FlxFunction* rv1;
    FlxFunction* rv2;
    FlxFunction* rho;
    virtual void task();
  public:
    FlxObjRndCorr_Set(bool dolog, FlxFunction* rv1, FlxFunction* rv2, FlxFunction* rho): FlxObjBase(dolog), rv1(rv1),rv2(rv2),rho(rho) {}
    ~FlxObjRndCorr_Set();
};

/**
* @brief object class: create transformation matrix in case of correlation
*
* rnd corr create;
*/
class FlxObjRndCorr_Create : public FlxObjBase {
  protected:
    virtual void task();
  public:
    FlxObjRndCorr_Create(bool dolog): FlxObjBase(dolog) {}
};

/**
* @brief object class: create random samples
*
* rnd smp;
*/
class FlxObjRndSmp : public FlxObjOutputBase {
  protected:
    bool genSMP;
    virtual void task();
  public:
    FlxObjRndSmp(bool dolog, std::string ostreamV, bool verbose, bool genSMP=true): FlxObjOutputBase(dolog,ostreamV,verbose),genSMP(genSMP) {}
};

/**
* @brief object class: create random samples
*
* rnd track record NUMB;
*/
class FlxObjRndTrackRecord : public FlxObjOutputBase {
  protected:
    FlxFunction* numb;
    virtual void task();
  public:
    FlxObjRndTrackRecord(bool dolog, FlxFunction* numb, std::string ostreamV, bool verbose): FlxObjOutputBase(dolog,ostreamV,verbose), numb(numb) {}
    ~FlxObjRndTrackRecord() { delete numb; }
};

/**
* @brief object class: stop reading of semi random samples
*
* rnd track replay "FILE";
*/
class FlxObjRndTrackReplay : public FlxObjBase {
  protected:
    FlxString* isname;
    virtual void task();
  public:
    FlxObjRndTrackReplay(bool dolog, FlxString* isname) : FlxObjBase(dolog), isname(isname) {};
    ~FlxObjRndTrackReplay() { delete isname; }
};

/**
* @brief object class: stop reading of semi random samples
*
* rnd track stop;
*/
class FlxObjRndTrackStop : public FlxObjBase {
  protected:
    virtual void task() {data->RndBox.RndCreator.replay_stop(true);};
  public:
    FlxObjRndTrackStop(bool dolog) : FlxObjBase(dolog) {};
};

/**
* @brief object class: create a new random variable
*
* rnd new ...;
*/
class FlxObjRndCreate : public FlxObjBase {
  protected:
    FlxFunction* index;
    FlxRnd_Generator_base* rG;
    virtual void task();
  public:
    FlxObjRndCreate(bool dolog, FlxFunction* index, FlxRnd_Generator_base* rG) : FlxObjBase(dolog), index(index), rG(rG) {}
    ~FlxObjRndCreate() {delete rG; delete index;};
};

/**
* @brief object class: create a new random field
*
* rnd rf new ...;
*/
class FlxObjRndRfCreate : public FlxObjBase {
  protected:
    FlxFunction* index;
    FlxRF_Generator_base* rG;
    virtual void task();
  public:
    FlxObjRndRfCreate(bool dolog, FlxFunction* index, FlxRF_Generator_base* rG) : FlxObjBase(dolog), index(index), rG(rG) {}
    ~FlxObjRndRfCreate() {delete rG; delete index;};
};

/**
* @brief object class: assign an element to a random field
*
* rnd rf elset ...;
*/
class FlxObjRndRfElSet : public FlxObjBase {
  protected:
    FlxFunction* index;				// index of the random field
    const flxDomain_base::geoel_type typeEL;	// type of the element (edge,face,box)
    FlxFunction* elindex;			// index of the element
    FlxKLelementProp_base* KLelProp;
    virtual void task();
  public:
    FlxObjRndRfElSet(bool dolog, FlxFunction* index, const flxDomain_base::geoel_type typeEL, FlxFunction* elindex, FlxKLelementProp_base* KLelProp) : FlxObjBase(dolog), index(index), typeEL(typeEL), elindex(elindex), KLelProp(KLelProp) {}
    ~FlxObjRndRfElSet() {delete index; delete elindex; delete KLelProp; };
};

/**
* @brief object class: assign an element to a random field
*
* rnd rf solve ...;
*/
class FlxObjRndRfSolve : public FlxObjOutputBase {
  protected:
    FlxFunction* index;
    FlxRFsolveProp* RFsolveProp;
    virtual void task();
  public:
    FlxObjRndRfSolve(bool dolog, FlxFunction* index, FlxRFsolveProp* RFsolveProp, std::string ostreamV, bool verbose) : FlxObjOutputBase(dolog,ostreamV,verbose), index(index), RFsolveProp(RFsolveProp) {}
    ~FlxObjRndRfSolve() { delete index; delete RFsolveProp; };
};

/**
* @brief object class: assign an element to a random field
*
* rnd rf err ...;
*/
class FlxObjRndRfErr : public FlxObjOutputBase {
  protected:
    FlxFunction* index;
    FlxFunction* errMeasure;
    virtual void task();
  public:
    FlxObjRndRfErr(bool dolog, FlxFunction* index, FlxFunction* errMeasure, std::string ostreamV, bool verbose) : FlxObjOutputBase(dolog,ostreamV,verbose), index(index), errMeasure(errMeasure) {}
    ~FlxObjRndRfErr() { delete index; delete errMeasure; };
};

/**
* @brief object class: assign an element to a random field
*
* rnd rf plot ...;
*/
class FlxObjRndRfPlot : public FlxObjOutputBase {
  protected:
    FlxFunction* index;
    FlxMtxFun* ppdim;
    FlxFunction* errMeasure;
    virtual void task();
  public:
    FlxObjRndRfPlot(bool dolog, FlxFunction* index, FlxMtxFun* ppdim, FlxFunction* errMeasure, std::string ostreamV, bool verbose): FlxObjOutputBase(dolog,ostreamV,verbose), index(index), ppdim(ppdim),errMeasure(errMeasure) {}
    ~FlxObjRndRfPlot() { delete index; delete ppdim; delete errMeasure; };
};

/**
* @brief object class: assign an element to a random field
*
* rnd rf plotcov ...;
*/
class FlxObjRndRfPlotCov : public FlxObjOutputBase {
  protected:
    FlxFunction* index;
    FlxFunPoint* p;
    FlxMtxFun* ppdim;
    virtual void task();
  public:
    FlxObjRndRfPlotCov(bool dolog, FlxFunction* index, FlxFunPoint* p, FlxMtxFun* ppdim, std::string ostreamV, bool verbose): FlxObjOutputBase(dolog,ostreamV,verbose), index(index),p(p), ppdim(ppdim) {}
    ~FlxObjRndRfPlotCov() { delete index; delete p; delete ppdim; };
};

/**
* @brief object class: set the values of the random variables
*
* rnd set FlxMtxFun;
*/
class FlxObjRndSet : public FlxObjBase {
  protected:
    FlxMtxFun* vec;
    bool origSpace;
    virtual void task();
  public:
    FlxObjRndSet( bool dolog, FlxMtxFun* vec, bool origSpace) : FlxObjBase(dolog), vec(vec), origSpace(origSpace) {}
    ~FlxObjRndSet() {delete vec;}
};

/**
* @brief object class: set a seed value
*
* rnd seed FlxFunction;
*/
class FlxObjRndSeed : public FlxObjBase {
  protected:
    FlxFunction* seedv;
    FlxFunction* icv;
    virtual void task();
  public:
    FlxObjRndSeed( bool dolog, FlxFunction* seedv, FlxFunction* icv) : FlxObjBase(dolog), seedv(seedv), icv(icv) {}
    ~FlxObjRndSeed() {delete seedv; delete icv;}
};

/**
* @brief object class: Counting Intervals
*
* IntervalCount (Number) { ... };
*/
class FlxObjIntervalCount : public FlxObjBase {
  protected:
    /**
    * @brief number of intervals
    */
    tuint Np;
    /**
    * @brief current number of intervals
    */
    tuint countI;
    /**
    * @brief the function used to calculate number of intervals
    */
    FlxFunction* funNp;
        /**
    * @brief the MCI_CODE
    */
    FlxObjBase* InternList;
    void task();
  public:
    FlxObjIntervalCount ( bool dolog, FlxFunction* funNp, FlxObjBase* InternList ) : FlxObjBase(dolog), countI(0),funNp(funNp),InternList(InternList) {}
    virtual ~FlxObjIntervalCount();
};

/**
* @brief object class: Counting Intervals
*
* Filter (i;MtxFun) { ... };
*/
class FlxObjFilter : public FlxObjBase {
  protected:
    /**
    * @brief pointer to the constant (in ConstBox)
    */
    tdouble* cv;
    /**
    * @brief the matrix to run through
    */
    FlxMtxFun* seqMtx;
    /**
    * @brief the FILTER_CODE
    */
    FlxObjBase* block;
    void task();
  public:
    FlxObjFilter ( bool dolog, tdouble* cv, FlxMtxFun* seqMtx, FlxObjBase* block ) : FlxObjBase(dolog), cv(cv),seqMtx(seqMtx),block(block) {}
    virtual ~FlxObjFilter();
};

/**
* @brief object class: set configuration-options for floating point conversion
*/
class FlxObjFloatingPointConversion : public FlxObjBase {
  private:
    int id;
    FlxFunction *fun;
    void task();
  public:
    FlxObjFloatingPointConversion ( bool dolog, FlxFunction* funV, int idV ) : FlxObjBase(dolog), id(idV), fun(funV) {}
    ~FlxObjFloatingPointConversion() { delete fun; }
};

/**
* @brief object class: set some default values at runtime
*/
class FlxObjSetVariousDefault : public FlxObjBase {
  private:
    const tuint ID;
    FlxFunction *fun;
    void task();
  public:
    /**
    * 1 set loglevel at runtime
    * 2 set default polynomial degree of FlxFunDeg
    */
    FlxObjSetVariousDefault ( const bool dolog, const int ID, FlxFunction* fun ) : FlxObjBase(dolog), ID(ID), fun(fun) {}
    ~FlxObjSetVariousDefault() { delete fun; }
};



// -------------------------- READERS ----------------------------------------------------------------------

/**
* @brief object read class: for FlxObjVar
*/
class FlxObjReadVar : public FlxObjReadFCVbase {
  public:
    FlxObjReadVar() { data->IgnoreBox.set_iL_recur("var"); }
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjMtxConst
*/
class FlxObjReadMtxConst : public FlxObjReadFCVbase {
  public:
    FlxObjBase* read ();
    FlxObjBase* read (const std::string cname);
};

/**
* @brief object read class: for FlxObjMtxConst
*/
class FlxObjReadMtxCoeff : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFun
*/
class FlxObjReadFun : public FlxObjReadFCVbase {
  public:
    FlxObjReadFun() { data->IgnoreBox.set_iL_recur("fun"); }
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjEnd
*/
class FlxObjReadEnd : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjExit
*/
class FlxObjReadExit : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjInteractive
*/
class FlxObjReadInteractive : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read (); 
};

/**
* @brief object read class: for FlxObjCalc
*/
class FlxObjReadCalc : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjPrintHLegendrePoly
*/
class FlxObjReadPrintHLegendrePoly : public FlxObjReadOutputBase {
  public:
    FlxObjReadPrintHLegendrePoly();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjPrintHGegenbauerPoly
*/
class FlxObjReadPrintHGegenbauerPoly : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjEcho
*/
class FlxObjReadEcho : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjWarranty
*/
class FlxObjReadWarranty : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjMtxCalc
*/
class FlxObjReadMtxCalc : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjTransformMtx2Octave
*/
class FlxObjReadTransformMtx2Octave : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjTransformMtx2Octave
*/
class FlxObjReadMtxConst_free : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjStreamSet
*/
class FlxObjReadStreamSet : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFileStream
*/
class FlxObjReadFileStream : public FlxObjReadBase {
  public:
    FlxObjReadFileStream();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjInputFileStream
*/
class FlxObjReadInputFileStream : public FlxObjReadBase {
  private:
    const tuint get_Columns(std::vector<int> &Cvec, const bool errSerious=true);
  public:
    FlxObjReadInputFileStream();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjInputVectorStream
*/
class FlxObjReadInputVectorStream : public FlxObjReadBase {
  public:
    FlxObjReadInputVectorStream();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjivstream_append
*/
class FlxObjReadivstream_append : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjistream_write
*/
class FlxObjReadistream_write : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFileStream
*/
class FlxObjReadReadFile : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFileStream
*/
class FlxObjReadDistributorStream : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjDefault
*/
class FlxObjReadDefault : public FlxObjReadBase {
  private:
    FlxObjBase* read_special(std::string& key);
  public:
    FlxObjBase* read();
};

/**
* @brief object read class: for FlxObjIf
*/
class FlxObjReadIf : public FlxObjReadCtrlStrct {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjWhile
*/
class FlxObjReadWhile : public FlxObjReadLoops {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFor
*/
class FlxObjReadFor : public FlxObjReadLoops {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjSFor
*/
class FlxObjReadSFor : public FlxObjReadCtrlStrct {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjSFor
*/
class FlxObjReadTime : public FlxObjReadCtrlStrct {
  private:
    std::string get_stream();
  public:
    FlxObjReadTime();
    FlxObjBase* read ();
};

/**
* @brief object read class: for Timer-Objects
*/
class FlxObjReadTimer : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read();
};

/**
* @brief object read class: for FunPlot-Objects
*/
class FlxObjReadFunPlot : public FlxObjReadOutputBase {
  public:
    FlxObjBase* read();
};

/**
* @brief object read class: for managed random variables
*/
class FlxObjReadRnd : public FlxObjReadOutputBase {
  public:
    FlxObjReadRnd();
    FlxObjBase* read();
};

/**
* @brief object read class: for FlxObjIntervalCount
*/
class FlxObjReadIntervalCount : public FlxObjReadCtrlStrct {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFilter
*/
class FlxObjReadFilter : public FlxObjReadCtrlStrct {
  public:
    FlxObjBase* read ();
};


// -------------------------- FUNCTIONS ----------------------------------------------------------------------

/**
* @brief arithmetic class: isread (reads a value from an input stream)
*/
class FunISread : public FunBase, public FlxDataBase {
  private:
    FlxString* strV;
    FlxIstream* istrm;
    std::string strS;
    void set_istrm();
  public:
    FunISread (FlxString* strV) : strV(strV), istrm(NULL) {};
    ~FunISread();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunISread : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious );
};

/**
* @brief arithmetic class: isread (reads a value from an input stream)
*/
class FunObjExec : public FunBase, public FlxDataBase {
  private:
    FunBase* fun;
    FlxObjBase* InternListLoop;
  public:
    FunObjExec (FunBase* fun, FlxObjBase* InternListLoop) : fun(fun), InternListLoop(InternListLoop) {};
    ~FunObjExec() { delete fun; delete InternListLoop; }
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

class FunReadObjExec : public FunReadFunBase, public FlxObjReadCtrlStrct {
  public:
    FunBase* read ( bool errSerious );
    FlxObjBase* read ();
};



#endif // fesslix_flxobjcommon_H
