/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxobjects_H
#define fesslix_flxobjects_H

#include "flxdata.h"

#if FLX_USE_GNUPLOT
  #include "extern/gnuplot_i.hpp"
#endif




class FlxOptionalParaBase : public FlxReaderBase3 {
  protected:
    std::string pName;
    bool is_set;
    virtual void set(void* valueP) = 0;
  public:
    FlxOptionalParaBase(std::string pName) : pName(pName), is_set(false) {}
    std::string get_name() const { return pName; };
    virtual ~FlxOptionalParaBase() {};
    virtual void read(bool errSerious = true);
    virtual void* read_value(bool errSerious) = 0;
    virtual void* get() = 0;
    virtual void set_default(void* defV) = 0;
    virtual void free_value(void* valueP) = 0;
};


template <class T> class FlxOptionalParaNumber : public FlxOptionalParaBase {
  private:
    T defv;
    T *value;
    void set(void* valueP) {
      T* valueV = static_cast<T*>(valueP);
      if ( value == NULL ) { value = new T( *valueV ); } 
      else { *value = *valueV; }
      delete valueV;
    }
  public:
    FlxOptionalParaNumber(T defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    ~FlxOptionalParaNumber() {  if ( value != NULL ) delete value;}
    void* read_value(bool errSerious) { return ( new T( reader->getNumber<T>(errSerious) ) ); }
    void* get() {
      if ( value == NULL || is_set == false ) {
	is_set = false;
	return &defv;
      } else {
	is_set = false;
	return value;
      }
    }
    void set_default(void* defV) {
      defv = *(static_cast<T*>(defV));
      GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv << "'." << std::endl;
    }
    void free_value(void* valueP) { 
      T* valueV = static_cast<T*>(valueP);
      if ( valueP != NULL) { delete valueV; }
    }
};

class FlxOptionalParaFun : public FlxOptionalParaBase {
  private:
    /**
    * @brief The default value; this value will be deleted by the destructor
    */
    FlxFunction* defv;
    /**
    * @brief The value; this value will be deleted by the destructor
    */
    FlxFunction *value;
    /**
    * @brief Set a new function;
    */
    void set(void* valueP);
  public:
    FlxOptionalParaFun(FlxFunction* defV, const std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    FlxOptionalParaFun(const tdouble defV, const std::string pName);
    ~FlxOptionalParaFun();
    /**
    * @brief Read the function - and allocate memory;
    */
    void* read_value(bool errSerious) { return ( new FlxFunction(funReader,errSerious)); }
    /**
    * @brief Return the function - the function is copied and has to be deleted by some other function;
    */
    void* get();
    /**
    * @brief Set a new default function;
    */
    void set_default(void* defV);
    /** 
    * @brief Delete valueP (used by flxObjDefault);
    */
    void free_value(void* valueP);
};

class FlxOptionalParaFunDeg : public FlxOptionalParaBase {
  private:
    /**
    * @brief The default value; this value will be deleted by the destructor
    */
    FlxFunDeg* defv;
    /**
    * @brief The value; this value will be deleted by the destructor
    */
    FlxFunDeg *value;
    /**
    * @brief Set a new function;
    */
    void set(void* valueP);
  public:
    FlxOptionalParaFunDeg(FlxFunDeg* defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    FlxOptionalParaFunDeg(tdouble defV, std::string pName);
    ~FlxOptionalParaFunDeg();
    /**
    * @brief Read the function - and allocate memory;
    */
    void* read_value(bool errSerious) { return ( new FlxFunDeg(reader,funReader,errSerious)); }
    /**
    * @brief Return the function - the function is copied and has to be deleted by some other function;
    */
    void* get();
    /**
    * @brief Set a new default function;
    */
    void set_default(void* defV);
    /** 
    * @brief Delete valueP (used by flxObjDefault);
    */
    void free_value(void* valueP);
};

class FlxOptionalParaFunDegDetect : public FlxOptionalParaBase {
  private:
    /**
    * @brief The default value; this value will be deleted by the destructor
    */
    flxFun_DegreeDetect* defv;
    /**
    * @brief The value; this value will be deleted by the destructor
    */
    flxFun_DegreeDetect *value;
    /**
    * @brief Set a new detector;
    */
    void set(void* valueP);
  public:
    FlxOptionalParaFunDegDetect(flxFun_DegreeDetect* defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    ~FlxOptionalParaFunDegDetect();
    /**
    * @brief Read the detector - and allocate memory;
    */
    void* read_value(bool errSerious) { return ( new flxFun_DegreeDetect(errSerious)); }
    /**
    * @brief Return the detector - the detector is copied and has to be deleted by some other function;
    */
    void* get();
    /**
    * @brief Set a new default detector;
    */
    void set_default(void* defV);
    /** 
    * @brief Delete valueP (used by flxObjDefault);
    */
    void free_value(void* valueP);
};

class FlxOptionalParaMeshRefine_1D : public FlxOptionalParaBase {
  private:
    /**
    * @brief The default value; this value will be deleted by the destructor
    */
    flxFEMrefine_1D* defv;
    /**
    * @brief The value; this value will be deleted by the destructor
    */
    flxFEMrefine_1D *value;
    /**
    * @brief Set a new detector;
    */
    void set(void* valueP);
  public:
    FlxOptionalParaMeshRefine_1D(flxFEMrefine_1D* defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    ~FlxOptionalParaMeshRefine_1D();
    /**
    * @brief Read the detector - and allocate memory;
    */
    void* read_value(bool errSerious) { return ( new flxFEMrefine_1D(errSerious)); }
    /**
    * @brief Return the detector - the detector is copied and has to be deleted by some other function;
    */
    void* get();
    /**
    * @brief Set a new default detector;
    */
    void set_default(void* defV);
    /** 
    * @brief Delete valueP (used by flxObjDefault);
    */
    void free_value(void* valueP);
};

class FlxOptionalParaFlxString : public FlxOptionalParaBase {
  private:
    /**
    * @brief The default value; this value will be deleted by the destructor
    */
    FlxString* defv;
    /**
    * @brief The value; this value will be deleted by the destructor
    */
    FlxString *value;
    /**
    * @brief Set a new FlxString;
    */
    void set(void* valueP);
  public:
    FlxOptionalParaFlxString(std::string defV, std::string pName);
    FlxOptionalParaFlxString(FlxString* defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    
    ~FlxOptionalParaFlxString();
    /**
    * @brief Read the FlxString - and allocate memory;
    */
    void* read_value(bool errSerious) { return ( new FlxString(false,errSerious)); }
    /**
    * @brief Return the FlxString - the FlxString is copied and has to be deleted by some other function;
    */
    void* get();
    /**
    * @brief Set a new default FlxString;
    */
    void set_default(void* defV);
    /** 
    * @brief Delete valueP (used by flxObjDefault);
    */
    void free_value(void* valueP);
};

class FlxOptionalParaMtxFun : public FlxOptionalParaBase {
  private:
    /**
    * @brief The default value; this value will be deleted by the destructor
    */
    FlxMtxFun* defv;
    /**
    * @brief The value; this value will be deleted by the destructor
    */
    FlxMtxFun *value;
    /**
    * @brief Set a new function;
    */
    void set(void* valueP);
  public:
    FlxOptionalParaMtxFun(FlxMtxFun* defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) {}
    ~FlxOptionalParaMtxFun();
    /**
    * @brief Read the function - and allocate memory;
    */
    void* read_value(bool errSerious) { return ( new FlxMtxFun(mtxReader,errSerious)); }
    /**
    * @brief Return the function - the function is copied and has to be deleted by some other function;
    */
    void* get();
    /**
    * @brief Set a new default function;
    */
    void set_default(void* defV);
    /** 
    * @brief Delete valueP (used by flxObjDefault);
    */
    void free_value(void* valueP);
};

class FlxOptionalParaStream : public FlxOptionalParaBase {
  protected:
    std::string defv;
    std::string *value;
    void set(void* valueP);
  public:
    FlxOptionalParaStream(std::string defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) { }
    virtual ~FlxOptionalParaStream();
    virtual void* read_value(bool errSerious);
    virtual void* get();
    virtual void set_default(void* defV);
    virtual void free_value(void* valueP);
};

class FlxOptionalParaBool : public FlxOptionalParaBase {
  private:
    bool defv;
    bool *value;
    void set(void* valueP);
  public:
    FlxOptionalParaBool(bool defV, std::string pName) : FlxOptionalParaBase(pName), defv(defV), value(NULL) { }
    ~FlxOptionalParaBool();
    void* read_value(bool errSerious);
    void* get();
    void set_default(void* defV);
    void free_value(void* valueP);
};

class FlxOptionalParaString : public FlxOptionalParaStream {
  public:
    FlxOptionalParaString(std::string defV, std::string pName) : FlxOptionalParaStream(defV,pName) {};
};

class FlxDefParaBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    std::map<std::string, FlxOptionalParaBase*> box;
  public:
    FlxDefParaBox();
    void insert (FlxOptionalParaBase* value);
    FlxOptionalParaBase* get(std::string name);
    ~FlxDefParaBox();
};

class FlxDefParaBoxBase {
  protected:
    static FlxDefParaBox* AllDefParaBox;
  public:
    virtual ~FlxDefParaBoxBase() {};
    static void set_AllDefParaBox ( FlxDefParaBox* AllDefParaBoxV ) { AllDefParaBox = AllDefParaBoxV; }
};

class FlxOptionalParaBox : public FlxDefParaBoxBase {
  private:
    std::map<std::string, FlxOptionalParaBase*> box;
  public:
    void insert ( std::string name, std::string defname);
    FlxOptionalParaBase* get ( std::string name );
};



/**
* @brief The base class for all other 'object read classes'
*/
class FlxObjReadBase : public FlxReaderBase3, public FlxDataBase, public FlxDefParaBoxBase {
  protected:
    FlxOptionalParaBox ParaBox;
    void read_optionalPara(bool errSerious = true);
    const bool get_doLog();
    FlxFunction* get_optPara_FlxFunction(std::string str);
    FlxFunDeg* get_optPara_FlxFunDeg(std::string str);
    FlxMtxFun* get_optPara_FlxMtxFun(std::string str);
    const bool get_optPara_bool(std::string str);
    FlxString* get_optPara_FlxString(std::string str);
    flxFun_DegreeDetect* get_optPara_DegreeDetect(std::string str);
    flxFEMrefine_1D* get_optPara_Refine_1D(std::string str);
  public:
    FlxObjReadBase();
    virtual ~FlxObjReadBase () {};
    /**
    * @brief This function has to be overriden by all 'object read classes' (that is the hart)
    */
    virtual FlxObjBase* read() = 0;
};

/**
* @brief object read class: base class for objects using output functionality
*/
class FlxObjReadLogBase : public FlxObjReadBase {
  protected:
    const bool get_verboseLog();
  public:
    FlxObjReadLogBase();
};

/**
* @brief object read class: base class for objects using output functionality
*/
class FlxObjReadOutputBase : public FlxObjReadBase {
  protected:
    const std::string get_stream();
    const bool get_verbose();
    const bool get_checkTOL();
  public:
    FlxObjReadOutputBase();
};


/**
* @brief A class for storing all FlxReadObjects (keywords in parameter-files)
*/
class FlxObjectReadBox {
  private:
    std::map<std::string, FlxObjReadBase*> box;
  public:
    FlxObjectReadBox();
    ~FlxObjectReadBox ();
    /**
    * @brief Get a FlxObject by name.
    * @param name The name of the FlxObject to return.
    * @return FlxObjBase*
    */
    FlxObjReadBase* get ( std::string name );
    /**
    * @brief Insert a Flx-Read-Object.
    * @param name An unique name
    * @param value The FlxReadObject to store
    */
    void insert ( std::string name, FlxObjReadBase* value);
};


/**
* @brief This class evaluates a single command in the parameter file
*/
class FlxEvaluateCmd : public FlxDataBase, public FlxReaderBase3 {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    FlxObjectReadBox ObjReadBox;
    void check_ending();
  public:
    FlxEvaluateCmd();
    FlxObjBase* evaluateCmd( );
    FlxObjectReadBox* get_ObjReadBox() { return &ObjReadBox; }
};

class FlxEvaluateCmdBase {
  protected:
    static FlxEvaluateCmd* EvaluateCmd;
  public:
    ~FlxEvaluateCmdBase() {};
    static void set_EvaluateCmd (FlxEvaluateCmd *EvaluateCmdV) {EvaluateCmd = EvaluateCmdV;};
};


class FlxCreateObjReaders {
  public:
    virtual void createObjReaders (FlxObjectReadBox* objReadBox ) = 0;
    virtual void createFunReaders (FlxData* dataBox ) {}
    virtual ~FlxCreateObjReaders() {}
};

/**
* @brief object class: base class for output objects
*/
class FlxObjLogBase : public FlxObjBase {
  protected:
    const bool verboseLog;
  public:
    FlxObjLogBase(bool dolog, bool verboseLog) : FlxObjBase(dolog), verboseLog(verboseLog) {}
    virtual ~FlxObjLogBase() { } 
};

/**
* @brief object class: base class for output objects
*/
class FlxObjOutputBase : public FlxObjBase {
  private:
    const std::string ostreamv;
    const bool verbose;
  protected:
    const bool checkTOL;
  public:
    FlxObjOutputBase( bool dolog, std::string ostreamV, bool verboseV=false, bool checkTOLv=true) : FlxObjBase(dolog), ostreamv(ostreamV), verbose(verboseV),checkTOL(checkTOLv) {}
    virtual ~FlxObjOutputBase() { }
    std::ostream& sout () { return data->OstreamBox.get(ostreamv); }    
    bool is_verbose() const {return verbose;}
};

/**
* @brief object class: define a constant
*
* const constName = FlxFunction
*/
class FlxObjConst : public FlxObjBase {
  private:
    std::string cname;
    FlxFunction *fun;
    tdouble* cptr;
    void task();
  public:
    FlxObjConst ( bool dolog, std::string cnameV, FlxFunction* funV ) : FlxObjBase(dolog), cname(cnameV), fun(funV), cptr(data->ConstantBox.get(cname,true)) {}
    ~FlxObjConst() { delete fun; }
};



/**
* @brief object read class: for FlxObjReadFCVbase
*/
class FlxObjReadFCVbase : public FlxObjReadOutputBase {
  protected:
    /**
    * @brief checks if a function or variable with the name 'thename' already exists - if it does, an exception is thrown
    * @param thename name to check
    * @param theFCVindex 'F'=function; 'C'='const'variable 'V'='var'variable
    */
    void isdefined(const std::string thename, const char theFCVindex, const bool errSerious = true) const;
};

/**
* @brief object read class: for FlxObjConst
*/
class FlxObjReadConst : public FlxObjReadFCVbase {
  public:
    FlxObjBase* read ();
    FlxObjBase* read (std::string cname);
};

/**
* @brief object class: procedure - sub
*
* sub name { ... };
*/
class FlxObjSub : public FlxObjBase {
  private:
    FlxObjBase* InternListSub;
    std::string sub_name;
    void task ();
  public:
    FlxObjSub ( bool dolog, std::string sub_nameV, FlxObjBase* InternListSubV ) 
      : FlxObjBase(dolog), InternListSub(InternListSubV), sub_name(sub_nameV) {}
    ~FlxObjSub() { if ( InternListSub != NULL) delete InternListSub; }    
};

/**
* @brief object class: procedure - sub (execution)
*
* sub_name(); 
* procedure sub_name();
*/
class FlxObjSubDo : public FlxObjBase {
  private:
    FlxObjBase* InternListSub;
    void task () { InternListSub->exec(); }
  public:
    FlxObjSubDo ( bool dolog, FlxObjBase* InternListSubV ) 
      : FlxObjBase(dolog), InternListSub(InternListSubV) {}
    ~FlxObjSubDo() { }    
};

/**
* @brief object read class: base class for readers of control structure objects
*/
class FlxObjReadCtrlStrct : public FlxObjReadLogBase, public FlxEvaluateCmdBase {
  protected:
    FlxObjBase* read_block (bool UseIgnoreList, bool errSerious=true);
};

/**
* @brief object read class: for FlxSub
*/
class FlxObjReadSub : public FlxObjReadCtrlStrct {
  public:
    FlxObjReadSub() { data->IgnoreBox.set_iL_recur("sub"); }
    FlxObjBase* read();
};

/**
* @brief object read class: for FlxSub
*/
class FlxObjReadSubDo : public FlxObjReadBase {
  public:
    FlxObjBase* read();
    FlxObjBase* read(FlxObjBase* objProcedure);
};

/**
* @brief object read base class: for Loops
*/
class FlxObjReadLoops : public FlxObjReadCtrlStrct {
  protected:
    tuint get_maxpasses();
  public:
    FlxObjReadLoops ();
};




#endif // fesslix_flxobjects_H