/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxobjfem_H
#define fesslix_flxobjfem_H

#include "flxobjects.h"

class FlxCreateObjReaders_FEM : public FlxCreateObjReaders {
  public:
    void createObjReaders (FlxObjectReadBox* objReadBox );
    void createFunReaders (FlxData* dataBox );
};

class flxDOFnV : public FlxDataBase {
  private:
    const FlxDof::dof dofDir;
    FlxDof::dof_fixed_state state;
    FlxFunction* state_value;
  public:
    flxDOFnV(const FlxDof::dof dofDir, const FlxDof::dof_fixed_state state, FlxFunction* state_value);
    ~flxDOFnV();
    void calc(flxDOFnPtr& start);
};

class flxDOFnVlist {
  private:
    std::map<FlxDof::dof, flxDOFnV*> box;
    void insert(const FlxDof::dof dofDir, const FlxDof::dof_fixed_state state, FlxFunction* state_value);
  public:
    flxDOFnVlist(ReadStream *reader, FlxFunctionReader *funReader, bool errSerious=true);
    ~flxDOFnVlist();
    /**
    * @brief calculates the dof-list of the node ... start must be a NULL-pointer ... and will contain the list
    */
    void calc(flxDOFnPtr& start) const;
};

class flxDOFeV : public FlxDataBase {
  private:
    const FlxDof::dof dofDir;
    FlxDof::dof_fixed_state state;
    FlxFunDeg* state_value;
    /**
    * @brief the default polynomial order of the edge to create
    */
    FlxFunction* porder;
  public:
    flxDOFeV(const FlxDof::dof dofDir, const FlxDof::dof_fixed_state state, FlxFunDeg* state_value, FlxFunction* porder);
    ~flxDOFeV();
    void calc(flxDOFePtr& start, const bool errSerious);
};

class flxDOFeVlist {
  private:
    std::map<FlxDof::dof, flxDOFeV*> box;
    void insert(const FlxDof::dof dofDir, const FlxDof::dof_fixed_state state, FlxFunDeg* state_value, FlxFunction* porder);
  public:
    flxDOFeVlist(ReadStream *reader, FlxFunctionReader *funReader, bool errSerious=true);
    ~flxDOFeVlist();
    /**
    * @brief calculates the dof-list of the edge ... start must be a NULL-pointer ... and will contain the list
    */
    void calc(flxDOFePtr& start, const bool errSerious) const;
};

class flxDOFfV : public FlxDataBase {
  private:
    const FlxDof::dof dofDir;
    const FlxDof::dof_fixed_state state;
    FlxFunDeg* state_value;
    /**
    * @brief the default polynomial order of the face to create
    */
    FlxFunction* porderLx;
    FlxFunction* porderLy;
    bool tensor;
  public:
    flxDOFfV(const FlxDof::dof dofDir, const FlxDof::dof_fixed_state state, FlxFunDeg* state_value, FlxFunction* porderLx, FlxFunction* porderLy, const bool tensor);
    ~flxDOFfV();
    void calc(flxDOFfPtr& start, const bool errSerious) const;
};

class flxDOFfVlist {
  private:
    std::map<FlxDof::dof, flxDOFfV*> box;
    void insert(const FlxDof::dof dofDir, const FlxDof::dof_fixed_state state, FlxFunDeg* state_value, FlxFunction* porderLx, FlxFunction* porderLy, const bool tensor);
  public:
    flxDOFfVlist(ReadStream *reader, FlxFunctionReader *funReader, bool errSerious=true);
    ~flxDOFfVlist();
    /**
    * @brief calculates the dof-list of the face ... start must be a NULL-pointer ... and will contain the list
    */
    void calc(flxDOFfPtr& start, const bool errSerious) const;
};

// -------------------------- OBJECTS ----------------------------------------------------------------------


/**
* @brief object class: define a node
*
* node index x y z fix enumFix [value] ... spring enumFix value ...
*/
class FlxObjNode : public FlxObjBase {
  private:
    FlxFunction *index;
    const FlxFunPoint *p;
    const flxDOFnVlist *dofVList;
    void task();
  public:
    FlxObjNode ( const bool dolog, FlxFunction* indexV, const FlxFunPoint* pV, const flxDOFnVlist *dofVList );
    ~FlxObjNode ();
};

/**
* @brief object class: define a edge
*
* edge index Node1 Node2 fix enumFix [value] ... spring enumFix value ...
*/
class FlxObjEdge : public FlxObjBase {
  private:
    FlxFunction *index;
    FlxFunction *node1;
    FlxFunction *node2;
    const flxDOFeVlist *dofVList;
    const flxFEMrefine_1D* mesh;
    FlxFunction* deforder;
    const bool condenseStf;
    void task();
  public:
    FlxObjEdge ( const bool dolog, FlxFunction* indexV, FlxFunction *node1, FlxFunction *node2, const flxDOFeVlist *dofVList, const flxFEMrefine_1D* mesh, FlxFunction* deforder,  const bool condenseStf );
    ~FlxObjEdge ();
};

/**
* @brief object class: define a face
*
* edge index Node1 Node2 fix enumFix [value] ... spring enumFix value ...
*/
class FlxObjFace : public FlxObjBase {
  private:
    FlxFunction *index;
    FlxFunction *edge1;
    FlxFunction *edge2;
    FlxFunction *edge3;
    FlxFunction *edge4;
    const flxDOFfVlist *dofVList;
    FlxFunction* deforder;
    FlxFunction* refine_min;
    FlxFunction* refine_max;
    FlxFunction* geo_fun;
    FlxFunction* cellAreaPoints;
    const bool deftensor;
    const bool condenseStf;
    void task();
  public:
    FlxObjFace ( const bool dolog, FlxFunction* indexV, FlxFunction *edge1, FlxFunction *edge2, FlxFunction *edge3, FlxFunction *edge4, const flxDOFfVlist *dofVList, FlxFunction* deforder, const bool deftensor, const bool condenseStf, FlxFunction* refine_min, FlxFunction* refine_max, FlxFunction* geo_fun, FlxFunction* cellAreaPoints);
    ~FlxObjFace ();
};

/**
* @brief object class: define a box
*
* edge index Node1 Node2 fix enumFix [value] ... spring enumFix value ...
*/
class FlxObjBox : public FlxObjBase {
  private:
    FlxFunction *index;
    FlxFunction *face1;
    FlxFunction *face2;
    FlxFunction *face3;
    FlxFunction *face4;
    FlxFunction *face5;
    FlxFunction *face6;
    void task();
  public:
    FlxObjBox ( const bool dolog, FlxFunction* indexV, FlxFunction *face1, FlxFunction *face2, FlxFunction *face3, FlxFunction *face4, FlxFunction *face5, FlxFunction *face6 );
    ~FlxObjBox ();
};

/**
* @brief object class: define an isotropic material
*
* material INDEX p1 VALUE p2 VALUE
*/
class FlxObjIsoMaterial : public FlxObjBase {
  private:
    FlxFunction* index;
    flxMaterial *mat;
    void task();
  public:
    FlxObjIsoMaterial ( bool dolog, FlxFunction* indexV, flxMaterial *mat ): FlxObjBase(dolog), index(indexV), mat(mat) {};
    ~FlxObjIsoMaterial ();
};

class FlxCrossSec_Generator_base;
/**
* @brief object class: define a cross section
*
* crosssec INDEX Manual A A Iy IY Iz IZ
*/
class FlxObjCrossSec : public FlxObjBase {
  private:
    FlxFunction* index;
    FlxCrossSec_Generator_base* cSG;
    void task();
  public:
    FlxObjCrossSec( bool dolog, FlxFunction* indexV, FlxCrossSec_Generator_base *cSG ): FlxObjBase(dolog), index(indexV), cSG(cSG) {};
    ~FlxObjCrossSec();
};

/**
* @brief object class: define a nodal load
*
* loadcase INDEX { factor= ; descr= ;};
*/
class FlxObjLoadCase : public FlxObjBase {
  private:
    FlxFunction* index;
    FlxFunction* factor;
    FlxString* descr;
    void task();
  public:
    FlxObjLoadCase( bool dolog, FlxFunction* indexV, FlxFunction* factorV, FlxString* descrV ): FlxObjBase(dolog), index(indexV), factor(factorV),descr(descrV) {};
    ~FlxObjLoadCase();
};

/**
* @brief object class: define a nodal load
*
* nload INDEX DOF VALUE
*/
class FlxObjNLoad : public FlxObjBase {
  private:
    FlxFunction* index;
    const FlxDof::dof dofV;
    FlxFunction* value;
    FlxFunction* loadcase;
    const bool evalVal;
    const bool replace;
    void task();
  public:
    FlxObjNLoad( const bool dolog, FlxFunction* indexV, const FlxDof::dof dofV, FlxFunction* valueV, const bool evalVal, const bool replaceV, FlxFunction* loadcaseV ): FlxObjBase(dolog), index(indexV), dofV(dofV),value(valueV),loadcase(loadcaseV),evalVal(evalVal),replace(replaceV) {};
    ~FlxObjNLoad();
};

/**
* @brief object class: define an element load
*
* eload ...
*/
class FlxObjELoad : public FlxObjBase {
  private:
    FlxFunction* edgeIdx;
    const bool global;
    const FlxDof::dof dofV;
    FlxFunDeg* value;
    FlxFunction* pos1;		// local coordinate of load-start [-1;1]
    FlxFunction* pos2;		// NULL in case of a point load
    FlxFunction* LCfun;

    void task();
  public:
    FlxObjELoad( bool dolog, FlxFunction* edgeIdx, const bool global, const FlxDof::dof dofV, FlxFunDeg* value, FlxFunction* pos1, FlxFunction* pos2, FlxFunction* LCfun);
    ~FlxObjELoad();
};

/**
* @brief object class: define a truss element
*
* truss INDEX EDGE_IDX material MATERIAL crosssec CROSSSEC;
*/
class FlxObjElTruss : public FlxObjBase {
  protected:
    FlxFunction* index;
    FlxFunction* edge_index;
    FlxFunction* material;
    FlxFunction* crossSec;
    bool allowFastIntegration;
    void task();
  public:
    FlxObjElTruss( const bool dolog, FlxFunction* index, FlxFunction* edge_index, FlxFunction* material, FlxFunction* crossSec, const bool allowFastIntegration);
    virtual ~FlxObjElTruss();
};

/**
* @brief object class: define a beam element
*
* elbeam INDEX N1 N2 material MATERIAL crosssec CROSSSEC;
*/
class FlxObjElBeam : public FlxObjBase {
  private:
    FlxFunction* index;
    FlxFunction* edge_index;
    FlxFunction* material;
    FlxFunction* crossSec;
    void task();
  public:
    FlxObjElBeam( const bool dolog, FlxFunction* index, FlxFunction* edge_index, FlxFunction* material, FlxFunction* crossSec);
    virtual ~FlxObjElBeam();
};

/**
* @brief object class: define a plane stress/strain element
*
* elplanestr INDEX N1 N2 N3 N4 material MATERIAL planestress BOOL;
*/
class FlxObjElPlaneStr : public FlxObjBase {
  private:
    FlxFunction* index;
    FlxFunction* face_index;
    FlxFunction* material;   
    FlxFunDeg* t;
    bool planeStress;
    void task();
  public:
    FlxObjElPlaneStr( bool dolog, FlxFunction* index, FlxFunction* face_index, FlxFunction* material, FlxFunDeg* t, bool planeStress);
    ~FlxObjElPlaneStr();
};

/**
* @brief object class: import geometry of plane stress/strain elements from file (h-FEM)
*
*/
class FlxObjImportData_ElPlaneStr : public FlxObjBase {
  private:
    FlxString* filename;
    FlxFunction* material;   
    FlxFunDeg* t;
    FlxFunction* lcfun;
    bool planeStress;
    FlxFunction* RFindex;
    FlxKLelementProp_base* KLelProp;
    void task();
  public:
    FlxObjImportData_ElPlaneStr( bool dolog, FlxString* filename, FlxFunction* material, FlxFunDeg* t, bool planeStress, FlxFunction* lcfun, FlxFunction* RFindex, FlxKLelementProp_base* KLelProp);
    ~FlxObjImportData_ElPlaneStr();
};

/**
* @brief object class: create new FEM-object
*/
class FlxObjControlFEM_new : public FlxObjOutputBase {
  protected:
    virtual void task();
  public:
    FlxObjControlFEM_new ( bool dolog, std::string ostreamV, bool verbose ) : FlxObjOutputBase(dolog,ostreamV,verbose) {}
};

/**
* @brief object class: assemble DOF
*/
class FlxObjControlFEM_assembleDOF : public FlxObjOutputBase {
  protected:
    virtual void task();
  public:
    FlxObjControlFEM_assembleDOF ( bool dolog, std::string ostreamV, bool verbose );
    ~FlxObjControlFEM_assembleDOF() {}
};

/**
* @brief object class: assemble Stiffness Matrix
*/
class FlxObjControlFEM_assembleStf : public FlxObjOutputBase {
  protected:
    bool printmtx;
    virtual void task();
  public:
    FlxObjControlFEM_assembleStf ( bool dolog, std::string ostreamV, bool verbose, bool printmtx ) : FlxObjOutputBase(dolog,ostreamV,verbose),printmtx(printmtx) {}
};

/**
* @brief object class: assemble Force Vector
*/
class FlxObjControlFEM_assembleForc : public FlxObjOutputBase {
  protected:
    std::vector<FlxFunction*> loadcases;
    bool printforceVec;
    virtual void task();
  public:
    FlxObjControlFEM_assembleForc ( bool dolog, std::string ostreamV, bool verbose, std::vector<FlxFunction*> loadcasesV, bool printforceVec ) : FlxObjOutputBase(dolog,ostreamV,verbose), loadcases(loadcasesV), printforceVec(printforceVec) {}
    ~FlxObjControlFEM_assembleForc();
};

/**
* @brief object class: solve the system
*/
class FlxObjControlFEM_solve : public FlxObjOutputBase {
  protected:
    FlxFunction* iter;
    FlxFunction* eps;
    FlxFunction* pcn;
    bool startZero;
    bool printsol;
    bool printmtx;
    virtual void task();
  public:
    FlxObjControlFEM_solve ( const bool dolog,FlxFunction* iter, FlxFunction* eps, FlxFunction* pcn, std::string ostreamV, const bool verbose, const bool printsol, const bool printmtx, const bool startZero ) : FlxObjOutputBase(dolog,ostreamV,verbose),iter(iter),eps(eps),pcn(pcn),startZero(startZero),printsol(printsol),printmtx(printmtx) {}
    virtual ~FlxObjControlFEM_solve();
};

/**
* @brief object class: update the element displacement vectors
*/
class FlxObjControlFEM_updateDispl : public FlxObjBase {
  protected:
    virtual void task();
  public:
    FlxObjControlFEM_updateDispl ( bool dolog ) : FlxObjBase(dolog) {}
};

/**
* @brief object class: print nodes
*/
class FlxObjControlFEM_printNodes : public FlxObjOutputBase {
  protected:
    bool printsol;
    virtual void task();
  public:
    FlxObjControlFEM_printNodes ( const bool dolog, std::string ostreamV, const bool printsol ) : FlxObjOutputBase(dolog,ostreamV), printsol(printsol) {}
};


class FlxObjElPlot : public FlxObjOutputBase {
  protected:
    const std::vector<FlxElementInfo::elInfo> V;
    FlxMtxFun* ppdim;
    FlxMtxFun* elfun;
    virtual void task();
  public:
    FlxObjElPlot(bool dolog,const std::vector<FlxElementInfo::elInfo>& V, FlxMtxFun* ppdim, std::string ostreamV, FlxMtxFun* elfun): FlxObjOutputBase(dolog,ostreamV), V(V), ppdim(ppdim), elfun(elfun) {}
    ~FlxObjElPlot();
};

// -------------------------- READERS ----------------------------------------------------------------------

/**
* @brief object read class: for FlxObjNode
*/
class FlxObjReadNode : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjEdge
*/
class FlxObjReadEdge : public FlxObjReadBase {
  public:
    FlxObjReadEdge();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjFace
*/
class FlxObjReadFace : public FlxObjReadBase {
  public:
    FlxObjReadFace();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjBox
*/
class FlxObjReadGeoBox : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjIsoMaterial
*/
class FlxObjReadIsoMaterial : public FlxObjReadBase {
  protected:
    virtual FlxFunction* read_matprop(char& c, bool errSerious = false);
  public:
    FlxObjReadIsoMaterial();
    virtual ~FlxObjReadIsoMaterial() {};
    virtual FlxObjBase* read ();
};

class FlxCrossSec_Generator_base : public FlxReaderBase2 {
  public:
    virtual ~FlxCrossSec_Generator_base() {};
    /**
    * @brief create an instance of a crosssection
    */
    virtual flxCrossSec* generate_CrossSec() = 0;
    // Types of Crosssections
      static flxCrossSec::cstype get_cst(std::string strcst, bool errSerious = true);
      static std::string get_cst(flxCrossSec::cstype cst);
    /**
    * @brief create an instance of a crosssection generator
    */
    static FlxCrossSec_Generator_base* createCS(flxCrossSec::cstype cst);   
};

class FlxCrossSec_Generator_Manual : public FlxCrossSec_Generator_base {
  private:
    FlxFunDeg* A;
    FlxFunDeg* Iy;
    FlxFunDeg* Iz;
    FlxFunDeg* It;
  public:
    FlxCrossSec_Generator_Manual();
    virtual ~FlxCrossSec_Generator_Manual() { delete A; delete Iy; delete Iz; delete It;}
    flxCrossSec* generate_CrossSec();
};

/**
* @brief object read class: for FlxObjCrossSec
*/
class FlxObjReadCrossSec : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjCrossSec
*/
class FlxObjReadLoadCase : public FlxObjReadBase {
  public:
    FlxObjReadLoadCase();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjCrossSec
*/
class FlxObjReadNLoad : public FlxObjReadBase {
  public:
    FlxObjReadNLoad();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjCrossSec
*/
class FlxObjReadELoad : public FlxObjReadBase {
  public:
    FlxObjReadELoad();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjElTruss
*/
class FlxObjReadElTruss : public FlxObjReadBase {
  protected:
    
  public:
    FlxObjReadElTruss();
    virtual ~FlxObjReadElTruss() {};
    FlxObjBase* read();
};

/**
* @brief object read class: for FlxObjElBeam
*/
class FlxObjReadElBeam : public FlxObjReadBase {
  public:
    FlxObjReadElBeam();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjElPlaneStr
*/
class FlxObjReadElPlaneStr : public FlxObjReadBase {
  public:
    FlxObjReadElPlaneStr();
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjControlFEM
*/
class FlxObjReadControlFEM : public FlxObjReadOutputBase {
  protected:
    FlxFunction* get_dfun(std::string dstr);
    bool get_printsol();
    bool get_printmtx();
    bool get_startZero();
    virtual FlxObjBase* read ();
    virtual FlxObjBase* read_new(bool errSerious=false);
    virtual FlxObjBase* read_assdof(bool errSerious=false);
    virtual FlxObjBase* read_assstf(bool errSerious=false);
    virtual FlxObjBase* read_assforc(bool errSerious=false);
    virtual FlxObjBase* read_solve(bool errSerious=false);
  public:
    FlxObjReadControlFEM();
    virtual ~FlxObjReadControlFEM() {};
};

/**
* @brief object read class: for FlxObjIsoMaterial (random material - variable)
*/
class FlxObjReadIsoRVmat : public FlxObjReadIsoMaterial {
  private:
    FlxFunction* read_matprop(char &c,bool errSerious=true);
  protected:
    void get_matprop(FlxFunctionPtr& E, FlxFunctionPtr& sD, FlxFunctionPtr& nu, bool errSerious=true);
  public:
    virtual ~FlxObjReadIsoRVmat() {};
    FlxObjBase* read ();
};

/**
* @brief object read class: for FlxObjIsoMaterial (random material - field)
*/
class FlxObjReadIsoRFmat : public FlxObjReadIsoRVmat {
  public:
    FlxObjBase* read ();
};

class FlxObjReadElPlot : public FlxObjReadOutputBase {
  private:
    virtual FlxObjBase* read ();
  public:
    FlxObjReadElPlot();
};

/**
* @brief object read class: for FlxObjElPlaneStr
*/
class FlxObjReadDataImport : public FlxObjReadBase {
  public:
    FlxObjReadDataImport();
    FlxObjBase* read ();
};


// -------------------------- Functions ---------------------------------------------------------------------

class FunFEM_base : public FunBase, public FlxDataBase {
  protected:
    FunBase* node;
    FlxDof::dof dofC;
    flxDOFn* get_dof(FunBase* nodeV, FlxDof::dof dofV);
    flxDOFn* get_dof();
    std::string write_dof(FunBase* nodeV, FlxDof::dof dofV);
    std::string write_dof();
  public:
    FunFEM_base(FunBase* node,FlxDof::dof dofC):node(node),dofC(dofC){}
    virtual ~FunFEM_base();
    const bool search_circref(FlxFunction* fcr) {return (node->search_circref(fcr));};
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
    virtual void dependOn_RV(bVec& bv);
    virtual const bool dependOn_Const(const tdouble* const thenumber);
};

/**
* @brief arithmetic read class of FunNodeU
*/
class FunReadFEM_base : public FunReadFunBase {
  protected:
    void read_node(FunBasePtr& node, FlxDof::dof& dofC, bool errSerious );
};

class FunNodeU : public FunFEM_base {
  public:
    FunNodeU(FunBase* node,FlxDof::dof dofC):FunFEM_base(node,dofC) {};
    const tdouble calc();
    const std::string write();
    const tuint precedence() { return 0; }
};

/**
* @brief arithmetic read class of FunNodeU
*/
class FunReadNodeU : public FunReadFEM_base {
  public:
    FunBase* read ( bool errSerious );
};

class FunNodeF : public FunFEM_base {
  public:
    FunNodeF(FunBase* node,FlxDof::dof dofC):FunFEM_base(node,dofC) {};
    const tdouble calc();
    const std::string write();
    const tuint precedence() { return 0; }
};

/**
* @brief arithmetic read class of FunNodeF
*/
class FunReadNodeF : public FunReadFEM_base {
  public:
    FunBase* read ( bool errSerious );
};


class FunFEMelInfo : public FunBase, public FlxDataBase {
  protected:
    FunBase* elF;
    FlxElementInfo::elInfo elI;
    FunBase* lxF;
    FunBase* lyF;
    FunBase* lzF;
  public:
    FunFEMelInfo(FunBase* elF, FlxElementInfo::elInfo elI, FunBase* lxF, FunBase* lyF, FunBase* lzF) : elF(elF), elI(elI), lxF(lxF), lyF(lyF), lzF(lzF) {}
    ~FunFEMelInfo();
    const tdouble calc();
    const std::string write();
    const tuint precedence() { return 0; }
    const bool search_circref(FlxFunction* fcr);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
};

/**
* @brief arithmetic read class of FunFEMelInfo
*/
class FunReadFEMelInfo : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious );
};

class FunGeoCalc : public FunBase, public FlxDataBase {
  protected:
    const flxDomain_base::geoel_type typeEL;	// type of the element (edge,face,box)
    FunBase* elindex;			// index of the element
    FunBase* fun;
    FunBase* lxF;
    FunBase* lyF;
    FunBase* lzF;
  public:
    FunGeoCalc(const flxDomain_base::geoel_type typeEL, FunBase* elindex, FunBase* fun, FunBase* lxF, FunBase* lyF, FunBase* lzF) : typeEL(typeEL), elindex(elindex), fun(fun), lxF(lxF), lyF(lyF), lzF(lzF) {}
    ~FunGeoCalc();
    const tdouble calc();
    const std::string write();
    const tuint precedence() { return 0; }
    const bool search_circref(FlxFunction* fcr);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
};

class FunReadGeoCalc : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious );
};

class FunGeoTimer : public FunGeoCalc {
  protected:
    FunBase* NumberRun;
  public:
    FunGeoTimer(const flxDomain_base::geoel_type typeEL, FunBase* elindex, FunBase* fun, FunBase* lxF, FunBase* lyF, FunBase* lzF,FunBase* NumberRun) : FunGeoCalc(typeEL,elindex,fun,lxF,lyF,lzF), NumberRun(NumberRun) {}
    ~FunGeoTimer();
    const tdouble calc();
    const std::string write();
    const bool search_circref(FlxFunction* fcr);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
//     void dependOn_RV(bVec& bv);
//     const bool dependOn_Const(const tdouble* const thenumber);
};

class FunReadGeoTimer : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious );
};

/**
* @brief arithmetic class:
*/
class FunHLegendrePoly : public FunBaseFun_multPara {
  public:
    FunHLegendrePoly (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "hlegendrepoly";}
};


class FunReadFunHLegendrePoly : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious ) { return new FunHLegendrePoly( read_parameters(2,errSerious) ); }
};

/**
* @brief arithmetic class:
*/
class FunHGegenbauerPoly : public FunBaseFun_multPara {
  public:
    FunHGegenbauerPoly (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "hgegenbauerpoly";}
};

class FunReadFunHGegenbauerPoly : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious ) { return new FunHGegenbauerPoly( read_parameters(2,errSerious) ); }
};

/**
* @brief arithmetic class:
*/
class FunRFeigenvalue : public FunBaseFun_multPara, public FlxDataBase {
  public:
    FunRFeigenvalue (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "rfeigenvalue";}
};

class FunReadFunFunRFeigenvalue : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious ) { return new FunRFeigenvalue( read_parameters(2,errSerious) ); }
};


#endif // fesslix_flxobjfem_H
