/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "sfem.h"
#include "flxfunction_ope_calc.h"
#include "flxfunction_fun_calc.h"
#include "flxmtxfun_fun_calc.h"
#include "flxmtxfun_ope_calc.h"

#include <iostream>

FlxRSM_Fun_base* FunReadFunRSM::RSMfunction = NULL;
FlxRndBox* SSS_bne_base::RndBox = NULL;


void FlxCreateObjReaders_RND::createObjReaders(FlxObjectReadBox* objReadBox) {
//    objReadBox->insert("sfem", new FlxObjReadSFEM());
//    objReadBox->insert("sfemtest", new FlxObjReadSFEMtest()); 
  objReadBox->insert("mci", new FlxObjReadMCI());
  objReadBox->insert("ips", new FlxObjReadIpS());
  objReadBox->insert("subsetsim", new FlxObjReadSubSetSim());
  objReadBox->insert("form", new FlxObjReadFORM());
  objReadBox->insert("form_pdf", new FlxObjReadFORM_pdf());
  objReadBox->insert("rsm", new FlxObjReadRSM());
  objReadBox->insert("rsm_set", new FlxObjReadRSM_set());
  objReadBox->insert("kde", new FlxObjReadKDE());
  objReadBox->insert("hlrv", new FlxObjReadHLRV());
  objReadBox->insert("statsmp", new FlxObjReadStatSmp());
  objReadBox->insert("sortsmp", new FlxObjReadSortSmp());
  objReadBox->insert("smpplot", new FlxObjReadSmpPlot());
}

void FlxCreateObjReaders_RND::createFunReaders(FlxData* dataBox)
{
  dataBox->ConstantBox.declareC("sss_iter");
  
  dataBox->FunBox.insert("rsm", new FunReadFunRSM() );
  dataBox->FunBox.insert("rsm_terms", new FunReadFunRSM_terms() );
  dataBox->FunBox.insert("hlrv", new FunReadFunHLRV() );
  dataBox->FunBox.insert("rf", new FunReadFunRF() );
  dataBox->FunBox.insert("rf_err", new FunReadFunRFerr() );
  
  dataBox->MtxFunBox.insert("rv_mean", new FunReadFunRVmean() );
  dataBox->MtxFunBox.insert("rv_sd", new FunReadFunRVstddev() );
  dataBox->MtxFunBox.insert("rv_vec", new FunReadFunRVvalue() );
  
  SSS_bne_base::RndBox = &(dataBox->RndBox);
}

void FlxObjMCI::FirstThingsFirst()
{
  GlobalVar.slog(4) << "mci: performing a Monte Carlo integration. (N=" << GlobalVar.Double2String(Np) << ")" << std::endl; 
}

void FlxObjMCI::log_AddResInfo(std::ostream& sout, const tdouble hits, const tdouble Npd)
{
  if (reliability) {
    // Bayesian estimation
      const tdouble alpha_prior = 0.5;
      const tdouble beta_prior = 0.5;
      FlxRnd_Beta betad(hits+alpha_prior,Npd-hits+beta_prior);
      sout << "  mean(Pf) (Bayesian):      " << GlobalVar.Double2String(betad.mean()) << std::endl;
      sout << "  C.o.V.(Pf) (Bayesian): " << GlobalVar.Double2String(betad.sd()/betad.mean()) << std::endl;
      sout << "  Estim. C.o.V. (MLE):   " << GlobalVar.Double2String(sqrt((1.-theResult)/(theResult*Npd))) << std::endl;
      // update _bne-const
	std::string res_name = data->ConstantBox.get(&theResult);
	res_name += "_bne";
	tdouble* theResult_bne = data->ConstantBox.get(res_name,true);
	*theResult_bne = betad.mean();
    SMtxBase* pcM = pc->calc();
    tdouble pcT;
    tuint i,j;
    for ( i=0; i<pcM->get_nrows(); ++i ) {
      for ( j=0; j<pcM->get_ncols(); ++j ) {
	pcT = pcM->operator()(i,j);
	// Calculate the confidence interval
	  sout << "  Confidence limits (two-sided) of probability of failure (" << GlobalVar.Double2String(pcT) << "):\t";
	  tdouble kc = 1-(1-pcT)/2;
	  sout << "Bayesian: [" << GlobalVar.Double2String(betad.Inv_cdf_x((1.-pcT)/2)) << "; " << GlobalVar.Double2String(betad.Inv_cdf_x(kc)) << "]\t";
	  kc = rv_InvPhi(kc);
	  tdouble lb = theResult - kc * sqrt(theResult/Npd);
	  if (lb<0.0) lb=0.0;
	  tdouble ub = theResult + kc * sqrt(theResult/Npd);
	  if (ub>1.0) ub=1.0;
	  sout << "Normal: [" << GlobalVar.Double2String(lb) << "; " << GlobalVar.Double2String(ub) << "]";
	  sout << std::endl;
      }
    }
    for ( i=0; i<pcM->get_nrows(); ++i ) {
      for ( j=0; j<pcM->get_ncols(); ++j ) {
	pcT = pcM->operator()(i,j);
	// Calculate the confidence interval
	  sout << "  Confidence limits (upper bound) of probability of failure (" << GlobalVar.Double2String(pcT) << "):\t";
	  sout << "Bayesian: " << GlobalVar.Double2String(betad.Inv_cdf_x(pcT)) << ";\t";
	  const tdouble kc = rv_InvPhi(pcT);
	  tdouble ub = theResult + kc * sqrt(theResult/Npd);
	  if (ub>1.0) ub=1.0;
	  sout << "Normal: " << GlobalVar.Double2String(ub);
	  sout << std::endl;
      }
    }
    delete pcM;
  }
}

void FlxObjMCI::task()
{
  Np = funNp->cast2tuint();
  FirstThingsFirst();
  #if FLX_KAHAN_MCI
    pdouble Integ, IntegI;
  #else
    tdouble Integ, IntegI;
  #endif
  tdouble hits, hitsI;
  // split the interval
    unsigned int NpI = Np;
    if (interv) {
      NpI = tuint(sqrt(tdouble(Np)));
    };
  // Perform the integration
    Integ=0.0; hits=0.0;
    // load the progress bar
      FlxProgress prg(std::cout);
      prg.start(NpI);
    if (NpI == Np) {
      for (unsigned int i = 0; i < Np; ++i) {
	Integrationstep(Integ,hits);
	prg.tick(i);
      }
    } else {
      for (unsigned int j = 0; j < NpI-1; ++j) {
	IntegI = 0.0; hitsI = 0.0;
	for (unsigned int i = 0; i < NpI; ++i) {
	  Integrationstep(IntegI,hitsI);
	}
	prg.tick(j);
	Integ+=IntegI;
	hits+=hitsI;
      }
      IntegI = 0.0;
      for (unsigned int i = 0; i < Np-(NpI-1)*NpI; ++i) {
	Integrationstep(IntegI,hitsI);
      }
      Integ+=IntegI;
      hits+=hitsI;
    }
    // stop the progress-bar
      prg.stop();
    Integ/=tdouble(Np); 
    #if FLX_KAHAN_MCI
      theResult = Integ.cast2double();
    #else
      theResult = Integ;
    #endif
    GlobalVar.slog(4) << " Result of the Integration: " << GlobalVar.Double2String(theResult);
    if (reliability) {
      GlobalVar.slog(4) << " (" << hits << " hits)";
    }
    GlobalVar.slog(4) << std::endl;
    if (reliability && (theResult > 1.5 || theResult <= 0.0) ) {
      GlobalVar.alert.alert("FlxObjMCI::log_AddResInfo","Result does not seem to be a probability !!!");
    }
    log_AddResInfo(GlobalVar.slog(4),hits,tdouble(Np));
    LastThingsLast();
}

FlxObjMCI::~FlxObjMCI()
{
  delete funNp;
  delete fung;
  delete InternList;
  delete pc;
}

FlxObjReadMCI::FlxObjReadMCI()
{
  // optional parameters
    // interval split
      AllDefParaBox->insert(new FlxOptionalParaBool(true,"mci::interv"));
      ParaBox.insert("interv", "mci::interv" );
    // determining probability of failure?
      AllDefParaBox->insert(new FlxOptionalParaBool(false,"mci::reliability"));
      ParaBox.insert("reliability", "mci::reliability" );
    // confidence probability
      std::vector<FunBase*>* vecV = new std::vector<FunBase*>;
      vecV->push_back(new FunNumber(0.90));
      vecV->push_back(new FunNumber(0.95));
      FlxMtxFun* mtxvalue = new FlxMtxFun( new SMtxValue(vecV,2,1) );
      AllDefParaBox->insert(new FlxOptionalParaMtxFun(mtxvalue,"mci::reliability::pc"));
      ParaBox.insert("pc", "mci::reliability::pc" );
}

void FlxObjReadMCI::read_MCIblock(tdoublePtr& theResult, FlxFunctionPtr& funNp, FlxFunctionPtr& fung, bool errSerious)
{
  reader->getChar('(',errSerious);
  // result-constant
    std::string cname = reader->getWord(true,errSerious);
    data->ConstantBox.declareC(cname);
    theResult = data->ConstantBox.get(cname);
    reader->getChar(';',errSerious);
  // function for number of points
    funNp = new FlxFunction(funReader,errSerious);
    reader->getChar(';',errSerious);
  // function to integrate
    fung = new FlxFunction(funReader,errSerious);
  reader->getChar(')',errSerious);
}

FlxObjBase* FlxObjReadMCI::read()
{
  tdouble* d1=NULL; FlxFunction* funNp=NULL; FlxFunction* fung=NULL;
  FlxObjBase* MCIblock = NULL;
  try {
    read_MCIblock(d1,funNp,fung,false);
    MCIblock = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjMCI(get_doLog(), *d1, funNp, fung, MCIblock, get_optPara_bool("interv"),get_verboseLog(),get_optPara_bool("reliability"),get_optPara_FlxMtxFun("pc"));
  } catch (FlxException &e) {
    if (funNp) delete funNp;
    if (fung) delete fung;
    if (MCIblock) delete MCIblock;
    throw;
  }
}

void FlxObjIpS::FirstThingsFirst()
{
  GlobalVar.slog(4) << "ips: performing an Importance sampling integration (N=" << GlobalVar.Double2String(Np) << ")" << std::endl; 
  sspace = sspaceG->generate_SS();
  GlobalVar.slog(4) << "  Sampling space: ";
  sspace->print_info(GlobalVar.slog(4), verboseLog);
  GlobalVar.slog(4) << std::endl;
}

void FlxObjIpS::log_AddResInfo(std::ostream& sout, const tdouble hits, const tdouble Npd)
{
  if (sspace->rely_confidence()>0.) {
    const tdouble f = sspace->rely_confidence();
    FlxObjMCI::log_AddResInfo(sout, hits, Npd*f);
  }
}

#if FLX_KAHAN_MCI
void FlxObjIpS::Integrationstep(pdouble& Integ, tdouble& hits)
#else
void FlxObjIpS::Integrationstep(tdouble& Integ, tdouble& hits)
#endif
{  
  tdouble foverh=0.0; 
  sspace->transform_space(data->RndBox,foverh);  
  InternList->exec();
  tdouble t = fung->calc()*foverh;
  Integ+=t;
  if (t>0) ++hits;
}

FlxObjBase* FlxObjReadIpS::read()
{
  tdouble* d1=NULL; FlxFunction* funNp=NULL; FlxFunction* fung=NULL;
  FlxRndSamplingSpace_Generator_base* sspaceG = NULL; FlxObjBase* MCIblock=NULL;
  try {
    read_MCIblock(d1,funNp,fung,false);
    // read sampling space
      reader->getChar('(',false);
      sspaceG = FlxRndSamplingSpace_Generator_base::createSS(FlxRndSamplingSpace_Generator_base::get_sst(reader->getWord(true,false),false),false);
      reader->getChar(')',false);
    MCIblock = read_block(true);
    read_optionalPara(false);
    return new FlxObjIpS(get_doLog(), *d1, funNp, fung, MCIblock, get_optPara_bool("interv"),get_verboseLog(),get_optPara_bool("reliability"),get_optPara_FlxMtxFun("pc"),sspaceG);
  } catch (FlxException &e) {
    if (funNp) delete funNp;
    if (fung) delete fung;
    if (sspaceG) delete sspaceG;
    if (MCIblock) delete MCIblock;
    throw;
  }
}

SSS_bne_MCI::SSS_bne_MCI(const tuint Y, const tuint N)
: Y(Y), N(N), distri(tdouble(Y)+0.5,tdouble(N-Y)+0.5)
{
  
}

const tdouble SSS_bne_MCI::sim_pf()
{
  const tdouble p = RndBox->RndCreator.gen_smp_uniform();
  return distri.Inv_cdf_x(p);
}

const tdouble SSS_bne_MCI::E_Pf() const
{
  return distri.mean();
}

const tdouble SSS_bne_MCI::sigma_Pf() const
{
  return distri.sd();
}

const tdouble SSS_bne_MCI::E2_Pf() const
{
  return distri.E2();
}

SSS_bne_MCMC::SSS_bne_MCMC(const tuint Nc, const tuint Nlc, cs* chain_structure, const tdouble g_limit)
:random_cpt(false),prior_cpt(0.0),level(1),N1(10000),N11(100000),N2(10),N3(500000),
Nc(Nc),Nlc(Nlc)
{
  // Initialize Nvec
    Nvec[0] = Nc*Nlc;
    for (tuint i=1;i<14;++i) Nvec[i] = 0;
    bool prev,prev2;
    cs* csp = chain_structure;
    for (tuint i=0;i<Nc;++i) {
      prev = ((csp->g <= g_limit)?true:false);
      if (prev) ++Nvec[1];	// increase Y_total by one
      ++csp;
      for (tuint j=1;j<Nlc;++j,++csp) {
	const bool now  = ((csp->g <= g_limit)?true:false);
	if (now) ++Nvec[1];
	// j-1
	  ++Nvec[(prev?4:2)];
	  if (now) {
	    ++Nvec[(prev?4:2)+1];
	  }
	// j-2
	if (j>1) {
	  ++Nvec[(prev2?10:6)+(prev?2:0)];
	  if (now) {
	    ++Nvec[(prev2?10:6)+(prev?2:0)+1];
	  }
	}
	prev2 = prev;
	prev = now;
      }
    }
  // Initialize Dvec
    Dvec[0] = tdouble(Nvec[1])/tdouble(Nvec[0]);
   initialize_1();
    
//   tdouble pcond[4];
//   pcond[0] = tdouble(Ncond[0])/tdouble(Ncond[0]+Ncond[2]);
//   pcond[1] = tdouble(Ncond[1])/tdouble(Ncond[1]+Ncond[3]);
//   pcond[2] = tdouble(Ncond[2])/tdouble(Ncond[0]+Ncond[2]);
//   pcond[3] = tdouble(Ncond[3])/tdouble(Ncond[1]+Ncond[3]);
//   tdouble plimit[2];
//   plimit[0] = 1./(1+(1-pcond[0])/pcond[1]);
//   plimit[1] = 1-plimit[0];
//   // eigen-decomposition
//     tdouble evpc[2];
//     const tdouble t1 = sqrt(pow2((pcond[1]-pcond[0]-1)/2)-(pcond[0]-pcond[1]));
//     const tdouble t2 = (pcond[0]+1-pcond[1])/2;
//     evpc[0] = t2+t1;
//     evpc[1] = t2-t1;
//     tdouble evec[4];		// eigenvectors in columns
//     evec[0] = 1/(1+(evpc[0]-pcond[0])/pcond[1]);
//     evec[1] = 1/(1+(evpc[1]-pcond[0])/pcond[1]);
//     evec[2] = 1-evec[0];
//     evec[3] = 1-evec[1];
//     tdouble eveci[4];		// inverse of evec
//     const tdouble t3 = (evec[0]*evec[3]-evec[1]*evec[2]);
//     eveci[0] = evec[3]/t3;
//     eveci[1] = -evec[1]/t3;
//     eveci[2] = -evec[2]/t3;
//     eveci[3] = evec[0]/t3;
//     
//     
//   std::cout << "ev=" << evpc[0] << "; " << evpc[1] << std::endl;
//   std::cout << "  " << pcond[0] << ";\t" << pcond[1] << std::endl;
//   std::cout << "  " << pcond[2] << ";\t" << pcond[3] << std::endl;
//   std::cout << "limit: " << plimit[0] << ";\t" << plimit[1] << std::endl;
//   std::cout << "limit* " << 1./(1.+(1.-pcond[3])/pcond[2]) << std::endl;
//   
//   tdouble tmpmtx1[4];
//   pow_mtx(50,tmpmtx1,evpc,evec,eveci);
//   std::cout << "***" << tmpmtx1[0] << ";\t" << tmpmtx1[1] << std::endl;
//   std::cout << "***" << tmpmtx1[2] << ";\t" << tmpmtx1[3] << std::endl;
}

void SSS_bne_MCMC::initialize_1()
{
  // Estimate the total number of sim_pf-calls:
    tuint NT = 0;
    for (tuint i=0;i<N2;++i) {
      NT += N1 + (N11-N1)/(N2-1)*i;
    }
    NT += N3;
  // load the progress bar
    FlxProgress prg(std::cout);
    prg.start(NT);
    NT = 0;
  tdouble d_err = 1.0;
  tuint i;
  for (i=0;i<N2 && d_err>1.e-3;++i) {
    const tdouble E_prev = Dvec[0];
    if (random_cpt==false) init_cpt();
    pdouble pf = 0.0;
    const tuint NN = N1 + (N11-N1)/(N2-1)*i;
    for (tuint j=0;j<NN;++j,++NT) {
      pf += sim_pf();
      prg.tick(NT);
    }
    pf /= NN;
    Dvec[0] = pf.cast2double();
    d_err = fabs((Dvec[0]-E_prev)/E_prev);
  }
  for (;i<N2;++i) {
    NT += N1 + (N11-N1)/(N2-1)*i;
  }
  // Estimate the final E_pf
    if (random_cpt==false) init_cpt();
    pdouble pf = 0.0;
    pdouble pf2 = 0.0;
    for (tuint i=0;i<N3;++i,++NT) {
      const tdouble p = sim_pf();
      pf += p;
      pf2 += pow2(p);
      prg.tick(NT);
    }
    pf /= N3;
    pf2 /= N3;
    E_pf = pf.cast2double();
    E_pf2 = pf2.cast2double();
    pf *= pf;
    pf2 -= pf;
    sd_pf = sqrt(pf2.cast2double());
    d_err = fabs((E_pf-Dvec[0])/Dvec[0]);
  // stop the progress-bar
    prg.stop();
}

void SSS_bne_MCMC::init_cpt()
{
  if (random_cpt) {
    {
      const tdouble alpha = Nvec[3] + Dvec[0];
      const tdouble beta  = Nvec[2]+1.-alpha;
      const tdouble p = iBeta_reg_inv(alpha,beta,RndBox->RndCreator.gen_smp_uniform());
      Dvec[1] = p;
    }{
      const tdouble alpha = Nvec[5] + Dvec[0];
      const tdouble beta  = Nvec[4]+1.-alpha;
      const tdouble p = iBeta_reg_inv(alpha,beta,RndBox->RndCreator.gen_smp_uniform());
      Dvec[2] = p;
    }
    if (level>2) {
      throw FlxException_NotImplemented("SSS_bne_MCI::init_cpt");
    }
  } else {
    // i-1
      if (prior_cpt<=GlobalVar.TOL() && Nvec[2]==0) {
	Dvec[1] = Dvec[0];
      } else {
	Dvec[1] = (tdouble(Nvec[3])+Dvec[0]*prior_cpt)/(tdouble(Nvec[2])+prior_cpt);
      }
      if (prior_cpt<=GlobalVar.TOL() && Nvec[4]==0) {
	Dvec[2] = Dvec[0];
      } else {
	Dvec[2] = (tdouble(Nvec[5])+Dvec[0]*prior_cpt)/(tdouble(Nvec[4])+prior_cpt);
      }
    // i-2
      if (level > 1) {
	for (tuint i=0;i<4;++i) {
	  if (prior_cpt<=GlobalVar.TOL() && Nvec[6+i*2]==0) {
	    Dvec[i+3] = Dvec[(i<=1?1:2)];
	  } else {
	    Dvec[i+3] = (tdouble(Nvec[6+i*2+1])+Dvec[(i<=1?1:2)]*prior_cpt)/(tdouble(Nvec[6+i*2])+prior_cpt);
	  }
	}
      }
  }
}

// void SSS_bne_MCMC::pow_mtx(const tuint n, tdouble* res, const tdouble*const evpc, const tdouble*const evec, const tdouble*const eveci)
// {
//   tdouble t[4];
//   t[0] = pow(evpc[0],n);
//   t[1] = pow(evpc[1],n);
//   t[2] = t[0];
//   t[3] = t[1];
//   t[0] *= evec[0];
//   t[1] *= evec[1];
//   t[2] *= evec[2];
//   t[3] *= evec[3];
//   res[0] = t[0]*eveci[0] + t[1]*eveci[2];
//   res[1] = t[0]*eveci[1] + t[1]*eveci[3];
//   res[2] = t[2]*eveci[0] + t[3]*eveci[2];
//   res[3] = t[2]*eveci[1] + t[3]*eveci[3];
// }

const tdouble SSS_bne_MCMC::sim_pf()
{
  if (random_cpt) init_cpt();
  bool bvec[3];
  tuint N_hits = 0;
  for (tuint i=0;i<Nc;++i) {
    bvec[1] = RndBox->RndCreator.gen_smp_binary(Dvec[0]);
      if (bvec[1]) ++N_hits;
    bvec[0] = RndBox->RndCreator.gen_smp_binary(Dvec[bvec[1]?2:1]);
      if (bvec[0]) ++N_hits;
    if (level==1) {
      for (tuint j=2;j<Nlc;++j) {
	bvec[1] = bvec[0];
	bvec[0] = RndBox->RndCreator.gen_smp_binary(Dvec[bvec[1]?2:1]);
	  if (bvec[0]) ++N_hits;
      }
    } else {	// level == 2
      for (tuint j=2;j<Nlc;++j) {
	bvec[2] = bvec[1];
	bvec[1] = bvec[0];
	bvec[0] = RndBox->RndCreator.gen_smp_binary( Dvec[(bvec[2]?5:3)+(bvec[1]?1:0)]);
	  if (bvec[0]) ++N_hits;
      }
    }
  }
  return tdouble(N_hits)/tdouble(Nc*Nlc);
}

const tdouble SSS_bne_MCMC::E_Pf() const
{
  return E_pf;
}

const tdouble SSS_bne_MCMC::sigma_Pf() const
{
  return sd_pf;
}

const tdouble SSS_bne_MCMC::E2_Pf() const
{
  return E_pf2;
}

SSS_bne_Box::SSS_bne_Box()
: N(0)
{
  box.reserve(10);
}

SSS_bne_Box::~SSS_bne_Box()
{
  for (tuint i=0;i<N;++i) {
    delete box[i];
  }
  N = 0;
  box.clear();
}

void SSS_bne_Box::add_bne(SSS_bne_base* el)
{
  box.push_back(el);
  ++N;
}

const tdouble SSS_bne_Box::sim_pf() const
{
  tdouble Pf = 1.0;
  for (tuint i=0;i<N;++i) {
    Pf *= box[i]->sim_pf();
  }
  return Pf;
}

const tdouble SSS_bne_Box::E_Pf() const
{
  tdouble Pf = 1.0;
  for (tuint i=0;i<N;++i) {
    Pf *= box[i]->E_Pf();
  }
  return Pf;
}

const tdouble SSS_bne_Box::Cov_Pf() const
{
  tdouble p1 = 1.0;
  pdouble p2 = 1.0;
  for (tuint i=0;i<N;++i) {
    const tdouble e1 = pow2(box[i]->E_Pf());
    tdouble e2 = box[i]->E2_Pf();
    p1 *= e1;
    p2 *= e2;
  }
  p2 -= p1;
  return p2.cast2double();
}

const SSS_bne_base& SSS_bne_Box::get_last() const
{
  return *(box[N-1]);
}

// #include <fstream>
// void SSS_bne_Box::write_test() const
// {
//   std::ofstream myfile;
//   for (tuint i=1;i<N;++i) {
//     std::ostringstream ssV;
//       ssV << "bne_" << i;
// //     std::cout << ssV.str() << std::endl;
//     myfile.open(ssV.str().c_str(), std::ios::out | std::ios::app);
//     SSS_bne_MCMC* SF = dynamic_cast<SSS_bne_MCMC*>(box[i]);
//     const tdouble *dvec = SF->write_test();
//     myfile << dvec[0] << "\t" << dvec[1] << "\t" << dvec[2] << std::endl;
//     myfile.close();
//   }
// }

SSS_sample_list::SSS_sample_list(const tuint N_RV, const tuint Ns, const tuint Nc, const tdouble cgkf)
: N_RV(N_RV), Ns(Ns), Nc(Nc), N_SRL(0), N_hits(0), g_kicked_out(0.0),consider_g_kicked_factor(cgkf)
{
  // allocate smpl_ref_list
    smpl_ref_list = new int[Nc];
    for (tuint i=0;i<Nc;++i) smpl_ref_list[i] = -1;
    g_list = new tdouble[Nc];
  // allocate smpl_list
    smpl_list = new tdouble[N_RV*Nc];
}

SSS_sample_list::~SSS_sample_list()
{
  delete [] smpl_ref_list;
  delete [] g_list;
  delete [] smpl_list;
}

void SSS_sample_list::clear()
{
  for (tuint i=0;i<Nc;++i) smpl_ref_list[i] = -1;
  N_SRL = 0;
  N_hits = 0;
}

void SSS_sample_list::find_multiples(tuint& Nsame_seq, tuint& Nmax_seq, tuint& Ntotal, tuint& Nconsid, const tdouble &gt) const
{
  Nsame_seq = 0;
  Nmax_seq = 1;
  Ntotal = 0;
  Nconsid = 0;
  #ifndef FLX_CV_1
    bool bv[Nc];
  #else
    bool *bv = new bool[Nc];
  #endif
  for (tuint i=0;i<N_SRL;++i) bv[i] = false;
  for (tuint i=0;i<N_SRL-1;++i) {
    if (g_list[i] <= gt) ++Nconsid;
    else continue;
    if (bv[i]) continue;
    tuint Nsmax_t = 1;
    const tdouble g_1 = g_list[i];
    const tdouble* const slp_1 = &smpl_list[smpl_ref_list[i]*N_RV];
    for (tuint j=i+1;j<N_SRL;++j) {
      if (g_list[i] > gt) continue;
      if (fabs(g_list[j]-g_1)<=GlobalVar.TOL()) {
	// make sure that the coordinates match as well
	  bool b = true;
	  const tdouble* const slp_2 = &smpl_list[smpl_ref_list[j]*N_RV];
	  for (tuint k=0;(k<N_RV)&&b;++k) {
	    if (fabs(slp_1[k]-slp_2[k])>GlobalVar.TOL()) b = false;
	  }
	if (b) {
	  if (!bv[i]) {
	    ++Nsame_seq;
	    bv[i] = true;
	  }
	  bv[j] = true;
	  ++Nsmax_t;
	}
      }
    }
    if (bv[i]) {
      if (Nsmax_t>Nmax_seq) Nmax_seq = Nsmax_t;
      Ntotal += Nsmax_t;
    }
  }
  #ifdef FLX_CV_1
    delete [] bv;
  #endif
}

const tuint SSS_sample_list::find_pos(const tdouble g) const
{
  if (N_SRL==0) return 0;
  if (g>=g_list[N_SRL-1]) return N_SRL;
  tuint start = 0;
  tuint length = N_SRL;
  while (length>1) {
    const tuint check = start + ((length+1)/2)-1;
    if (g<g_list[check]) {
      // start = start;
      length = check+1-start;
    } else {
      length = start+length-check-1;
      start = check+1;
    }
  };
  return start; 
}

void SSS_sample_list::register_sample(const tdouble g, const tdouble*const y_ptr)
{
  if (g<=0) ++N_hits;
  if (N_hits > Nc) {
    if (g_kicked_out>g) g_kicked_out = g;
    return;
  }
  const tuint pos = find_pos(g);
  if (pos>=Nc) {
    if (g_kicked_out>g) {	// we should remember that g-value
      g_kicked_out = g;
    }
    return;
  }
  // memorize the threshold which will be kicked out
  int free_id = -1;
  if (N_SRL==Nc) {
    free_id = smpl_ref_list[Nc-1];
    g_kicked_out = g_list[Nc-1];
  }
  else {
    free_id = N_SRL;
    ++N_SRL;
  }
  // create a gap in the list
    {
      register int rt = smpl_ref_list[pos];
      for (tuint i = pos+1;i<N_SRL;++i) {
	register int rt2 = smpl_ref_list[i];
	smpl_ref_list[i] = rt;
	rt = rt2;
      }
    }{
      register tdouble rt = g_list[pos];
      for (tuint i = pos+1;i<N_SRL;++i) {
	register tdouble rt2 = g_list[i];
	g_list[i] = rt;
	rt = rt2;
      }
    }
  // fill the gap :)
    g_list[pos] = g;
    smpl_ref_list[pos] = free_id;
    tdouble* dp = smpl_list+N_RV*free_id;
    for (tuint i=0;i<N_RV;++i) {
      *(dp++) = y_ptr[i];
    }
}

SSS_csm_MCMC_base::SSS_csm_MCMC_base(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne)
: N_RV(N_RV), ID_Ns(ID_Ns), Nc(Nc),
  y_prev(N_RV),y_prop(N_RV),yp_prev(&y_prev[0]),yp_prop(&y_prop[0]),
  RndBox(RndBox),adptvCtrl(adptvCtrl),bn_box(bn_box),bne(bne),fung(fung),smpl_ref_list(NULL),g_list(NULL),smpl_list(NULL)
{
  // allocate chain structure
    chain_structure = new cs[ID_Ns*Nc];
}

SSS_csm_MCMC_base::~SSS_csm_MCMC_base()
{
  delete [] chain_structure;
}


void SSS_csm_MCMC_base::register_samples(const SSS_sample_list* const smpl_cond_v, SSS_sample_list* smpl_new_v)
{
  smpl_cond = smpl_cond_v;
  smpl_new = smpl_new_v;
  smpl_ref_list = smpl_cond_v->get_reference_smpl_ref_list();
  g_list = smpl_cond_v->get_reference_g_list();
  smpl_list = smpl_cond_v->get_reference_smpl_list();
  g_treshold = smpl_cond_v->get_treshold_g();
  
  smpl_new->clear();
}

void SSS_csm_MCMC_base::register_new_smpl_line(const tuint curr_ID)
{
  g_last = g_list[curr_ID];
  
  const tdouble* sp = &(smpl_list[smpl_ref_list[curr_ID]*N_RV]);
  for (tuint i=0;i<N_RV;++i,++sp) {
    yp_prev[i] = *sp;
  }
  
  smpl_new->register_sample(g_last,yp_prev);
  
  chain_structure[curr_ID*ID_Ns].changed = false;
  chain_structure[curr_ID*ID_Ns].g = g_last;
}

void SSS_csm_MCMC_base::generate_smpl(const SSS_sample_list*const smpl_cond_v, SSS_sample_list* smpl_new_v, tdouble& acr, tdouble& nacr1, tdouble& velo, FlxProgress& prg)
{
  register_samples(smpl_cond_v,smpl_new_v);
  tuint lvl2_nac_global = 0;
  tuint lvl1_nac_global = 0;
  tdouble velo_global = 0.0;
  tuint adupd_counter = 0;
  tuint adupd_acrN = 0;
  tuint adupd_counter_N = Nc/adptvCtrl.get_maxUpdatesPerCStep();
  tVec y_smpl(N_RV);
  // activate progress-bar
    prg.start(Nc);
  for (tuint curr_ID=0;curr_ID<Nc;++curr_ID) {
    register_new_smpl_line(curr_ID);
    tuint lvl2_nac = 0;	// not accepted samples (level 2)
    tdouble velo = 0.0;
    for (tuint curr_ID_Ns=1;curr_ID_Ns<ID_Ns;++curr_ID_Ns) {
      // propose a sample-vector
	const bool diff = propose_sample(y_smpl);
      // check proposed sample-vector
	bool changed = false;
	if (diff) {
	  RndBox.set_y_Vec(y_prop);
	  const tdouble g=fung->calc(); 
	  if (g<g_treshold) {
	    changed = true;
	    g_last = g;
	    tdouble vt = 0.0;
	    for (tuint i=0;i<N_RV;++i) {
	      vt += pow2(yp_prev[i]-yp_prop[i]);
	    }
	    velo += sqrt(vt);
	    y_prev = y_prop;
	  } else {
	    ++lvl2_nac;
	  } 
	} else {
	  ++lvl1_nac_global;
	}
	chain_structure[curr_ID*ID_Ns+curr_ID_Ns].changed = changed;
	chain_structure[curr_ID*ID_Ns+curr_ID_Ns].g = g_last;
	smpl_new->register_sample(g_last,yp_prev);
    }
    lvl2_nac_global += lvl2_nac;
    velo /= (ID_Ns-1)*sqrt(tdouble(N_RV));
    velo_global += velo;
    // adaptive control
      adupd_counter++;
      adupd_acrN += ID_Ns-1-lvl2_nac;
      if (adupd_counter>=adupd_counter_N) {
	switch (adptvCtrl.requires_adptv_step(tdouble(adupd_acrN)/tdouble(adupd_counter*(ID_Ns-1)))) {
	  case 0:
	    break;
	  case 1:
	    adptv_acr_up();
	    break;
	  case 2:
	  adptv_acr_down();
	    break;
	};
	adupd_counter = 0;
	adupd_acrN = 0;
      }
    prg.tick(curr_ID);
  }
  // deactivate progress-bar
    prg.stop();
  velo_global /= Nc;
  velo = velo_global;
  acr = 1.0-tdouble(lvl2_nac_global)/tdouble((ID_Ns-1)*Nc);
  if (lvl1_nac_global==0) nacr1 = 0.0;
  else nacr1 = tdouble(lvl1_nac_global)/tdouble((ID_Ns-1)*Nc);
  // evaluate the chain structure
  if (bne) {
    const tdouble g_limit = smpl_new->get_treshold_g();
    SSS_bne_MCMC* bne = new SSS_bne_MCMC(Nc,ID_Ns,chain_structure,(g_limit>0.?g_limit:0.0));
    bn_box.add_bne(bne);
  }
}

SSS_csm_MCMC_base* SSS_csm_MCMC_base::new_object(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, const std::string& smplTypeStr, const std::string& kernelName, const tdouble h, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl_V, SSS_bne_Box& bn_box, const bool bne, const bool errSerious)
{
  if (smplTypeStr=="cwmh") {
    return new SSS_csm_cwMH_MCMC(N_RV,ID_Ns,Nc,fung,kernelName,h,RndBox,adptvCtrl_V,bn_box,bne);
  } else if (smplTypeStr=="csusmh") {
    return new SSS_csm_csusMH_MCMC(N_RV,ID_Ns,Nc,fung,h,RndBox,adptvCtrl_V,bn_box,bne);
  } else if (smplTypeStr=="mci") {
    return new SSS_csm_mci(N_RV,ID_Ns,Nc,fung,RndBox,adptvCtrl_V,bn_box,bne);
  } else {
    std::ostringstream ssV;
    ssV << "Unknown keyword of 'smpl_type': " << smplTypeStr;
    FlxError(errSerious,"SSS_csm_MCMC_base::new_object",ssV.str()); return NULL;
  }
}


SSS_csm_cwMH_MCMC::SSS_csm_cwMH_MCMC(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, const std::string& kernelName, const tdouble h, FlxRndBox& RndBox,SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne)
: SSS_csm_MCMC_base(N_RV,ID_Ns,Nc,fung,RndBox,adptvCtrl,bn_box,bne),kernel(FlxRndKernel_base::read(kernelName,false))
{
  kernel->set_h(h);
}

SSS_csm_cwMH_MCMC::~SSS_csm_cwMH_MCMC()
{
  delete kernel;
}

const bool SSS_csm_cwMH_MCMC::propose_sample(tVec& y_smpl)
{
  
  tdouble* yp_prev_p = yp_prev;
  tdouble* yp_prop_p = yp_prop;
  const tdouble* const yp_smpl = &y_smpl[0];
  RndBox.RndCreator.gen_smp(y_smpl);
  bool diff = false;	// true if sample is different from seed
  // for each coordinate ...
    for (tuint i=0;i<N_RV;++i,++yp_prev_p,++yp_prop_p) {
      const tdouble xi = kernel->transform_y2x(yp_smpl[i])+*yp_prev_p;
      tdouble r = rv_phi(xi)/rv_phi(*yp_prev_p);
      if (r>=1.0) {
	*yp_prop_p = xi;
	diff = true;
      } else {
	const tdouble pr = RndBox.RndCreator.gen_smp_uniform();
	if (pr<=r) {
	  *yp_prop_p = xi;
	  diff = true;
	} else {
	  *yp_prop_p = *yp_prev_p;
	}
      }
    }
  return diff;
}

void SSS_csm_cwMH_MCMC::write_adaptive_info(std::ostream& sout)
{
  sout << "\th=" << kernel->get_h();
}

void SSS_csm_cwMH_MCMC::adptv_acr_up()
{
  kernel->set_h(kernel->get_h()*(1.0-adptvCtrl.get_factor()));
}

void SSS_csm_cwMH_MCMC::adptv_acr_down()
{
  kernel->set_h(kernel->get_h()*(1.0+adptvCtrl.get_factor()));
}

SSS_csm_mci::SSS_csm_mci(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, FlxRndBox& RndBox,SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne)
: SSS_csm_MCMC_base(N_RV,ID_Ns,Nc,fung,RndBox,adptvCtrl,bn_box,bne)
{
  
}

const bool SSS_csm_mci::propose_sample(tVec& y_smpl)
{
  RndBox.RndCreator.gen_smp(y_prop);
  return true;
}

SSS_csm_csusMH_MCMC::SSS_csm_csusMH_MCMC(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, const tdouble rho, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne)
: SSS_csm_MCMC_base(N_RV,ID_Ns,Nc,fung,RndBox,adptvCtrl,bn_box,bne),rho(rho),sD(sqrt(1.0-pow2(rho)))
{
}

const bool SSS_csm_csusMH_MCMC::propose_sample(tVec& y_smpl)
{
  // for each coordinate ...
    tdouble* yp_prev_p = yp_prev;
    tdouble* yp_prop_p = yp_prop;
    const tdouble* const yp_smpl = &y_smpl[0];
    RndBox.RndCreator.gen_smp(y_smpl);
    for (tuint i=0;i<N_RV;++i,++yp_prev_p,++yp_prop_p) {
      *yp_prop_p = yp_smpl[i]*sD+*yp_prev_p*rho;
    }
  return true;		// a new state has been proposed for sure
}

void SSS_csm_csusMH_MCMC::write_adaptive_info(std::ostream& sout)
{
  sout << "\trho=" << rho;
}

void SSS_csm_csusMH_MCMC::adptv_acr_up()
{
  rho = rho * (1.0+adptvCtrl.get_factor());
  if (rho>0.99) rho = 0.99;
  sD = sqrt(1.0-pow2(rho));
}

void SSS_csm_csusMH_MCMC::adptv_acr_down()
{
  rho = rho * (1.0-adptvCtrl.get_factor());
  sD = sqrt(1.0-pow2(rho));
}

SSS_adaptive_ctrl::SSS_adaptive_ctrl(FlxFunction* factor, FlxFunction* lower, FlxFunction* upper, FlxFunction* maxUpdatesPerCStep)
:factor_d(0.0),lower_d(0.0),upper_d(0.0),
factor(factor),lower(lower),upper(upper),maxUpdatesPerCStep(maxUpdatesPerCStep)
{

}

SSS_adaptive_ctrl::~SSS_adaptive_ctrl()
{
  delete maxUpdatesPerCStep;
  delete upper;
  delete lower;
  delete factor;
}

void SSS_adaptive_ctrl::eval()
{
  factor_d = factor->cast2positive_or0();
  lower_d = lower->cast2positive_or0();
  upper_d = upper->cast2positive_or0();
  if (upper_d < lower_d) {
    std::ostringstream ssV;
    ssV << "Lower bound must be smaller than the upper bound.";
    throw FlxException("SSS_adaptive_ctrl::eval",ssV.str());
  }
}

const int SSS_adaptive_ctrl::requires_adptv_step(const tdouble acr) const
{
  if (acr<lower_d) return 1;
  else if (acr>upper_d) return 2;
  else return 0;
}

const tuint SSS_adaptive_ctrl::get_maxUpdatesPerCStep()
{
  return maxUpdatesPerCStep->cast2tuint();
}

FlxObjSubSetSim::FlxObjSubSetSim(bool dolog, tdouble& theResult, FlxFunction* funNsM, FlxFunction* funNc, FlxFunction* fung, FlxString* smplTypeStr, FlxString* kernelName, FlxFunction* h, SSS_adaptive_ctrl* adptvCtrl, const tdouble cgkf, const bool bne, bool verboseLog)
: FlxObjBase(dolog), theResult(theResult), theResult_bne(NULL), funNsM(funNsM),funNc(funNc), fung(fung),smplTypeStr(smplTypeStr),kernelName(kernelName),h(h),adptvCtrl(adptvCtrl),cgkf(cgkf),bne(bne),verboseLog(verboseLog)
{
  if (bne) {
    std::string res_name = data->ConstantBox.get(&theResult);
    res_name += "_bne";
    theResult_bne = data->ConstantBox.get(res_name,true);
    res_name += "_cov";
    theResult_bne_cov = data->ConstantBox.get(res_name,true);
  }
}

FlxObjSubSetSim::~FlxObjSubSetSim()
{
  delete funNsM;
  delete funNc;
  delete fung;
  delete smplTypeStr;
  delete kernelName;
  delete h;
  delete adptvCtrl;
}

void FlxObjSubSetSim::task()
{
  GlobalVar.slog(4) << "subsetsim: performing a subset simulation." << std::endl; 
  const tuint Nc = funNc->cast2tuint(false);
  const tuint NsM = funNsM->cast2tuint(false);
  const tuint Ns = Nc*NsM;
  const tuint N_runs_max = 100;
  tuint N_runs = 0;
  tdouble* iterN = data->ConstantBox.get("sss_iter",true);
    *iterN = 0;
  FlxRndBox& RndBox = data->RndBox;
  SSS_bne_Box bn_box;
  SSS_csm_MCMC_base* condgen = SSS_csm_MCMC_base::new_object(RndBox.get_NumbRnd(),NsM,Nc,fung,smplTypeStr->eval_word(true),kernelName->eval_word(true),h->cast2positive(false),RndBox,*adptvCtrl,bn_box,bne,false);
  SSS_sample_list smpl_1(RndBox.get_NumbRnd(),Ns,Nc,cgkf);
  SSS_sample_list smpl_2(RndBox.get_NumbRnd(),Ns,Nc,cgkf);
  SSS_sample_list* sampling = &smpl_1;
  SSS_sample_list* conditioning = &smpl_1;
  const tdouble*const y_ptr = &(RndBox.get_y_Vec()[0]);
  GlobalVar.slog(4) << "  Ns=" << GlobalVar.Double2String(Ns) << "; p=" << GlobalVar.Double2String(1.0/tdouble(NsM)) << std::endl; 
  // register progress-bar
    std::ostream &op = std::cout;
    FlxProgress prg(op);
      
  // do MCI in a first step
    GlobalVar.slog(4) << "  iter " << N_runs << ":\t";
    // register the progress-bar
      op.flush();
      prg.start(Nc);
    for (tuint i=0;i<Ns;++i) {
      RndBox.gen_smp(); 
      const tdouble g=fung->calc(); 
      sampling->register_sample(g,y_ptr);
      if (i%NsM==0) prg.tick(i/NsM);
    }
    // deactivate the progress-bar
      prg.stop();
    if (bne) {
      if (conditioning->get_g_lastPoint()>0.0) {
	bn_box.add_bne(new SSS_bne_MCI(Nc,Ns));
      } else {
	bn_box.add_bne(new SSS_bne_MCI(sampling->get_N_hits(),Ns));
      }
    }
    sampling = &smpl_2;
    GlobalVar.slog(4) << "gt=" << GlobalVar.Double2String(conditioning->get_treshold_g());
    if (bne) GlobalVar.slog(4) << "\tC.o.V.=" << GlobalVar.Double2String(bn_box.get_last().sigma_Pf()/bn_box.get_last().E_Pf());
    ++N_runs;
    
  // perform a conditioned sampling
    while (conditioning->get_g_lastPoint()>0.0 && N_runs<N_runs_max) {
      *iterN = N_runs;
      // detect multiple occurrences
	if (verboseLog && N_runs>1) {
	  tuint Nsame_seq, Nmax_seq, Ntotal,Nconsid;
	  tdouble gt = conditioning->get_g_lastPoint()+1.;
	  conditioning->find_multiples(Nsame_seq,Nmax_seq,Ntotal,Nconsid,gt);
	  GlobalVar.slog(4) << "\tmultiple:" << Nsame_seq << "/" << Nmax_seq << "/" << tuint(round_flx((1.0-tdouble(Ntotal)/tdouble(Nc))*100)) << "% ";
// 	  if (N_runs>2) {
	    gt = conditioning->get_g_lastPoint();
	    sampling->find_multiples(Nsame_seq,Nmax_seq,Ntotal,Nconsid,gt);
	    GlobalVar.slog(4) << "\t[seed(" << tuint(round_flx((tdouble(Nconsid)/tdouble(Nc))*100)) << "%):" << Nsame_seq << "/" << Nmax_seq;
	    const tuint ti = tuint(round_flx((1.0-tdouble(Ntotal)/tdouble(Nconsid))*100));
	    if (ti!=100) {
	      GlobalVar.slog(4) << "/" << ti << "%";
	    }
	    GlobalVar.slog(4) << "]";
// 	  }
	}
	GlobalVar.slog(4) << std::endl;
      // generate a new set of samples
	GlobalVar.slog(4) << "  iter " << N_runs << ":";
	// adaptivity ...
	  adptvCtrl->eval();
	  if (adptvCtrl->is_adaptive()) {
	    condgen->write_adaptive_info(GlobalVar.slog(4));
	  }
	  // visibility of progress-bar
	    GlobalVar.slog(4) << '\t';
	    op.flush();
	tdouble acr, nacr1, velo;
	condgen->generate_smpl(conditioning,sampling,acr,nacr1,velo,prg);
	// make sure the list is full
	  #if FLX_DEBUG
	  if (sampling->is_listFull() == false) {
	    std::ostringstream ssV;
	    ssV << "Not enough seed-values found.";
	    throw FlxException("FlxObjSubSetSim::task_1",ssV.str());
	  }
	  #endif
      // interchange the lists 
	SSS_sample_list* tsl = conditioning;
	conditioning = sampling;
	sampling = tsl;
      // make sure progress-bar is not visible anymore
	  prg.stop(); op.flush();
      ++N_runs;
      GlobalVar.slog(4) << "acr=" << GlobalVar.Double2String(acr);
      if (nacr1>GlobalVar.TOL()) {
	GlobalVar.slog(4) << "(" << acr-nacr1 << ")";
      }
      GlobalVar.slog(4) << "\tvelo=" << GlobalVar.Double2String(velo);
      if (conditioning->get_g_lastPoint()>0.) {
	GlobalVar.slog(4) << "\tgt=" << GlobalVar.Double2String(conditioning->get_treshold_g());
      } else {
	GlobalVar.slog(4) << "\tpi=" << GlobalVar.Double2String(conditioning->get_Pf());
      }
      if (bne) GlobalVar.slog(4) << "\tC.o.V.=" << GlobalVar.Double2String(bn_box.get_last().sigma_Pf()/bn_box.get_last().E_Pf());
    }
    GlobalVar.slog(4) << std::endl;
  if (conditioning->get_g_lastPoint()>0.0) {
    std::ostringstream ssV;
    ssV << "Simulation did not converge within " << N_runs << " runs.";
    throw FlxException_NeglectInInteractive("FlxObjSubSetSim::task_2",ssV.str());
  }
  const tdouble Pf = (1.0/pow(tdouble(NsM),int(N_runs)-1))*conditioning->get_Pf();
  theResult = Pf;
  GlobalVar.slog(4) << "  result=" << GlobalVar.Double2String(Pf) << std::endl;
  // Bayesian estimate:
  if (bne) {
    const tdouble bn_e = bn_box.E_Pf();
    *theResult_bne = bn_e;
    const tdouble bn_CoV = sqrt(bn_box.Cov_Pf())/bn_e;
    *theResult_bne_cov = bn_CoV;
    GlobalVar.slog(4) << "  Bayesian estimate: E[Pf]=" << GlobalVar.Double2String(bn_e) << "; C.o.V.: " << GlobalVar.Double2String(bn_CoV) << std::endl;
  }
  // free allocated dynamic memory
    delete condgen;
}

FlxObjReadSubSetSim::FlxObjReadSubSetSim()
{
  AllDefParaBox->insert(new FlxOptionalParaFlxString("cwmh","sss::smpl_type"));
  ParaBox.insert("smpl_type", "sss::smpl_type" );
  AllDefParaBox->insert(new FlxOptionalParaFlxString("uniform","sss::kernel"));
  ParaBox.insert("kernel", "sss::kernel" );
  AllDefParaBox->insert(new FlxOptionalParaFun(1.0,"sss::kernel_h"));
  ParaBox.insert("kernel_h", "sss::kernel_h" );
  AllDefParaBox->insert(new FlxOptionalParaFun(0.0,"sss::cgkf"));
  ParaBox.insert("cgkf", "sss::cgkf" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(0.0,"sss::adaptive_factor"));
  ParaBox.insert("adaptive_factor", "sss::adaptive_factor" );
  AllDefParaBox->insert(new FlxOptionalParaFun(0.3,"sss::adaptive_lower"));
  ParaBox.insert("adaptive_lower", "sss::adaptive_lower" );
  AllDefParaBox->insert(new FlxOptionalParaFun(0.5,"sss::adaptive_upper"));
  ParaBox.insert("adaptive_upper", "sss::adaptive_upper" );
  AllDefParaBox->insert(new FlxOptionalParaFun(10,"sss::adaptive_numbstep"));
  ParaBox.insert("adaptive_numbstep", "sss::adaptive_numbstep" );
  
  // determining probability of failure?
    AllDefParaBox->insert(new FlxOptionalParaBool(false,"sss::bne"));
    ParaBox.insert("bne", "sss::bne" );
}

FlxObjBase* FlxObjReadSubSetSim::read () {
  tdouble* d1=NULL; FlxFunction* funNsM=NULL; FlxFunction* funNc=NULL; FlxFunction* fung=NULL;
  try {    
    reader->getChar('(',false);
    // result-constant
      const std::string cname = reader->getWord(true,false);
      data->ConstantBox.declareC(cname);
      d1 = data->ConstantBox.get(cname);
      reader->getChar(';',false);
    // function for number of samples to condition on
      funNc = new FlxFunction(funReader,false);
      reader->getChar(';',false);
    // function for factor of number of samples per iteration more than Nc
      funNsM = new FlxFunction(funReader,false);
      reader->getChar(';',false);
    // function to integrate
      fung = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    // determine cgk-factor
      tdouble cgkf;
      {
	FlxFunction* cgkf_fun = get_optPara_FlxFunction("cgkf");
	cgkf = cgkf_fun->cast2positive_or0(true);
	delete cgkf_fun;
	if (cgkf>=1.0) {
	  std::ostringstream ssV;
	  ssV << "'cgkf' must be a value in the range [0;1[ and not " << cgkf << ".";
	  throw FlxException_NeglectInInteractive("FlxObjReadSubSetSim::read",ssV.str());
	}
      }
    return new FlxObjSubSetSim(get_doLog(), *d1, funNsM, funNc, fung,get_optPara_FlxString("smpl_type"),get_optPara_FlxString("kernel"),get_optPara_FlxFunction("kernel_h"), new SSS_adaptive_ctrl(get_optPara_FlxFunction("adaptive_factor"),get_optPara_FlxFunction("adaptive_lower"),get_optPara_FlxFunction("adaptive_upper"),get_optPara_FlxFunction("adaptive_numbstep")),cgkf,get_optPara_bool("bne"),get_verboseLog());
  } catch (FlxException &e) {
    if (funNsM) delete funNsM;
    if (funNc) delete funNc;
    if (fung) delete fung;
    throw;
  }
}
 
const bool FlxObjReadFORM_base::get_verboseLog()
{
  return *(static_cast<bool*>(ParaBox.get("vlog")->get() ));
}

const tuint FlxObjReadFORM_base::get_maxiter()
{
  FlxFunction *f1 = get_optPara_FlxFunction("maxiter");
  tuint i1 = f1->cast2tuint();
  delete f1;
  return i1;
}

const int FlxObjReadFORM_base::get_fd_method()
{
  FlxFunction *f1 = get_optPara_FlxFunction("fd_method");
  int i1 = f1->cast2int();
  delete f1;
  return i1;
}

const int FlxObjReadFORM_base::get_opt_method()
{
  FlxFunction *f1 = get_optPara_FlxFunction("opt_method");
  int i1 = f1->cast2int();
  delete f1;
  return i1;
}

FlxObjReadFORM_base::FlxObjReadFORM_base()
{
  // optional parameters
    // maxiter
      AllDefParaBox->insert(new FlxOptionalParaFun(100,"form::maxiter"));
      ParaBox.insert("maxiter", "form::maxiter" );
    // fdstep
      AllDefParaBox->insert(new FlxOptionalParaFun(1.0,"form::fdstep"));
      ParaBox.insert("fdstep", "form::fdstep" );
    // eps1
      AllDefParaBox->insert(new FlxOptionalParaFun(1e-5,"form::eps1"));
      ParaBox.insert("eps1", "form::eps1" ); 
    // eps2
      AllDefParaBox->insert(new FlxOptionalParaFun(1e-5,"form::eps2"));
      ParaBox.insert("eps2", "form::eps2" );
    // ihlrf_lambda_start
      AllDefParaBox->insert(new FlxOptionalParaFun(1.0,"form::ihlrf_lambda_start"));
      ParaBox.insert("ihlrf_lambda_start", "form::ihlrf_lambda_start" );
    // ihlrf_epsilon
      AllDefParaBox->insert(new FlxOptionalParaFun(0.4,"form::ihlrf_epsilon"));
      ParaBox.insert("ihlrf_epsilon", "form::ihlrf_epsilon" );
    // ihlrf_reduce
      AllDefParaBox->insert(new FlxOptionalParaFun(0.5,"form::ihlrf_reduce"));
      ParaBox.insert("ihlrf_reduce", "form::ihlrf_reduce" );
    // xstart
      FlxMtxFun* mtxvalue = new FlxMtxFun(new FunRVmean());
      AllDefParaBox->insert(new FlxOptionalParaMtxFun(mtxvalue,"form::xstart"));
      ParaBox.insert("xstart", "form::xstart" );
    // dx_min
      mtxvalue = new FlxMtxFun(new FunMtxMult(new FunNewMtx(new FunNumber(1.0),new FunNumber(1.0),new FunNumber(1e-6)),new FunRVstddev()));
      AllDefParaBox->insert(new FlxOptionalParaMtxFun(mtxvalue,"form::dxmin"));
      ParaBox.insert("dxmin", "form::dxmin" ); 
    // dxdyanalytical
      AllDefParaBox->insert(new FlxOptionalParaBool(true,"form::dxdyanalytical"));
      ParaBox.insert("dxdyanalytical", "form::dxdyanalytical" );
    // fd_method
      AllDefParaBox->insert(new FlxOptionalParaFun(1,"form::fd_method"));
      ParaBox.insert("fd_method", "form::fd_method" );
    // opt_method
      AllDefParaBox->insert(new FlxOptionalParaFun(2,"form::opt_method"));
      ParaBox.insert("opt_method", "form::opt_method" );
}

FlxObjReadFORM_pdf::FlxObjReadFORM_pdf()
{
  // optional parameters
    // lbound
      AllDefParaBox->insert(new FlxOptionalParaFun(100,"form_pdf::intervals"));
      ParaBox.insert("intervals", "form_pdf::intervals" );
    // Default verbose-setting
      AllDefParaBox->insert(new FlxOptionalParaBool(false,"form_pdf::verbose"));
      ParaBox.insert("vlog", "form_pdf::verbose" );
}

FlxObjBase* FlxObjReadFORM_pdf::read()
{
  reader->getChar('(',false);
  FlxFunction* rvfun = new FlxFunction(funReader,false);
  FlxFunction *lbound=NULL, *ubound=NULL;
  try {
    reader->getChar(';',false);
    reader->getChar('[',false);
    lbound = new FlxFunction(funReader,false);
    reader->getChar(';',false);
    ubound = new FlxFunction(funReader,false);
    reader->getChar(']',false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjFORM_pdf(get_doLog(),rvfun,lbound,ubound,get_optPara_FlxFunction("intervals"),get_optPara_FlxMtxFun("xstart"),get_optPara_FlxFunction("fdstep"), get_optPara_FlxFunction("eps1"), get_optPara_FlxFunction("eps2"), get_optPara_FlxFunction("ihlrf_lambda_start"), get_optPara_FlxFunction("ihlrf_epsilon"), get_optPara_FlxFunction("ihlrf_reduce"), get_maxiter(),get_verboseLog(),get_stream(),get_optPara_bool("dxdyanalytical"), get_optPara_FlxMtxFun("dxmin"), get_fd_method(), get_opt_method() );
  } catch (FlxException &e) {
    delete rvfun;
    if (lbound) delete lbound;
    if (ubound) delete ubound;
    throw;
  }
}

FlxObjReadFORM::FlxObjReadFORM()
{
  // optional parameters
    // betaDP -> name of const variable where to store
      AllDefParaBox->insert(new FlxOptionalParaFlxString("","form::betadp"));
      ParaBox.insert("betadp", "form::betadp" ); 
    // Default verbose-setting
      AllDefParaBox->insert(new FlxOptionalParaBool(true,"flxlog::verbose"));
      ParaBox.insert("vlog", "flxlog::verbose" );
}

FlxObjBase* FlxObjReadFORM::read()
{
  reader->getChar('(',false);
  // result-constant
    std::string cnamey = reader->getWord(false,false);
    data->ConstMtxBox.declareC(cnamey);
    reader->getChar(';',false);
    std::string cnamex = reader->getWord(false,false);
    data->ConstMtxBox.declareC(cnamex);
    reader->getChar(';',false);
  FlxFunction* lsf = new FlxFunction(funReader,false);
  try {
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjFORM( get_doLog(), cnamey, cnamex, get_optPara_FlxMtxFun("xstart"), lsf, get_optPara_FlxFunction("fdstep"), get_optPara_FlxFunction("eps1"), get_optPara_FlxFunction("eps2"), get_optPara_FlxFunction("ihlrf_lambda_start"), get_optPara_FlxFunction("ihlrf_epsilon"), get_optPara_FlxFunction("ihlrf_reduce"), get_maxiter(), get_optPara_FlxString("betadp"), get_verboseLog(), get_optPara_bool("dxdyanalytical"), get_optPara_FlxMtxFun("dxmin"), get_fd_method(), get_opt_method() );
  } catch (FlxException &e) {
    delete lsf;
    throw;
  }
}

FlxObjFORM_base::~FlxObjFORM_base()
{
  if (LSF!=NULL) delete LSF;
  delete fdstep;
  delete eps1;
  delete eps2;
  delete iHLRF_lambda_start;
  delete iHLRF_epsilon;
  delete iHLRF_reduce;
  delete dx_min;
}

void FlxObjFORM_base::update_Start()
{
  DIM = data->RndBox.get_NumbRnd();
  if (DIM==0) {
    std::ostringstream ssV;
    ssV << "FORM cannot be executed because no random variables exist.";
    throw FlxException("FlxObjFORM_base::update_Start_1", ssV.str() );
  }
  data->RndBox.corr_create(true);
}

tVec FlxObjFORM_base::do_FORM(tVec& x, tVec& y, tdouble& beta_new)
{
  // check parameters
    switch (fd_method) {
      case 1:
      case 2:
      case 3:
	break;
      default:
	std::ostringstream ssV;
	ssV << "Finite difference method with ID '" << fd_method << "' not known.";
	throw FlxException("FlxObjFORM_base::do_FORM_1", ssV.str() );
    }
    switch (opt_method) {
      case 1:
      case 2:
	break;
      case 3:
	throw FlxException_NotImplemented("FlxObjFORM_base::do_FORM_2");
	break;
      default:
	std::ostringstream ssV;
	ssV << "Optimization algorithm with ID '" << fd_method << "' not known.";
	throw FlxException("FlxObjFORM_base::do_FORM_3", ssV.str() );
    }
  const tVec dxmin = dx_min->calcVec();
  // calculate z_0  (value of LSF at y=0)
    tdouble z_0, z_old;
    tVec d(y.size());		// iHLRF direction vector
    tVec y_new(y.size());	// iHLRF new position vector
    d = 0.0;
    data->RndBox.set_y_Vec(d);
    z_0 = LSF->calc();
    if (NormMax(data->RndBox.get_x_Vec()-x)<=GlobalVar.TOL()) {
      z_old = z_0;
      y = d;
    } else {
      data->RndBox.set_x_Vec(x);
      z_old = LSF->calc();
      y = data->RndBox.get_y_Vec();
    }
  // Write log-message
    if (verboseLog) {
      GlobalVar.slog(4) << " Start point: x=" << x << "; y=" << y << ";" << std::endl;
    }
    tuint loopc = 0;
  beta_new = Norm2(y);
  
  tVec xhelp(DIM), dzdy(DIM);
  tdouble beta_old, zo, zu,fd,fd_max,dx,lambda=-9999.;
  const tdouble sqrtEps = GlobalVar.sqrtEps;	// square root of machine precision
  FlxMtxSparsLTri* dxdy=NULL;
  try {
    do {
      beta_old = beta_new;
      fd = fdstep->calc()*sqrtEps;
      fd_max = fd;
      
      if (dxdyAnalytical) {
	// calc dx/dy
	  dxdy = data->RndBox.calc_Jinv();
	// calculate dz/dx
	for (tuint i = 0; i < DIM; ++i) {
	  xhelp = x;
	  // choose an appropriate step-size for finite-difference
	    if (fd*fabs(x[i])<dxmin[i] && fd*fabs(z_old)<dxmin[i]) {
	      if (fd_method!=3) {	// forward (central)
		xhelp[i] += dxmin[i];
		dx = xhelp[i]-x[i];
	      } else {			// backward
		xhelp[i] -= dxmin[i];
		dx = x[i]-xhelp[i];
	      }
	      register tdouble tr = dx/x[i];
	      if (tr>fd_max) fd_max = tr;
	    } else if (fabs(xhelp[i]) >= fabs(z_old)) {	// x-help determines step-size
	      if (fd_method!=3) {	// forward (central)
		xhelp[i] = (x[i]>=0.0)?((1.0+fd)*x[i]):((1.0-fd)*x[i]);
		dx = xhelp[i]-x[i];
	      } else {			// backwards
		xhelp[i] = (x[i]>=0.0)?((1.0-fd)*x[i]):((1.0+fd)*x[i]);
		dx = x[i]-xhelp[i];
	      }
	    } else {			// z_old determines step-size
	      if (fd_method!=3) {	// forward (central)
		xhelp[i] += fd*fabs(z_old);
		dx = xhelp[i]-x[i];
	      } else {			// backwards
		xhelp[i] -= fd*fabs(z_old);
		dx = x[i]-xhelp[i];
	      }
	      register tdouble tr = dx/x[i];
	      if (tr>fd_max) fd_max = tr;
	    }
	    #if FLX_DEBUG
	      if (dx<=0.0) throw FlxException_Crude("FlxObjFORM_base::do_FORM_4");
	    #endif
	  if ( data->RndBox.check_xVec( xhelp ) ) {	// check if xhelp is in domain
	    data->RndBox.set_x_Vec(xhelp);
	    // calc LSF
	      zo = LSF->calc();
	    if (fd_method!=2) {
	      if (fd_method==1) {
		dzdy[i] = (zo-z_old)/dx;
	      } else {
		dzdy[i] = (z_old-zo)/dx;
	      }
	    } else {		// central difference
	      xhelp = x+x-xhelp;
	      if ( data->RndBox.check_xVec( xhelp ) ) {
		data->RndBox.set_x_Vec(xhelp);
		// calc LSF
		  zu = LSF->calc();
		// calc dz/dx(i)
		  dzdy[i] = (zo-zu)/(2.0*dx);
	      } else {
		dzdy[i] = (zo-z_old)/dx;
	      }
	    }
	  } else {
	    if (fd_method!=3) {		// forward (central) -> use backwards
	      xhelp[i] = x[i] - dx;
	    } else {			// backwards -> use forwards
	      xhelp[i] = x[i] + dx;
	    }
	    #if FLX_DEBUG
	      if (data->RndBox.check_xVec( xhelp ) == false) {
		throw FlxException_Crude("FlxObjFORM_base::do_FORM_5");
	      }
	    #endif
	    data->RndBox.set_x_Vec(xhelp);
	    zu = LSF->calc();
	    if (fd_method!=3) {
	      dzdy[i] = (z_old-zu)/dx;
	    } else {
	      dzdy[i] = (zu-z_old)/dx;
	    }
	  }
	}
	// calculate dz/dy
	  dxdy->TransMultVec(dzdy);
	  delete dxdy; dxdy = NULL;
      } else {
	throw FlxException_NotImplemented("FlxObjFORM_base::do_FORM_6");
// 	// calculate dz/dy
// 	for (tuint i = 0; i < DIM; ++i) {
// 	  xhelp = y;
// 	  if (std::fabs(y[i]) < fd ) xhelp[i] = y[i] + fd;   // make sure its different from zero
// 	  else xhelp[i] = (1.0+fd)*y[i];
// 	  dx = xhelp[i]-y[i];
// 	  if ( data->RndBox.check_xVec( xhelp ) ) {
// 	    data->RndBox.set_y_Vec(xhelp);
// 	    // calc LSF
// 	      zo = LSF->calc();
// 	    xhelp = y+y-xhelp;
// 	    if ( data->RndBox.check_xVec( xhelp ) ) {
// 	      data->RndBox.set_y_Vec(xhelp);
// 	      // calc LSF
// 		zu = LSF->calc();
// 	      // calc dz/dx(i)
// 		dzdy[i] = (zo-zu)/(2.0*dx);
// 	    } else {
// 	      dzdy[i] = (zo-z_old)/dx;
// 	    }
// 	  } else {
// 	    xhelp[i] = y[i] - dx;
// 	    data->RndBox.set_y_Vec(xhelp);
// 	    zu = LSF->calc();
// 	    dzdy[i] = (z_old-zu)/dx;
// 	  }
// 	}
      }
      
      if (NormMax(dzdy) <= GlobalVar.TOL() ) {
	for (size_t i = 0; i < DIM; ++i) {
	  y[i] = rv_normal();
	}
	if (verboseLog) {
	  GlobalVar.alert.alert("FlxObjFORM_base::do_FORM_7", "Warning: zero_Gradient -> Initialize random seed." );
	}
      } else {
	dx=(dot(dzdy,y)-z_old)/(dot(dzdy,dzdy));  // this is just a factor
	if (opt_method==1) {
	  y = dzdy;
	  y *= dx;
	} else {
	  d = dzdy;
	  d *= dx;
	  d -= y;
	  // determine c^k
	    tdouble c = Norm2(y)/Norm2(dzdy);
	    if (fabs(z_old)>=0.001*fabs(z_0)) {
	      const tdouble c2 = 0.5*Norm2_NOroot(y+d)/fabs(z_old);
	      if (c2>c) c = c2;
	    }
	    c *= 2.0;
	  lambda = iHLRF_lambda_start->cast2positive();	// start value for the step-size
	  const tdouble epsilon = iHLRF_epsilon->cast2positive();		// parameter of Armijo-rule
	  const tdouble reduce = iHLRF_reduce->cast2positive();
	  const tdouble m_0 = reduce*Norm2_NOroot(y)+c*fabs(z_old);	// current value of merit function
	    if (epsilon>=1.0) {
	      std::ostringstream ssV;
	      ssV << "'iHLRF_epsilon' has to be a value within the interval ]0;1[, and not (" << GlobalVar.Double2String(epsilon) << ")";
	      throw FlxException_NeglectInInteractive("FlxObjFORM_base::do_FORM_8", ssV.str() );
	    }
	  tdouble m_n, m_ne, z_next;		// next value (estimate) of merit-function
	  const tuint maxIter_LS = 20;
	  tuint i;
	  for (i=0;i<maxIter_LS;++i) {		// maximum number of interations
	    // compute estimator
	      m_ne = m_0 + epsilon*lambda*dot(y+c*sign(z_old)*dzdy,d);
	    // compute LSF-value
	      y_new = y + lambda*d;
	      data->RndBox.set_y_Vec(y_new);
	      z_next = LSF->calc();
	      m_n = 0.5*Norm2_NOroot(y_new)+c*fabs(z_next);
	    if (m_ne > m_n) break;	// get me out of here
	    lambda *= reduce;
	  }
	  if (i==maxIter_LS) {	// no-convergence
	    std::ostringstream ssV;
	    ssV << "Maximum number of line-search iterations reached (" << maxIter_LS << ").";
	    throw FlxException_NeglectInInteractive("FlxObjFORM_base::do_FORM_9", ssV.str() );
	  }
	  z_old = z_next;
	  y = y_new;
	}
      }
      
      beta_new = Norm2(y);
      
      // check if we are within reasonable intervals
	bool ok = true;		// false, if y has been reassigned
	if (beta_new>0) {
	  dx = rv_Phi(-beta_new);
	  if (dx==0.0) dx=1.0;
	} else {
	  dx = rv_Phi(beta_new);
	}
	if (dx == 0.0 || dx == 1.0 ) {
	  if ( dx == 0.0 ) {
	    dx = rv_InvPhi(1e-6);	// this is our new beta
	    if (verboseLog) {
	      GlobalVar.alert.alert("FlxObjFORM_base::do_FORM_10", "Warning: Interval correction (beta=" + GlobalVar.Double2String(beta_new) + ")" );
	    }
	  } else if ( dx == 1.0 ) {
	    dx = -1*rv_InvPhi(1e-6);
	    if (verboseLog) {
	      GlobalVar.alert.alert("FlxObjFORM_base::do_FORM_11", "Warning: Interval correction (beta=" + GlobalVar.Double2String(beta_new) + ")" );
	    }
	  }
	  y*=dx/beta_new;
	  beta_new = Norm2(y);  // equal to: beta_new = dx;
	  ok = false;
	}
      
      if (opt_method!=2 || !ok) {
	data->RndBox.set_y_Vec(y);
      }
      x = data->RndBox.get_x_Vec();
      // calc LSF
	if (opt_method!=2 || !ok) {
	  z_old = LSF->calc();		// actually this is z_new
	}
	
      // check for maximum number of iterations 
	loopc++;
	if (loopc > maxIter) {
	  std::ostringstream ssV;
	  ssV << "Maximum number of iterations reached (" << maxIter << ").";
	  throw FlxException_NeglectInInteractive("FlxObjFORM_base::do_FORM_12", ssV.str() );
	}
	
      if (verboseLog) {
	GlobalVar.slog(4) << "  Iter " << loopc << ": lsf=" << z_old << "; beta=" << beta_new << ";" << std::endl;
	GlobalVar.slog(4) << "    y=" << y << std::endl;
	GlobalVar.slog(4) << "    x=" << x << std::endl;
	if (std::fabs(fd_max-fd)<=GlobalVar.TOL()) {
	  GlobalVar.slog(4)  << "    fd=" << GlobalVar.Double2String(fd);
	} else {
	  GlobalVar.slog(4)  << "    fd=" << GlobalVar.Double2String(fd) << "->" << GlobalVar.Double2String(fd_max);
	}
	if (opt_method==2)  {
	  GlobalVar.slog(4)  << "; lambda=" << GlobalVar.Double2String(lambda);
	}
	GlobalVar.slog(4) << std::endl; 
      }
      
    } while ( std::fabs(beta_new-beta_old) > eps1->calc() || fabs(z_old) > eps2->calc() );
    
    if ( dot(dzdy,y) > 0 ) {
      beta_new *= -1;
    }
    return dzdy;
  } catch (FlxException &e) {
    delete dxdy;
    throw;
  }
}


void FlxObjFORM::task()
{ 
  update_Start();
  // calculate the start vector
    tVec x = xstart->calcVec();
    if (x.size() != DIM) {
      std::ostringstream ssV;
      ssV << "Vector sizes do not match.\n\tDIM_required=" << DIM << "; DIM_x=" << x.size();
      throw FlxException("FlxObjFORM::task_2", ssV.str() );
    }
    tVec y(DIM);
  // Write log-message
      GlobalVar.slog(4) << "form: performing FORM analysis. " << std::endl;
  // Perform the FORM interation
    tdouble beta;
    do_FORM(x,y,beta);
  
  // Deal with the results
    data->ConstMtxBox.insert(cnamey,new SMtxVector(new tVec(y),DIM));
    data->ConstMtxBox.insert(cnamex,new SMtxVector(new tVec(x),DIM));
    std::string betaDP_constName = betaDP->eval_word(true,true);
    if (betaDP_constName!="") {
      data->ConstantBox.insert(betaDP_constName,beta);
    }
  
  GlobalVar.slog(3) << "form: determined design point: " << "betaDP=" << GlobalVar.Double2String(beta);
  if (verboseLog) {
    GlobalVar.slog(3) << "; x=" << x << "; y=" << y;
  }
  GlobalVar.slog(3) << "." << std::endl;
  GlobalVar.slog(3) << "  Estimated probability of failure: " << GlobalVar.Double2String(rv_Phi(-beta)) << std::endl;
  GlobalVar.slog(3) << "  Probability of failure for sure within [0; " << GlobalVar.Double2String(1-rv_cdf_ChiSquare(DIM,pow2(beta))) << "]" << std::endl;
}

FlxObjFORM::~FlxObjFORM()
{
  delete xstart;
  delete betaDP;
}

void FlxObjFORM_pdf::task()
{
  FunNumber* fn = new FunNumber(0.0);
  FunSub* fa = new FunSub(fn,rvfun);
  FlxFunction* lsf = new FlxFunction(fa);
  try {
  LSF = lsf;
  update_Start();
  // calculate the start vector
    tVec x = xstart->calcVec();
    if (x.size() != DIM) {
      std::ostringstream ssV;
      ssV << "Vector sizes do not match.\n\tDIM_required=" << DIM << "; DIM_x=" << x.size();
      throw FlxException("FlxObjFORM_pdf::task_1", ssV.str() );
    }
    tVec y(DIM);
  // evaluate the bounds
    tdouble lbound = lboundF->calc();
    tdouble ubound = uboundF->calc();
    if (ubound<=lbound) {
      std::ostringstream ssV;
      ssV << "The lower bound has to be smaller than the upper bound.";
      throw FlxException("FlxObjFORM_pdf::task_2", ssV.str() );
    }
    tuint ninterval = nintervalF->cast2tuint(false);
    tdouble du = (ubound-lbound)/ninterval;
  // Perform the FORM interation
    tdouble beta;
    tdouble u = lbound;
    while (u<=ubound) {
      fn->set_thenumber(u);
      tVec dzdy = do_FORM(x,y,beta);
      sout() << GlobalVar.Double2String(u,true) << "\t";
      sout() << GlobalVar.Double2String( rv_phi(beta)/Norm2(dzdy) ) << "\t"; 	// PDF
      sout() << GlobalVar.Double2String( rv_Phi(beta) );	//CDF
      sout() << std::endl;      
      u+=du;
    }
  } catch (FlxException &e) {
    delete fn;
    fa->set_childs_zero();
    delete lsf;
    LSF = NULL;
    throw;
  }
  delete fn;
  fa->set_childs_zero();
  delete lsf;
  LSF = NULL;
}

FlxObjFORM_pdf::FlxObjFORM_pdf(bool dolog, FlxFunction* rvfun, FlxFunction* lboundF, FlxFunction* uboundF, FlxFunction* nintervalF, FlxMtxFun* xstartV, FlxFunction* fdstepV, FlxFunction* eps1, FlxFunction* eps2, FlxFunction* iHLRF_lambda_start, FlxFunction* iHLRF_epsilon, FlxFunction* iHLRF_reduce, unsigned int maxIter, bool verbose, std::string ostreamV, bool dxdyAnalytical, FlxMtxFun* dx_min, int fd_method, int opt_method)
: FlxObjFORM_base(dolog,NULL,fdstepV,eps1,eps2,iHLRF_lambda_start,iHLRF_epsilon,iHLRF_reduce,maxIter,verbose,ostreamV,dxdyAnalytical,dx_min,fd_method,opt_method), rvfun(rvfun), lboundF(lboundF), uboundF(uboundF), nintervalF(nintervalF), xstart(xstartV)
{
  
}

FlxObjFORM_pdf::~FlxObjFORM_pdf()
{
  delete rvfun;
  delete lboundF;
  delete uboundF;
  delete nintervalF;
  delete xstart;
  
}

SMtxBase* FunRVmean::calc()
{
  tVec* mvec = new tVec( data->RndBox.get_mean_Vec(output_just_user) );
  if (mvec->size()<=0) {
    std::ostringstream ssV_2;
    ssV_2 << "There exists no ";
    if (output_just_user) {
      ssV_2 << "user defined";
    } else {
      ssV_2 << "managed";
    }
    ssV_2 << "random variables.";
    throw FlxException("FunRVmean::calc", ssV_2.str() );  
  }
  return new SMtxVector(mvec, mvec->size() );
}

SMtxBase* FunRVstddev::calc()
{
  tVec* mvec = new tVec( data->RndBox.get_sd_Vec(output_just_user) );
  if (mvec->size()<=0) {
    std::ostringstream ssV_2;
    ssV_2 << "There exists no ";
    if (output_just_user) {
      ssV_2 << "user defined";
    } else {
      ssV_2 << "managed";
    }
    ssV_2 << "random variables.";
    throw FlxException("FunRVstddev::calc", ssV_2.str() );  
  }
  return new SMtxVector(mvec, mvec->size() );
}

FunMtxBase* FunReadFunRVmean::read(bool errSerious)
{
  bool output_just_user = false;
  if (reader->whatIsNextChar()=='u' || reader->whatIsNextChar() == 'U' ) {
    output_just_user = true;
  }
  return new FunRVmean(output_just_user);
}

FunMtxBase* FunReadFunRVstddev::read(bool errSerious)
{
  bool output_just_user = false;
  if (reader->whatIsNextChar()=='u' || reader->whatIsNextChar() == 'U' ) {
    output_just_user = true;
  }
  return new FunRVstddev(output_just_user);
}

const std::string FunRVvalue::write()
{
  if (bx) {
    return "rv_vec(x)";
  } else {
    return "rv_vec(y)";
  }
}

SMtxBase* FunRVvalue::calc()
{
  tVec* mvec;
  if (bx) {
    mvec = new tVec( data->RndBox.get_x_Vec_UD() );
  } else {
    mvec = new tVec( data->RndBox.get_y_Vec_UD() );
  }
  if (mvec->size()<=0) {
    std::ostringstream ssV_2;
    ssV_2 << "There exists no user defined random variables.";
    throw FlxException("FunRVvalue::calc", ssV_2.str() );  
  }
  return new SMtxVector(mvec, mvec->size() );
}

FunMtxBase* FunReadFunRVvalue::read(bool errSerious)
{
  bool bx;
  char c = reader->getChar();
  if (c == 'x' || c == 'X') {
    bx = true;
  } else if (c == 'y' || c == 'Y') {
    bx = false;
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Expected 'x' or 'y' - and not '" << c << "'.";
    FlxError(errSerious,"FunReadFunRVvalue::read", ssV_2.str() ); bx=false;
  }
  return new FunRVvalue(bx);
}

FlxRSM_Fun_base::~FlxRSM_Fun_base()
{
  if (paraVec != NULL) delete paraVec;
}

void FlxRSM_Fun_base::set_paraVec(tVec* paraVecV)
{
  if (paraVecV->size() != get_NumbOfTerms()) {
    std::ostringstream ssV_2;
    ssV_2 << "Parameter vector has wrong dimension.";
    throw FlxException("FlxRSM_Fun_base::set_paraVec", ssV_2.str() );  
  }
  if (paraVec != NULL) delete paraVec;
  paraVec = paraVecV;
}

void FlxRSM_Fun_base::set_para(tuint index, tdouble paraV)
{
  if (paraVec->size() <= index) {
    std::ostringstream ssV_2;
    ssV_2 << "Index (" << index << ") out of range [0;" << paraVec->size()-1 << "].";
    throw FlxException("FlxRSM_Fun_base::set_para", ssV_2.str() );  
  }
  (*paraVec)[index] = paraV;
}

const tdouble FlxRSM_Fun_base::get_para(tuint index)
{
  if (paraVec->size() <= index) {
    std::ostringstream ssV_2;
    ssV_2 << "Index (" << index << ") out of range [0;" << paraVec->size()-1 << "].";
    throw FlxException("FlxRSM_Fun_base::get_para", ssV_2.str() );  
  }
  return (*paraVec)[index];
}

tdouble FlxRSM_Fun_base::calc_solution(const tVec& yVec, const tVec& xVec)
{
  if (paraVec == NULL) {
    std::ostringstream ssV_2;
    ssV_2 << "A response surface has not yet been defined.";
    throw FlxException("FlxRSM_Fun_base::calc_solutionVec", ssV_2.str() );  
  }
//   std::cout << yVec << "; " << xVec << std::endl;
  tVec t((*paraVec).size());
  calc_termVec(yVec,xVec,t);
  return dot(*paraVec,t);
}

FlxRSM_Fun_base* FlxRSM_Fun_base::createRSMfun(bool errSerious, const std::string ftype)
{
  std::string funType;
  if (ftype=="") {
    funType = reader->getWord(true,errSerious);
  } else {
    funType = ftype;
  }
  if (funType == "p2womt") {
    return new FlxRSM_Fun_p2womt();
  } else if ( funType == "p2wmt" ) {
    return new FlxRSM_Fun_p2wmt();
  } else if ( funType == "p1" ) {
    return new FlxRSM_Fun_p1();
  } else if ( funType == "pc" ) {
    reader->getChar('(',errSerious);
    FlxFunction* orderF = new FlxFunction(funReader,errSerious);
    try {
      reader->getChar(')',errSerious);
      return new sfem::FlxRSM_Fun_pc(orderF);
    } catch (FlxException &e) {
      delete orderF;
      throw;
    }
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Type of RSM-function not known.";
    FlxError(errSerious,"FlxRSM_Fun_base::createRSMfun", ssV_2.str() ); return NULL; // dummy-return
  }
}

void FlxRSM_Fun_p2womt::calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec)
{
  #if FLX_DEBUG
    if (termVec.size() != get_NumbOfTerms()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR. (" << termVec.size() << ", " << get_NumbOfTerms() << ")";
      throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_0", ssV_2.str() );  
    }
  #endif
  if (xVec.size() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "Number of random variables has changed.";
    throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_1", ssV_2.str() );  
  }
  termVec[0]=1.0;
  for (tuint i = 1; i <= DIM; ++i) {
    termVec[i] = xVec[i-1];
    termVec[DIM+i] = pow2( xVec[i-1]);
  }
}

FlxRSM_Fun_base* FlxRSM_Fun_p2womt::copy()
{
  return new FlxRSM_Fun_p2womt(DIM, new tVec(*paraVec));
}

void FlxRSM_Fun_p2wmt::calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec)
{
  tuint k = FlxRSM_Fun_p2womt::get_NumbOfTerms();
  FlxRSM_Fun_p2womt::calc_termVec(yVec,xVec, termVec);
  for (tuint i = 0; i < DIM; ++i) {
    for (tuint j = i+1; j < DIM; ++j, ++k) {
      termVec[k] = xVec[i]*xVec[j];
    }
  }
}

FlxRSM_Fun_base* FlxRSM_Fun_p2wmt::copy()
{
  return new FlxRSM_Fun_p2wmt(DIM, new tVec(*paraVec));
}

FlxRSM_Fun_base* FlxRSM_Fun_p1::copy()
{
  return new FlxRSM_Fun_p1(DIM, new tVec(*paraVec));
}

void FlxRSM_Fun_p1::calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec)
{
  #if FLX_DEBUG
    if (termVec.size() != get_NumbOfTerms()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR. (" << termVec.size() << ", " << get_NumbOfTerms() << ")";
      throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_0", ssV_2.str() );  
    }
  #endif
  if (xVec.size() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "Number of random variables has changed.";
    throw FlxException("FlxRSM_Fun_p2womt::calc_termVec_1", ssV_2.str() );  
  }
  termVec[0]=1.0;
  for (tuint i = 1; i <= DIM; ++i) {
    termVec[i] = xVec[i-1];
  }
}

FlxRSM_SmpPoints_base* FlxRSM_SmpPoints_base::createRSMpts(bool errSerious)
{
  std::string ptsType = reader->getWord(true);
  if (ptsType == "joa") {
    return new FlxRSM_SmpPoints_joa(errSerious);
  } else if (ptsType == "cc") { 
    return new FlxRSM_SmpPoints_cc(errSerious);
  } else if (ptsType == "ff") { 
    return new FlxRSM_SmpPoints_ff(errSerious);
  } else if (ptsType == "p1") { 
    return new FlxRSM_SmpPoints_p1(errSerious);
  } else if (ptsType == "rnd") { 
    return new FlxRSM_SmpPoints_rnd(errSerious);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Type of RSM-sampling rule not known.";
    FlxError(errSerious,"FlxRSM_SmpPoints_base::createRSMpts", ssV_2.str() ); return NULL; // dummy-return
  }
}

FlxRSM_SmpPoints_joa::FlxRSM_SmpPoints_joa(bool errSerious) : centerY(NULL)
{
  reader->getWord("centery",errSerious);
  centerYF = new FlxMtxFun(mtxReader,errSerious);
  try {
    reader->getWord("k",errSerious);
    kF = new FlxFunction(funReader,errSerious);
  } catch ( FlxException &e) {
    delete centerYF;
    throw;
  }
}

FlxRSM_SmpPoints_joa::~FlxRSM_SmpPoints_joa()
{
  delete centerYF;
  delete kF;
  if (centerY != NULL) delete centerY;
}

void FlxRSM_SmpPoints_joa::updateStart()
{
  FlxRSM_SmpPoints_base::updateStart();
  if (centerY != NULL) delete centerY;
  centerY = new tVec(centerYF->calcVec());
  if (centerY->size() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "'centerY' has wrong dimension (" << centerY->size() << " instead of " << DIM << ")";
    throw FlxException("FlxRSM_SmpPoints_joa::updateStart", ssV_2.str() );  
  }
  k = kF->calc();
  index = 0;
}

void FlxRSM_SmpPoints_joa::get_nextPoint(tVec& y)
{
  #if FLX_DEBUG
    if (DIM != y.size()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR.";
      throw FlxException("FlxRSM_SmpPoints_joa::get_nextPoint_0", ssV_2.str() );  
    }
  #endif
  ++index;
  y = *centerY;
  if (index > 1) {
    if (index > get_NumbOfPoints()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR.";
      throw FlxException("FlxRSM_SmpPoints_joa::get_nextPoint_1", ssV_2.str() );  
    }
    tuint j = index/2-1;
    if (index % 2 == 0) {
      y[j] += k;
    } else {
      y[j] -= k;
    }
  }
//   std::cout << "index: " << y << std::endl;
}

void FlxRSM_SmpPoints_cc::get_nextPoint(tVec& y)
{
   if ( index < FlxRSM_SmpPoints_joa::get_NumbOfPoints() ) {
     FlxRSM_SmpPoints_joa::get_nextPoint(y);
     return;
   }
   y = *centerY + Comb->Next() * k2;
}

FlxRSM_SmpPoints_cc::FlxRSM_SmpPoints_cc(bool errSerious) : FlxRSM_SmpPoints_joa(), Comb(NULL)
{
  reader->getWord("k2",errSerious);
  kF2 = new FlxFunction(funReader,errSerious);
}

FlxRSM_SmpPoints_cc::~FlxRSM_SmpPoints_cc()
{
  delete kF2;
  if (Comb != NULL) delete Comb;
}

void FlxRSM_SmpPoints_cc::updateStart()
{
  FlxRSM_SmpPoints_joa::updateStart();
  k2 = kF2->calc();
  if (Comb != NULL) delete Comb;
  Comb = new FlxCombination(-1.0, 1.0, 2.0, DIM);
}

void FlxObjRSM::task()
{ 
  GlobalVar.slog(4) << "rsm: generating a response surface" << std::endl;
  RSMfunction->updateStart();
  SmpPoints->updateStart();
  const tuint Pneed = RSMfunction->get_NumbOfTerms();
  const tuint Pprov = SmpPoints->get_NumbOfPoints();
  
  FlxMtx A(Pprov, Pneed);
  GlobalVar.slog(4) << "rsm: terms in approximation: " << GlobalVar.Double2String(Pneed) << "; points provided: " << GlobalVar.Double2String(Pprov) << std::endl;
  if (Pprov<Pneed) {
    std::ostringstream ssV;
    ssV << "You have to provide at least as many points as terms are in the approximation." << std::endl << "\tPoints provided: " << GlobalVar.Double2String(Pprov) << "; Terms in the approximation: " << GlobalVar.Double2String(Pneed);
    throw FlxException("FlxObjRSM::task", ssV.str() );
  }
  
  tVec b(Pprov);
  tVec y(data->RndBox.get_NumbRnd());
  tVec x(data->RndBox.get_NumbRnd());
  tVec t(Pneed);
  tVec *p = new tVec(Pneed);
  
  // Assemble A and b
  for (tuint i = 0; i < Pprov; ++i) {
    SmpPoints->get_nextPoint(y);
    data->RndBox.set_y_Vec(y);
    data->RndBox.get_x_Vec(x);
    RSMfunction->calc_termVec(y,x,t);
    for (tuint j = 0; j < Pneed; ++j) {
	A(i,j) = t[j];
    }
    InternList->exec(); 
    b[i] = fung->calc();
  };
  // Solve LSE
    solMeth->solve(A,b,*p);
    RSMfunction->set_paraVec(p);
  FunReadFunRSM::set_RSMfunction(RSMfunction->copy());
  GlobalVar.slog(4) << "rsm: response surface initialized." << std::endl;
}

FlxObjRSM::~FlxObjRSM()
{
  delete RSMfunction;
  delete SmpPoints;
  delete solMeth;
  delete fung;
  delete InternList;
}

FlxObjBase* FlxObjReadRSM::read()
{
  FlxFunction* fung=NULL; FlxRSM_Fun_base* RSMfun=NULL;FlxRSM_SmpPoints_base* SmpPts=NULL;
  FlxRSM_Solve_base* solMeth=NULL; FlxObjBase* RSMblock=NULL;
  try {
    reader->getChar('(',false); 
    fung = new FlxFunction(funReader,false);
    reader->getChar(';',false); 
    RSMfun = FlxRSM_Fun_base::createRSMfun(false);
    reader->getChar(';',false); 
    SmpPts = FlxRSM_SmpPoints_base::createRSMpts(false);
    reader->getChar(';',false); 
    solMeth = FlxRSM_Solve_base::createRSMsolve(false);
    reader->getChar(')',false); 
    RSMblock = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjRSM(get_doLog(),RSMfun,SmpPts,solMeth,fung,RSMblock);
  } catch (FlxException &e) {
    if (fung!=NULL) delete fung;
    if (RSMfun!=NULL) delete RSMfun;
    if (SmpPts!=NULL) delete SmpPts;
    if (solMeth!=NULL) delete solMeth;
    if (RSMblock!=NULL) delete RSMblock;
    throw;
  }
}

FlxObjRSM_set_fun::~FlxObjRSM_set_fun()
{
  delete RSMfunction;
}

void FlxObjRSM_set_fun::task()
{
  GlobalVar.slog(4) << "rsm: generating an empty response surface" << std::endl;
  RSMfunction->updateStart();
  tuint Pneed = RSMfunction->get_NumbOfTerms();
  GlobalVar.slog(4) << "rsm: terms in approximation: " << GlobalVar.Double2String(Pneed) << std::endl;
  tVec *p = new tVec(Pneed);
  RSMfunction->set_paraVec(p);
  FunReadFunRSM::set_RSMfunction(RSMfunction->copy());
}

void FlxObjRSM_set_para::task()
{
  FlxRSM_Fun_base* RSMfunction = FunReadFunRSM::get_RSMfunction();
  RSMfunction->set_para(index->cast2tuintW0(),value->calc());
}

FlxObjRSM_set_para::~FlxObjRSM_set_para()
{
  delete index;
  delete value;
}

void FlxObjRSM_set_PCproject::task()
{
  sfem::FlxRSM_Fun_pc* pcRSM = dynamic_cast<sfem::FlxRSM_Fun_pc*>(RSMfunction);
  RSMfunction->updateStart();
  const tuint Pneed = RSMfunction->get_NumbOfTerms();
  GlobalVar.slog(4) << "rsm: generating a polynomial chaos response surface "
    << "(order=" << pcRSM->get_order()  
    << "; terms=" << Pneed
    << ")" << std::endl;
  tVec *p = new tVec(Pneed);
  if (data->RndBox.get_NumbRnd()==1) {
    sfem::FunSFEM_pc_term* pct = new sfem::FunSFEM_pc_term(pcRSM->pc_reference(),0);
    FunMult* fn = new FunMult(rv_fun,pct);
    tdouble cv;
    for (tulong i=0;i<Pneed;++i) {
      pct->set_j(i);
      cv = data->RndBox.calc_expectation_numerical(fn,true);
      cv/=pcRSM->pc_reference().get_ExpPB2(i);
      (*p)[i] = cv;
    }
    delete pct;
    fn->set_childs_zero();
    delete fn;
  } else {
    calc_PC_projection_numerical_MCI pcpnmci(pcRSM->pc_reference());
    pcpnmci.calc_expectation(*p,rv_fun,intervalsF->cast2tuint(),smpPintervF->cast2tuint(),smpRedistF->cast2tuint(),diffusionF->calc());
  }
  RSMfunction->set_paraVec(p);
  FunReadFunRSM::set_RSMfunction(RSMfunction->copy());
}

FlxObjRSM_set_PCproject::~FlxObjRSM_set_PCproject()
{
  delete rv_fun;
  delete RSMfunction;
  delete intervalsF;
  delete smpPintervF;
  delete smpRedistF;
  delete diffusionF;
}

void FlxObjRSM_set_print_coeff::task()
{
  FlxRSM_Fun_base* rsmfun = FunReadFunRSM::get_RSMfunction();
  for (tuint i = 0; i < rsmfun->get_NumbOfTerms(); ++i) {
    if (is_verbose()) {
      sout() << i << "\t";
    }
    sout() << GlobalVar.Double2String(rsmfun->get_para(i)) << std::endl;
  }
}

void FlxObjRSM_set_coeff_vec::task()
{
  FlxRSM_Fun_base* rsmfun = FunReadFunRSM::get_RSMfunction();
  tVec* t = new tVec(rsmfun->get_paraVec());
  SMtxVector *vec = new SMtxVector(t,t->size());
  data->ConstMtxBox.insert(vname->eval_word(),vec);
}

FlxObjReadRSM_set::FlxObjReadRSM_set()
{
  AllDefParaBox->insert(new FlxOptionalParaFun(608,"rsm::project::intervals"));
  ParaBox.insert("intervals", "rsm::project::intervals" );
  AllDefParaBox->insert(new FlxOptionalParaFun(100,"rsm::project::samplesperinterval"));
  ParaBox.insert("samplesperinterval", "rsm::project::samplesperinterval" );
  AllDefParaBox->insert(new FlxOptionalParaFun(1e6,"rsm::project::redistsmp"));
  ParaBox.insert("redistsmp", "rsm::project::redistsmp" );
  AllDefParaBox->insert(new FlxOptionalParaFun(0.9,"rsm::project::diffusion"));
  ParaBox.insert("diffusion", "rsm::project::diffusion" );
}

FlxObjBase* FlxObjReadRSM_set::read()
{
  std::string strT = reader->getWord(true,false);
  if (strT=="fun") {	// define the type of the rsm-function
    FlxRSM_Fun_base* RSMfun=NULL;
    try {
      RSMfun = FlxRSM_Fun_base::createRSMfun(false);
      read_optionalPara(false);
      return new FlxObjRSM_set_fun(get_doLog(),RSMfun);
    } catch (FlxException &e) {
      if (RSMfun!=NULL) delete RSMfun;
      throw;
    }
  } else if (strT == "para") {
    reader->getChar('(',false);
    FlxFunction* index = new FlxFunction(funReader,false);
    FlxFunction* valueP=NULL;
    try {
      reader->getChar(')',false);
      reader->getChar('=',false);
      valueP = new FlxFunction(funReader,false);
      return new FlxObjRSM_set_para(get_doLog(),index,valueP);
    } catch (FlxException &e) {
      delete index;
      if (valueP!=NULL) delete valueP;
      throw;
    }
  } else if (strT=="pc_project") {		// polynomial chaos projection
    FlxRSM_Fun_base* RSMfun=NULL;
    FlxFunction* rvFun=NULL;
    try {
      RSMfun = FlxRSM_Fun_base::createRSMfun(false,"pc");
      reader->getChar('=');
      rvFun = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjRSM_set_PCproject(get_doLog(),RSMfun,rvFun,get_optPara_FlxFunction("intervals"),get_optPara_FlxFunction("samplesperinterval"),get_optPara_FlxFunction("redistsmp"),get_optPara_FlxFunction("diffusion"));
    } catch (FlxException &e) {
      if (RSMfun!=NULL) delete RSMfun;
      if (rvFun!=NULL) delete rvFun;
      throw;
    }
  } else if (strT == "print_coeff") {
    read_optionalPara(false);
    return new FlxObjRSM_set_print_coeff(get_doLog(), get_stream(), get_verbose());
  } else if (strT == "coeff_vec") {
    reader->getChar('(',false);
    FlxString* vname = new FlxString(false,false);
    try {
      reader->getChar(')',false);
      read_optionalPara(false);
      return new FlxObjRSM_set_coeff_vec(get_verbose(),vname);
    } catch (FlxException &e) {
      delete vname;
      throw;
    }
  } else {
    std::ostringstream ssV;
    ssV << "Unknown keyword '" << strT << "'.";
    throw FlxException_NeglectInInteractive("FlxObjReadRSM_set::read", ssV.str() );
  }
}

FlxRSM_Solve_base* FlxRSM_Solve_base::createRSMsolve(bool errSerious)
{
  std::string ptsType = reader->getWord(true,errSerious);
  if (ptsType == "transpose") {
    return new FlxRSM_Solve_CG(errSerious);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "Type of RSM-solution method not known.";
    FlxError(errSerious,"FlxRSM_Solve_base::createRSMsolve", ssV_2.str() ); return NULL; // dummy-return 
  }
}

void FlxRSM_Solve_CG::solve(const FlxMtx& A, const tVec& b, tVec& sol)
{
// std::cout << "A: " << A << std::endl << "b: " << b << std::endl;
  FlxMtxSym ATA(sol.size());
  tVec ATb(sol.size());
  A.TransposeMmultM(ATA);
  A.TransposeMmultVec(b,ATb);
  tnlong iter = iterF->cast2tnlong();
  tdouble eps = epsF->calc();
  if (eps <= 0) {
    std::ostringstream ssV_2;
    ssV_2 << "'eps' (" << eps << ") must be larger than zero.";
    throw FlxException("FlxRSM_Solve_CG::solve_1", ssV_2.str() );  
  }
  int pcn = pcnF->cast2int();
  if ( ! ATA.solve_CG(sol,ATb,eps,iter,pcn,false) ) {
    std::ostringstream ssV_2;
    ssV_2 << "System could not be solved.";
    std::ostringstream ssV_3;
    ssV_2 << "iter=" << iter << "; eps=" << eps;
    throw FlxException("FlxRSM_Solve_CG::solve_2", ssV_2.str(), ssV_3.str() );  
  }
  if (print) {
    GlobalVar.slog(4) << "  Solution: iter=" << GlobalVar.Double2String(iter) << "; eps=" << GlobalVar.Double2String(eps) << "(" << GlobalVar.Double2String(eps/(Norm2(ATb))) << ")";
    if (printS) { GlobalVar.slog(4) << "; sol=" << sol; }
    GlobalVar.slog(4) << std::endl;
  }
}

FlxRSM_Solve_CG::~FlxRSM_Solve_CG()
{
  delete iterF;
  delete epsF;
  delete pcnF;
}

FlxRSM_Solve_CG::FlxRSM_Solve_CG(bool errSerious) :
iterF(new FlxFunction(new FunNumber(1e4))),
epsF(new FlxFunction(new FunNumber(1e-8))),
pcnF(new FlxFunction(new FunNumber(1))),
print(true),printS(false)
{
  if (reader->whatIsNextChar() == ')') return;
  try {
    while ( true ) {
      std::string strV = reader->getWord(true,errSerious);
      reader->getChar('=',errSerious);
      if (strV == "eps") {
	delete epsF; epsF=NULL;
	epsF = new FlxFunction(funReader,errSerious);
      } else if ( strV == "iter" ) {
	delete iterF; iterF=NULL;
	iterF = new FlxFunction(funReader,errSerious);
      } else if ( strV == "pcn" ) {
	delete pcnF; pcnF=NULL;
	pcnF = new FlxFunction(funReader,errSerious);
      } else if ( strV == "print" ) {
	print = reader->getBool(errSerious);
      } else if ( strV == "printsol" ) {
	printS = reader->getBool(errSerious);
      } else {
	std::ostringstream ssV_2;
	ssV_2 << "Unknown keyword (" << strV << ")";
	FlxError(errSerious,"FlxRSM_Solve_CG::FlxRSM_Solve_CG_1", ssV_2.str() );  
      }
      if (reader->whatIsNextChar() != ',' ) break;
      reader->getChar(',',errSerious);
    }
  } catch (FlxException &e) {
    if (epsF!=NULL) delete epsF;
    if (iterF!=NULL) delete iterF;
    if (pcnF!=NULL) delete pcnF;
    throw;
  }
}

void FunRSM::dependOn_RV(bVec& bv)
{
  bv=true;
}

const bool FunRSM::dependOn_Const(tdouble* thenumber)
{
  return false;
}

const tdouble FunRSM::calc()
{	
  return (FunReadFunRSM::get_RSMfunction())->calc_solution(data->RndBox.get_y_Vec(),data->RndBox.get_x_Vec());
}

const tdouble FunRSM_terms::calc()
{
  return (FunReadFunRSM::get_RSMfunction())->get_NumbOfTerms();
}

const std::string FunHLRV::write()
{
  return "hlrv(" + strV->write() + ")";
}

const tdouble FunHLRV::calc()
{
  if (hlrv==NULL) {
    hlrv = data->RndBox.get(strV->eval_word(true));
    delete strV;
    strV = NULL;
  }
  return hlrv->get_Realization();
}

FunHLRV::~FunHLRV()
{
  if (strV != NULL) delete strV;
}

void FunHLRV::dependOn_RV(bVec& bv)
{
  std::ostringstream ssV_2;
  ssV_2 << "It is not allowed to check a 'High Level Random Variable' for dependence on another random variable.";
  throw FlxException("FunHLRV::dependOn_RV", ssV_2.str() ); 
}

const bool FunHLRV::dependOn_Const(const tdouble*const thenumber)
{
  throw FlxException_NotImplemented("FunHLRV::dependOn_Const");
}

const bool FunHLRV::search_circref(FlxFunction* fcr)
{
  if (strV != NULL) {
    return strV->search_circref(fcr);
  } else {
    return false;
  }
}

FunBase* FunReadFunHLRV::read(bool errSerious)
{  
  FlxString* strV = new FlxString(false,errSerious);
  return new FunHLRV( strV );
}

void FunRF::dependOn_RV(bVec& bv)
{
  const tuint ist = rfb->get_RndBox_index();
  const tuint DIM = rfb->get_DIM();
  for (tuint i=0;i<DIM;++i) {
    bv[i+ist] = true;
  }
}

const std::string FunRF::write()
{
  std::ostringstream ssV;
  ssV << "rf(" << RFindex << ")";
  return ssV.str();
}

FunBase* FunReadFunRF::read(bool errSerious)
{
  FunBase* fb = FunctionList->read(errSerious);
  try {
    const tuint RFindex = tuint_from(fb->calc(),"Index of random field",errSerious);
    delete fb; fb = NULL;
    FlxRandomField_base* rfb = RndBox->RFbox->get(RFindex,errSerious);
    if (rfb->is_automatic()) {
      FlxError(errSerious,"FunReadFunRF::read", "Random field is an automatic random field and must not be accessed via 'rf'." );
    }
    return new FunRF( RFindex, rfb );
  } catch (FlxException& e) {
    if (fb) delete fb;
    throw;
  }
}

void FunReadFunRSM::set_RSMfunction(FlxRSM_Fun_base* RSMfunctionV)
{
  if (RSMfunction != NULL ) delete RSMfunction;
  RSMfunction = RSMfunctionV;
}

FlxRSM_Fun_base* FunReadFunRSM::get_RSMfunction()
{
  if (RSMfunction == NULL) {
    std::ostringstream ssV_2;
    ssV_2 << "A response surface has not yet been defined.";
    throw FlxException("FunReadFunRSM::get_RSMfunction", ssV_2.str() ); 
  }
  return RSMfunction;
}

FlxCombination::FlxCombination(tdouble start, tdouble end, tdouble step, tuint dim)
: start(start), end(end), step(step), dim(dim), cv(dim), runs(false)
{
  cv = start;
}

tVec FlxCombination::Next()
{
  if (!runs) {
    runs = true;
    return cv;
  }
  for (tuint i = 0; i < dim; ++i) {
    if (cv[i] >= end) {
      cv[i] = start;
    } else {
      cv[i]+=step;
      return cv;
    }
  }
  std::ostringstream ssV_2;
  ssV_2 << "No combinations can be created anymore.";
  throw FlxException("FlxCombination::Next", ssV_2.str() ); 
}

void FlxRSM_SmpPoints_p1::get_nextPoint(tVec& y)
{
  #if FLX_DEBUG
    if (DIM != y.size()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR.";
      throw FlxException("FlxRSM_SmpPoints_p1::FlxRSM_SmpPoints_p1_1", ssV_2.str() );  
    }
  #endif
  y = *centerY;
  tVec mu = data->RndBox.get_mean_Vec();
  if (y.size() != mu.size()) {
    std::ostringstream ssV_2;
    ssV_2 << "ERROR. (" << y.size() << "; " << mu.size() << ")";
    throw FlxException("FlxRSM_SmpPoints_p1::FlxRSM_SmpPoints_p1_3", ssV_2.str() );  
  }
  if (index > 0) {
    if (index >= get_NumbOfPoints()) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR.";
      throw FlxException("FlxRSM_SmpPoints_p1::FlxRSM_SmpPoints_p1_2", ssV_2.str() );  
    }
    if ( y[index-1] > mu[index-1] ) {
      y[index-1] -= k;
    } else {
      y[index-1] += k;
    }
  }
  ++index;
}


FlxRSM_SmpPoints_ff::FlxRSM_SmpPoints_ff(bool errSerious)
: FlxRSM_SmpPoints_joa(), Comb(NULL)
{
  reader->getWord("sphere",errSerious);
  sphere = reader->getBool(errSerious);
}

FlxRSM_SmpPoints_ff::~FlxRSM_SmpPoints_ff()
{
  if (Comb != NULL) delete Comb;
}	

void FlxRSM_SmpPoints_ff::updateStart()
{
  FlxRSM_SmpPoints_joa::updateStart();
  if (Comb != NULL) delete Comb;
  Comb = new FlxCombination(-1.0, 1.0, 1.0, DIM);
}

void FlxRSM_SmpPoints_ff::get_nextPoint(tVec& y)
{
  if (sphere) {
    tVec t = Comb->Next();
    tdouble l = Norm2(t);
    if (l != 0.0) {
      y = *centerY + t * (k/l);
    } else {
      y = *centerY;
    }
  } else {
    y = *centerY + Comb->Next() * k;
  }
}

FlxRSM_SmpPoints_rnd::FlxRSM_SmpPoints_rnd(bool errSerious)
: FlxRSM_SmpPoints_joa()
{
  reader->getWord("numb",errSerious);
  numb = new FlxFunction(funReader,errSerious);
}

void FlxRSM_SmpPoints_rnd::get_nextPoint(tVec& y)
{
  if (y.size() != centerY->size()) {
    std::ostringstream ssV_2;
    ssV_2 << "ERROR.";
    throw FlxException("FlxRSM_SmpPoints_rnd::get_nextPoint", ssV_2.str() ); 
  }
  y = *centerY;
  for (tuint i = 0; i < y.size(); ++i) {
    y[i] += rv_normal() * k;
  }
}

FlxObjReadKDE::FlxObjReadKDE()
{
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"kde::do_cdf"));
  ParaBox.insert("do_cdf", "kde::do_cdf" );
  AllDefParaBox->insert(new FlxOptionalParaFun(-10,"kde::lbound"));
  ParaBox.insert("lbound", "kde::lbound" );
  AllDefParaBox->insert(new FlxOptionalParaFun(10,"kde::ubound"));
  ParaBox.insert("ubound", "kde::ubound" );
  AllDefParaBox->insert(new FlxOptionalParaFun(100,"kde::ninterval"));
  ParaBox.insert("ninterval", "kde::ninterval" );
}

FlxObjBase* FlxObjReadKDE::read()
{
  FlxString* HLRVstr = NULL; FlxFunction* N=NULL; FlxRndKernel_base* kernel=NULL; FlxFunction* h=NULL;
  try {
    reader->getChar('(',false);
    HLRVstr = new FlxString(false,false);
    reader->getChar(',',false);
    N = new FlxFunction(funReader,false);
    reader->getChar(',',false);
    kernel = FlxRndKernel_base::read(false);
    reader->getChar(',',false);
    h = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjKDE(get_doLog(),HLRVstr, N, kernel, h, get_optPara_FlxFunction("lbound"), get_optPara_FlxFunction("ubound"), get_optPara_FlxFunction("ninterval"), get_optPara_bool("do_cdf"), get_stream());
  } catch (FlxException &e) {
    if (HLRVstr!=NULL) delete HLRVstr;
    if (N!=NULL) delete N;
    if (kernel!=NULL) delete kernel;
    if (h!=NULL) delete h;
    throw;
  }
}

FlxObjKDE::FlxObjKDE( bool dolog, FlxString* HLRVstring, FlxFunction* N, FlxRndKernel_base* kernel, FlxFunction* h, FlxFunction* lbound, FlxFunction* ubound, FlxFunction* Ninterval, bool do_cdf, std::string ostreamV) 
: FlxObjOutputBase(dolog,ostreamV), HLRVstring(HLRVstring), N(N), kernel(kernel), h(h), lbound(lbound), ubound(ubound), Ninterval(Ninterval), do_cdf(do_cdf)
{
  
}

FlxObjKDE::~FlxObjKDE()
{
  delete HLRVstring;
  delete N;
  delete kernel;
  delete h;
  delete lbound;
  delete ubound;
  delete Ninterval;
}

void FlxObjKDE::task()
{
  std::string hlrv_name = HLRVstring->eval_word(true);
  FlxHLRVbase* hlrv = data->RndBox.get(hlrv_name);
  
  tdouble dlb = lbound->calc();  	// lower bound
  tdouble dub = ubound->calc();		// upper bound
  if (dlb >= dub) {
    std::ostringstream ssV_2;
    ssV_2 << "Lower bound (" << dlb << ") must not be larger than upper bound (" << dub << ").";
    throw FlxException_NeglectInInteractive("FlxObjKDE::task_1", ssV_2.str() ); 
  }
  tnlong LNi = Ninterval->cast2tuint(false); 		// number of intervals
  #if FLX_KAHAN_KDE
    pdouble dss(dub);
    dss-=dlb;
    dss/=tdouble(LNi);
  #else
    tdouble dss = (dub-dlb)/LNi; 				// stepsize
  #endif
  kernel->set_h(h->cast2positive(false));
  tnlong LN = N->cast2tnlong();				// number of samples
  
  // generate random realizations
  tdouble *arr = new tdouble[LN];			// array
  for (tnlong i = 0; i < LN; ++i) {
    arr[i] = hlrv->get_Realization();
  }
  
  #if FLX_KAHAN_KDE
    pdouble x = dlb;
    pdouble p;
  #else
    tdouble x = dlb;
    tdouble p;
  #endif
  if (!do_cdf) {
    for (tnlong i = 0; i <= LNi; ++i) {
      p = 0.0;
      for (tnlong j = 0; j < LN; ++j) {
	#if FLX_KAHAN_KDE
	  p+=kernel->pdf(x.cast2double(), arr[j]);
	#else
	  p+=kernel->pdf(x, arr[j]);
	#endif
      }
      p/=tdouble(LN);
      #if FLX_KAHAN_KDE
	sout() << GlobalVar.Double2String(x.cast2double()) << " " << GlobalVar.Double2String(p.cast2double()) << std::endl;
      #else
	sout() << GlobalVar.Double2String(x) << " " << GlobalVar.Double2String(p) << std::endl;
      #endif
      x += dss;
    }
  } else  {
    for (tnlong i = 0; i <= LNi; ++i) {
      p = 0.0;
      for (tnlong j = 0; j < LN; ++j) {
	#if FLX_KAHAN_KDE
	  p+=kernel->cdf(x.cast2double(), arr[j]);
	#else
	  p+=kernel->cdf(x, arr[j]);
	#endif
      }
      p/=tdouble(LN);
      #if FLX_KAHAN_KDE
	sout() << GlobalVar.Double2String(x.cast2double()) << " " << GlobalVar.Double2String(p.cast2double()) << std::endl;
      #else
	sout() << GlobalVar.Double2String(x) << " " << GlobalVar.Double2String(p) << std::endl;
      #endif
      x += dss;
    }
  }
  
  delete[] arr;
  
  GlobalVar.slog(4) << "kde: performed kernel density estimation (" << (do_cdf?"cdf":"pdf") << ") of the HLRV '" << hlrv_name << "'." << std::endl;
  GlobalVar.slog(4) << "  Kernel: "; kernel->print_info(GlobalVar.slog(4)); GlobalVar.slog(4) << " (Number of samples: " << GlobalVar.Double2String(LN) << ")" << std::endl;
  GlobalVar.slog(4) << "  Plot interval: [" << GlobalVar.Double2String(dlb) << "; " << GlobalVar.Double2String(dub) << "]; stepsize=" 
    #if FLX_KAHAN_KDE
      << dss.cast2double() << std::endl;
    #else
      << dss << std::endl;
    #endif
}


FlxObjHLRV::~FlxObjHLRV()
{
  delete hlrv_name;
  if ( fun ) delete fun;
  if ( Icode ) delete Icode;
}

void FlxObjHLRV::task()
{
  data->RndBox.insert(hlrv_name->eval_word(true),new FlxHLRVfun(fun,Icode));
  fun = NULL;
  Icode = NULL;
}

FlxObjBase* FlxObjReadHLRV::read()
{
  FlxString* hlrv_name=NULL; FlxFunction* hlrv_fun=NULL; FlxObjBase* Icode=NULL;
  try {
    FlxString* hlrv_name = new FlxString(false,false);
    reader->getChar('=',false);
    FlxFunction* hlrv_fun = new FlxFunction(funReader,false);
    FlxObjBase* Icode = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjHLRV(get_doLog(),hlrv_name, hlrv_fun, Icode);
  } catch (FlxException &e) {
    if (hlrv_name) delete hlrv_name;
    if (hlrv_fun) delete hlrv_fun;
    if (Icode) delete Icode;
    throw;
  }
}

FlxHLRVfun::~FlxHLRVfun()
{
  delete fun;
  delete Icode;
}

const tdouble FlxHLRVfun::get_Realization()
{
  RndBox->gen_smp();
  Icode->exec(); 
  return fun->calc();
}

const std::string FlxHLRVfun::print_info()
{
  return "random function '" + fun->write() + "'";
}

FlxObjReadStatSmp::FlxObjReadStatSmp()
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e6,"statsmp::np"));
  ParaBox.insert("np", "statsmp::np" );
  
  AllDefParaBox->insert(new FlxOptionalParaFlxString("","statsmp::addname"));
  ParaBox.insert("addname", "statsmp::addname" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(1,"statsmp::optionp"));
  ParaBox.insert("optionp", "statsmp::optionp" );
}

FlxObjBase* FlxObjReadStatSmp::read()
{
  FlxString* flxStr = NULL;
  FlxFunction* optionF = NULL;
  try {
    reader->getChar('(',false);
    flxStr = new FlxString(false,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    optionF = get_optPara_FlxFunction("optionp");
    int optionp = optionF->cast2int();
    delete optionF; optionF=NULL;
    return new FlxObjStatSmp(get_doLog(), get_stream(), flxStr, get_optPara_FlxString("addname") ,get_optPara_FlxFunction("np"), optionp);
  } catch (FlxException &e) {
    if (flxStr!=NULL) delete flxStr;
    if (optionF!=NULL) delete optionF;
    throw;
  }
}

void FlxObjStatSmp::task()
{
  bool doubleSample;
  if (optionP==1) doubleSample = false;
  else if (optionP==2) doubleSample = true;
  else {
    throw FlxException_Crude("FlxObjStatSmp::task_1");
  }
  
  std::string as = add_name->eval_word(true,true);
  FlxIstream &is = data->IstreamBox.get(isname->eval_word());
  tdouble d,d2;
  tnlong N=0;
  tuint NpB = tuint(sqrt(tdouble(Np->cast2tuint(false))));
  qdouble sx1(NpB), sx1_2(NpB);
  qdouble sx2(NpB), sx2_2(NpB);
  qdouble sx3(NpB), sx3_2(NpB);
  qdouble sxy(NpB);
  tdouble xmin,xmax,xmin2,xmax2;
  if (is.get_value(d,true)) {
    xmax = d; xmin = d;
    if (doubleSample) {
      if (!is.get_value(d2,true)) {
	std::ostringstream ssV_2;
	ssV_2 << "The two sets have not the same size.";
	throw FlxException_NeglectInInteractive("FlxObjStatSmp::task_2", ssV_2.str() );
      }
      xmax2 = d2; xmin2 = d2;
    }
    do {
      ++N;
      sx1+=d;
      sx2+=pow2(d);
      sx3+=pow(d,3);
      if (d<xmin) xmin=d;
      if (d>xmax) xmax=d;
      if (doubleSample) {
	if (N>1) {
	  if (!is.get_value(d2,true)) {
	    std::ostringstream ssV_2;
	    ssV_2 << "The two sets have not the same size.";
	    throw FlxException_NeglectInInteractive("FlxObjStatSmp::task_3", ssV_2.str() );
	  }
	}
	sx1_2+=d2;
	sx2_2+=pow2(d2);
	sx3_2+=pow(d2,3);
	sxy+=d*d2;
	if (d2<xmin2) xmin2=d2;
	if (d2>xmax2) xmax2=d2;
      }
    } while (is.get_value(d,true));
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "No sample points.";
    throw FlxException_NeglectInInteractive("FlxObjStatSmp::task_4", ssV_2.str() );
  }
  // mean
    pdouble t = sx1.cast2pdouble();
    t/=N;
    pdouble mean = t;
    pdouble mean2, t_2;
    if (doubleSample) {
      t_2 = sx1_2.cast2pdouble();
      t_2/=N;
      mean2 = t_2;
    }
  // variance
    t*=mean;
    pdouble t2 = sx2.cast2pdouble();
    t2-=t*N;
    pdouble variance = t2;
    variance/=(N-1.0);
    pdouble variance2;
    if (doubleSample) {
      t_2*=mean2;
      t2 = sx2_2.cast2pdouble();
      t2-=t_2*N;
      variance2 = t2;
      variance2/=(N-1.0);
    }
  // skewness
    t2 = t;
    t2*=mean;
    t2*=2.0*N;
    t2+=sx3.cast2pdouble();
    t=sx2.cast2pdouble();
    t*=mean;
    t*=3.0;
    t2-=t;
    t=variance;
    t*=N*sqrt(variance.cast2double());
    tdouble skewness = t2.cast2double()/t.cast2double();
    tdouble skewness2;
    if (doubleSample) {
      t2 = t_2;
      t2*=mean2;
      t2*=2.0*N;
      t2+=sx3_2.cast2pdouble();
      t_2=sx2_2.cast2pdouble();
      t_2*=mean2;
      t_2*=3.0;
      t2-=t_2;
      t_2=variance2;
      t_2*=N*sqrt(variance2.cast2double());
      skewness2 = t2.cast2double()/t_2.cast2double();
    }
    tdouble covariance,corrcoeff;
    if (doubleSample) {
      t2 = sxy.cast2pdouble();
      t = mean;
      t *= mean2;
      t *= N;
      t2 -= t;
      covariance = t2.cast2double() / (N-1.0);
      corrcoeff = covariance / sqrt( variance.cast2double() * variance2.cast2double() );
    }
  sout() << "Evaluation of the statistical properties of the given samples:" << std::endl;
  sout() << "  number of samples:               " << GlobalVar.Double2String(N) << std::endl;
  if (doubleSample) {
  sout() << std::endl;
  sout() << "  Sample Set 1" << std::endl;
  sout() << "  ------------" << std::endl;
  }
  sout() << "  sample mean:                     " << GlobalVar.Double2String(mean.cast2double()) << "\t(estim. c.o.v.: " << GlobalVar.Double2String(sqrt(variance.cast2double())/(sqrt(tdouble(N))*fabs(mean.cast2double()))) << ")" << std::endl;
  sout() << "  sample standard deviation:       " << GlobalVar.Double2String(sqrt(variance.cast2double())) << std::endl;
  if (mean.cast2double()!=0.0) {
  sout() << "  sample coefficient of variation: " << GlobalVar.Double2String(sqrt(variance.cast2double())/fabs(mean.cast2double())) << std::endl;
  }
  sout() << "  sample variance:                 " << GlobalVar.Double2String(variance.cast2double()) << std::endl;
  sout() << "  sample skewness:                 " << GlobalVar.Double2String(skewness) << std::endl;
  sout() << "  smallest sample:                 " << GlobalVar.Double2String(xmin) << std::endl;
  sout() << "  largest sample:                  " << GlobalVar.Double2String(xmax) << std::endl;
  sout() << "  sample range:                    " << GlobalVar.Double2String(xmax-xmin) << std::endl;
  if (doubleSample) {    
  sout() << std::endl;
  sout() << "  Sample Set 2" << std::endl;
  sout() << "  ------------" << std::endl;
  sout() << "  sample mean:                     " << GlobalVar.Double2String(mean2.cast2double()) << "\t(estim. c.o.v.: " << GlobalVar.Double2String(sqrt(variance.cast2double())/(sqrt(tdouble(N))*fabs(mean.cast2double()))) << ")" << std::endl;
  sout() << "  sample standard deviation:       " << GlobalVar.Double2String(sqrt(variance2.cast2double())) << std::endl;
  if (mean2.cast2double()!=0.0) {
  sout() << "  sample coefficient of variation: " << GlobalVar.Double2String(sqrt(variance2.cast2double())/fabs(mean2.cast2double())) << std::endl;
  }
  sout() << "  sample variance:                 " << GlobalVar.Double2String(variance2.cast2double()) << std::endl;
  sout() << "  sample skewness:                 " << GlobalVar.Double2String(skewness2) << std::endl;
  sout() << "  smallest sample:                 " << GlobalVar.Double2String(xmin2) << std::endl;
  sout() << "  largest sample:                  " << GlobalVar.Double2String(xmax2) << std::endl;
  sout() << "  sample range:                    " << GlobalVar.Double2String(xmax2-xmin2) << std::endl;
  sout() << std::endl;
  sout() << "  Sample Dependence" << std::endl;
  sout() << "  -----------------" << std::endl;
  sout() << "  sample correlation coefficient:  " << GlobalVar.Double2String(corrcoeff) << std::endl;
  sout() << "  sample covariance:               " << GlobalVar.Double2String(covariance) << std::endl;
  }
  if (!as.empty()) {
    data->ConstantBox.insert(as + "_n",N);
    data->ConstantBox.insert(as + "_mean",mean.cast2double());
    data->ConstantBox.insert(as + "_sd",sqrt(variance.cast2double()));
    if (mean.cast2double()!=0.0) {
    data->ConstantBox.insert(as + "_coeffofvar",sqrt(variance.cast2double())/fabs(mean.cast2double()));
    }
    data->ConstantBox.insert(as + "_var",variance.cast2double());
    data->ConstantBox.insert(as + "_skewness",skewness);
    data->ConstantBox.insert(as + "_min",xmin);
    data->ConstantBox.insert(as + "_max",xmax);
    data->ConstantBox.insert(as + "_range",xmax-xmin);
    if (doubleSample) {
      data->ConstantBox.insert(as + "_mean2",mean2.cast2double());
      data->ConstantBox.insert(as + "_sd2",sqrt(variance2.cast2double()));
      if (mean2.cast2double()!=0.0) {
      data->ConstantBox.insert(as + "_coeffofvar2",sqrt(variance2.cast2double())/fabs(mean2.cast2double()));
      }
      data->ConstantBox.insert(as + "_var2",variance2.cast2double());
      data->ConstantBox.insert(as + "_skewness2",skewness2);
      data->ConstantBox.insert(as + "_min2",xmin2);
      data->ConstantBox.insert(as + "_max2",xmax2);
      data->ConstantBox.insert(as + "_range2",xmax2-xmin2);
      data->ConstantBox.insert(as + "_corrcoeff",corrcoeff);
      data->ConstantBox.insert(as + "_covar",covariance);
    }
  }
}

FlxObjStatSmp::FlxObjStatSmp(bool dolog, std::string ostreamV, FlxString* isname, FlxString* add_name, FlxFunction* Np, int optionP)
: FlxObjOutputBase(dolog,ostreamV), isname(isname), add_name(add_name), Np(Np), optionP(optionP)
{

}

FlxObjStatSmp::~FlxObjStatSmp()
{
  delete isname;
  delete add_name;
  delete Np;
}

FlxObjReadSortSmp::FlxObjReadSortSmp()
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e6,"sortsmp::np"));
  ParaBox.insert("np", "sortsmp::np" );
}

FlxObjBase* FlxObjReadSortSmp::read()
{
  FlxString* flxStr = NULL;
  try {
    reader->getChar('(',false);
    flxStr = new FlxString(false,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjSortSmp(get_doLog(), get_stream(), flxStr, get_optPara_FlxFunction("np"));
  } catch (FlxException &e) {
    if (flxStr!=NULL) delete flxStr;
    throw;
  }
}

FlxObjSortSmp::FlxObjSortSmp(bool dolog, std::string ostreamV, FlxString* isname, FlxFunction* Np)
: FlxObjOutputBase(dolog,ostreamV), isname(isname), Np(Np)
{
  
}

FlxObjSortSmp::~FlxObjSortSmp()
{
  delete isname;
  delete Np;
}

void FlxObjSortSmp::task()
{
  std::string isn = isname->eval_word();
  FlxIstream &is = data->IstreamBox.get(isn);
  FlxIstream_vector *isv = dynamic_cast<FlxIstream_vector*>(&is);
  if (isv!=NULL) {
    isv->sortStream();
    GlobalVar.slog(3) << "sortsmp: sorted vector-input stream '" << isn << "'." << std::endl;
  } else {
    const tulong BS = Np->cast2tnlong(false);
    
    std::vector<tdouble> sortVec;
    sortVec.reserve(BS);
    tdouble d;
    while (is.get_value(d,true)) {
      sortVec.push_back(d);
    };
    std::sort(sortVec.begin(),sortVec.end());
    std::vector<tdouble>::iterator it;
    for (it=sortVec.begin(); it!=sortVec.end(); ++it) {
      sout() << GlobalVar.Double2String(*it) << std::endl;
    }
  }
}


FlxObjReadSmpPlot::FlxObjReadSmpPlot()
{
  AllDefParaBox->insert(new FlxOptionalParaBool(true,"smpplot::autobound"));
  ParaBox.insert("autobound", "smpplot::autobound" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(0.0,"smpplot::xmin"));
  ParaBox.insert("xmin", "smpplot::xmin" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(1.0,"smpplot::xmax"));
  ParaBox.insert("xmax", "smpplot::xmax" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(1.0,"smpplot::binestimator"));
  ParaBox.insert("binestimator", "smpplot::binestimator" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(0.0,"smpplot::nbins"));
  ParaBox.insert("nbins", "smpplot::nbins" );
}

FlxObjBase* FlxObjReadSmpPlot::read()
{
  FlxString* flxStr = NULL;
  FlxFunction* typeFun = NULL;
  FlxFunction* beFun = NULL;
  try {
    reader->getChar('(',false);
    flxStr = new FlxString(false,false);
    reader->getChar(',',false);
    reader->getWord("type",false);
    reader->getChar('=',false);
    typeFun = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    beFun = get_optPara_FlxFunction("binestimator");
    int binEstimator = beFun->cast2int();
    delete beFun;
    return new FlxObjSmpPlot(get_doLog(),get_stream(),flxStr,typeFun,get_optPara_bool("autobound"),get_optPara_FlxFunction("xmin"),get_optPara_FlxFunction("xmax"),binEstimator,get_optPara_FlxFunction("nbins"));
  } catch (FlxException &e) {
    if (flxStr!=NULL) delete flxStr;
    if (typeFun!=NULL) delete typeFun;
    if (beFun!=NULL) delete beFun;
    throw;
  }
}

FlxObjSmpPlot::FlxObjSmpPlot(bool dolog, std::string ostreamV, FlxString* isname, FlxFunction* typeFun, bool autoBound, FlxFunction* xminU, FlxFunction* xmaxU, int binEstimator, FlxFunction* NbinsU)
: FlxObjOutputBase(dolog, ostreamV), isname(isname), typeFun(typeFun), autoBound(autoBound), xminU(xminU), xmaxU(xmaxU), binEstimator(binEstimator), NbinsU(NbinsU)
{

}

FlxObjSmpPlot::~FlxObjSmpPlot()
{
  delete isname;
  delete typeFun;
  delete xminU;
  delete xmaxU;
  delete NbinsU;
}

void FlxObjSmpPlot::task()
{
  int type = typeFun->cast2int();
  std::string isname_str = isname->eval_word(true);
  FlxIstream_vector *isv = dynamic_cast<FlxIstream_vector*>(&(data->IstreamBox.get(isname_str)));
  if (isv==NULL) {
    std::ostringstream ssV;
    ssV << "The input stream'" << isname_str << "' is not a vector-input stream.";
    throw FlxException_NeglectInInteractive("FlxObjSmpPlot::task_1", ssV.str() );
  }
  isv->reset_stream();
  tnlong N = isv->get_total_size();
  
  if (type==1) {	// cumulative frequency diagram
    tdouble pdx = 1.0/N;
    tdouble px = 0.0;
    tdouble d;
    isv->get_value(d,true);
    tdouble d_prev = d;
    bool b = true;
    while (b) {
      d_prev = d;
      sout() << GlobalVar.Double2String(d,true) << "\t" << GlobalVar.Double2String(px) << std::endl;
      while (d==d_prev) {
	px+=pdx;
	if ((b=isv->get_value(d,true))==false) break;
      }
      sout() << GlobalVar.Double2String(d_prev,true) << "\t" << GlobalVar.Double2String(px) << std::endl;
    }
  } else if (type==2 || type==3 || type==4 || type==5 ) {	// histogram
    // get the bounds
      tdouble min, max;
      if (autoBound) {
	min=isv->get_first_value();
	max=isv->get_last_value();
	if (min>=max) {
	std::ostringstream ssV;
	ssV << "Set vector-input stream appears not to be sorted.";
	throw FlxException_NeglectInInteractive("FlxObjSmpPlot::task_3", ssV.str() );
      }
      } else {
	min=xminU->calc();
	max=xmaxU->calc();
	if (min>=max) {
	std::ostringstream ssV;
	ssV << "xmin(" << min << ") has to be smaller than xmax(" << max << ").";
	throw FlxException_NeglectInInteractive("FlxObjSmpPlot::task_4", ssV.str() );
      }
      }
    // get the number of bins
      tuint Nbins = NbinsU->cast2tuintW0(false);
      if (Nbins==0) {
	if (binEstimator==1) {
	  Nbins = tuint(round_flx( sqrt((tdouble)N) )) ;
	} else if (binEstimator==2) {
	  Nbins = tuint(round_flx( 1.0 + 3.3 * log10(tdouble(N)) )) ;
	} else {
	  std::ostringstream ssV;
	  ssV << "'binEstimator' has to be eigher 1 or 2 and not '" << binEstimator << "'.";
	  throw FlxException_NeglectInInteractive("FlxObjSmpPlot::task_5", ssV.str() );
	}
	if (Nbins<5) Nbins = 5;
	else if (Nbins > 25) Nbins = 25;
      }
    tdouble dx = (max-min)/Nbins;
    tdouble xm = min + dx/2.0;
    tdouble xe = min + dx;
    tnlong xn = 0;
    
    tdouble d;
    bool stream_empty = false;
    if (!isv->get_value(d,true)) stream_empty = true;
    if (type == 5 ) {
      sout() << GlobalVar.Double2String(xm-dx,true) << "\t" << GlobalVar.Double2String(0.0) << std::endl;
    }
    for (tuint i=1;i<=Nbins;++i) {
      do {
	if (!stream_empty) {
	  if (d<xe) {
	    ++xn;
	  } else {
	    break;
	  }
	} else break;
	if (!isv->get_value(d,true)) stream_empty = true;
      } while (true);
      if (i==Nbins && !stream_empty) {
	while (d<=xe) {
	  ++xn;
	  if (!isv->get_value(d,true)) {
	    stream_empty = true;
	    break;
	  }
	}
      }
      sout() << GlobalVar.Double2String(xm,true) << "\t";
      if (type==2) {
	sout() << GlobalVar.Double2String(xn);
      } else if (type == 3) {
	sout() << GlobalVar.Double2String(xn/(tdouble)N);
      } else if (type == 4 || type==5 ) {
	sout() << GlobalVar.Double2String(xn/(N*dx));
      }
      if (type != 5 ) {
	sout() << "\t" << GlobalVar.Double2String(dx);
      }
      sout()  << std::endl;;
      xe+=dx;
      xm+=dx;
      xn=0;
      if (stream_empty) break;
    }
    if (type == 5 ) {
      sout() << GlobalVar.Double2String(xm,true) << "\t" << GlobalVar.Double2String(0.0) << std::endl;
    }
  } else {
    std::ostringstream ssV;
    ssV << "Unknown type of plot (" << type << ").";
    throw FlxException_NeglectInInteractive("FlxObjSmpPlot::task_2", ssV.str() );
  }
  
}

const tdouble FunRFerr::calc()
{
  const tuint idx = tuint_from(ParaListP[0]->calc(),"Index of random field");
  const int erM = int(round_flx(ParaListP[1]->calc()));
  FlxRandomField_base* rf = data->RFbox.get(idx);
  return rf->get_domain_error(erM);
}

