/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef flx_rnd_H
#define flx_rnd_H

#include "flxobjects.h"

class FlxCreateObjReaders_RND : public FlxCreateObjReaders {
  public:
    void createObjReaders (FlxObjectReadBox* objReadBox );
    void createFunReaders (FlxData* dataBox );
};


//======================== Monte Carlo Integration ===========================

/**
* @brief object class: Monte Carlo Integration
*
* MCI (const-name, Np, g) { ... };
*/
class FlxObjMCI : public FlxObjBase {
  protected:
    /**
    * @brief number of sampling points
    */
    unsigned int Np;
    /**
    * @brief pointer to the variable storing the result of the calculation (the integral)
    */
    tdouble& theResult;
    /**
    * @brief the function used to calculate Np
    */
    FlxFunction* funNp;
    /**
    * @brief the function to integrate
    */
    FlxFunction* fung;
    /**
    * @brief the MCI_CODE
    */
    FlxObjBase* InternList;
    /**
    * @brief true: split the sampling into intervals
    */
    bool interv;
    /**
    * @brief log verbose output
    */
    bool verboseLog;
    /**
    * @brief Integration in order to obtain reliability 
    */
    bool reliability;
    /**
    * @brief probability of exceedance (used for confidence studies)
    */
    FlxMtxFun* pc;
    /**
    * @brief do one Integration-Step (needed for deriving Importance Sampling)
    */
    #if FLX_KAHAN_MCI
      virtual void Integrationstep(pdouble& Integ, tdouble &hits) { 
    #else
      virtual void Integrationstep(tdouble& Integ, tdouble &hits) { 
    #endif
	data->RndBox.gen_smp(); InternList->exec(); tdouble t=fung->calc(); Integ+=t; if (t>0) ++hits; }
    /**
    * @brief needed to calculate FlxMtxFun (faster) (needed for deriving Importance Sampling)
    */
    virtual void FirstThingsFirst();
    /**
    * @brief needed for deriving Importance Sampling
    */
    virtual void LastThingsLast() {}
    /**
    * @brief log additional information after integration
    */
    virtual void log_AddResInfo(std::ostream &sout, const tdouble hits, const tdouble Npd);
    
    void task();
  public:
    FlxObjMCI ( bool dolog, tdouble& theResult, FlxFunction* funNp, FlxFunction* fung, FlxObjBase* InternList, bool interv, bool verboseLog, bool reliability, FlxMtxFun* pc ) 
      : FlxObjBase(dolog),  Np(0),theResult(theResult), funNp(funNp), fung(fung), InternList(InternList),interv(interv), verboseLog(verboseLog), reliability(reliability), pc(pc) {}
    virtual ~FlxObjMCI();
};

/**
* @brief object read class: for FlxObjMCI
*/
class FlxObjReadMCI : public FlxObjReadCtrlStrct {
  protected:
    /**
    * @brief read the settings of the Monte Carlo integration
    */
    void read_MCIblock(tdoublePtr& theResult, FlxFunctionPtr& funNp, FlxFunctionPtr& fung, bool errSerious=true);
  public:
    FlxObjReadMCI();
    FlxObjBase* read ();
};

//======================== Importance Sampling ===========================

/**
* @brief object class: Importance Sampling
*
* IPS (const-name, Np, g) SMPL_SPACE { ... };
*/
class FlxObjIpS : public FlxObjMCI {
  protected:
    FlxRndSamplingSpace_Generator_base* sspaceG;
    FlxRndSamplingSpace_base* sspace;
    #if FLX_KAHAN_MCI
      virtual void Integrationstep(pdouble& Integ, tdouble &hits);
    #else
      virtual void Integrationstep(tdouble& Integ, tdouble &hits);
    #endif
    void FirstThingsFirst();
    void LastThingsLast() {delete sspace; }
    void log_AddResInfo(std::ostream &sout, const tdouble hits, const tdouble Npd);
  public:
    FlxObjIpS ( bool dolog, tdouble& theResult, FlxFunction* funNp, FlxFunction* fung, FlxObjBase* InternList, bool interv, bool verboseLog, bool reliability, FlxMtxFun* pc, FlxRndSamplingSpace_Generator_base* sspaceG) 
      : FlxObjMCI(dolog, theResult,funNp,fung,InternList,interv,verboseLog,reliability,pc), sspaceG(sspaceG) {}
    ~FlxObjIpS() { delete sspaceG; }
};

/**
* @brief object read class: for FlxObjIpS
*/
class FlxObjReadIpS : public FlxObjReadMCI {
  public:
    FlxObjBase* read ();
};


//======================== Subset Simulation ===========================

/**
* @brief object class: manages the samples of the subset-simulation
*
* (at least) two instances of this object are created during the actual subset simulation
*/

class SSS_sample_list {
  private:
    /**
    * @brief number of random variables in the list
    */
    const tuint N_RV;
    /**
    * @brief number of samples per SSS-iteration-step
    */
    const tuint Ns;
    /**
    * @brief number of samples to condition on in subsequent steps
    */
    const tuint Nc;
    /**
    * @brief number of entries in smpl_ref_list
    */
    tuint N_SRL;
    /**
    * @brief number hits with g<=0 counted
    */
    tuint N_hits;
    
    /**
    * @brief sorted list - contains the id of the corresponding smpl_list-entry - in ascenting order (w.r.t. g)
    */
    int* smpl_ref_list;
    /**
    * @brief sorted list - contains the g-value - in ascending order
    */
    tdouble* g_list;
    /**
    * @brief a list of samples referenced to in smpl_ref_list
    */
    tdouble* smpl_list;
    tdouble g_kicked_out;
    const tdouble consider_g_kicked_factor;
    
    /**
    * @brief find the smallest index in smpl_ref_list which is larger than g
    */
    const tuint find_pos(const tdouble g) const;
    
  public:
    /**
     * @param cgkf consider g kicked -> factor of it [0;1[
    */
    SSS_sample_list(const tuint N_RV, const tuint Ns, const tuint Nc, const tdouble cgkf);
    ~SSS_sample_list();
    
    /**
    * @brief register a newly generated sample
    * @param g result of the limit state function evaluated at y_ptr
    * @param y_ptr the current vector of random numbers - in uncorrelated standard normal space
    */
    void register_sample(const tdouble g, const tdouble* const y_ptr);
    /**
    * @brief return the largest g in the list
    */
    const tdouble get_treshold_g() const { return (N_SRL==Nc)?(g_list[N_SRL-1]*(1.-consider_g_kicked_factor)+(g_kicked_out*consider_g_kicked_factor)):(g_list[N_SRL-1]); }
    const tdouble get_g_lastPoint() const { return g_list[N_SRL-1]; }
    const tdouble get_g_kickedOut() const { return g_kicked_out; }
    const bool is_listFull() const { return Nc==N_SRL; }
    /**
    * @brief return observed probability of g<=0
    */
    const tdouble get_Pf() const { return tdouble(N_hits)/tdouble(Ns); }
    const tuint get_N_hits() const { return N_hits; }
    /**
    * @brief clear the list
    */
    void clear();
    /**
    * @brief dectects how many values appear multiple times
    * @param Nsame_seq the number of sequences appearing multiple times
    * @param Nmax_seq the maximum length of a sequence appearing multiple times
    * @param Ntotal the total number of entries appearing more than once
    * @param Nconsid the total number of samples considered (<=gt!!!)
    * @param gt only values <=gt are considered in this study
    */
    void find_multiples(tuint& Nsame_seq, tuint& Nmax_seq, tuint& Ntotal, tuint& Nconsid, const tdouble &gt) const;
    
    const int* const get_reference_smpl_ref_list() const { return smpl_ref_list; }
    const tdouble* const get_reference_g_list() const { return g_list; }
    const tdouble* const get_reference_smpl_list() const { return smpl_list; }
};


/**
* @brief object class: bayesian estimator for subset simulation in post-processing
*/
class SSS_bne_base {
  protected:
    
  public:
    /**
    * @brief obtain a sample of the probability of failure
    */
    virtual const tdouble sim_pf() = 0;
    /**
    * @brief Expectation of the probability of failure (Bayesian estimate)
    */
    virtual const tdouble E_Pf() const = 0;
    /**
    * @brief Standard deviation of the probability of failure (Bayesian estimate)
    */
    virtual const tdouble sigma_Pf() const = 0;
    /**
    * @brief Expectation of the squared probability of failure (Bayesian estimate)
    */
    virtual const tdouble E2_Pf() const = 0;
    
    static FlxRndBox* RndBox;
};

class SSS_bne_MCI : public SSS_bne_base {
  protected:
    /**
    * @brief observed Y hits within N samples
    */
    const tuint Y;	// number of hits
    const tuint N;	// number of samples
    const FlxRnd_Beta distri;
  public:
    SSS_bne_MCI(const tuint Y, const tuint N);
    virtual const tdouble sim_pf();
    virtual const tdouble E_Pf() const;
    virtual const tdouble sigma_Pf() const;
    virtual const tdouble E2_Pf() const;
};

struct cs { bool changed; tdouble g; };
class SSS_bne_MCMC : public SSS_bne_base {
  protected:
    const bool random_cpt;		// random generation of cpt's
    const tdouble prior_cpt;		// use a prior in cpt (define weighting factor)
    const tuint level;			// level of cpt's: 1-2
    const tuint N1;			// number of samples to estimate E[Pf] in one interation steps (lower bound)
    const tuint N11;			// "" (upper bound)
    const tuint N2;			// number of interation steps to estimate E[Pf]
    const tuint N3;			// number of samples to estimate the final E[Pf]
    /**
    * @brief number of seeds per conditioning step
    */
    const tuint Nc;
    /**
    * @brief length of one Markov chain
    */
    const tuint Nlc;
    /**
    * @brief vector storing some integer values
    * 	0: N_total
    * 	1: Y_total
    * 	2: N|i-1:0
    * 	3: Y|i-1:0
    * 	4: N|i-1:1
    * 	5: Y|i-1:1
    */
    tuint Nvec[14];
    /**
    * @brief vector storing some double values
    * 	f - failure
    * 	0: Pr(f)
    * 	1: Pr(f|i-1:0)
    * 	2: Pr(f|i-1:1)
    */
    tdouble Dvec[7];
    /**
    * @brief estimate of E[f]
    */
    tdouble E_pf;
    tdouble E_pf2;	// estimate of square of pf
    tdouble sd_pf;	// estimate of the standard deviation of pf
    
    /**
    * @brief determines E[f]
    */
    void initialize_1();
    void init_cpt();
//     void pow_mtx(const tuint n, tdouble* res, const tdouble* const evpc, const tdouble* const evec, const tdouble* const eveci);
  public:
    SSS_bne_MCMC(const tuint Nc, const tuint Nlc, cs* chain_structure, const tdouble g_limit);
    
    virtual const tdouble sim_pf();
    virtual const tdouble E_Pf() const;
    virtual const tdouble sigma_Pf() const;
    virtual const tdouble E2_Pf() const;
//     const tdouble* write_test() const { return Dvec; }
};

class SSS_bne_Box {
  private:
    /**
    * @brief the number of entries in the box
    */
    tuint N;
    std::vector<SSS_bne_base*> box;
  public:
    SSS_bne_Box();
    ~SSS_bne_Box();
    
    void add_bne(SSS_bne_base* el);
    const tdouble sim_pf() const;
    const tdouble E_Pf() const;
    const tdouble Cov_Pf() const;
    const SSS_bne_base& get_last() const;
//     void write_test() const;
};

/**
* @brief object class: base-class for managing the samples of the subset-simulation
*/
class SSS_adaptive_ctrl;
class SSS_csm_MCMC_base {
  protected:
    /**
    * @brief number of random variables in the list
    */
    const tuint N_RV;
  private:
    /**
    * @brief number of samples per SSS-iteration-step
    */
    const tuint ID_Ns;
    /**
    * @brief number of samples to condition on in subsequent steps
    */
    const tuint Nc;
    /**
    * @brief current threshold value
    */
    tdouble g_treshold;
    /**
    * @brief previously evaluated value of g
    */
    tdouble g_last;
    /**
    * @brief vector containing the previous random samples
    */
    tVec y_prev;
    
    cs* chain_structure;
  protected:
    /**
    * @brief vector containing the proposed random samples
    */
    tVec y_prop;
    tdouble* const yp_prev;
    tdouble* const yp_prop;
    
    FlxRndBox& RndBox;
    SSS_adaptive_ctrl& adptvCtrl;
    
  private:
    /**
    * @brief collects information for a Bayesian estimate of the result
    */
    SSS_bne_Box& bn_box;
    const bool bne;
    const SSS_sample_list* smpl_cond;
    SSS_sample_list* smpl_new;
    FlxFunction* fung;
    /**
    * @brief REFERENCE TO SSS_sample_list::
    */
    const int* smpl_ref_list;
    const tdouble* g_list;
    const tdouble* smpl_list;
    
    /**
    * @brief execute before an entire sampling block is generated
    */
    void register_samples(const SSS_sample_list* const smpl_cond_v, SSS_sample_list* smpl_new_v);
    /**
    * @brief register a new sample point (from the list) ... i.e. use as new seed
    * @param curr_ID the ID of the sample to use as seed
    * @note update curr_ID first!
    */
    void register_new_smpl_line(const tuint curr_ID);
    
    virtual const bool propose_sample(tVec& y_smpl) = 0;
    
  public:
    SSS_csm_MCMC_base(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne);
    virtual ~SSS_csm_MCMC_base();
    
    /**
    * @brief generate all samples for one conditioning step in SSS
    */
    void generate_smpl(const SSS_sample_list* const smpl_cond_v, SSS_sample_list* smpl_new_v, tdouble& acr, tdouble& nacr1, tdouble& velo, FlxProgress& prg);
    /**
    * @brief outputs information concerning the adaptive setting
    */
    virtual void write_adaptive_info(std::ostream& sout) = 0;
    virtual void adptv_acr_up() = 0;
    virtual void adptv_acr_down() = 0;
    
    static SSS_csm_MCMC_base* new_object(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, const std::string& smplTypeStr, const std::string& kernelName, const tdouble h, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl_V, SSS_bne_Box& bn_box, const bool bne, const bool errSerious=true);
};

/**
* @brief object class: samples MCMC with component-wise MH-algorithm ... manages the samples of the subset-simulation
*/
class SSS_csm_cwMH_MCMC : public SSS_csm_MCMC_base {
  private:
    /**
    * @brief the kernel to use for MCMC-sampling
    */
    FlxRndKernel_base* kernel;
    
    const bool propose_sample(tVec& y_smpl);
  public:
    SSS_csm_cwMH_MCMC(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, const std::string& kernelName, const tdouble h, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne);
    ~SSS_csm_cwMH_MCMC();
    
    void write_adaptive_info(std::ostream& sout);
    void adptv_acr_up();
    void adptv_acr_down();
};

/**
* @brief object class: samples MCMC with component-wise MH-algorithm ... manages the samples of the subset-simulation
*/
class SSS_csm_mci : public SSS_csm_MCMC_base {
  private:
    const bool propose_sample(tVec& y_smpl);
  public:
    SSS_csm_mci(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne);
    ~SSS_csm_mci() {}
    
    void write_adaptive_info(std::ostream& sout) {}
    void adptv_acr_up() {}
    void adptv_acr_down() {}
};

/**
* @brief object class: conditional sampling from U-Space (MCMC) ... manages the samples of the subset-simulation
*/
class SSS_csm_csusMH_MCMC : public SSS_csm_MCMC_base {
  private:
    tdouble rho;
    tdouble sD;
    
    const bool propose_sample(tVec& y_smpl);
  public:
    SSS_csm_csusMH_MCMC(const tuint N_RV, const tuint ID_Ns, const tuint Nc, FlxFunction* fung, const tdouble rho, FlxRndBox& RndBox, SSS_adaptive_ctrl& adptvCtrl, SSS_bne_Box& bn_box, const bool bne);
    
    void write_adaptive_info(std::ostream& sout);
    void adptv_acr_up();
    void adptv_acr_down();
};


/**
* @brief controls the adaptive process
*/
class SSS_adaptive_ctrl {
  private:
    tdouble factor_d;
    tdouble lower_d;
    tdouble upper_d;
    FlxFunction* factor;
    FlxFunction* lower;
    FlxFunction* upper;
    FlxFunction* maxUpdatesPerCStep;
  public:
    SSS_adaptive_ctrl(FlxFunction* factor, FlxFunction* lower, FlxFunction* upper, FlxFunction* maxUpdatesPerCStep);
    ~SSS_adaptive_ctrl();
    
    const tdouble get_factor() const { return factor_d; }
    const bool is_adaptive() const { return factor_d>GlobalVar.TOL(); }
    /**
    * @brief controls the adaptive process
    */
    void eval();
    /**
    * @brief checks if an adaptive adaption is required
    * @returns 0: no; 1: yes, acr too small; 2: yes: acr too large
    * @param acr acceptance rate in last sub-step
    */
    const int requires_adptv_step(const tdouble acr) const;
    /**
    * @brief returns the number of adaptive updates per conditioning step
    */
    const tuint get_maxUpdatesPerCStep();
};


/**
* @brief object class: Subset simulation
*
* subsetsim (const-name, NpS, NpC, g) COND_SMP_METHOD { ... };
*/
class FlxObjSubSetSim : public FlxObjBase {
  private:
    /**
    * @brief pointer to the variable storing the result of the calculation (the integral)
    */
    tdouble& theResult;
    tdouble* theResult_bne;
    tdouble* theResult_bne_cov;
    /**
    * @brief number of samples at each step
    */
    FlxFunction* funNsM;
    /**
    * @brief number of samples to condition on (in each step)
    */
    FlxFunction* funNc;
    /**
    * @brief the function to integrate
    */
    FlxFunction* fung;
    // options describing the sampling procedure
      /**
      * @brief the type of the sampling procedure
      */
      FlxString* smplTypeStr;
      /**
      * @brief the name of the kernel to use for MCMC-sampling
      */
      FlxString* kernelName;
      /**
      * @brief the initial bandwidth of the kernel
      */
      FlxFunction* h;
    SSS_adaptive_ctrl* adptvCtrl;
    /**
    * @brief factor with which to consider the last kicked out g ... [0;1[
    */
    const tdouble cgkf;
    /**
    * @brief if true: do a Bayesian estimate !!!
    */
    const bool bne;
    /**
    * @brief log verbose output
    */
    bool verboseLog;
    
    void task();
  public:
    FlxObjSubSetSim ( bool dolog, tdouble& theResult, FlxFunction* funNsM, FlxFunction* funNc, FlxFunction* fung, FlxString* smplTypeStr, FlxString* kernelName, FlxFunction* h, SSS_adaptive_ctrl* adptvCtrl, const tdouble cgkf, const bool bne, bool verboseLog );
    virtual ~FlxObjSubSetSim();
    
};

/**
* @brief object read class: for FlxObjIpS
*/
class FlxObjReadSubSetSim : public FlxObjReadMCI {
  public:
    FlxObjReadSubSetSim();
    FlxObjBase* read ();
};


//======================== FORM ===========================


class FlxObjFORM_base : public FlxObjOutputBase {
  protected:
    tuint DIM;  
    FlxFunction* LSF;
    FlxFunction* fdstep;
    FlxFunction* eps1;
    FlxFunction* eps2;
    FlxFunction* iHLRF_lambda_start;
    FlxFunction* iHLRF_epsilon;
    FlxFunction* iHLRF_reduce;
    FlxMtxFun* dx_min;
    unsigned int maxIter;
    /**
    * @brief id of the finite difference method
    * 1: forward
    * 2: central
    * 3: backward
    */
    int fd_method;
    /**
    * @brief id of the optimization method
    * 1: HL-RF
    * 2: iHL-RF
    * 3: iGP
    */
    int opt_method;
    bool dxdyAnalytical;
    const bool verboseLog;
    void update_Start();
    tVec do_FORM(tVec &x, tVec &y, tdouble &beta_new);
  public:
    FlxObjFORM_base ( bool dolog, FlxFunction* LSFv, FlxFunction* fdstepV, FlxFunction* eps1, FlxFunction* eps2, FlxFunction* iHLRF_lambda_start, FlxFunction* iHLRF_epsilon, FlxFunction* iHLRF_reduce, unsigned int maxIter, bool verbose, std::string ostreamV, bool dxdyAnalytical, FlxMtxFun* dx_min, int fd_method, int opt_method ) 
     : FlxObjOutputBase(dolog,ostreamV), LSF(LSFv), fdstep(fdstepV), eps1(eps1), eps2(eps2), iHLRF_lambda_start(iHLRF_lambda_start),iHLRF_epsilon(iHLRF_epsilon), iHLRF_reduce(iHLRF_reduce), dx_min(dx_min), maxIter(maxIter), fd_method(fd_method),opt_method(opt_method), dxdyAnalytical(dxdyAnalytical), verboseLog(verbose) {}
    ~FlxObjFORM_base();
};


/**
* @brief object class
*
* form ( Y, X, LSF );
*/
class FlxObjFORM : public FlxObjFORM_base {
  private:
    std::string cnamey;
    std::string cnamex;
    FlxMtxFun* xstart;
    FlxString* betaDP;
    void task();
  public:
    FlxObjFORM ( bool dolog, std::string cnamey, std::string cnamex, FlxMtxFun* xstartV, FlxFunction* LSFv, FlxFunction* fdstepV, FlxFunction* eps1, FlxFunction* eps2, FlxFunction* iHLRF_lambda_start, FlxFunction* iHLRF_epsilon, FlxFunction* iHLRF_reduce, unsigned int maxIter, FlxString* betaDP, bool verbose, bool dxdyAnalytical, FlxMtxFun* dx_min, int fd_method, int opt_method ) 
     : FlxObjFORM_base(dolog,LSFv,fdstepV,eps1,eps2,iHLRF_lambda_start,iHLRF_epsilon,iHLRF_reduce,maxIter,verbose,"cout",dxdyAnalytical,dx_min,fd_method,opt_method), cnamey(cnamey), cnamex(cnamex), xstart(xstartV), betaDP(betaDP) {}
    ~FlxObjFORM();
};

class FlxObjReadFORM_base : public FlxObjReadOutputBase {
  protected:
    const tuint get_maxiter();
    const int get_fd_method();
    const int get_opt_method();
    const bool get_verboseLog();
  public:
    FlxObjReadFORM_base();
};

/**
* @brief object read class: for FlxObjFORM
*/
class FlxObjReadFORM : public FlxObjReadFORM_base {
  public:
    FlxObjReadFORM();
    FlxObjBase* read ();
};

/**
* @brief object class
*
* form_pdf (  );
*/
class FlxObjFORM_pdf : public FlxObjFORM_base {
  private:
    FlxFunction* rvfun;
    FlxFunction* lboundF;
    FlxFunction* uboundF;
    FlxFunction* nintervalF;
    FlxMtxFun* xstart;
    void task();
  public:
    FlxObjFORM_pdf ( bool dolog, FlxFunction* rvfun, FlxFunction* lboundF, FlxFunction* uboundF, FlxFunction* nintervalF, FlxMtxFun* xstartV, FlxFunction* fdstepV, FlxFunction* eps1, FlxFunction* eps2, FlxFunction* iHLRF_lambda_start, FlxFunction* iHLRF_epsilon, FlxFunction* iHLRF_reduce, unsigned int maxIter, bool verbose, std::string ostreamV, bool dxdyAnalytical, FlxMtxFun* dx_min, int fd_method, int opt_method );
    ~FlxObjFORM_pdf();
};

/**
* @brief object read class: for FlxObjFORM_pdf
*/
class FlxObjReadFORM_pdf : public FlxObjReadFORM_base {
  public:
    FlxObjReadFORM_pdf();
    FlxObjBase* read ();
};


//======================== Response Surface ===========================


class FlxRSM_Fun_base : public FlxReaderBase3, public FlxDataBase {
  protected:
    tuint DIM;
    tVec* paraVec;
  public:
    FlxRSM_Fun_base() : DIM( 0 ), paraVec(NULL) {}
    FlxRSM_Fun_base(tuint DIM, tVec* paraVec) : DIM(DIM), paraVec(paraVec) {}
    virtual ~FlxRSM_Fun_base();
    virtual FlxRSM_Fun_base* copy() = 0;
    virtual tuint get_NumbOfTerms() const = 0;
    virtual void calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec) = 0;
    void set_paraVec(tVec* paraVecV);
    void set_para(tuint index, tdouble paraV);
    tdouble calc_solution(const tVec& yVec, const tVec& xVec);
    virtual void updateStart() { DIM = data->RndBox.get_NumbRnd(); };
    const tdouble get_para(tuint index);
    const tVec &get_paraVec() { return *paraVec; }
    
    static FlxRSM_Fun_base* createRSMfun(bool errSerious=true, const std::string ftype="" );
};

class FlxRSM_Fun_p1 : public FlxRSM_Fun_base {
  public:
    FlxRSM_Fun_p1() : FlxRSM_Fun_base() {}
    FlxRSM_Fun_p1(tuint DIM, tVec* paraVec) : FlxRSM_Fun_base(DIM,paraVec) {}
    FlxRSM_Fun_base* copy();
    virtual tuint get_NumbOfTerms() const { return DIM+1; }
    void calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec);
};

class FlxRSM_Fun_p2womt : public FlxRSM_Fun_base {
  public:
    FlxRSM_Fun_p2womt() : FlxRSM_Fun_base() {}
    FlxRSM_Fun_p2womt(tuint DIM, tVec* paraVec) : FlxRSM_Fun_base(DIM,paraVec) {}
    FlxRSM_Fun_base* copy();
    virtual tuint get_NumbOfTerms() const { return 2*DIM+1; }
    void calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec);
};

class FlxRSM_Fun_p2wmt : public FlxRSM_Fun_p2womt {
  public:
    FlxRSM_Fun_p2wmt() : FlxRSM_Fun_p2womt() {}
    FlxRSM_Fun_p2wmt(tuint DIM, tVec* paraVec) : FlxRSM_Fun_p2womt(DIM,paraVec) {}
    FlxRSM_Fun_base* copy();
    virtual tuint get_NumbOfTerms() const { return 2*DIM+1+(DIM*DIM-DIM)/2; }
    void calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec);
};


class FlxCombination {
  private:
    tdouble start;
    tdouble end;
    tdouble step;
    tuint dim;
    tVec cv;
    bool runs;
  public:
    FlxCombination(tdouble start, tdouble end, tdouble step, tuint dim);
    tVec Next();
};

class FlxRSM_SmpPoints_base : public FlxReaderBase3, public FlxDataBase {
  protected:
    tuint DIM;
  public:
    FlxRSM_SmpPoints_base() : DIM( 0 ) {}
    virtual ~FlxRSM_SmpPoints_base() {};
    virtual tuint get_NumbOfPoints() = 0;
    virtual void get_nextPoint(tVec& y) = 0;
    virtual void updateStart() { DIM = data->RndBox.get_NumbRnd(); };
    
    static FlxRSM_SmpPoints_base* createRSMpts(bool errSerious=true);
};

class FlxRSM_SmpPoints_joa : public FlxRSM_SmpPoints_base {
  protected:
    tuint index;
    FlxMtxFun* centerYF;
    FlxFunction* kF;
    tVec* centerY;
    tdouble k;
  public:
    FlxRSM_SmpPoints_joa(bool errSerious=true);
    virtual ~FlxRSM_SmpPoints_joa();
    virtual void updateStart();
    virtual tuint get_NumbOfPoints() { return 2*DIM+1; }
    virtual void get_nextPoint(tVec& y);
};

class FlxRSM_SmpPoints_cc : public FlxRSM_SmpPoints_joa {
  private:
    FlxFunction* kF2;
    tdouble k2;
    FlxCombination* Comb;
  public:
    FlxRSM_SmpPoints_cc(bool errSerious=true);
    ~FlxRSM_SmpPoints_cc();
    void updateStart();
    tuint get_NumbOfPoints() { return 2*DIM+1+pow_b2(DIM); }
    void get_nextPoint(tVec& y);
};

class FlxRSM_SmpPoints_ff : public FlxRSM_SmpPoints_joa {
  private:
    FlxCombination* Comb;
    bool sphere;
  public:
    FlxRSM_SmpPoints_ff(bool errSerious=true);
    ~FlxRSM_SmpPoints_ff();
    void updateStart();
    tuint get_NumbOfPoints() { return pow_int(3,DIM); }
    void get_nextPoint(tVec& y);
};

class FlxRSM_SmpPoints_p1 : public FlxRSM_SmpPoints_joa {
  public:
    FlxRSM_SmpPoints_p1(bool errSerious=true) : FlxRSM_SmpPoints_joa(errSerious) { };
    tuint get_NumbOfPoints() { return DIM+1; }
    void get_nextPoint(tVec& y);
};

class FlxRSM_SmpPoints_rnd : public FlxRSM_SmpPoints_joa {
  private:
    FlxFunction* numb;
  public:
    FlxRSM_SmpPoints_rnd(bool errSerious=true);
    ~FlxRSM_SmpPoints_rnd() {delete numb;}
    tuint get_NumbOfPoints() { return numb->cast2tuint(); }
    void get_nextPoint(tVec& y);
};

class FlxRSM_Solve_base : public FlxReaderBase3, public FlxDataBase {
  public:
    virtual ~FlxRSM_Solve_base() {};
    virtual void solve(const FlxMtx& A, const tVec& b, tVec& sol) = 0;
    static FlxRSM_Solve_base* createRSMsolve(bool errSerious=true);
};

class FlxRSM_Solve_CG : public FlxRSM_Solve_base {
  private:
    FlxFunction* iterF;
    FlxFunction* epsF;
    FlxFunction* pcnF;
    bool print;
    bool printS;
  public:
    FlxRSM_Solve_CG(bool errSerious=true);
    ~FlxRSM_Solve_CG();
    void solve(const FlxMtx& A, const tVec& b, tVec& sol);
};

/**
* @brief object class: Response Surface
*
* RSM (g; RSMfunction; SmpPoints; SolMeth) { RSM_CODE};
*/
class FlxObjRSM : public FlxObjBase {
  protected:
    /**
    * @brief the approximation function
    */
    FlxRSM_Fun_base* RSMfunction;
    /**
    * @brief the sample points (or better the rule for sampling points)
    */
    FlxRSM_SmpPoints_base* SmpPoints;
    /**
    * @brief the solution method
    */
    FlxRSM_Solve_base* solMeth;
    /**
    * @brief the function to approximate
    */
    FlxFunction* fung;
    /**
    * @brief the RSM_CODE
    */
    FlxObjBase* InternList;
    
    void task();
  public:
    FlxObjRSM ( bool dolog, FlxRSM_Fun_base* RSMfunction, FlxRSM_SmpPoints_base* SmpPoints, FlxRSM_Solve_base* solMeth, FlxFunction* fung, FlxObjBase* InternList ) 
      : FlxObjBase(dolog), RSMfunction(RSMfunction),SmpPoints(SmpPoints),solMeth(solMeth),fung(fung),InternList(InternList) {}
    virtual ~FlxObjRSM();
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM : public FlxObjReadCtrlStrct {
  public:
    FlxObjBase* read ();
};


/**
* @brief object class: set a response surface function
*
* RSM_set fun RSM_function_TYPE;
*/
class FlxObjRSM_set_fun : public FlxObjBase {
  protected:
    /**
    * @brief the approximation function
    */
    FlxRSM_Fun_base* RSMfunction;
    
    void task();
  public:
    FlxObjRSM_set_fun ( bool dolog, FlxRSM_Fun_base* RSMfunction ) : FlxObjBase(dolog), RSMfunction(RSMfunction) {}
    virtual ~FlxObjRSM_set_fun();
};

/**
* @brief object class: set the coefficients of a response surface function
*
* RSM_set para index=value;
*/
class FlxObjRSM_set_para : public FlxObjBase {
  protected:
    FlxFunction* index;
    FlxFunction* value;
    void task();
  public:
    FlxObjRSM_set_para ( bool dolog, FlxFunction* index, FlxFunction* value ) : FlxObjBase(dolog), index(index), value(value) {}
    virtual ~FlxObjRSM_set_para();
};

/**
* @brief object class: set a polynomial chaos response surface function (using the projection technique)
*
* RSM_set para index=value;
*/
class FlxObjRSM_set_PCproject : public FlxObjBase {
  protected:
    FlxRSM_Fun_base* RSMfunction;
    FlxFunction* rv_fun;
    FlxFunction* intervalsF;
    FlxFunction* smpPintervF;
    FlxFunction* smpRedistF;
    FlxFunction* diffusionF;
    void task();
  public:
    FlxObjRSM_set_PCproject ( bool dolog, FlxRSM_Fun_base* RSMfunction, FlxFunction* rv_fun, FlxFunction* intervalsF, FlxFunction* smpPintervF, FlxFunction* smpRedistF, FlxFunction* diffusionF ) : FlxObjBase(dolog), RSMfunction(RSMfunction), rv_fun(rv_fun), intervalsF(intervalsF), smpPintervF(smpPintervF), smpRedistF(smpRedistF), diffusionF(diffusionF) {}
    virtual ~FlxObjRSM_set_PCproject();
};

/**
* @brief object class
*/
class FlxObjRSM_set_print_coeff : public FlxObjOutputBase {
  protected:
    void task();
  public:
    FlxObjRSM_set_print_coeff ( bool dolog, std::string ostreamV, bool verboseV) : FlxObjOutputBase(dolog,ostreamV,verboseV) {}
};

/**
* @brief object class
*/
class FlxObjRSM_set_coeff_vec : public FlxObjBase {
  protected:
    FlxString* vname;
    void task();
  public:
    FlxObjRSM_set_coeff_vec ( bool dolog, FlxString* vname) : FlxObjBase(dolog), vname(vname) {}
    ~FlxObjRSM_set_coeff_vec() { delete vname; }
};

/**
* @brief object read class: for FlxObjRSM
*/
class FlxObjReadRSM_set : public FlxObjReadOutputBase {
  public:
    FlxObjReadRSM_set();
    FlxObjBase* read ();
};


// -------------------------- Kernel Density Estimation ----------------------------------------------------------------------

/**
* @brief object read class: for FunPlot-Objects
*/
class FlxObjReadKDE : public FlxObjReadOutputBase {
  public:
    FlxObjReadKDE();
    FlxObjBase* read();
};

/**
* @brief object class: plot some functions (for GNUplot)
*
* KDE ( N, KERNEL, BANDWIDTH ) { ... };
*/
class FlxObjKDE : public FlxObjOutputBase {
  protected:
    FlxString* HLRVstring;
    FlxFunction* N;
    FlxRndKernel_base* kernel;
    FlxFunction* h;
    FlxFunction* lbound;
    FlxFunction* ubound;
    FlxFunction* Ninterval;
    bool do_cdf;
    virtual void task();
  public:
    FlxObjKDE( bool dolog, FlxString* HLRVstring, FlxFunction* N, FlxRndKernel_base* kernel, FlxFunction* h, FlxFunction* lbound, FlxFunction* ubound, FlxFunction* Ninterval, bool do_cdf, std::string ostreamV );
    ~FlxObjKDE();
};


// -------------------------- High Level Random Variables ----------------------------------------------------------------------

/**
* @brief Returns as realization a realization from an evaluation of a FlxFunction
*/
class FlxHLRVfun : public FlxHLRVbase {
  private:
    FlxFunction* fun;
    FlxObjBase* Icode;
  public:
    FlxHLRVfun(FlxFunction* fun, FlxObjBase* Icode) : fun(fun), Icode(Icode) {};
    ~FlxHLRVfun();
    virtual const tdouble get_Realization();
    virtual const std::string print_info();
};

/**
* @brief object class: plot some functions (for GNUplot)
* 
* HLRV NAME = FlxFunction { CODE };
*/
class FlxObjHLRV : public FlxObjBase {
  protected:
    FlxString* hlrv_name;
    FlxFunction* fun;
    FlxObjBase* Icode;
    virtual void task();
  public:
    FlxObjHLRV( const bool dolog, FlxString* hlrv_name, FlxFunction* fun, FlxObjBase* Icode ) : FlxObjBase(dolog), hlrv_name(hlrv_name), fun(fun), Icode(Icode) {}
    ~FlxObjHLRV();
};

/**
* @brief object read class: for FunPlot-Objects
*/
class FlxObjReadHLRV : public FlxObjReadCtrlStrct {
  public:
    FlxObjReadHLRV() { data->IgnoreBox.set_iL_recur("hlrv"); }
    FlxObjBase* read();
};


// -------------------------- Statistical Commands ----------------------------------------------------------------------


/**
* @brief object read class: for statsmp-Objects
*/
class FlxObjReadStatSmp : public FlxObjReadOutputBase {
  public:
    FlxObjReadStatSmp();
    FlxObjBase* read();
};

/**
* @brief object class: descriptive statistics of a sample set
*/
class FlxObjStatSmp : public FlxObjOutputBase {
  protected:
    /**
    * @brief the name of the input stream
    */
    FlxString* isname;
    /**
    * @brief the string to add to the name of the const-variables (to store calculated values)
    */
    FlxString* add_name;
    /**
    * @brief required for qdouble-summation
    */
    FlxFunction* Np;
    /**
    * @brief the option (1: single sample set; 2: 2 correlation of 2 sample sets)
    */
    int optionP;
    
    void task();
  public:
    FlxObjStatSmp( bool dolog, std::string ostreamV, FlxString* isname, FlxString* add_name, FlxFunction* Np, int optionP );
    ~FlxObjStatSmp();
};


/**
* @brief object read class: for sortsmp-Objects
*/
class FlxObjReadSortSmp : public FlxObjReadOutputBase {
  public:
    FlxObjReadSortSmp();
    FlxObjBase* read();
};

/**
* @brief object class: sort numbers in a file
*/
class FlxObjSortSmp : public FlxObjOutputBase {
  protected:
    /**
    * @brief the name of the input stream
    */
    FlxString* isname;
    /**
    * @brief maxiumum block-size for sorting
    */
    FlxFunction* Np;
    
    void task();
  public:
    FlxObjSortSmp( bool dolog, std::string ostreamV, FlxString* isname, FlxFunction* Np );
    ~FlxObjSortSmp();
};

/**
* @brief object read class: for smpplot-Objects
*/
class FlxObjReadSmpPlot : public FlxObjReadOutputBase {
  public:
    FlxObjReadSmpPlot();
    FlxObjBase* read();
};

/**
* @brief object class: plotting histograms, ...
*/
class FlxObjSmpPlot : public FlxObjOutputBase {
  protected:
    /**
    * @brief the name of the input stream
    */
    FlxString* isname;
    /**
    * @brief type of the plot
    */
    FlxFunction* typeFun;
    
    bool autoBound;
    FlxFunction* xminU;
    FlxFunction* xmaxU;
    int binEstimator;
    FlxFunction* NbinsU;
    
    void task();
  public:
    FlxObjSmpPlot( bool dolog, std::string ostreamV, FlxString* isname, FlxFunction* typeFun, bool autoBound, FlxFunction* xminU, FlxFunction* xmaxU, int binEstimator, FlxFunction* NbinsU );
    ~FlxObjSmpPlot();
};

// -------------------------- FlxFunction ----------------------------------------------------------------------


class FunRSM : public FunBaseFun_multPara, public FlxDataBase {
  public:
    FunRSM (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write() { return "rsm()"; }
    const std::string write_v() { return "rsm"; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(tdouble *thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunRSM : public FunReadFunBase {
  private:
    static FlxRSM_Fun_base* RSMfunction;
  public:
    ~FunReadFunRSM() { if (RSMfunction!= NULL) { delete RSMfunction; RSMfunction = NULL; } }
    FunBase* read ( bool errSerious ) { return new FunRSM( read_parameters(0,errSerious) );}
    
    static void set_RSMfunction(FlxRSM_Fun_base* RSMfunctionV);
    static FlxRSM_Fun_base* get_RSMfunction();
};

class FunRSM_terms : public FunBaseFun_multPara, public FlxDataBase {
  public:
    FunRSM_terms (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write() { return "rsm_terms()"; }
    const std::string write_v() { return "rsm_terms"; }
    void dependOn_RV(bVec& bv) {}
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunRSM_terms : public FunReadFunBase {
    FunBase* read ( bool errSerious ) { return new FunRSM_terms( read_parameters(0,errSerious) );}
};

class FunHLRV : public FunBase, public FlxDataBase {
  private:
    FlxString* strV;
    FlxHLRVbase* hlrv;
  public:
    FunHLRV (FlxString* strV) : strV(strV), hlrv(NULL) {};
    ~FunHLRV();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunHLRV : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious );
};

/**
* @brief arithmetic class: numerical integration of a function
*/
class FunRF : public FunBase {
  private:
    const tuint RFindex;
    FlxRandomField_base* rfb;
  public:
    FunRF(const tuint RFindex, FlxRandomField_base* rfb) : RFindex(RFindex), rfb(rfb)  {}
    ~FunRF() {}
    const tdouble calc() { return rfb->get_realization(); }
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const tdouble expectation() { return rfb->get_mean_fun()->calc(); }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunRF : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious );
};

/**
* @brief arithmetic class: rf_err(INDEX,MEASURE)
*/
class FunRFerr : public FunBaseFun_multPara, public FlxDataBase {
  public:
    FunRFerr (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "rf_err";}
};

class FunReadFunRFerr : public FunReadFunBase {
  public:
    FunBase* read ( bool errSerious ) { return new FunRFerr( read_parameters(2,errSerious) ); }
};


// -------------------------- Mtx-FUNCTIONS ----------------------------------------------------------------------

/**
* @brief arithmetic class: returns the mean values of the random variables
*/
class FunRVmean : public FunMtxBase, public FlxDataBase {
  private:
    bool output_just_user;
  public:
    FunRVmean(bool output_just_user=false) : output_just_user(output_just_user) {};
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr) { return false;};
    const std::string write() { return "rv_mean()";}
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
};

class FunReadFunRVmean : public FunReadMtxFunBase {
  public:
    FunMtxBase* read ( bool errSerious );
};

/**
* @brief arithmetic class: returns the mean values of the random variables
*/
class FunRVstddev : public FunMtxBase, public FlxDataBase {
  private:
    bool output_just_user;
  public:
    FunRVstddev(bool output_just_user=false) : output_just_user(output_just_user) {};
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr) { return false;};
    const std::string write() { return "rv_sd()";}
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
};

class FunReadFunRVstddev : public FunReadMtxFunBase {
  public:
    FunMtxBase* read ( bool errSerious );
};

/**
* @brief arithmetic class: returns the current values of the random variables
*/
class FunRVvalue : public FunMtxBase, public FlxDataBase {
  private:
    bool bx;
  public:
    FunRVvalue(bool bx) : bx(bx) {};
    SMtxBase* calc ();
    const bool search_circref(FlxFunction* fcr) { return false;};
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
};

class FunReadFunRVvalue : public FunReadMtxFunBase {
  public:
    FunMtxBase* read ( bool errSerious );
};

#endif // flx_rnd_H