/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_python_H
#define fesslix_python_H

#include "flxobjects.h"

#if FLX_USE_PYTHON


class FlxCreateObjReaders_Python : public FlxCreateObjReaders {
  public:
    void createObjReaders (FlxObjectReadBox* objReadBox );
    void createFunReaders (FlxData* dataBox );
};



#ifdef fesslix_python_CPP

// -------------------------- OBJECTS ----------------------------------------------------------------------


/**
* @brief object class: register a python-function as FlxFunction
*/
class FlxObjPythonFun : public FlxObjBase {
  private:
    const std::string funname; 
    const std::string pymodname; 
    const std::string pyfunname;
    void task();
  public:
    FlxObjPythonFun ( const bool dolog, const std::string& funname, const std::string& pymodname, const std::string& pyfunname ) : FlxObjBase(dolog), funname(funname), pymodname(pymodname), pyfunname(pyfunname) {}
    ~FlxObjPythonFun() { }
};

/**
* @brief object class: register a python-function as FlxFunction which returns a Python-variable (float)
*/
class FlxObjPythonVarGet : public FlxObjBase {
  private:
    const std::string funname; 
    const std::string pymodname; 
    const std::string pyvarname;
    void task();
  public:
    FlxObjPythonVarGet ( const bool dolog, const std::string& funname, const std::string& pymodname, const std::string& pyvarname ) : FlxObjBase(dolog), funname(funname), pymodname(pymodname), pyvarname(pyvarname) {}
    ~FlxObjPythonVarGet() { }
};

/**
* @brief object class: set a Python-variable (float)
*/
class FlxObjPythonVarSet : public FlxObjBase {
  private:
    FlxFunction* fun; 
    const std::string pymodname; 
    const std::string pyvarname;
    void task();
  public:
    FlxObjPythonVarSet ( const bool dolog, FlxFunction* fun, const std::string& pymodname, const std::string& pyvarname ) : FlxObjBase(dolog), fun(fun), pymodname(pymodname), pyvarname(pyvarname) {}
    ~FlxObjPythonVarSet() { delete fun; }
};

/**
* @brief object class: load a Python-module
*/
class FlxObjPythonModule : public FlxObjBase {
  private: 
    const std::string pymodname; 
    void task();
  public:
    FlxObjPythonModule ( const bool dolog, const std::string& pymodname ) : FlxObjBase(dolog), pymodname(pymodname) {}
    ~FlxObjPythonModule() { }
};

/**
* @brief object class: register a python search-path (loading modules)
*/
class FlxObjPythonPath : public FlxObjBase {
  private:
    FlxString* pathStr;
    void task();
  public:
    FlxObjPythonPath ( const bool dolog, FlxString* pathStr) : FlxObjBase(dolog), pathStr(pathStr) {}
    ~FlxObjPythonPath() { delete pathStr; }
};

/**
* @brief object class: execute a simple Python command
*/
class FlxObjPythonCmd : public FlxObjBase {
  private:
    FlxString* cmdStr;
    void task();
  public:
    FlxObjPythonCmd ( const bool dolog, FlxString* cmdStr) : FlxObjBase(dolog), cmdStr(cmdStr) {}
    ~FlxObjPythonCmd() { delete cmdStr; }
};





// -------------------------- READERS ----------------------------------------------------------------------


/**
* @brief object read class: for FlxSub
*/
class FlxObjReadPython : public FlxObjReadCtrlStrct {
  public:
    FlxObjReadPython() { } // data->IgnoreBox.set_iL_recur("python"); }
    FlxObjBase* read();
};


// -------------------------- FUNCTIONS ----------------------------------------------------------------------


/**
* @brief arithmetic class: represents a Python function which can be registered by the user
*/
class FunPython : public FunBaseFun_multPara, public FlxDataBase {
  private:
    const std::string funname;
    PyObject* pyfun;
  public:
    FunPython (const std::string& funname, PyObject* pyfun, std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV), funname(funname), pyfun(pyfun) {};
    ~FunPython() {}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write_v();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) {}
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunPython : public FunReadFunBase, public FlxDataBase {
  private:
    const std::string funname;
    PyObject* pyfun;
  public:
    FunReadFunPython(const std::string& funname, const std::string& pymodname, const std::string& pyfunname);
    FunBase* read ( bool errSerious );
};

/**
* @brief arithmetic class: represents a Python function which can be registered by the user (returns a single float-variable)
*/
class FunPythonVar : public FunBase, public FlxDataBase {
  private:
    const std::string funname;
    const std::string pymodname;
    const std::string pyvarname;
  public:
    FunPythonVar (const std::string& funname, const std::string& pymodname, const std::string& pyvarname) : funname(funname), pymodname(pymodname), pyvarname(pyvarname) {};
    ~FunPythonVar() {}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) {}
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

class FunReadFunPythonVar : public FunReadFunBase, public FlxDataBase {
  private:
    const std::string varname;
    const std::string pymodname;
    const std::string pyvarname;
  public:
    FunReadFunPythonVar(const std::string& varname, const std::string& pymodname, const std::string& pyvarname);
    FunBase* read ( bool errSerious );
};




#endif
#endif
#endif // fesslix_python_H

