/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxrandom.h"

int FlxRnd_base::Current_rv_index = -1;
#if FLX_DEBUG
  int FlxRndBox::Cinst = 0;
#endif


const FlxRnd_base::rvtype FlxRnd_Generator_base::get_rvt(std::string strrvt, const bool errSerious)
{
  std::transform(strrvt.begin(), strrvt.end(), strrvt.begin(), (int(*)(int)) std::tolower);
  if (strrvt == "stdn" ) {
    return FlxRnd_base::rvStdN;
  } else if (strrvt == "normal" ) {
    return FlxRnd_base::rvNormal;
  } else if (strrvt == "lognormal" ) {
    return FlxRnd_base::rvLogN;
  } else if (strrvt == "gumbel" ) {
    return FlxRnd_base::rvGumbel;
  } else if (strrvt == "uniform" ) {
    return FlxRnd_base::rvUnifrom;
  } else if (strrvt == "exponential" ) {
    return FlxRnd_base::rvExponential;
  } else if (strrvt == "beta" ) {
    return FlxRnd_base::rvBeta;
  } else {
    std::ostringstream ssV;
    ssV << "Unkown type of random variable '" << strrvt << "'.";
    FlxError(errSerious,"FlxRnd_Generator_base::get_rvt", ssV.str() ); return FlxRnd_base::rvStdN; // DUMMY-RETURN
  }
}

const std::string FlxRnd_Generator_base::get_rvt(const FlxRnd_base::rvtype rvt)
{
  switch(rvt) {
    case FlxRnd_base::rvStdN:
      return "stdn";
      break;
    case FlxRnd_base::rvNormal:
      return "normal";
      break;
    case FlxRnd_base::rvLogN:
      return "lognormal";
      break;
    case FlxRnd_base::rvGumbel:
      return "gumbel";
      break;
    case FlxRnd_base::rvUnifrom:
      return "uniform";
      break;
    case FlxRnd_base::rvExponential:
      return "exponential";
      break;
    default:
      return "";
  }
}

FlxRnd_Generator_base* FlxRnd_Generator_base::createRV(const FlxRnd_base::rvtype rvt)
{
  switch(rvt) {
    case FlxRnd_base::rvStdN:
      return new FlxRnd_Generator_StdN();
    case FlxRnd_base::rvNormal:
      return new FlxRnd_Generator_Normal();
    case FlxRnd_base::rvLogN:
      return new FlxRnd_Generator_LogNormal();
    case FlxRnd_base::rvGumbel:
      return new FlxRnd_Generator_Gumbel();
    case FlxRnd_base::rvUnifrom:
      return new FlxRnd_Generator_Uniform();
    case FlxRnd_base::rvExponential:
      return new FlxRnd_Generator_Exponential();
    case FlxRnd_base::rvBeta:
      return new FlxRnd_Generator_Beta();
    default:
      std::ostringstream ssV;
      ssV << "ERROR.";
      throw FlxException("FlxRnd_Generator_base::createRV", ssV.str() );
  }
}


FlxRnd_Generator_Normal::FlxRnd_Generator_Normal()
: mu(NULL),sigma(NULL)
{
  try {
    reader->getWord("mu");
    mu = new FlxFunction(funReader);
    reader->getWord("sd");
    sigma = new FlxFunction(funReader);
  } catch (FlxException& e) {
    if (mu) delete mu; 
    if (sigma) delete sigma;
    throw;
  }
}

FlxRnd_Generator_LogNormal::FlxRnd_Generator_LogNormal()
:lambda(NULL),zeta(NULL),epsilon(NULL)
{
  try {
    const std::string strV = reader->getWord(true);
    if (strV == "lambda") {
      lambda = new FlxFunction(funReader);
      reader->getWord("zeta");
      zeta = new FlxFunction(funReader);
      is_mean = false;
    } else if (strV == "mu") {
      lambda = new FlxFunction(funReader);
      reader->getWord("sd");
      zeta = new FlxFunction(funReader);
      is_mean = true;
    } else {
      std::ostringstream ssV;
      ssV << "Keyword '" << strV << "' not known.";
      throw FlxException("FlxRnd_LogNormal::FlxRnd_LogNormal_1", ssV.str() );
    }
    
    if (reader->whatIsNextChar()!=';') {
      reader->getWord("epsilon");
      epsilon = new FlxFunction(funReader);
    } else {
      epsilon = new FlxFunction(new FunNumber(0.0));
    }
  } catch (FlxException& e) {
    if (lambda) delete lambda; 
    if (zeta) delete zeta; 
    if (epsilon) delete epsilon;
    throw;
  }
}

FlxRnd_base* FlxRnd_Generator_LogNormal::generate_RV(const bool errSerious)
{
  tdouble epsilonD = epsilon->calc();
  tdouble lambdaD;
  tdouble zetaD;
  if (is_mean) {
    const tdouble mu = lambda->calc();
    const tdouble sigma = zeta->cast2positive();
    const tdouble t = std::log(pow2(sigma)/pow2(mu-epsilonD)+1);
    lambdaD = std::log(mu-epsilonD) - 0.5*t;
    zetaD = std::sqrt(t);
  } else {
    lambdaD = lambda->calc();
    zetaD = zeta->calc();
  }
  return new FlxRnd_LogNormal(lambdaD,zetaD,epsilonD);
}

FlxRnd_Generator_Gumbel::FlxRnd_Generator_Gumbel()
: u(NULL),alpha(NULL)
{
  try {
    std::string strV = reader->getWord(true);
    if (strV == "u") {
      u = new FlxFunction(funReader);
      reader->getWord("alpha");
      alpha = new FlxFunction(funReader);
      is_mean = false;
    } else if (strV == "mu") {
      u = new FlxFunction(funReader);
      reader->getWord("sd");
      alpha = new FlxFunction(funReader);
      is_mean = true;
    } else {
      std::ostringstream ssV;
      ssV << "Keyword '" << strV << "' not known.";
      throw FlxException("FlxRnd_LogNormal::FlxRnd_LogNormal_1", ssV.str() );
    }
  } catch (FlxException& e) {
    if (u) delete u;
    if (alpha) delete alpha;
    throw;
  }
}

FlxRnd_base* FlxRnd_Generator_Gumbel::generate_RV(const bool errSerious)
{
  tdouble alphaD;
  tdouble uD;
  if (is_mean) {
    tdouble mu = u->calc();
    tdouble sigma = alpha->cast2positive();
    alphaD = PI/(std::sqrt(6.)*sigma);
    uD = mu - GAMMA/alphaD;
  } else {
    uD = u->calc();
    alphaD = alpha->calc();
  }
  return new FlxRnd_Gumbel(uD,alphaD);
}

FlxRnd_Generator_Uniform::FlxRnd_Generator_Uniform()
: a(NULL),b(NULL)
{
  try {
    reader->getWord("a");
    a = new FlxFunction(funReader);
    reader->getWord("b");
    b = new FlxFunction(funReader);
  } catch (FlxException& e) {
    if (a) delete a;
    if (b) delete b;
    throw;
  }
}

FlxRnd_base* FlxRnd_Generator_Uniform::generate_RV(const bool errSerious)
{
  const tdouble av = a->calc();
  const tdouble bv = b->calc();
  if (bv<=av) {
    std::ostringstream ssV;
    ssV << "'b' must be larger than 'a'";
    FlxError(errSerious, "FlxRnd_Generator_Uniform::generate_RV", ssV.str() );
  }
  return new FlxRnd_Uniform(av,bv);
}

FlxRnd_Generator_Exponential::FlxRnd_Generator_Exponential()
: lambda(NULL)
{
  try {
    reader->getWord("lambda");
    lambda = new FlxFunction(funReader);
  } catch (FlxException& e) {
    if (lambda) delete lambda;
    throw;
  }
}

FlxRnd_base* FlxRnd_Generator_Exponential::generate_RV(const bool errSerious)
{
  const tdouble lambdaV = lambda->cast2positive(errSerious);
  return new FlxRnd_Exponential(lambdaV);
}

FlxRnd_Generator_Beta::FlxRnd_Generator_Beta()
:alpha(NULL),beta(NULL),a(NULL),b(NULL)
{
  try {
    reader->getWord("alpha");
    alpha = new FlxFunction(funReader);
    reader->getWord("beta");
    beta = new FlxFunction(funReader);
    if (reader->whatIsNextChar()!=';') {
      reader->getWord("a");
      a = new FlxFunction(funReader);
      reader->getWord("b");
      b = new FlxFunction(funReader);
    }
  } catch (FlxException& e) {
    if (alpha) delete alpha;
    if (beta) delete beta;
    if (a) delete a;
    if (b) delete b;
    throw;
  }  
}

FlxRnd_Generator_Beta::~FlxRnd_Generator_Beta()
{
  delete alpha; 
  delete beta; 
  if (a) delete a; 
  if (b) delete b;
}

FlxRnd_base* FlxRnd_Generator_Beta::generate_RV(const bool errSerious)
{
  const tdouble alphaV = alpha->cast2positive(errSerious);
  const tdouble betaV = beta->cast2positive(errSerious);
  const tdouble aV = ((a==NULL)?0.:(a->calc()));
  const tdouble bV = ((b==NULL)?1.:(b->calc()));
  if (bV<=aV) {
    std::ostringstream ssV;
    ssV << "'b' must be larger than 'a'";
    FlxError(errSerious, "FlxRnd_Generator_Beta::generate_RV", ssV.str() );
  }
  return new FlxRnd_Beta(alphaV,betaV,aV,bV);
}

void FlxRndSamplingSpace_base::gen_smp(FlxRndCreator rndcreator, tVec& y)
{
  for (size_t i = 0; i < y.size(); ++i) y[i] = rndcreator.gen_smp();
}

void FlxRndSamplingSpace_base::calc_foverh(tdouble& foverh, const tVec& z)
{
  size_t DIM = z.size();
  // get h
    tdouble h;
    get_h(h,z);
  // get f
    tdouble f = 1.0;
    for (size_t i=0;i<DIM;++i) {
      f*=rv_phi(z[i]);
    }
}

void FlxRndSamplingSpace_base::transform_space(FlxRndBox& rndbox, tdouble& foverh)
{
  size_t DIM = rndbox.get_NumbRnd();
  // generate uncorrelated standard normal samples
    tVec y(DIM);
    gen_smp(rndbox.RndCreator,y);
  // get vector in z-space (sampling space)
    tVec z(DIM);
    y2z(y,z);
  // calc foverh
    calc_foverh(foverh,z);
  // set z as new y-vector
    rndbox.set_y_Vec(z);
}

void FlxRndSamplingSpace_uni::y2z(const tVec& y, tVec& z) const
{
  for (size_t i=0;i<z.size();++i) {
    z[i] = rv->y2x(y[i]);
  }
}

void FlxRndSamplingSpace_uni::get_h(tdouble& h, const tVec& z) const
{
  h = 1.0;
  for (size_t i = 0; i<z.size();++i) {
    h*=rv->pdf_x(z[i]);
  }
}

void FlxRndSamplingSpace_uni::print_info(std::ostream& sout, bool verbose)
{
  sout << "uniform " ;
  rv->print_para(sout);
}

FlxRndSamplingSpace_Generator_base* FlxRndSamplingSpace_Generator_base::createSS(FlxRndSamplingSpace_base::sstype sst, bool errSerious)
{
  switch(sst) {
    case FlxRndSamplingSpace_base::ssuni:
      return new FlxRndSamplingSpace_Generator_Uni(errSerious);
      break;
    case FlxRndSamplingSpace_base::ssnormal:
      return new FlxRndSamplingSpace_Generator_Normal(errSerious);
      break;
    case FlxRndSamplingSpace_base::sstail:
      return new FlxRndSamplingSpace_Generator_TailStdN(errSerious);
      break;
    default:
      std::ostringstream ssV;
      ssV << "ERROR.";
      throw FlxException("FlxRndSamplingSpace_Generator_base::createSS", ssV.str() );
  }
}

std::string FlxRndSamplingSpace_Generator_base::get_rvt(FlxRndSamplingSpace_base::sstype sst)
{
  switch(sst) {
    case FlxRndSamplingSpace_base::ssuni:
      return "uni";
      break;
    case FlxRndSamplingSpace_base::ssnormal:
      return "normal";
      break;
    case FlxRndSamplingSpace_base::sstail:
      return "tailstdn";
      break;
    default:
      return "";
  }
}

FlxRndSamplingSpace_base::sstype FlxRndSamplingSpace_Generator_base::get_sst(std::string strsst, bool errSerious)
{
  std::transform(strsst.begin(), strsst.end(), strsst.begin(), (int(*)(int)) std::tolower);
  if (strsst == "uni" ) {
    return FlxRndSamplingSpace_base::ssuni;
  } else if (strsst == "normal" ) {
    return FlxRndSamplingSpace_base::ssnormal;
  } else if (strsst == "tailstdn" ) {
    return  FlxRndSamplingSpace_base::sstail;
  } else {
    std::ostringstream ssV;
    ssV << "Unkown type of sampling space '" << strsst << "'.";
    FlxError(errSerious,"FlxRndSamplingSpace_Generator_base::get_sst", ssV.str() ); return FlxRndSamplingSpace_base::ssuni; // DUMMY-RETURN
  }
}

FlxRndSamplingSpace_Generator_Uni::FlxRndSamplingSpace_Generator_Uni(bool errSerious)
{
  rvG = FlxRnd_Generator_base::createRV(FlxRnd_Generator_base::get_rvt(reader->getWord(true,errSerious),errSerious));
}

FlxRndSamplingSpace_base* FlxRndSamplingSpace_Generator_Uni::generate_SS()
{
  return new FlxRndSamplingSpace_uni( rvG->generate_RV() );
}

FlxRndSamplingSpace_Generator_TailStdN::FlxRndSamplingSpace_Generator_TailStdN(bool errSerious)
{
  reader->getWord("betadp",false);
  betaDP_F = new FlxFunction(funReader,errSerious);
}

FlxRndSamplingSpace_base* FlxRndSamplingSpace_Generator_TailStdN::generate_SS()
{
  return new FlxRndSamplingSpace_TailStdN(betaDP_F->calc(),RndBox->get_NumbRnd());
}

FlxRndSamplingSpace_base* FlxRndSamplingSpace_Generator_Normal::generate_SS()
{
  return new FlxRndSamplingSpace_normal(muMF->calcVec(),sigmaMF->calcVec(),RndBox->get_NumbRnd());  
}

FlxRndSamplingSpace_Generator_Normal::FlxRndSamplingSpace_Generator_Normal(bool errSerious)
{
  reader->getWord("mu",errSerious);
  muMF = new FlxMtxFun(mtxReader,errSerious);
  try {
  reader->getWord("sd",errSerious);
  sigmaMF = new FlxMtxFun(mtxReader,errSerious);
  } catch (FlxException &e) {
    delete muMF;
    if (sigmaMF!=NULL) delete sigmaMF;
    throw;
  }
}

void FlxRndSamplingSpace_normal::print_info(std::ostream& sout, bool verbose)
{
  sout << "Normal";
  if (verbose) {
    sout << " - mean=" << mu << "; sigma=" << sigma;
  }
}

FlxRndSamplingSpace_normal::FlxRndSamplingSpace_normal(const tVec& mu, const tVec& sigma, const tuint DIM)
: mu(mu),sigma(sigma)
{
  if (mu.size() != sigma.size()) {
    std::ostringstream ssV;
    ssV << "Vector sizes do not match.";
    throw FlxException("FlxRndSamplingSpace_normal::FlxRndSamplingSpace_normal_1", ssV.str() );
  }
  if (DIM != mu.size() ) {
    std::ostringstream ssV;
    ssV << "Size of vectors 'mu' and 'sigma' (" << DIM << ") out of range (" << mu.size() << ").";
    throw FlxException("FlxRndSamplingSpace_normal::FlxRndSamplingSpace_normal_2", ssV.str() );
  }
}

void FlxRndSamplingSpace_normal::y2z(const tVec& y, tVec& z) const
{
  for (size_t i=0;i<z.size();++i) {
    z[i] = y[i]*sigma[i]+mu[i];
  }
}

void FlxRndSamplingSpace_normal::get_h(tdouble& h, const tVec& z) const
{
  h = 1.0;
  for (size_t i = 0; i<z.size();++i) {
    h*=rv_phi((z[i]-mu[i])/sigma[i])/sigma[i];
  }
}

void FlxRndSamplingSpace_TailStdN::calc_foverh(tdouble& foverh, const tVec& z)
{
  // calc f/h
    foverh=cF;
}

FlxRndSamplingSpace_TailStdN::FlxRndSamplingSpace_TailStdN(const tdouble betaDP, const tuint DIM)
: betaDP2(pow2(betaDP)),cF(1-rv_cdf_ChiSquare(DIM,pow2(betaDP)))
{
  if (betaDP < 0.0) {
    std::ostringstream ssV;
    ssV << "'betaDP' (" << betaDP << ") has to be larger than zero.";
    throw FlxException("FlxRndSamplingSpace_TailStdN::FlxRndSamplingSpace_TailStdN", ssV.str() );
  }
}

void FlxRndSamplingSpace_TailStdN::print_info(std::ostream& sout, bool verbose)
{
  sout << "Tails of the Normal - Normal defined outside the interval [-" << sqrt(betaDP2) << "; " << sqrt(betaDP2) << "]";
  if (verbose) {
    sout << std::endl << "   to get the same accuracy with Monte Carlo, you would need " << GlobalVar.Double2String(1/cF) << " times more samples.";
  }
}

void FlxRndSamplingSpace_TailStdN::y2z(const tVec& y, tVec& z) const
{
  const tdouble ly2 = Norm2_NOroot(y);
  tuint DIM = y.size();
  const tdouble lz2 = rv_InvCDF_ChiSquare(DIM,1-cF*rv_cdf_ChiSquare(DIM,ly2));
  z = sqrt(lz2/ly2)*y;
}

FlxRndKernel_base* FlxRndKernel_base::read(const bool errSerious)
{
  return FlxRndKernel_base::read(reader->getWord(false,errSerious),errSerious);
}

FlxRndKernel_base* FlxRndKernel_base::read(std::string kernelStr, const bool errSerious)
{
  std::transform(kernelStr.begin(), kernelStr.end(), kernelStr.begin(), (int(*)(int)) std::tolower);
  if (kernelStr == "gauss") {
    return new FlxRndKernel_Gauss();
  } else if (kernelStr == "uniform") {
    return new FlxRndKernel_Uniform();
  } else {
    std::ostringstream ssV;
    ssV << "Unknown keyword '" << kernelStr << "'.";
    FlxError(errSerious,"FlxRndKernel_base::read", ssV.str() ); return NULL; // DUMMY-RETURN
  }
}

void FlxRndKernel_base::set_h(const tdouble h_value)
{
  if (h_value <= 0.0) {
    std::ostringstream ssV;
    ssV << "Bandwidth has to be positive.";
    throw FlxException("FlxRndKernel_base::set_h", ssV.str() );
  } else {
    h = h_value;
    hinv = 1/h_value;
  }    
}

const tdouble FlxRndKernel_Uniform::cdf(const tdouble x, const tdouble xs) const
{
  const tdouble xt = (x-xs)/h;
  if (xt<=-1.0) return 0.0;
  if (xt>=1.0) return 1.0;
  return xt/2+0.5;
}


