/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxrandom_H
#define fesslix_flxrandom_H

#include "flxmtxfun.h"

//======================= Preliminary Definitions ============================================

class FlxBoxBase2 : public FlxBoxBase, public FlxMtxBoxBase {
  public:
    virtual ~FlxBoxBase2() {};
};


//======================= Generators of Random Variables ============================================

class FlxRnd_Generator_base : public FlxReaderBase2 {
  public:
    virtual ~FlxRnd_Generator_base() {};
    /**
    * @brief create an instance of a random variable
    */
    virtual FlxRnd_base* generate_RV(const bool errSerious=true) = 0;
    // Types of Random Variables
      static const FlxRnd_base::rvtype get_rvt(std::string strrvt, const bool errSerious = true);
      static const std::string get_rvt(const FlxRnd_base::rvtype rvt);
    /**
    * @brief create an instance of a random variable generator
    */
    static FlxRnd_Generator_base* createRV(const FlxRnd_base::rvtype rvt);   
};

class FlxRnd_Generator_StdN : public FlxRnd_Generator_base {
  public:
    FlxRnd_base* generate_RV(const bool errSerious=true) { return new FlxRnd_StdN(); }
};

class FlxRnd_Generator_Normal : public FlxRnd_Generator_base {
  private:
    FlxFunction* mu;
    FlxFunction* sigma;
  public:
    FlxRnd_Generator_Normal();
    virtual ~FlxRnd_Generator_Normal() { delete mu; delete sigma; }
    FlxRnd_base* generate_RV(const bool errSerious=true) { return new FlxRnd_Normal(mu->calc(),sigma->cast2positive()); }
};

class FlxRnd_Generator_LogNormal : public FlxRnd_Generator_base {
  private:
    bool is_mean;
    FlxFunction* lambda;
    FlxFunction* zeta;
    FlxFunction* epsilon;
  public:
    FlxRnd_Generator_LogNormal();
    virtual ~FlxRnd_Generator_LogNormal() { delete lambda; delete zeta; delete epsilon; }
    FlxRnd_base* generate_RV(const bool errSerious=true);
};

class FlxRnd_Generator_Gumbel : public FlxRnd_Generator_base {
  private:
    bool is_mean;
    FlxFunction* u;
    FlxFunction* alpha;
  public:
    FlxRnd_Generator_Gumbel();
    virtual ~FlxRnd_Generator_Gumbel() { delete u; delete alpha; }
    FlxRnd_base* generate_RV(const bool errSerious=true);
};

class FlxRnd_Generator_Uniform : public FlxRnd_Generator_base {
  private:
    FlxFunction* a;
    FlxFunction* b;
  public:
    FlxRnd_Generator_Uniform();
    virtual ~FlxRnd_Generator_Uniform() { delete a; delete b; }
    FlxRnd_base* generate_RV(const bool errSerious=true);
};

class FlxRnd_Generator_Exponential : public FlxRnd_Generator_base {
  private:
    FlxFunction* lambda;
  public:
    FlxRnd_Generator_Exponential();
    virtual ~FlxRnd_Generator_Exponential() { delete lambda; }
    FlxRnd_base* generate_RV(const bool errSerious=true);
};

class FlxRnd_Generator_Beta : public FlxRnd_Generator_base {
  private:
    FlxFunction* alpha;
    FlxFunction* beta;
    FlxFunction* a;
    FlxFunction* b;
  public:
    FlxRnd_Generator_Beta();
    virtual ~FlxRnd_Generator_Beta();
    FlxRnd_base* generate_RV(const bool errSerious=true);
};

//======================= Sampling Spaces ============================================

class FlxRndSamplingSpace_base {
  protected:
    /**
    * @brief generates samples in uncorrelated standard normal space
    */
    void gen_smp(FlxRndCreator rndcreator, tVec& y);
    /**
    * @brief transform from uncorrelated standard normal space to uncorrelated sampling space
    */
    virtual void y2z(const tVec& y, tVec& z) const = 0;
    /**
    * @brief calculates the h-coefficient for Importance Sampling
    */
    virtual void get_h(tdouble &h, const tVec& z) const = 0;
    /**
    * @brief calculates the f/h-factor for Importance Sampling
    */
    virtual void calc_foverh(tdouble& foverh, const tVec& z);
  public:
    virtual ~FlxRndSamplingSpace_base() {}
    /**
    * @brief generates y-realizations in the sampling sapce
    */
    void transform_space(FlxRndBox& rndbox, tdouble& foverh);
    virtual void print_info(std::ostream &sout, bool verbose) = 0;
    virtual const tdouble rely_confidence() const { return -1.; }
    // Types of SamplingSpaces
      enum sstype { 
	ssuni, ssnormal,sstail
      };
};

class FlxRndSamplingSpace_Generator_base : public FlxReaderBase3, protected FlxBoxBase2 {
  public:
    virtual ~FlxRndSamplingSpace_Generator_base() {};
    /**
    * @brief create an instance of a sampling space
    */
    virtual FlxRndSamplingSpace_base* generate_SS() = 0;
    
    // Types of sampling spaces
      static FlxRndSamplingSpace_base::sstype get_sst(std::string strsst, bool errSerious=true);
      static std::string get_rvt(FlxRndSamplingSpace_base::sstype sst);
    /**
    * @brief create an instance of a sampling space
    */
    static FlxRndSamplingSpace_Generator_base* createSS(FlxRndSamplingSpace_base::sstype sst, bool errSerious=true);
};

class FlxRndSamplingSpace_normal : public FlxRndSamplingSpace_base {
  private:
    const tVec mu;
    const tVec sigma;
  protected:
    void y2z(const tVec& y, tVec& z) const;
    void get_h(tdouble &h, const tVec& z) const;
  public:
    FlxRndSamplingSpace_normal(const tVec& mu, const tVec &sigma, const tuint DIM);
    void print_info(std::ostream &sout, bool verbose);
};

class FlxRndSamplingSpace_Generator_Normal : public FlxRndSamplingSpace_Generator_base {
  private:
    FlxMtxFun* muMF;
    FlxMtxFun* sigmaMF;
  public:
    FlxRndSamplingSpace_Generator_Normal(bool errSerious);
    ~FlxRndSamplingSpace_Generator_Normal() { delete muMF; delete sigmaMF; }
    FlxRndSamplingSpace_base* generate_SS();
};

class FlxRndSamplingSpace_TailStdN : public FlxRndSamplingSpace_base {
  private:
    const tdouble betaDP2;
    tdouble cF;
  protected:
    void y2z(const tVec& y, tVec& z) const;
    void get_h(tdouble &h, const tVec& z) const {};  // not explicitely needed
    void calc_foverh(tdouble& foverh, const tVec& z);
  public:
    FlxRndSamplingSpace_TailStdN(const tdouble betaDP, const tuint DIM);
    void print_info(std::ostream &sout, bool verbose); 
    const tdouble rely_confidence() const { return (1./cF); }
};

class FlxRndSamplingSpace_Generator_TailStdN : public FlxRndSamplingSpace_Generator_base {
  private:
    FlxFunction* betaDP_F;
  public:
    FlxRndSamplingSpace_Generator_TailStdN(bool errSerious);
    ~FlxRndSamplingSpace_Generator_TailStdN() { delete betaDP_F; }
    FlxRndSamplingSpace_base* generate_SS();
};

class FlxRndSamplingSpace_uni : public FlxRndSamplingSpace_base {
  private:
    const FlxRnd_base* const rv;
  public:
    FlxRndSamplingSpace_uni(FlxRnd_base* rv) : rv(rv) {}
    ~FlxRndSamplingSpace_uni() { delete rv; }
    void y2z(const tVec& y, tVec& z) const;
    void get_h(tdouble &h, const tVec& z) const;
    void print_info(std::ostream &sout, bool verbose);
};

class FlxRndSamplingSpace_Generator_Uni : public FlxRndSamplingSpace_Generator_base {
  private:
    FlxRnd_Generator_base* rvG;
  public:
    FlxRndSamplingSpace_Generator_Uni(bool errSerious);
    ~FlxRndSamplingSpace_Generator_Uni() { delete rvG; }
    FlxRndSamplingSpace_base* generate_SS();
};


//======================= KERNELS ============================================

class FlxRndKernel_base : public FlxReaderBase2 {
  protected:
    tdouble h;
    tdouble hinv;	// bandwidth of the kernel
  public:
    FlxRndKernel_base() : h(1.0), hinv(1.0) {}
    virtual ~FlxRndKernel_base() {};
    
    /**
    * @brief set the bandwidth of the kernel
    */
    void set_h(const tdouble h_value);
    const tdouble get_h() const { return h;}
    /**
    * @brief kernel-PDF centered at xs evaluated at x
    */
    virtual const tdouble pdf(const tdouble x, const tdouble xs) const = 0;
    /**
    * @brief kernel-PDF centered at xs evaluated at x
    */
    virtual const tdouble cdf(const tdouble x, const tdouble xs) const = 0;
    /**
    * @brief transfrom from standard normal space to kernel-space (centered around zero)
    */
    virtual const tdouble transform_y2x(const tdouble y) const = 0;
    virtual void print_info(std::ostream& sout) const = 0;
    
    static FlxRndKernel_base* read(const bool errSerious=true);
    static FlxRndKernel_base* read(std::string kernelStr, const bool errSerious=true);
};

class FlxRndKernel_Gauss : public FlxRndKernel_base {
  public:
    const tdouble pdf(const tdouble x, const tdouble xs) const { return hinv*rv_phi((x-xs)*hinv); }
    const tdouble cdf(const tdouble x, const tdouble xs) const { return rv_Phi((x-xs)*hinv); }
    const tdouble transform_y2x(const tdouble y) const { return y*h; }
    void print_info(std::ostream& sout) const { sout << "Gaussian kernel; h=" << GlobalVar.Double2String(hinv); }
};

class FlxRndKernel_Uniform : public FlxRndKernel_base {
  public:
    const tdouble pdf(const tdouble x, const tdouble xs) const { return hinv/2; }
    const tdouble cdf(const tdouble x, const tdouble xs) const;
    const tdouble transform_y2x(const tdouble y) const { return ((y>0)?(1.-rv_Phi(-y)*2):(rv_Phi(y)*2-1.))*h; }
    void print_info(std::ostream& sout) const { sout << "uniform kernel; h=" << GlobalVar.Double2String(hinv); }
};


#endif // fesslix_flxrandom_H
