/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxrandomFEM.h"

flxDomain_base* FlxRandomField_base::Current_Element = NULL;
tdouble *FlxRandomField_base::LX = NULL;
tdouble *FlxRandomField_base::GX = NULL;
tdouble *FlxRandomField_base::GY = NULL;
tdouble *FlxRandomField_base::GZ = NULL;
tdouble *FlxRandomField_base::GX2 = NULL;
tdouble *FlxRandomField_base::GY2 = NULL;
tdouble *FlxRandomField_base::GZ2 = NULL;
tdouble *FlxRandomField_base::DELTAX = NULL;
tdouble *FlxRandomField_base::DELTAY = NULL;
tdouble *FlxRandomField_base::DELTAZ = NULL;
tdouble *FlxRandomField_base::DELTAP = NULL;


#include "flxMtx_Eigen.h"

  
FlxRandomField_base::~FlxRandomField_base()
{
  for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
    delete pos->second;
  }
  if (Eigenvalues!=NULL) delete Eigenvalues;
}

void FlxRandomField_base::set_automatic()
{
  if (!is_assembled) {
    if (!automaticV) {
      automaticV = true;
      for (tuint i=0;i<DIM;++i) {
	RndBox->get(RndBox_index)->set_RandomField_automatic();
      }
      RndBox->corret_boxM_size_Rf(DIM);
    } else {
      std::ostringstream ssV;
      ssV << "This random field has already been declared as automatic.";
      throw FlxException("FlxRandomField_base::set_automatic_1", ssV.str() );
    }
  } else {
    std::ostringstream ssV;
    ssV << "This random field has already been assembled.";
    throw FlxException("FlxRandomField_base::set_automatic_2", ssV.str(), "It is no longer possible to declare a random field as automatic after it has been assembled." );
  }
}

void FlxRandomField_base::update_RV_index()
{
  if (!is_assembled) return;
  x_smp = &(RndBox->get(RndBox_index)->get_x_smp());
  y_smp = &(RndBox->get(RndBox_index)->get_y_smp());
}

void FlxRandomField_base::set_element(flxDomain_base* pel, flxElement_KL_base* klb, const bool report, const tuint rfindex, const std::string& elID)
{
  std::pair<flxDomain_base*, flxElement_KL_base*>Element(pel, klb);
  if ( ! boxEl.insert(Element).second ) {
    delete Element.second;
    std::ostringstream ssV;
    ssV << "The specified 'element' (" << elID << ") is already part of the random field";
    GlobalVar.alert.alert("FlxRandomField_base::set_element",ssV.str());
  } else {
    if (report) Element.second->log_report(rfindex,elID,get_Type());
  }
}

const tdouble FlxRandomField_base::get_domain() const
{
  double domain=0.0;
  for (std::map<flxDomain_base*, flxElement_KL_base*>::const_iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
    domain+=pos->second->calc_domain(true);
  }
  return domain;
}

const tdouble FlxRandomField_base::get_domain_error(int errMeasure)
{
  if (is_assembled) {
    if ((get_Type() == flxElement_KL_base::rfKLfemGauss || get_Type() == flxElement_KL_base::rfKLfemGaussProj) && errMeasure == 1) {
      FlxRandomField_KL_FEM_Gauss* rft = dynamic_cast<FlxRandomField_KL_FEM_Gauss*>(this);
      if ( ! ConstantBox->dependOn_GlobalSpatialVar(rft->get_sD())) {
	tdouble t = 0.0;
	for (tuint i=0;i<DIM;++i) {
	  t+=pow2(((*Eigenvalues)[i]));
	}
	t/=pow2(rft->get_sD()->calc());
	const tdouble domain = get_domain();
	return (domain - t)/domain;
      }
    }
    flxElement_KL_base* cekl = Current_Element_KL;
    tdouble domain=0.0;
    tdouble errI=0.0;
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      Current_Element_KL = pos->second;
      domain+=Current_Element_KL->calc_domain(true);
      errI+=get_domain_error_el(errMeasure);
    }
    Current_Element_KL = cekl;
    return errI/domain;
  } else {
    std::ostringstream ssV;
    ssV << "The requested random field has not yet been assembled.";
    throw FlxException("FlxRandomField_base::get_domain_error", ssV.str() );
  }
}

// define a type
  struct elh {
    tuint ID;
    flxElement_KL_base* ptr;
  };
  struct elhc {
    bool operator ()(elh const& a, elh const& b) const {
	if (a.ID < b.ID) return true;
	return false;
    }
  };  
void FlxRandomField_base::plot(std::ostream& sout, FlxMtxFun* ppdim, const int errMeasure)
{
  if (is_assembled) {
    std::vector<elh> v;
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      elh elht;
      elht.ID = pos->first->defID;
      elht.ptr = pos->second;
      v.push_back(elht);      
    }
    std::sort(v.begin(), v.end(), elhc());
    flxElement_KL_base* cekl = Current_Element_KL;
    for (std::vector<elh>::iterator pos = v.begin(); pos != v.end(); ++pos) {
      Current_Element_KL = (*pos).ptr;
      plot_el(sout,ppdim,errMeasure);
    }
    Current_Element_KL = cekl;
  } else {
    std::ostringstream ssV;
    ssV << "The requested random field has not yet been assembled.";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plot", ssV.str() );
  }
}

void FlxRandomField_base::plotcov(std::ostream& sout, const flxPoint& pp, FlxMtxFun* ppdim)
{
  if (is_assembled) {
    flxElement_KL_base* cekl = Current_Element_KL;
    // detect relevant element and compute constant terms
      bool b1 = false;
      std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos;
      for (pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
	if ((b1=(pos->second->map2local(pp)))) break;
      }
      if (!b1) {
	std::ostringstream ssV;
	ssV << "Specified point is not element of the random field domain.";
	throw FlxException("FlxRandomField_base::plotcov_1", ssV.str() );
      }
      tVec tr = *Eigenvalues;	// assign square-root of eigenvalues
      Current_Element_KL = pos->second;
      pp.set_global();
      for (tuint i=0;i<DIM;++i) {
	tr[i] *= calc_Hi(i+1,false);
      }
    // do the actual plotting
      for (pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
	Current_Element_KL = pos->second;
	plotcov_el(sout,pp,ppdim,tr);
      }
    Current_Element_KL = cekl;
  } else {
    std::ostringstream ssV;
    ssV << "The requested random field has not yet been assembled.";
    throw FlxException("FlxRandomField_base::plotcov_2", ssV.str() );
  }
}

FlxFunction* FlxRandomField_base::get_mean_fun()
{
  throw FlxException_NotImplemented("FlxRandomField_base::get_mean_fun");
}

const tdouble FlxRandomField_base::calc_Hi(int i, bool globalIndex)
{
  throw FlxException_NotImplemented("FlxRandomField_base::calc_Hi");
}

void FlxRandomField_base::set_Current_Element_KL()
{
  if (Current_Element==NULL) {
    Current_Element_KL = NULL;
    return;
  }
  std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.find(Current_Element);
  if ( pos != boxEl.end() ) {
    Current_Element_KL = pos->second;
  } else {
    Current_Element_KL = NULL;
  }
}

void FlxRandomField_base::set_Const()
{
  LX = ConstantBox->get("lx",true);
  GX = ConstantBox->get("gx",true);
  GY = ConstantBox->get("gy",true);
  GZ = ConstantBox->get("gz",true);
  GX2 = ConstantBox->get("gx2",true);
  GY2 = ConstantBox->get("gy2",true);
  GZ2 = ConstantBox->get("gz2",true);
  DELTAX = ConstantBox->get("deltax",true);
  DELTAY = ConstantBox->get("deltay",true);
  DELTAZ = ConstantBox->get("deltaz",true);
  DELTAP = ConstantBox->get("deltap",true);
}

FlxRandomField_KL_FEM_Gauss::FlxRandomField_KL_FEM_Gauss(FlxFunction* mu, FlxFunction* sD, FlxFunction* rho, const int ev_solver)
: mu(mu), sD(sD), rho(rho), ev_solver(ev_solver)
{
  if (mu->dependOn_RV()) {
    std::ostringstream ssV;
    ssV << "The mean value must not depend on a random variable!";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D", ssV.str() );
  }
}

FlxRandomField_KL_FEM_Gauss::~FlxRandomField_KL_FEM_Gauss()
{
  delete mu;
  delete sD;
  delete rho;
}

const tdouble FlxRandomField_KL_FEM_Gauss::get_realization() const
{
  if (is_assembled) {
    if (Current_Element_KL) {
      if ( Current_Element_KL->calc_geo_fun()+GlobalVar.TOL() >= 0.0 ) {
	tdouble t = 0.0;
	for (tuint i=0;i<DIM;++i) {
	  t+=((*Eigenvalues)[i]*Current_Element_KL->calc_djNj( &((Eigenvectors[i])[0]),*LX,*(LX-1),*(LX-2))) * (y_smp[i]);
	}
	t+=mu->calc();
	return t;
      } else {
	return mu->calc();
      }
    } else {
      std::ostringstream ssV;
      ssV << "Out of domain of definiton of the random field";
      throw FlxException("FlxRandomField_KL_FEM_Gauss::get_realization_1", ssV.str() );
    }
  } else {
    std::ostringstream ssV;
    ssV << "The requested random field has not yet been assembled.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss::get_realization_2", ssV.str() );
  }
}

const tdouble FlxRandomField_KL_FEM_Gauss::get_domain_error_el(const int errMeasure) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    if (errMeasure == 1 || errMeasure == 101) {	
      GlobalVar.alert.alert("FlxRandomField_KL_FEM_Gauss::get_domain_error_el_0","This feature has not yet been implemented.");
      return 0.0;
    } else if (errMeasure == 2) {	
      return Current_Element_KL->err_cov_domain();
    } else {
      std::ostringstream ssV;
      ssV << "Unknown error measure (" << errMeasure << ")";
      throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_domain_error_el_4", ssV.str() );
    }
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_domain_error_el_1", ssV.str() );
  }
  #endif
}

void FlxRandomField_KL_FEM_Gauss::plot_el(std::ostream& sout, FlxMtxFun* ppdim, const int errMeasure) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    Current_Element_KL->plot_el(sout,ppdim,errMeasure);
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plot_el_1", ssV.str() );
  }
  #endif
}

void FlxRandomField_KL_FEM_Gauss::plotcov_el(std::ostream& sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif 
    Current_Element_KL->plotcov_el(sout,pp,ppdim,tr);
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plotcov_el_1", ssV.str() );
  }
  #endif
}

void FlxRandomField_KL_FEM_Gauss::assembleElementDOFs()
{
  tnlong DOFindex = 0;
  nodalM.clear();
  edgeM.clear();
  faceM.clear();
  // collect the DOFs
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->countDOFs(nodalM,edgeM,faceM);
    }
  // assign the indices
    for (std::map<const flxVertex*,tnlong>::iterator pos = nodalM.begin(); pos != nodalM.end(); ++pos) {
      pos->second = DOFindex++;
    }
    for (std::map<const flxEdge*,eiMap_sec>::iterator pos = edgeM.begin(); pos != edgeM.end(); ++pos) {
      eiMap_sec &ems = pos->second;
      ems.dofindex = DOFindex;
      DOFindex += ems.count;
    }
    for (std::map<const flxFace*,fiMap_sec>::iterator pos = faceM.begin(); pos != faceM.end(); ++pos) {
      fiMap_sec &ems = pos->second;
      ems.dofindex = DOFindex;
      DOFindex += ems.count;
    }
  // assemble the Dofs
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->assembleDOFs(nodalM,edgeM,faceM);
    }
  Ndof = DOFindex;
}

const tnlong FlxRandomField_KL_FEM_Gauss::solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout)
{
  std::ostream& slogS = GlobalVar.slog(5);
  slogS << "Random field - solve:" << std::endl;
  if (is_assembled) {
    std::ostringstream ssV;
    ssV << "The random field has already been assembled.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss::solve_1", ssV.str() );
  }
  if ( boxEl.size() == 0 ) {
    std::ostringstream ssV;
    ssV << "At least one element required for this type of random field.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss::solve_2", ssV.str() );
  }
  assembleElementDOFs();
  if (Ndof == 0) {
    std::ostringstream ssV;
    ssV << "Problem has no DOFs.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss::solve_3", ssV.str() );
  }
  DIM = RFsolveProp->get_dim();
  
  // prepare elements
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->set_ptrFunSigma(&(sD->get_ptrFun()));
      pos->second->set_ptrFunRho(&(rho->get_ptrFun()));
      pos->second->set_ptrBoxEl(&boxEl);
    }
  
  // Create M
    GlobalVar.slog(4) << "  ... assemble 'mass' matrix" << std::endl;
    FlxMtxSym* Mmtx = new FlxMtxSym(Ndof);
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->calc_local_MassMatrix_KLE(*Mmtx);
    }
//     std::cout << *Mmtx << std::endl;
  
  // Create D-Matrix
    GlobalVar.slog(4) << "  ... assemble 'correlation' matrix (" << Ndof << " DOFs)" << std::endl;
    tuint Nel = boxEl.size();
    std::vector<flxElement_KL_base*> ElVec(Nel);
    tuint i=0;
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      ElVec[i++] = pos->second;
    }
    FlxMtxSym* Cmtx = new FlxMtxSym(Ndof);
    
    slogS << "    ... elements: ";
    for (i=1;i<=Nel;++i) {
      if (i>1) slogS << ", ";
      slogS << Nel-i+1;
      slogS.flush();
      ElVec[Nel-i]->calc_local_corrMtx_KLE(*Cmtx,ElVec);
      ElVec.pop_back();
    }
    slogS << std::endl;
    
    if (RFsolveProp->get_octaveOutput()) {
      sout << "%Octave-Input" << std::endl;
      sout << "%============" << std::endl;
      sout << "A=";
      Cmtx->output_OctaveMtx(sout,true,false);
      sout << ";" << std::endl << "B=";
      Mmtx->output_OctaveMtx(sout,true,false);
      sout << ";" << std::endl << "qz(A,B)" << std::endl;
    }
    
//     std::cout << *Cmtx << std::endl;
//     FlxMtxLTri T(Cmtx->ncols());
//     T.CholeskyDec(*Cmtx);
    
  // Create Storage for Eigenvalues and Eigenvectors
    Eigenvalues = new tVec(DIM);
    Eigenvectors.clear();
    for (tuint i = 0; i < DIM; i++) {
      tVec Vh(Ndof);
      Eigenvectors.push_back(Vh);
    }
    
  // Solve the generalized eigenvalue problem
    GlobalVar.slog(4) << "  ... solve the eigenvalue problem" << std::endl;
    try {
      MtxEigenValue(*Cmtx,*Mmtx,DIM,*Eigenvalues,Eigenvectors,ev_solver);
    } catch (FlxException &e) {
      delete Cmtx;
      delete Mmtx;
      throw;
    }
    delete Cmtx;
    delete Mmtx;
    
  // Normalize the Eigenfunctions
    GlobalVar.slog(4) << "  ... normalize the eigenfunctions" << std::endl;
    for (tuint i=0;i<DIM;i++) {
      tVec& EV = Eigenvectors[i];
      checkTOL(EV);
      tdouble D=0.0;
      for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
	flxElement_KL_base& elKL = *(pos->second);
	D+=elKL.calc_djNj2(&EV[0]);
      }
      D=1/sqrt(D);
      EV*=D;
      // Compute square-roots of Eigenvalues
	if ((*Eigenvalues)[i]<=0.0) {
	  std::ostringstream ssV;
	  ssV << "The " << i+1 << "th eigenvalue is negative.";
	  throw FlxException("FlxRandomField_KL_FEM_Gauss::solve_99", ssV.str() );
	}
	(*Eigenvalues)[i] = sqrt((*Eigenvalues)[i]);
    }
  // update elements
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->set_ptrEigenvectors(&Eigenvectors);
      pos->second->set_ptrEigenvalues(Eigenvalues);
    }
  
  // create DIM random variables
    std::vector<FlxRnd_base*> rv(DIM);
    for (tuint i=0;i<DIM;++i) {
      rv[i] = new FlxRnd_StdN();
      rv[i]->set_RandomField();
    }
    RndBox_index = RndBox->insert_RFv(rv);
    is_assembled = true;
    update_RV_index();
    return Ndof;
}

FlxFunction* FlxRandomField_KL_FEM_Gauss::get_mean_fun()
{
  return new FlxFunction(*mu);
}

const tdouble FlxRandomField_KL_FEM_Gauss::calc_Hi(int i, bool globalIndex)
{
  // correct the index -> set to local index
  if (globalIndex && i > 0) {
    if (i-1>=(int)RndBox_index && i<=(int)(RndBox_index+DIM)) {
      i -= RndBox_index;
    } else {
      return 0.0;
    }
  }
  // check for special case -2;
    if (i==-2) i = (int)DIM;
    
  if (i<0) {
    throw FlxException_Crude("FlxRandomField_KL_FEM_Gauss::calc_Hi_0");
  } else if (i==0) {
    return mu->calc();
  } else if (i>(int)DIM) {
    throw FlxException_Crude("FlxRandomField_KL_FEM_Gauss::calc_Hi_1");
  } else  {
    if (Current_Element_KL->calc_geo_fun() >= 0.0) {
      return (*Eigenvalues)[i-1]*Current_Element_KL->calc_djNj( &((Eigenvectors[i-1])[0]),*LX,*(LX-1),*(LX-2));
    } else {
      return 0.0;
    }
  }
}

const tdouble FlxRandomField_KL_FEM_Gauss::get_Eigenvalue(const int i, bool errSerious)
{
  if (i>=(int)DIM) {
    std::ostringstream ssV;
    ssV << "Index is too large. Not enough eigenvalues.";
    FlxError(errSerious,"FlxRandomField_KL_FEM_Gauss::get_Eigenvalue", ssV.str() );
  }
  return pow2((*Eigenvalues)[i]);
}

void FlxRandomField_KL_FEM_Gauss::print_para(std::ostream& sout) const
{
  sout << "Gaussian random field (FEM-KL-expansion)";
}

FlxRandomField_KL_FEM_Gauss_proj::FlxRandomField_KL_FEM_Gauss_proj(FlxFunction* mu, FlxFunction* sD, FlxFunction* rho, const int ev_solver)
: FlxRandomField_KL_FEM_Gauss(mu, sD, rho, ev_solver)
{
  
}

FlxRandomField_KL_FEM_Gauss_proj::~FlxRandomField_KL_FEM_Gauss_proj()
{
  
}

void FlxRandomField_KL_FEM_Gauss_proj::print_para(std::ostream& sout) const
{
  sout << "Gaussian random field (FEM-KL-expansion - projection)";
}

const tnlong FlxRandomField_KL_FEM_Gauss_proj::solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout)
{
  if (is_assembled) {
    std::ostringstream ssV;
    ssV << "The random field has already been assembled.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss_proj::solve_1", ssV.str() );
  }
  if ( boxEl.size() == 0 ) {
    std::ostringstream ssV;
    ssV << "At least one element required for this type of random field.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss_proj::solve_2", ssV.str() );
  }
  assembleElementDOFs();
  if (Ndof == 0) {
    std::ostringstream ssV;
    ssV << "Problem has no DOFs.";
    throw FlxException("FlxRandomField_KL_FEM_Gauss_proj::solve_3", ssV.str() );
  }
  DIM = RFsolveProp->get_dim();
  
  // prepare elements
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->set_ptrFunSigma(&(sD->get_ptrFun()));
      pos->second->set_ptrFunRho(&(rho->get_ptrFun()));
      pos->second->set_ptrBoxEl(&boxEl);
    }
    
  // Create M-Matrix
    GlobalVar.slog(4) << "  ... assemble 'mass' matrix" << std::endl;
    FlxMtxSym* Mmtx = new FlxMtxSym(Ndof);
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->calc_local_MassMatrix_KLE(*Mmtx);
    }    
    FlxMtxLTri* L = new FlxMtxLTri(Ndof);
    L->CholeskyDec(*Mmtx);
    
  // Create C-Matrix
    GlobalVar.slog(4) << "  ... assemble 'correlation' matrix (" << Ndof << " DOFs)" << std::endl;
    // arrange the nodes in a certain order
      const tuint Nnb = nodalM.size();
      std::vector<const flxVertex*> NVec(Nnb);
      std::vector<tnlong> NiVec(Nnb);
      tuint i=0,j;
      for (std::map<const flxVertex*, tnlong>::const_iterator pos = nodalM.begin(); pos != nodalM.end(); ++pos) {
	NiVec[i] = pos->second;
	NVec[i++] = pos->first;
      }
    FlxMtxSym* Cmtx = new FlxMtxSym(Ndof);
    tdouble sD1;
    // perform the nodal projection
      for (i=0;i<Nnb;++i) {
	NVec[i]->get_point().set_global();
	NVec[i]->get_point().set_global2();
	sD1 = sD->calc();
	for (j=0;j<=i;++j) {
	  NVec[j]->get_point().set_global_dist();
	  Cmtx->operator()(NiVec[i],NiVec[j]) = sD1*sD->calc()*rho->calc();
	}
      }
    // deal with the internal DOFs
      if (Nnb!=Ndof) {
	throw FlxException_NotImplemented("FlxRandomField_KL_FEM_Gauss_proj::solve_3");
      }
    // M = L^T C L
      MtxProd_BTKB(*L,*Cmtx,*Mmtx);
      delete Cmtx;
      Cmtx = Mmtx;
    
  // Create Storage for Eigenvalues and Eigenvectors
    Eigenvalues = new tVec(DIM);
    Eigenvectors.clear();
    for (tuint i = 0; i < DIM; i++) {
      tVec Vh(Ndof);
      Eigenvectors.push_back(Vh);
    }
    
    if (RFsolveProp->get_octaveOutput()) {
      sout << "%Octave-Input" << std::endl;
      sout << "%============" << std::endl;
      sout << "A=";
      Cmtx->output_OctaveMtx(sout,true,false);
      sout << ";" << std::endl;
    }
    
  // Solve the standard eigenvalue problem
    try {
      MtxEigenValue(*Cmtx,DIM,*Eigenvalues,Eigenvectors,ev_solver);
    } catch (FlxException &e) {
      delete Cmtx;
      delete L;
      throw;
    }
    delete Cmtx;
    L->Invert();

  // Normalize the Eigenfunctions
    for (tuint i=0;i<DIM;i++) {
      tVec& EV = Eigenvectors[i];
      L->TransMultVec(EV);
      checkTOL(EV);
      tdouble D=0.0;
      for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
	flxElement_KL_base& elKL = *(pos->second);
	D+=elKL.calc_djNj2(&EV[0]);
      }
      D=1/sqrt(D);
      EV*=D;
      // Compute square-roots of Eigenvalues
	if ((*Eigenvalues)[i]<=0.0) {
	  std::ostringstream ssV;
	  ssV << "The " << i+1 << "th eigenvalue is negative.";
	  throw FlxException("FlxRandomField_KL_FEM_Gauss_proj::solve_99", ssV.str() );
	}
	(*Eigenvalues)[i] = sqrt((*Eigenvalues)[i]);
    }
    delete L;
  // update elements
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->set_ptrEigenvectors(&Eigenvectors);
      pos->second->set_ptrEigenvalues(Eigenvalues);
    }
  
  // create DIM random variables
    std::vector<FlxRnd_base*> rv(DIM);
    for (tuint i=0;i<DIM;++i) {
      rv[i] = new FlxRnd_StdN();
      rv[i]->set_RandomField();
    }
    RndBox_index = RndBox->insert_RFv(rv);
    is_assembled = true;
    update_RV_index();
    return Ndof;
}

const tdouble FlxRandomField_KL_FEM_Gauss_proj::get_domain_error_el(const int errMeasure) const
{
  int errM = errMeasure;
  if (errM==1) errM = 101;
  return FlxRandomField_KL_FEM_Gauss::get_domain_error_el(errM);
}

void FlxRandomField_KL_FEM_Gauss_proj::plot_el(std::ostream& sout, FlxMtxFun* ppdim, const int errMeasure) const
{
  int errM = errMeasure;
  if (errM==1) errM = 101;
  FlxRandomField_KL_FEM_Gauss::plot_el(sout, ppdim, errM);
}

FlxRandomField_EOLE_Gauss::FlxRandomField_EOLE_Gauss(FlxFunction* mu, FlxFunction* sD, FlxFunction* rho, const int ev_solver)
: FlxRandomField_KL_FEM_Gauss(mu, sD, rho, ev_solver),sdVec(NULL),Hx(NULL)
{
  
}

FlxRandomField_EOLE_Gauss::~FlxRandomField_EOLE_Gauss()
{
  if (sdVec!=NULL) delete sdVec;
  if (Hx!=NULL) delete Hx;
}

const tdouble FlxRandomField_EOLE_Gauss::get_domain_error_el(const int errMeasure) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    if (errMeasure == 1) {	
      return Current_Element_KL->err_var_domain_EOLE(*Hx,pointVec,*sdVec);
    } else if (errMeasure == 2) {	
      return Current_Element_KL->err_cov_domain_EOLE(pointVec,*sdVec);
    } else {
      std::ostringstream ssV;
      ssV << "Unknown error measure (" << errMeasure << ")";
      throw FlxException("FlxRandomField_EOLE_Gauss::get_domain_error_el_4", ssV.str() );
    }
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_EOLE_Gauss::get_domain_error_el_1", ssV.str() );
  }
  #endif
}

const tdouble FlxRandomField_EOLE_Gauss::get_error(int errMeasure) const
{
  if (errMeasure==1) {
    const tdouble sD_curr_x_2 = pow2(sD->calc());
    tdouble t = sD_curr_x_2;
    for (tuint i=0;i<DIM;++i) {	
      t -= pow2( dot( Eigenvectors[i], *Hx ) * (*Eigenvalues)[i] );
    }
    // speed-test KL vs. EOLE
//       flxElement_KL_3D* tmp = dynamic_cast<flxElement_KL_3D*>(boxEl.begin()->second);
//       return tmp->err_var_domain_EOLE(*Hx,pointVec,*sdVec);
    return t/sD_curr_x_2;
  } else if (errMeasure==2) {
    *GX2 = *GX; 
    *GY2 = *GY;
    *GZ2 = *GZ;
    tVec tr(DIM);
    tVec Hx2(Ndof);
    return Current_Element_KL->get_error_2i_EOLE(Hx2,tr,pointVec,*sdVec);
  } else {
    std::ostringstream ssV;
    ssV << "Unknown error measure (" << errMeasure << ")";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_error_2", ssV.str() );
  }
}

void FlxRandomField_EOLE_Gauss::plot_el(std::ostream& sout, FlxMtxFun* ppdim, const int errMeasure) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    std::vector<flxPoint> pointVec;
    std::vector< tuint > geoVec;
    Current_Element_KL->get_PlotPoints(pointVec,geoVec,ppdim);
    tuint k = 0;
    if (geoVec.size() > 0 && geoVec[0]==0) ++k;
    for (tuint i=0;i<pointVec.size();++i) {
      if (k<geoVec.size() && geoVec[k]==i) {
	sout << std::endl;
	++k;
	while (k < geoVec.size() && geoVec[k]==i) ++k;
      }
      pointVec[i].set_global();
      assemble_Hx();
      sout << GlobalVar.Double2String(*GX) << '\t' << GlobalVar.Double2String(*GY) << '\t' << GlobalVar.Double2String(*GZ) << '\t';
      sout << GlobalVar.Double2String(get_error(errMeasure));
      for (tuint j=0;j<DIM;++j) {
	sout << '\t' << dot( Eigenvectors[j], *Hx ) * (*Eigenvalues)[j];
      }
      sout << std::endl;
    }
    sout << std::endl << std::endl;
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plot_el_1", ssV.str() );
  }
  #endif
}

void FlxRandomField_EOLE_Gauss::plotcov_el(std::ostream& sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    std::vector<flxPoint> pointVec;
    std::vector< tuint > geoVec;
    // determine sd(pp)
      pp.set_global();
      tdouble sd_pp_X = sD->cast2positive();
    tdouble sd_cur_X, cov_ana, cov_EOLE;
    Current_Element_KL->get_PlotPoints(pointVec,geoVec,ppdim);
    tuint i,j,k=0;
    if (geoVec.size() > 0 && geoVec[0]==0) ++k;
    for (i=0;i<pointVec.size();++i) {
      if (k<geoVec.size() && geoVec[k]==i) {
	sout << std::endl;
	++k;
	while (k < geoVec.size() && geoVec[k]==i) ++k;
      }
      pointVec[i].set_global();
      sd_cur_X = assemble_Hx();
      pp.set_global2_dist();
      sout << GlobalVar.Double2String(*GX) << '\t' << GlobalVar.Double2String(*GY) << '\t' << GlobalVar.Double2String(*GZ) << '\t';
      cov_EOLE = 0.0;
      for (j=0;j<DIM;++j) {
	cov_EOLE += dot( Eigenvectors[j], *Hx ) * tr[j] ;
      }
      cov_ana = sd_cur_X * sd_pp_X * rho->calc();
      sout << cov_EOLE << '\t' << GlobalVar.Double2String( cov_ana ) << '\t' << GlobalVar.Double2String( cov_ana-cov_EOLE );
      sout << std::endl;
    }
    sout << std::endl;
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plot_el_1", ssV.str() );
  }
  #endif
}

FlxFunction* FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_mean_fun()
{
  return new FlxFunction(*mu);
}

const tdouble FlxRandomField_EOLE_Gauss::get_realization() const
{
  if (is_assembled) {
    if (Current_Element_KL) {
      if ( Current_Element_KL->calc_geo_fun()+GlobalVar.TOL() >= 0.0 ) {
	assemble_Hx();
	tdouble t = 0.0;
	for (tuint i=0;i<DIM;++i) {
	  t+= dot( Eigenvectors[i], *Hx ) * (*Eigenvalues)[i] * (y_smp[i]) ;
	}
	t+=mu->calc();
	return t;
      } else {
	return mu->calc();
      }
    } else {
      std::ostringstream ssV;
      ssV << "Out of domain of definiton of the random field";
      throw FlxException("FlxRandomField_EOLE_Gauss::get_realization_1", ssV.str() );
    }
  } else {
    std::ostringstream ssV;
    ssV << "The requested random field has not yet been assembled.";
    throw FlxException("FlxRandomField_EOLE_Gauss::get_realization_2", ssV.str() );
  }
}

void FlxRandomField_EOLE_Gauss::print_para(std::ostream& sout) const
{
  sout << "Gaussian random field (EOLE-expansion)";
}

void FlxRandomField_EOLE_Gauss::assembleElementDOFs()
{
  pointVec.clear();
  for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
    pos->second->assembleEOLEpointVec(pointVec);
  }
  Ndof = pointVec.size();
}

const tdouble FlxRandomField_EOLE_Gauss::assemble_Hx() const
{
  tdouble sD_curr_x = sD->calc();
  for (tuint i=0;i<Ndof;++i) {
    pointVec[i].set_global2_dist();
    (*Hx)[i] = sD_curr_x * (*sdVec)[i] * rho->calc();
  }
  return sD_curr_x;
}

const tnlong FlxRandomField_EOLE_Gauss::solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout)
{
  if (is_assembled) {
    std::ostringstream ssV;
    ssV << "The random field has already been assembled.";
    throw FlxException("FlxRandomField_EOLE_Gauss::solve_1", ssV.str() );
  }
  if ( boxEl.size() == 0 ) {
    std::ostringstream ssV;
    ssV << "At least one element required for this type of random field.";
    throw FlxException("FlxRandomField_EOLE_Gauss::solve_2", ssV.str() );
  }
  assembleElementDOFs();	// update 
  if (Ndof == 0) {
    std::ostringstream ssV;
    ssV << "Problem has no DOFs.";
    throw FlxException("FlxRandomField_EOLE_Gauss::solve_3", ssV.str() );
  }
  DIM = RFsolveProp->get_dim();
  
  // prepare elements
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->set_ptrFunSigma(&(sD->get_ptrFun()));
      pos->second->set_ptrFunRho(&(rho->get_ptrFun()));
      pos->second->set_ptrBoxEl(&boxEl);
    }
    
  // Create C-Matrix
    GlobalVar.slog(4) << "  ... assemble 'correlation' matrix (" << Ndof << " DOFs)" << std::endl;
    FlxMtxSym* Cmtx = new FlxMtxSym(Ndof);
    sdVec = new tVec(Ndof);
    Hx = new tVec(Ndof);
    tuint i,j;
    for (i=0;i<Ndof;++i) {
      (pointVec[i]).set_global();
      sdVec->operator[](i) = sD->calc();		// update the standard deviation
      for (j=0;j<=i;++j) {
	(pointVec[j]).set_global2_dist();
	Cmtx->operator()(i,j) = sdVec->operator[](i)*sdVec->operator[](j)*rho->calc();
      }
    }
  
  // Create Storage for Eigenvalues and Eigenvectors
    Eigenvalues = new tVec(DIM);
    Eigenvectors.clear();
    for (tuint i = 0; i < DIM; i++) {
      tVec Vh(Ndof);
      Eigenvectors.push_back(Vh);
    }
    
    if (RFsolveProp->get_octaveOutput()) {
      sout << "%Octave-Input" << std::endl;
      sout << "%============" << std::endl;
      sout << "A=";
      Cmtx->output_OctaveMtx(sout,true,false);
      sout << ";" << std::endl;
    }
  // Solve the standard eigenvalue problem
    try {
      MtxEigenValue(*Cmtx,DIM,*Eigenvalues,Eigenvectors,ev_solver);
    } catch (FlxException &e) {
      delete Cmtx;
      throw;
    }
    delete Cmtx;
    
  // Normalize the Eigenvectors
    for (tuint i=0;i<DIM;i++) {
      tVec& EV = Eigenvectors[i];
      EV/=Norm2(EV);
      checkTOL(EV);
      (*Eigenvalues)[i] = 1.0 / sqrt((*Eigenvalues)[i]);
    }
  // update elements
    for (std::map<flxDomain_base*, flxElement_KL_base*>::iterator pos = boxEl.begin(); pos != boxEl.end(); ++pos) {
      pos->second->set_ptrEigenvectors(&Eigenvectors);
      pos->second->set_ptrEigenvalues(Eigenvalues);
    }
  
  // create DIM random variables
    std::vector<FlxRnd_base*> rv(DIM);
    for (tuint i=0;i<DIM;++i) {
      rv[i] = new FlxRnd_StdN();
      rv[i]->set_RandomField();
    }
    RndBox_index = RndBox->insert_RFv(rv);
    is_assembled = true;
    update_RV_index();
    return Ndof;
}

const tdouble FlxRandomField_EOLE_Gauss::calc_Hi(int i, bool globalIndex)
{
  // correct the index -> set to local index
  if (globalIndex && i > 0) {
    if (i-1>=(int)RndBox_index && i<=(int)(RndBox_index+DIM)) {
      i -= RndBox_index;
    } else {
      return 0.0;
    }
  }
  // check for special case -2;
    if (i==-2) i = DIM;
    
  if (i<0) {
    throw FlxException_Crude("FlxRandomField_KL_FEM_Gauss::calc_Hi_0");
  } else if (i==0) {
    return mu->calc();
  } else if (i>(int)DIM) {
    throw FlxException_Crude("FlxRandomField_KL_FEM_Gauss::calc_Hi_1");
  } else  {
    if (Current_Element_KL->calc_geo_fun() >= 0.0) {
      assemble_Hx();
      return dot( Eigenvectors[i-1], *Hx ) * (*Eigenvalues)[i-1];
    } else {
      return 0.0;
    }
  }
}

const tdouble FlxRandomField_EOLE_Gauss::get_Eigenvalue(const int i, bool errSerious)
{
  if (i>=(int)DIM) {
    std::ostringstream ssV;
    ssV << "Index is too large. Not enough eigenvalues.";
    FlxError(errSerious,"FlxRandomField_KL_FEM_Gauss::get_Eigenvalue", ssV.str() );
  }
  const tdouble A = get_domain();
  return 1.0 / pow2((*Eigenvalues)[i]) / Ndof * A;
}

const tnlong FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout)
{
  if (is_assembled) {
    std::ostringstream ssV;
    ssV << "The random field has already been assembled.";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::solve_1", ssV.str() );
  }
  const tdouble x_startO = 1e-5;
  const tdouble err = 1e-14;
  DIM = RFsolveProp->get_dim();
  if ( boxEl.size() != 1 ) {
    std::ostringstream ssV;
    ssV << "Not more than one element allowed for this type of random field.";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::solve_2", ssV.str() );
  }
  etkl = dynamic_cast<flxElement_KL_1D*>(boxEl.begin()->second);
  if ( etkl == NULL ) {
    std::ostringstream ssV;
    ssV << "This type of random field just works with 1D-elements.";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::solve_3", ssV.str() );
  }
  if (etkl->has_geometry()) {
    std::ostringstream ssV;
    ssV << "This type of random field cannot deal with geometries.";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::solve_4", ssV.str() );
  }
  Eigenvalues = new tVec(DIM);
  alpha_i = new tVec(DIM);
  omega_i = new tVec(DIM);
  domain_l = etkl->calc_length() / 2.0;
  const tdouble lt =  corrL;
  const tdouble lti = 1.0/lt;
  // solve the actual problem
  tdouble x_startR = x_startO;
  tdouble x, x_old, f, fp;
  tnlong lc;
//   std::cout << "L=" << domain_l << std::endl;
//   std::cout << "lt=" << lt << std::endl;
  for (tuint i=0;i<DIM;++i) {
    if (i%2==0) {	// i+1 is odd
      // compute omega_i
	x_old =(tdouble(i/2)+0.5)*PI/domain_l;
	x = x_old - x_startR;
	f = lti - x*tan(domain_l*x);
	lc = 0;
	while (fabs(f) > err && x_old!=x && ++lc < 1e4) {
	  fp = -tan(domain_l*x)-domain_l*x/pow2(cos(domain_l*x));
	  x_old = x;
	  x -= f/fp;
	  f = lti - x*tan(domain_l*x);
	}
	(*omega_i)[i] = x;
	x_startR = (tdouble(i/2)+0.5)*PI/domain_l-x;
      // compute alpha_i
	(*alpha_i)[i] = 1.0/sqrt(domain_l+sin(2.0*domain_l*(*omega_i)[i])/(2.0*(*omega_i)[i]));
    } else {		// i+1 is even
      // compute omega_i
	x_old =(tdouble((i-1)/2)+0.5)*PI/domain_l;
	x = x_old + x_startO;
	f = tan(domain_l*x)*lti + x;
	lc = 0;
	while (fabs(f) > err && x_old!=x &&  ++lc < 1e4 ) {
	  fp = domain_l/(lt*pow2(cos(domain_l*x)))+1;
	  x_old = x;
	  x -= f/fp;
	  f = tan(domain_l*x)*lti + x;
	}
	(*omega_i)[i] = x;
      // compute alpha_i
	(*alpha_i)[i] = 1.0/sqrt((domain_l-(sin(2.0*domain_l*(*omega_i)[i]))/(2.0*(*omega_i)[i])));
    }
    // compute the square root of the eigenvalues
      (*Eigenvalues)[i] = sqrt((2*lt)/(1.0+pow2((*omega_i)[i]*lt)));
//       std::cout << "  " << i+1 << ": " << std::endl;
//       std::cout << "    omega=" << (*omega_i)[i] << std::endl;
//       std::cout << "    alpha=" << (*alpha_i)[i] << std::endl;
//       std::cout << "    eigenvalue=" << pow2((*Eigenvalues)[i]) << std::endl;
  }
  // create DIM random variables
    std::vector<FlxRnd_base*> rv(DIM);
    for (tuint i=0;i<DIM;++i) {
      rv[i] = new FlxRnd_StdN();
      rv[i]->set_RandomField();
    }
    RndBox_index = RndBox->insert_RFv(rv);
    is_assembled = true;
    update_RV_index();
    return 0;
}

const tdouble FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::calc_Hi(int i, bool globalIndex)
{
  // correct the index -> set to local index
    if (globalIndex && i>0) i -= RndBox_index;
  // check for special case -2;
    if (i==-2) i = (int)DIM;
    
  if (i<0) {
    throw FlxException_Crude("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::calc_Hi_0");
  } else if (i==0) {
    return mu->calc();
  } else if (i>(int)DIM) {
    throw FlxException_Crude("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::calc_Hi_1");
  } else  {
    if (i%2==0) {
      return sD*(((*Eigenvalues)[i-1])*(((*alpha_i)[i-1] * sin((*omega_i)[i-1]*(*LX)*domain_l))));
    } else {
      return sD*(((*Eigenvalues)[i-1])*(((*alpha_i)[i-1] * cos((*omega_i)[i-1]*(*LX)*domain_l))));
    }
  }
}

const tdouble FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_Eigenvalue(const int i, bool errSerious)
{
  if (i>=(int)DIM) {
    std::ostringstream ssV;
    ssV << "Index is too large. Not enough eigenvalues.";
    FlxError(errSerious,"FlxRandomField_KL_FEM_Gauss::get_Eigenvalue", ssV.str() );
  }
  return pow2(sD*(*Eigenvalues)[i]);
}

const tdouble FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_realization() const
{
  if (is_assembled) {
    if (Current_Element_KL) {
      tdouble t = 0.0;
      for (tuint i=0;i<DIM;i=i+2) {	// i+1 is odd
	t+=(((*Eigenvalues)[i])*(((*alpha_i)[i] * cos((*omega_i)[i]*(*LX)*domain_l)))) * (y_smp[i]);
      }
      for (tuint i=1;i<DIM;i=i+2) {	// i+1 is even
	t+=(((*Eigenvalues)[i])*(((*alpha_i)[i] * sin((*omega_i)[i]*(*LX)*domain_l)))) * (y_smp[i]);
      }
      t*=sD;
      t+=mu->calc();
      return t;
    } else {
      std::ostringstream ssV;
      ssV << "Out of domain of definiton of the random field";
      throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_realization_1", ssV.str() );
    }
  } else {
    std::ostringstream ssV;
    ssV << "The requested random field has not yet been assembled.";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_realization_2", ssV.str() );
  }
}

const tdouble FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_error(const int errMeasure) const
{
  if (errMeasure==1) {
    tdouble t = 0.0;
    for (tuint i=0;i<DIM;i=i+2) {	// i+1 is odd
      t+=pow2(((*Eigenvalues)[i])*(((*alpha_i)[i] * cos((*omega_i)[i]*domain_l*(*LX)))));
    }
    for (tuint i=1;i<DIM;i=i+2) {	// i+1 is even
      t+=pow2(((*Eigenvalues)[i])*(((*alpha_i)[i] * sin((*omega_i)[i]*domain_l*(*LX)))));
    }
    return 1.0-t;
  } else if (errMeasure==2) {
    // refine iVec
      flxElement_KL_1D* curElKL = dynamic_cast<flxElement_KL_1D*>(Current_Element_KL);
      #if FLX_DEBUG
	if (curElKL==NULL) {
	  throw FlxException_Crude("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_error_1");
	}
      #endif
      const std::vector<tdouble> &iVecR = curElKL->get_IntervalVec_err2(0);
    // get Gauss-points
      const tuint GN = GI->pDegree2GPs(curElKL->get_pOcovOrder());		// of course this won't be an exact integration
      const tdouble* GP = GI->get_GP(GN);
      const tdouble* GW = GI->get_GW(GN);
    return get_error_cov(iVecR,GN,GP,GW);
  } else {
    std::ostringstream ssV;
    ssV << "Unknown error measure (" << errMeasure << ")";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_error_2", ssV.str() );
  }
}

const tdouble FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_error_cov(const std::vector< tdouble >& iVec, const tuint GN, const tdouble* GP, const tdouble* GW) const
{
  // precompute some coefficients
    tVec tr = *Eigenvalues;
    for (tuint i=0;i<DIM;++i) { tr[i] *= tr[i] * pow2((*alpha_i)[i]); }
    for (tuint i=0;i<DIM;i=i+2) {	// i+1 is odd
      tr[i] *= cos((*omega_i)[i]*(*LX)*domain_l);
    }
    for (tuint i=1;i<DIM;i=i+2) {	// i+1 is odd
      tr[i] *= sin((*omega_i)[i]*(*LX)*domain_l);
    }
  // integral rho(x,x') over \Omega
    tdouble tt = corrL*(2.0 - exp(-1.0*(*LX+1.0)*domain_l/corrL) - exp((*LX-1.0)*domain_l/corrL) );
  // integration (staggered Gaussian)
    tdouble dx,lx,r,d=0.0;
    const tdouble lxT = *LX;
    register tuint j,k;
    register tdouble dd;
    for (tuint i=1;i<iVec.size();++i) {
      dx = (iVec[i]-iVec[i-1])/2.0;
      r = 0.0;
      for (j=0;j<GN;++j) {
	lx = (GP[j]+1.0)*dx+iVec[i-1];
	etkl->update_globalConst(lx);
	dd = exp(-1*(fabs(lxT-(*LX))*domain_l)/corrL);
	for (k=0;k<DIM;k=k+2) {
	  dd -= tr[k] * cos((*omega_i)[k]*(*LX)*domain_l);
	}
	for (k=1;k<DIM;k=k+2) {
	  dd -= tr[k] * sin((*omega_i)[k]*(*LX)*domain_l);
	}
	r+= fabs(dd) * GW[j];
      }
      r*=dx;
      d+=r;
    }
    d*=domain_l;
    *LX = lxT;
    return d/tt;
}

const tdouble FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_domain_error_el(const int errMeasure) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    if (errMeasure==1) {
      tdouble t = 0.0;
      for (tuint i=0;i<DIM;++i) {
	t+=pow2(((*Eigenvalues)[i]));
      }
      return (domain_l*2.0-t);
    } else if (errMeasure == 2) {
      flxElement_KL_1D* curElKL = dynamic_cast<flxElement_KL_1D*>(Current_Element_KL);
      #if FLX_DEBUG
	if (curElKL==NULL) {
	  throw FlxException_Crude("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_error_1");
	}
      #endif
      // refine inner interval vector
	const std::vector<tdouble> &iVecI = curElKL->get_IntervalVec_err2(0);
      // refine outer interval vector
	const std::vector<tdouble> &iVecO = curElKL->get_IntervalVec_err2(2);
	const size_t iVecOs = iVecO.size();
      // get Gauss-points (for inner AND outer integral)
	const tuint GN = GI->pDegree2GPs(curElKL->get_pOcovOrder());		// of course this won't be an exact integration
	const tdouble* GP = GI->get_GP(GN);
	const tdouble* GW = GI->get_GW(GN);
      // perform the integration
	tdouble res=0.0, dx, r, lx;
	register tuint j;
	for (tuint i=1;i<iVecOs;++i) {
	  dx = (iVecO[i]-iVecO[i-1])/2.0;
	  r = 0.0;
	  for (j=0;j<GN;++j) {
	    lx = (GP[j]+1.0)*dx+iVecO[i-1];
	    etkl->update_globalConst(lx);
	    r += get_error_cov(iVecI,GN,GP,GW)*GW[j];
	  }
// 	  std::cout << lx << ": " << r/GW[0] << std::endl;
	  r*=dx;
	  res+=r;
	}
	res*=domain_l;
      return res;
    } else {
      std::ostringstream ssV;
      ssV << "Unknown error measure (" << errMeasure << ").";
      throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_domain_error_el_1", ssV.str() );
    }
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::get_domain_error_el_2", ssV.str() );
  }
  #endif
}

void FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plot_el(std::ostream& sout, FlxMtxFun* ppdim, int errMeasure) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif
    iVec ppdim_help = ppdim->cast2tuintNo0();
    if (ppdim_help.size()<1) {
      std::ostringstream ssV;
      ssV << "The dimension of the vector has to be at least '1'.";
      throw FlxException("flxElement_KL_1D::get_PlotPoints", ssV.str() );
    }
    const tdouble d = 2.0/ppdim_help[0];
    const tdouble DU = d/2.0+1.0;
    for (tdouble i=-1.0;i<=DU;i+=d) {
      etkl->update_globalConst(i);
      sout << GlobalVar.Double2String(*GX) << '\t' << GlobalVar.Double2String(*GY) << '\t' << GlobalVar.Double2String(*GZ) << '\t';
      sout << GlobalVar.Double2String(get_error(errMeasure));
      for (tuint j=0;j<DIM;++j) {
	if (j%2==0) {
	  sout << '\t' << GlobalVar.Double2String((((*alpha_i)[j] * cos((*omega_i)[j]*domain_l*(*LX)))));
	} else {
	  sout << '\t' << GlobalVar.Double2String((((*alpha_i)[j] * sin((*omega_i)[j]*domain_l*(*LX)))));
	}
      }
      sout << std::endl;
    }
    sout << std::endl;
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plot_el_1", ssV.str() );
  }
  #endif
}

void FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plotcov_el(std::ostream& sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const
{
  #if FLX_DEBUG
  if (Current_Element_KL!=NULL) {
  #endif  
    iVec ppdim_help = ppdim->cast2tuintNo0();
    if (ppdim_help.size()<1) {
      std::ostringstream ssV;
      ssV << "The dimension of the vector has to be at least '1'.";
      throw FlxException("flxElement_KL_1D::get_PlotPoints", ssV.str() );
    }
    const tdouble d = 2.0/ppdim_help[0];
    *GX2 = pp.get_x();
    *GY2 = pp.get_y();
    *GZ2 = pp.get_z();
    const tdouble DU = d/2.0+1.0;
    for (tdouble i=-1.0;i<=DU;i+=d) {
      etkl->update_globalConst(i);
      *DELTAX = fabs(*GX2-*GX);
      *DELTAY = fabs(*GY2-*GY);
      *DELTAZ = fabs(*GZ2-*GZ);
      *DELTAP = sqrt(pow2(*DELTAX)+pow2(*DELTAY)+pow2(*DELTAZ));
      sout << GlobalVar.Double2String(*GX) << '\t' << GlobalVar.Double2String(*GY) << '\t' << GlobalVar.Double2String(*GZ) << '\t';
      tdouble t = 0.0;
      for (tuint i=0;i<DIM;i=i+2) {	// i+1 is odd
	t+=tr[i]*sD*((*alpha_i)[i] * cos((*omega_i)[i]*domain_l*(*LX)));
      }
      for (tuint i=1;i<DIM;i=i+2) {	// i+1 is even
	t+=tr[i]*sD*((*alpha_i)[i] * sin((*omega_i)[i]*domain_l*(*LX)));
      }
      const tdouble tt = pow2(sD) * exp(-1*(*DELTAP)/corrL);
      sout << GlobalVar.Double2String(t) << '\t' << GlobalVar.Double2String(tt) << '\t' << GlobalVar.Double2String(tt-t);
      sout << std::endl;
    }
    sout << std::endl;
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Out of domain of definiton of the random field";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::plotcov_el_1", ssV.str() );
  }
  #endif
}

void FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::print_para(std::ostream& sout) const
{
  sout << "Gaussian random field (Analytical-KL-expansion; exponential autocorrelation function)";
}

FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D(FlxFunction* mu, tdouble sD, tdouble corrL)
: mu(mu), sD(sD), corrL(corrL),domain_l(1.0),alpha_i(NULL),omega_i(NULL),etkl(NULL)
{
  if (mu->dependOn_RV()) {
    std::ostringstream ssV;
    ssV << "The mean value must not depend on a random variable!";
    throw FlxException("FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D", ssV.str() );
  }
}

FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D::~FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D()
{
  delete mu;
  delete alpha_i;
  delete omega_i;
}

FlxRFBox::FlxRFBox()
{
  flxElement_base::set_RFbox(this);
  flxEdgeList::set_RFbox(this);
  flxFaceList::set_RFbox(this);
  flxBoxList::set_RFbox(this);
}

FlxRFBox::~FlxRFBox()
{
  for (std::map<tuint, FlxRandomField_base*>::iterator pos = boxM.begin(); pos != boxM.end(); ++pos) {
    delete pos->second;
  }
}

void FlxRFBox::update_RV_index()
{
  for (std::map<tuint, FlxRandomField_base*>::iterator pos = boxM.begin(); pos != boxM.end(); ++pos) {
    pos->second->update_RV_index();
  }
}

void FlxRFBox::set_Current_Element_KL(flxDomain_base* El)
{
  FlxRandomField_base::set_Current_Element(El);
  for (std::map<tuint, FlxRandomField_base*>::iterator pos = boxM.begin(); pos != boxM.end(); ++pos) {
    pos->second->set_Current_Element_KL();
  }
}

void FlxRFBox::insert(const tuint index, FlxRandomField_base* value)
{
  std::pair<tuint, FlxRandomField_base*>Element(index, value);
  if ( ! boxM.insert(Element).second ) {;
    delete value;
    std::ostringstream ssV;
    ssV << "Random field '" << index << "' has already been defined.";
    throw FlxException("FlxRFBox::insert_1", ssV.str() );
  }
  GlobalVar.slog(3) << "Random field: defined new random field with index '" << index << "'." << std::endl;
  GlobalVar.slog(4) << "  Type of RF: ";
  value->print_para(GlobalVar.slog(4));
  GlobalVar.slog(4) << std::endl;
}

FlxRandomField_base* FlxRFBox::get(const tuint index, const bool errSerious)
{
  std::map<tuint, FlxRandomField_base*>::iterator pos = boxM.find(index);
  if ( pos != boxM.end() ) {
    return pos->second;
  } else {
    std::ostringstream ssV;
    ssV << "Random field with index '" << index << "' has not been defined.";
    FlxError(errSerious, "FlxRFBox::get", ssV.str() ); return NULL; // dummy-return
  }
}

std::string FlxRF_Generator_base::get_rft(flxElement_KL_base::rftype rft)
{
  switch(rft) {
    case flxElement_KL_base::rfKLfemGauss:
      return "kl_fem_gauss";
      break;
    case flxElement_KL_base::rfKLfemGaussProj:
      return "kl_fem_gauss_proj";
      break;
    case flxElement_KL_base::rfKLAnaExpCorrGauss1D:
      return "kl_analytical_expcorr_gauss_1d";
      break;
    case flxElement_KL_base::rfEOLEgauss:
      return "kl_eole_gauss";
      break;
    default:
      return "";
  }
}

flxElement_KL_base::rftype FlxRF_Generator_base::get_rft(std::string strrft, bool errSerious)
{
  std::transform(strrft.begin(), strrft.end(), strrft.begin(), (int(*)(int)) std::tolower);
  if (strrft == "kl_fem_gauss" ) {
    return flxElement_KL_base::rfKLfemGauss;
  } else if (strrft == "kl_fem_gauss_proj") {
    return flxElement_KL_base::rfKLfemGaussProj;
  } else if (strrft == "kl_analytical_expcorr_gauss_1d" ) {
    return flxElement_KL_base::rfKLAnaExpCorrGauss1D;
  } else if (strrft == "kl_eole_gauss") {
    return flxElement_KL_base::rfEOLEgauss;
  } else {
    std::ostringstream ssV;
    ssV << "Unkown type of random field '" << strrft << "'.";
    FlxError(errSerious,"FlxRF_Generator_base::get_rft", ssV.str() ); return flxElement_KL_base::rfKLfemGauss; // dummy-return
  }
}

FlxRF_Generator_base* FlxRF_Generator_base::createRF(flxElement_KL_base::rftype rft)
{
  switch(rft) {
    case flxElement_KL_base::rfKLfemGauss:
      return new FlxRF_Generator_KL_FEM_Gauss();
      break;
    case flxElement_KL_base::rfKLfemGaussProj:
      return new FlxRF_Generator_KL_FEM_Gauss_proj();
      break;
    case flxElement_KL_base::rfKLAnaExpCorrGauss1D:
      return new FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D();
      break;
    case flxElement_KL_base::rfEOLEgauss:
      return new FlxRF_Generator_EOLE_Gauss();
      break;
    default:
      std::ostringstream ssV;
      ssV << "ERROR.";
      throw FlxException("FlxRF_Generator_base::createRF", ssV.str() );
  }
}

FlxRF_Generator_KL_FEM_Gauss::FlxRF_Generator_KL_FEM_Gauss()
:mu(NULL),sD(NULL),rho(NULL),evtype(NULL)
{
  try {
    reader->getWord("mu");
    mu = new FlxFunction(funReader);
    reader->getWord("sd");
    sD = new FlxFunction(funReader);
    reader->getWord("rhocor");
    rho = new FlxFunction(funReader);
    if (reader->whatIsNextString(6,true) == "evtype") {
      reader->getWord("evtype");
      evtype = new FlxFunction(funReader);
    } else {
      evtype = new FlxFunction(new FunNumber(1.0));
    }
  } catch (FlxException &e) {
    if (mu!=NULL) delete mu;
    if (sD!=NULL) delete sD;
    if (rho!=NULL) delete rho;
    if (evtype!=NULL) delete evtype;
    throw;
  }
}

FlxRF_Generator_KL_FEM_Gauss::~FlxRF_Generator_KL_FEM_Gauss()
{
  delete mu; 
  delete sD; 
  delete rho;
  delete evtype;
}

FlxRandomField_base* FlxRF_Generator_KL_FEM_Gauss::generate_RF(bool errSerious)
{
  return new FlxRandomField_KL_FEM_Gauss(new FlxFunction(*mu),new FlxFunction(*sD),new FlxFunction(*rho),evtype->cast2int());
}

FlxRandomField_base* FlxRF_Generator_KL_FEM_Gauss_proj::generate_RF(bool errSerious)
{
  return new FlxRandomField_KL_FEM_Gauss_proj(new FlxFunction(*mu),new FlxFunction(*sD),new FlxFunction(*rho),evtype->cast2int());
}

FlxRandomField_base* FlxRF_Generator_EOLE_Gauss::generate_RF(bool errSerious)
{
  return new FlxRandomField_EOLE_Gauss(new FlxFunction(*mu),new FlxFunction(*sD),new FlxFunction(*rho),evtype->cast2int());
}

FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D::FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D()
{
  reader->getWord("mu");
  mu = new FlxFunction(funReader);
  reader->getWord("sd");
  sD = new FlxFunction(funReader);
  reader->getWord("corrlength");
  corrL = new FlxFunction(funReader);
}

FlxRandomField_base* FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D::generate_RF(bool errSerious)
{
  if (ConstantBox->dependOn_GlobalSpatialVar(sD) || ConstantBox->dependOn_GlobalSpatialVar(corrL) ) {
    std::ostringstream ssV;
    ssV << "The standard deviation or the correlation length must not depend on spatial global variables.";
    FlxError(errSerious, "FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D::generate_RF_1", ssV.str() );
  }
  tdouble sdt = sD->cast2positive(errSerious);
  tdouble clt = corrL->cast2positive(errSerious);
  return new FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D(new FlxFunction(*mu),sdt,clt);
}

FlxRFsolveProp::FlxRFsolveProp(bool errSerious)
:dim(NULL),octaveOutput(false),calcerr(true)
{
  try {
    bool b=true;
    do {
      if (reader->whatIsNextString(3,true) == "dim") {
	reader->getWord("dim",false);
	reader->getChar('=',false);
	if (dim!=NULL) {
	  std::ostringstream ssV;
	  ssV << "'dim' has aleready been defined.";
	  throw FlxException("FlxRFsolveProp::FlxRFsolveProp", ssV.str() );
	}
	dim = new FlxFunction(funReader,false);
      } else if (reader->whatIsNextString(12,true) == "octaveoutput") {
	reader->getWord("octaveoutput",false);
	reader->getChar('=',false);
	octaveOutput = reader->getBool(false);
      } else if (reader->whatIsNextString(7,true) == "calcerr") {
	reader->getWord("calcerr",false);
	reader->getChar('=',false);
	calcerr = reader->getBool(false);
      } else {
	b=false;
      }
    } while (b);
  } catch (FlxException &e) {
    if (dim) {
      delete dim;
      dim=NULL;
    }
    throw;
  }
}

FlxRFsolveProp::~FlxRFsolveProp()
{
  if (dim!=NULL) delete dim;
}

const tuint FlxRFsolveProp::get_dim()
{
  if (dim!=NULL) {
    return dim->cast2tuint();
  } else {
    std::ostringstream ssV;
    ssV << "'dim' has not been defined";
    throw FlxException("FlxRFsolveProp::get_dim", ssV.str() );
  }
}



// -------------------------- IsoRFMaterial ----------------------------------------------------------------------


flxIsoRFMat::flxIsoRFMat(FlxRandomField_base* RFptr, FlxFunction* nu, FlxFunction* density, const tuint deg_E)
: flxIsotropicMaterial('E',RFptr->get_mean_fun(),'n',nu,density,deg_E), RFptr(RFptr)
{
  FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D* klanalytgauss1d = dynamic_cast<FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D*>(RFptr);
  if (klanalytgauss1d==NULL) {
    FlxRandomField_KL_FEM_Gauss* klfemgauss = dynamic_cast<FlxRandomField_KL_FEM_Gauss*>(RFptr);
    if (klfemgauss==NULL) {
      std::ostringstream ssV;
      ssV << "The specified random field is not a Gaussian random field.";
      throw FlxException("flxIsoRFMat::flxIsoRFMat", ssV.str() );
    }
  }
}

flxIsoRFMat::flxIsoRFMat(const flxIsoRFMat& FlxM): flxIsotropicMaterial(FlxM), RFptr(FlxM.RFptr)
{
  RFptr->set_automatic();
}

flxIsoRFMat* flxIsoRFMat::copy()
{
  return new flxIsoRFMat(*this);
}

void flxIsoRFMat::log_info(std::ostream& lout)
{
  lout << "  isorfmat: ";
  lout << v1c << "=" << v1->write() << "; " << v2c << "=" << v2->write() << "; ";
  lout << "Random in E: ";
  RFptr->print_para(lout);
  lout << std::endl;
}

const bool flxIsoRFMat::depend_E_on_Const(const tdouble* const thenumber) {
  if (ConstantBox->is_SpatialVar(thenumber)) return true;
  else return flxIsotropicMaterial::depend_E_on_Const(thenumber);
}

const tdouble flxIsoRFMat::get_E() const
{
  const int& Current_rv_index = FlxRnd_base::reference_Current_rv_index();
  if (!RFptr->is_Assembled() || Current_rv_index==0) return flxIsotropicMaterial::get_E();
  if (Current_rv_index == -1) {
    return RFptr->get_realization();
  } else {
    return RFptr->calc_Hi(Current_rv_index,true);
  }
}

const tdouble flxIsoRFMat::get_G() const
{
  const tdouble E = get_E();
  const tdouble nu = get_nu();
  return check_G(E/(2*(1+nu)));
}

const tdouble flxIsoRFMat::get_lambda() const
{
  const tdouble E = get_E();
  const tdouble nu = get_nu();
  return E*nu/((1+nu)*(1-2*nu));
}


// -------------------------- IsoRVMaterial ----------------------------------------------------------------------

flxIsoRVMat::flxIsoRVMat(FlxFunction* E, FlxFunction* sD, FlxFunction* nu, FlxFunction* density)
: flxIsotropicMaterial('E',E,'n',nu,density,0),sD(sD)
{

}


flxIsoRVMat::flxIsoRVMat(const flxIsoRVMat& FlxM)
: flxIsotropicMaterial(FlxM), sD(new FlxFunction(*(FlxM.sD)))
{
  RndBox_index = RndBox->insert(new FlxRnd_StdN());
}

flxIsoRVMat* flxIsoRVMat::copy()
{
  return new flxIsoRVMat(*this);
}

void flxIsoRVMat::log_info(std::ostream& lout)
{
  lout << "  isorvmat: ";
  lout << v1c << "=" << v1->write() << "; " << v2c << "=" << v2->write() << "; ";
  lout << "Random in E: ";
  lout << "Normal(mu=" << v1->write() << "; sd=" << sD->write() << "); ";
  lout << std::endl;
}

const tdouble flxIsoRVMat::get_E() const
{
  const int& Current_rv_index = FlxRnd_base::reference_Current_rv_index();
  if (Current_rv_index < 0) {
    tdouble d = flxIsotropicMaterial::get_E();
    d+=sD->cast2positive()*RndBox->get_x_Vec()[RndBox_index];
    return d;
  } else if (Current_rv_index == 0) {
    return flxIsotropicMaterial::get_E();
  } else if (static_cast<tuint>(Current_rv_index)-1 == RndBox_index) {
    return sD->cast2positive();
  } else return 0.0;
}

const tdouble flxIsoRVMat::get_G() const
{
  const tdouble E = get_E();
  const tdouble nu = get_nu();
  return check_G(E/(2*(1+nu)));
}

const tdouble flxIsoRVMat::get_lambda() const
{
  const tdouble E = get_E();
  const tdouble nu = get_nu();
  return E*nu/((1+nu)*(1-2*nu));
}


