/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxrandomFEM_H
#define fesslix_flxrandomFEM_H

#include "flxfem.h"



//======================= Random Fields ============================================

class FlxRFsolveProp;
class FlxRFBox;
class FlxRandomField_base : public FlxBoxBase2 {
  protected:
    static tdouble *LX,*GX,*GY,*GZ,*GX2,*GY2,*GZ2,*DELTAX,*DELTAY,*DELTAZ,*DELTAP;
    /**
    * @brief true, if the random field has been assembled (solved)
    */
    bool is_assembled;
    /**
    * @brief if it is an automatic random field, it cannot be accessed within an FlxFunction
    */
    bool automaticV;
    /**
    * @brief the dimension of the random field - zero if is_assembled=false
    */
    tuint DIM;
    /**
    * @brief pointer to first random variable of the field in uncorrelated standard normal space
    */
    const tdouble* y_smp;
    /**
    * @brief pointer to first random variable of the field in original space
    */
    const tdouble* x_smp;
    /**
    * @brief Eigenvalues
    */
    tVec* Eigenvalues;	// stores the square roots of the eigenvalues
    /**
    * @brief index of first random variable in RandomBox
    */
    tuint RndBox_index;
    /**
    * @brief stores the elements of the random field ... they define the domain !!!
    */
    std::map<flxDomain_base*, flxElement_KL_base*> boxEl;
    flxElement_KL_base* Current_Element_KL;
    /**
    * @brief Current Element (has to be updated before 'get_realization' is called)
    */
    static flxDomain_base* Current_Element;
    
    const tdouble get_domain() const;
    /**
    * @brief integrate the error over the domain of the current element
    */
    virtual const tdouble get_domain_error_el(const int errMeasure) const = 0;
    /**
    * @brief plot the randomfield within the current element
    */
    virtual void plot_el(std::ostream &sout, FlxMtxFun* ppdim, const int errMeasure) const = 0;
    /**
    * @brief plot the covariance function of the randomfield within the current element
    */
    virtual void plotcov_el(std::ostream &sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const = 0;
    /**
    * @brief ... will be called by 'FlxRFBox'
    */
    static void set_Current_Element(flxDomain_base* El) { Current_Element = El; }
  public:
    FlxRandomField_base() : is_assembled(false), automaticV(false), DIM(0), y_smp(NULL), x_smp(NULL), Eigenvalues(NULL), RndBox_index(0), Current_Element_KL(NULL) {};
    virtual ~FlxRandomField_base();
    void set_UserVar() { automaticV=false;}
    const bool is_automatic() { return automaticV; }
    const bool is_Assembled() { return is_assembled; }
    void set_automatic();
    /**
    * @brief updates the pointers 'y_smp' and 'x_smp'
    */
    void update_RV_index();
    /**
    * @brief assign an element to the random field
    * @param pel the FEM-parent element
    * @param KLelProp element properties to assign
    * @param report true: log element info
    * @param rfindex if report=true - index of the random field
    */
    void set_element(flxDomain_base* pel, flxElement_KL_base* klb, const bool report=false, const tuint rfindex=0, const std::string& elID=0);
    /**
    * @brief get the current realization of the random field (based on current_element and global variables)
    */
    virtual const tdouble get_realization() const = 0;
    /**
    * @brief get the mean error
    * @param errMeasure	1: error variance; 2: covariance error
    */
    const tdouble get_domain_error(int errMeasure);
    
    /**
    * @brief returns a string describing the RF
    */
    virtual void print_para(std::ostream &sout) const = 0;
    /**
    * @brief assembles the random field
    * @return number of DOFs in the system
    */
    virtual const tnlong solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout) = 0;
    /**
    * @brief plots the random field
    */
    void plot(std::ostream &sout, FlxMtxFun* ppdim, const int errMeasure);
    /**
    * @brief plots the covariance function of the random field
    */
    void plotcov(std::ostream &sout, const flxPoint& pp, FlxMtxFun* ppdim);    
    
    /**
    * @brief returns a function describing the mean - memory has to be deallocated !!!
    */
    virtual FlxFunction* get_mean_fun();
    /**
    * @brief calculates a term of the linear expansion - without the random variable
    * @param i the index 
    *		-2: shape function with highest order
    * 		-1: not allowed here
    * 		0: mean value
    * 		>1: mode -> if globalIndex=true: global mode; otherwise  local mode
    * @note uses global local coordinates (LX,LY,LZ)
    */
    virtual const tdouble calc_Hi(int i, bool globalIndex);
    /**
    * @brief returns the ith eigenvalue (counting starts with 0)
    */
    virtual const tdouble get_Eigenvalue(const int i, bool errSerious=true) = 0;
    
    const tuint get_RndBox_index() { return RndBox_index; }
    const tuint get_DIM() { return DIM; }
    /**
    * @brief set the current KL-element
    */
    void set_Current_Element_KL();
    

    virtual flxElement_KL_base::rftype get_Type() const = 0;
    static void set_Const();
    
    friend class FlxRFBox;
};


class FlxRandomField_KL_FEM_Gauss : public FlxRandomField_base {
  protected:
    /**
    * @brief mean value
    */
    FlxFunction* mu;
    /**
    * @brief standard deviation
    */
    FlxFunction* sD;
    /**
    * @brief correlation
    */
    FlxFunction* rho;
    /**
    * @brief Eigenvectors (DIM*Ndof)
    */
    std::vector<tVec> Eigenvectors;
    /**
    * @brief Number of DOFs of the problem
    */
    tnlong Ndof;
    /**
    * @brief specifies the type used for solving the eigenvalue problem
    */
    const int ev_solver;
    /**
    * @brief stores topological information - only required for projection methods!!!
    */
    niMap nodalM;
    eiMap edgeM;
    fiMap faceM;

    virtual const tdouble get_domain_error_el(const int errMeasure) const;
    virtual void plot_el(std::ostream &sout, FlxMtxFun* ppdim, const int errMeasure) const;
    virtual void plotcov_el(std::ostream &sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const;
    /**
    * @brief Assembles the DOFs of the elements in the field
    */
    void assembleElementDOFs();
  public:
    FlxRandomField_KL_FEM_Gauss(FlxFunction* mu, FlxFunction* sD, FlxFunction* rho, const int ev_solver);
    virtual ~FlxRandomField_KL_FEM_Gauss();
    virtual const tdouble get_realization() const;
    virtual void print_para(std::ostream &sout) const;
    virtual flxElement_KL_base::rftype get_Type() const {return flxElement_KL_base::rfKLfemGauss;}
    virtual const tnlong solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout);
    virtual FlxFunction* get_mean_fun();
    virtual const tdouble calc_Hi(int i, bool globalIndex);
    virtual const tdouble get_Eigenvalue(const int i, bool errSerious=true);
    
    FlxFunction* get_sD() { return sD; }
};

class FlxRandomField_KL_FEM_Gauss_proj : public FlxRandomField_KL_FEM_Gauss {
  protected:
//     virtual const tdouble get_domain_error_el(const int errMeasure) const;
//     virtual void plot_el(std::ostream &sout, FlxFunction* ppdim, const int errMeasure) const;
//     virtual void plotcov_el(std::ostream &sout, flxPoint* pp, FlxFunction* ppdim, const tVec& tr) const;
    virtual const tdouble get_domain_error_el(const int errMeasure) const;
    virtual void plot_el(std::ostream &sout, FlxMtxFun* ppdim, const int errMeasure) const;
  public:
    FlxRandomField_KL_FEM_Gauss_proj(FlxFunction* mu, FlxFunction* sD, FlxFunction* rho, const int ev_solver);
    virtual ~FlxRandomField_KL_FEM_Gauss_proj();
    virtual void print_para(std::ostream &sout) const;
    virtual flxElement_KL_base::rftype get_Type() const {return flxElement_KL_base::rfKLfemGaussProj;}
    virtual const tnlong solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout);
};

class FlxRandomField_EOLE_Gauss : public FlxRandomField_KL_FEM_Gauss {
  protected:
    std::vector<flxPoint> pointVec;
    /**
    * @brief standard deviation of points in pointVec
    */
    tVec* sdVec;
    /**
    * @brief vector needed to compute the eigenfunctions ... depends on current x !!!
    */
    tVec* Hx;
    
    virtual const tdouble get_domain_error_el(const int errMeasure) const;
    /**
    * @brief get the error at a point within the field (based on current_element and global variables)
    */
    const tdouble get_error(int errMeasure) const;
    virtual void plot_el(std::ostream &sout, FlxMtxFun* ppdim, const int errMeasure) const;
    virtual void plotcov_el(std::ostream &sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const;
    /**
    * @brief Assembles the DOFs of the elements in the field ... meaning a LOT of points from each element
    */
    void assembleElementDOFs();
    /**
    * @brief Updates the Hx-vector ... note GX,GY and GZ have to be set !!!
    */
    const tdouble assemble_Hx() const;
  public:
    FlxRandomField_EOLE_Gauss(FlxFunction* mu, FlxFunction* sD, FlxFunction* rho, const int ev_solver);
    ~FlxRandomField_EOLE_Gauss();
    virtual const tdouble get_realization() const;
    virtual void print_para(std::ostream &sout) const;
    virtual flxElement_KL_base::rftype get_Type() const {return flxElement_KL_base::rfEOLEgauss;}
    virtual const tnlong solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout);
    virtual const tdouble calc_Hi(int i, bool globalIndex);
    virtual const tdouble get_Eigenvalue(const int i, bool errSerious=true);
};

class FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D : public FlxRandomField_base {
  protected:
    /**
    * @brief mean value
    */
    FlxFunction* mu;
    /**
    * @brief standard deviation
    */
    tdouble sD;
    /**
    * @brief correlation length
    */
    tdouble corrL;
    /**
    * @brief length of the domain over 2 - after the field has been assembled
    */
    tdouble domain_l;
    
    tVec* alpha_i;
    tVec* omega_i;
    
    flxElement_KL_1D* etkl;
    
    virtual const tdouble get_domain_error_el(const int errMeasure) const;
    /**
    * @brief get the error at a point within the field (based on current_element and global variables)
    */
    const tdouble get_error(const int errMeasure) const;
    /**
    * @brief calculates the cov_error by integartion. LX has to be set (LX will be conserved)
    */
    const tdouble get_error_cov(const std::vector<tdouble> &iVec, const tuint GN, const tdouble* GP, const tdouble* GW) const;
    virtual void plot_el(std::ostream &sout, FlxMtxFun* ppdim, const int errMeasure) const;
    virtual void plotcov_el(std::ostream &sout, const flxPoint& pp, FlxMtxFun* ppdim, const tVec& tr) const;
  public:
    FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D(FlxFunction* mu, tdouble sD, tdouble corrL);
    ~FlxRandomField_KL_Analytical_ExpCorr_Gauss_1D();
    virtual const tdouble get_realization() const;
    virtual void print_para(std::ostream &sout) const;
    virtual flxElement_KL_base::rftype get_Type() const {return flxElement_KL_base::rfKLAnaExpCorrGauss1D;}
    virtual const tnlong solve(FlxRFsolveProp* RFsolveProp,std::ostream& sout);
    virtual FlxFunction* get_mean_fun();
    virtual const tdouble calc_Hi(int i, bool globalIndex);
    virtual const tdouble get_Eigenvalue(const int i, bool errSerious=true);
};


//======================= FlxRFBox ============================================

class FlxRFBox {
  protected:
    /**
    * @brief stores all random fields
    */
    std::map<tuint, FlxRandomField_base*> boxM;
  public:
    FlxRFBox();
    ~FlxRFBox();
    /**
    * @brief insert an user defined random variable
    */
    void insert (const tuint index, FlxRandomField_base* value);
    
    FlxRandomField_base* get(const tuint index, const bool errSerious=true);
    
    /**
    * @brief updates the pointers 'y_smp' and 'x_smp' of all random fields
    */
    void update_RV_index();
    /**
    * @brief set the current KL-element of all random fields
    */
    void set_Current_Element_KL(flxDomain_base* El);
};


//======================= Generators of Random Fields ============================================

class FlxRF_Generator_base : public FlxReaderBase2, public FlxBoxBase2 {
  public:
    virtual ~FlxRF_Generator_base() {};
    /**
    * @brief create an instance of a random field
    */
    virtual FlxRandomField_base* generate_RF(bool errSerious=true) = 0;
    
    // Types of Random Fields
      static flxElement_KL_base::rftype get_rft(std::string strrft, bool errSerious = true);
      static std::string get_rft(flxElement_KL_base::rftype rft);
    /**
    * @brief create an instance of a random field generator
    */
    static FlxRF_Generator_base* createRF(flxElement_KL_base::rftype rft);   
};

class FlxRF_Generator_KL_FEM_Gauss : public FlxRF_Generator_base {
  protected:
    FlxFunction* mu;
    FlxFunction* sD;
    FlxFunction* rho;
    FlxFunction* evtype;
  public:
    FlxRF_Generator_KL_FEM_Gauss();
    virtual ~FlxRF_Generator_KL_FEM_Gauss();
    FlxRandomField_base* generate_RF(bool errSerious=true);
};

class FlxRF_Generator_KL_FEM_Gauss_proj : public FlxRF_Generator_KL_FEM_Gauss {
  protected:
    
  public:
    FlxRandomField_base* generate_RF(bool errSerious=true);
};

class FlxRF_Generator_EOLE_Gauss : public FlxRF_Generator_KL_FEM_Gauss {
  protected:
    
  public:
    FlxRandomField_base* generate_RF(bool errSerious=true);
};

class FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D : public FlxRF_Generator_base {
  protected:
    FlxFunction* mu;
    FlxFunction* sD;
    FlxFunction* corrL;
  public:
    FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D();
    ~FlxRF_Generator_KL_Analytical_ExpCorr_Gauss_1D() { delete mu; delete sD; delete corrL; }
    FlxRandomField_base* generate_RF(bool errSerious=true);
};


/**
* @brief stores properties to create KL-elements
*/
class FlxRFsolveProp : public FlxReaderBase2 {
  private:
    FlxFunction* dim;
    bool octaveOutput;
    bool calcerr;
  public:
    FlxRFsolveProp(bool errSerious=true);
    ~FlxRFsolveProp();
    const tuint get_dim();
    const bool get_octaveOutput() { return octaveOutput; }
    const bool get_calcerr() { return calcerr; }
};


//======================= Random Field Material ============================================
  
class flxIsoRFMat : public flxIsotropicMaterial, public FlxBoxBase2 {
  private:
    FlxRandomField_base* RFptr;
  public:
    flxIsoRFMat(FlxRandomField_base* RFptr, FlxFunction* nu, FlxFunction* density, const tuint deg_E);
    flxIsoRFMat(const flxIsoRFMat& FlxM);
    virtual ~flxIsoRFMat() {};
    virtual flxIsoRFMat* copy();
    virtual void log_info(std::ostream& lout);
    const bool depend_E_on_Const(const tdouble* const thenumber);
    virtual const tdouble get_E() const;
    virtual const tdouble get_G() const;
    virtual const tdouble get_lambda() const;
};


//======================= Random Variable Material ============================================

class flxIsoRVMat : public flxIsotropicMaterial, public FlxBoxBase2 {
  private:
    /**
    * @brief standard deviation
    */
    FlxFunction* sD;
    /**
    * @brief index of first random variable in RandomBox
    */
    tuint RndBox_index;
  public:
    flxIsoRVMat(FlxFunction* E, FlxFunction* sD, FlxFunction* nu, FlxFunction* density);
    flxIsoRVMat(const flxIsoRVMat& FlxM);
    virtual ~flxIsoRVMat() { delete sD; };
    virtual flxIsoRVMat* copy();
    virtual void log_info(std::ostream& lout);
    virtual const tdouble get_E() const;
    virtual const tdouble get_G() const;
    virtual const tdouble get_lambda() const;
};



#endif // fesslix_flxrandomFEM_H
