/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxrandom_base_H
#define fesslix_flxrandom_base_H

#include "flxfunction_data.h"
#include "flxMtx.h"

class calc_PC_projection_numerical_MCI;

//======================= Preliminary Definitions ============================================

class FlxRndBox;

class FlxBoxBaseR {
  protected:
    static FlxRndBox *RndBox;
    static GaussIntegration* GI;
  public:
    ~FlxBoxBaseR() {};
    static void set_Boxes( FlxRndBox *RndBoxV);
    static void set_GI(GaussIntegration* GIv) { GI = GIv;}
};


//======================= Calculate Expectation numerically 1D ============================================

class FlxRnd_base;

class calc_expectation_numerical_base {
  private:
    virtual tdouble calc(const tdouble& x, const tdouble& y) = 0;
    pdouble calc_Interval(const tdouble& x_o, const tdouble& x_u, const tnlong& N, const FlxRnd_base& rnd);
  public:
    virtual ~calc_expectation_numerical_base() {}
    /**
    * @brief calculates the expectation of calc(x)
    * @param N_Interv Number of intervals
    * @param N Number of samples to calculate - distributed over intervals
    * if (N==0): just N_Interv are calculated
    */
    tdouble calc_expectation(const tnlong& N_Interv, const tnlong& N, const tdouble& pRedistribute, const FlxRnd_base& rnd);
};

class calc_expectation_numerical_nth_moment : public calc_expectation_numerical_base {
  private:
    int n;
    tdouble mu;
    virtual tdouble calc(const tdouble& x, const tdouble& y) { return pow(x-mu,n); }
  public:
    calc_expectation_numerical_nth_moment(int n, tdouble mu=0.0) : n(n), mu(mu) {}
};

class FunBase;

/**
* @brief calculates the expectation value of a function numerically
* NOTE: does not delete fun !!!
*/
class calc_expectation_numerical_Fun : public calc_expectation_numerical_base, public FlxBoxBaseR {
  private:
    FunBase* fun;
    const FlxRnd_base* rndB;
    tdouble* orig_x_Ptr;
    tdouble* orig_y_Ptr;
    virtual tdouble calc(const tdouble& x, const tdouble& y);
  public:
    calc_expectation_numerical_Fun(FunBase* fun) : fun(fun) {}
    tdouble calc_expectation(const tnlong& N_Interv, const tnlong& N, const tdouble& pRedistribute, const FlxRnd_base& rnd);
};

//======================= Calculate Expectation numerically MCI ============================================

class FlxRnd_Uniform;

/**
* @brief calculates the expectation value of a function numerically
* ... for functions depending on more than one variable
*/
class calc_expectation_numerical_MCI : public FlxBoxBaseR {
  private:
    /**
    * @brief The function to calculate the expectation from
    */
    FunBase* funR;
    /**
    * @brief Dimension of the problem
    */
    tuint DIM;
    /**
    * @brief vector with references to realizations in uncorrelated standard normal space
    */
    tdouble** y_smp;
    /**
    * @brief vector with references to realizations in original space
    */
    tdouble** x_smp;
    /**
    * @brief stores mapping of random variables
    */
    iVec* Index_V;
    /**
    * @brief in case L=NULL -> no correlation
    */
    FlxMtxSparsLTri* L;
    /**
    * @brief uniform distribution
    */
    FlxRnd_Uniform* uD;
    /**
    * @brief transforms from uncorrelated stdN-space to correlated stdN-space
    */
    inline void y2w(const tVec& y, tVec& w) const;
    /**
    * @brief create standard normal y vector
    */
    inline void create_y(tVec& y);
    /**
    * @brief set standard normal y vector in RndBox
    */
    inline void set_y(const tVec& y);
    /**
    * @brief set standard normal x vector in RndBox
    */
    inline void set_x(const tVec& x);
    /**
    * @brief update x from y
    */
    inline void update_x(const tVec& y, tVec &tempVec);
    /**
    * @brief one MCI-Integrationstep
    */
    inline tdouble Integrationstep(tVec &tempVec);
  public:
    /**
    * @brief calculates the expectation of fun
    * @param fun The function to calculate the expectation from
    * @param bv a vector with information about the random variables involved in this function
    * @param N_Interv number of intervals to split the sampling
    * @param N_mciI_1 number of MCI-samples per interval (for the first round)
    * @param N_mciI_2 number of MCI-samples in general (for the second round) - adaptive sampling
    * @return returns the expectation of fun
    */
    tdouble calc_expectation(FunBase* fun, const bVec& bv, const tnlong& N_Interv, const tnlong& N_mciI_1, const tnlong& N_mciI_2, const tdouble& pRedistribute);
};

//======================= Random Variables ============================================

class FlxRFBox;
class FlxRnd_base {
  protected:
    /**
    * @brief uncorrelated standard normal space
    */
    const tdouble* y_smp;
    /**
    * @brief original space
    */
    const tdouble* x_smp;
    bool automaticV;
    /**
    * @brief true, if the random variable belongs to a random field
    */
    bool isRandomField; 
    bool isRandomField_automatic; // belongs to an automatic RF
    /**
    * @brief stores the numerical moments
    */
    tuint num_moment_array_size;
    tuint num_moment_central_array_size;
    tdouble* num_moment_array;
    tdouble* num_moment_central_array;
    /**
    * @brief Current global index of the random variable to use
    * -2: shape function of highest order (in case of random fields) ... random realization otherwise 
    * -1: random realization
    * 0: mean value
    * 1 .. n_RV: Eigenmodes
    */
    static int Current_rv_index;
  public:
    FlxRnd_base();
    virtual ~FlxRnd_base();
    const tdouble pdf_y(const tdouble y) const { return rv_phi(y);}
    const tdouble cdf_y(const tdouble y) const { return rv_Phi(y);}
    const tdouble& get_x_smp() const { return *x_smp; }
    const tdouble& get_y_smp() const {return *y_smp;}
    void set_UserVar() { automaticV=false;}
    const bool is_automatic() const { return automaticV; }
    const bool is_RandomField() const { return isRandomField; }
    const bool is_RandomField_automatic() const { return isRandomField_automatic; }
    void set_RandomField() { isRandomField=true; }
    void set_RandomField_automatic() { isRandomField_automatic=true; }
    virtual const tdouble y2x(const tdouble y) const = 0;
    virtual const tdouble x2y(const tdouble x) const = 0;
    virtual const tdouble pdf_x(const tdouble x) const = 0;
    virtual const tdouble mean() const = 0;
    /**
    * @brief Calculates the standard deviation of the variable
    */
    virtual const tdouble sd() const = 0;
    virtual const tdouble cdf_x(const tdouble x) const = 0;
    /**
    * @brief Calculates the expectation value of X^n
    */
    virtual const tdouble moment(const int nth);
    /**
    * @brief Calculates the expectation value of (X-mean)^n
    */
    virtual const tdouble moment_central(const int nth);
    /**
    * @brief returns a string describing the RV
    */
    virtual void print_para(std::ostream &sout) const = 0;
    /**
    * @brief checks if the x-value is in the physical domain
    */
    virtual const bool check_value(const tdouble &xval) const { return true; }
    
    static void set_Current_rv_index(int index) { Current_rv_index = index;};
    static const int get_Current_rv_index() { return Current_rv_index;};
    static int& reference_Current_rv_index() {return Current_rv_index;};
    
    // Types of Random Variables (do NOT change the ordering)
      enum rvtype { 
	rvStdN, rvNormal,rvLogN, rvGumbel, rvUnifrom, rvExponential, rvBeta
      };
      
    virtual const rvtype get_Type() const = 0;
    
    friend class FlxRndBox;
};

class FlxRnd_StdN : public FlxRnd_base {
  public:
    virtual const tdouble y2x(const tdouble y) const { return y;};
    virtual const tdouble x2y(const tdouble x) const { return x;};
    virtual const tdouble pdf_x(const tdouble x) const { return pdf_y(x); }
    virtual const tdouble cdf_x(const tdouble x) const { return cdf_y(x); }
    virtual const tdouble mean() const { return 0.0; }
    virtual const tdouble sd() const { return 1.0; }
    virtual const tdouble moment(const int nth);
    virtual const tdouble moment_central(const int nth) { return moment(nth);}
    virtual void print_para(std::ostream &sout) const { sout << "standard normal"; }
    virtual const rvtype get_Type() const {return rvStdN;}
};

class FlxRnd_Normal : public FlxRnd_base {
  protected:
    const tdouble mu;
    const tdouble sigma;
  public:
    FlxRnd_Normal(const tdouble mu, const tdouble sigma):mu(mu),sigma(sigma){ };
    FlxRnd_Normal(const FlxRnd_Normal& Distri) : mu(Distri.mu), sigma(Distri.sigma) {}
    virtual const tdouble y2x(const tdouble y) const { return y*sigma+mu;}
    virtual const tdouble x2y(const tdouble x) const {return (x-mu)/sigma;}
    virtual const tdouble pdf_x(const tdouble x) const { return pdf_y(x2y(x))/sigma; }
    virtual const tdouble cdf_x(const tdouble x) const { return cdf_y(x2y(x)); }
    virtual const tdouble mean() const { return mu; }
    virtual const tdouble sd() const { return sigma; }
    virtual const tdouble moment(const int nth);
    virtual const tdouble moment_central(const int nth);
    virtual void print_para(std::ostream &sout) const { sout << "normal: mean=" << mu << "; sd=" << sigma << ";"; }
    virtual const rvtype get_Type() const {return rvNormal;}
};

class FlxRnd_LogNormal : public FlxRnd_base {
  private:
    const tdouble lambda;
    const tdouble zeta;
    const tdouble epsilon;
  public:
    FlxRnd_LogNormal(const tdouble lambda, const tdouble zeta, const tdouble epsilon) : lambda(lambda),zeta(zeta),epsilon(epsilon) {}
    FlxRnd_LogNormal(const FlxRnd_LogNormal& Distri);
    virtual const tdouble y2x(const tdouble y) const { return exp(y*zeta+lambda)+epsilon;}
    virtual const tdouble x2y(const tdouble x) const;
    virtual const tdouble pdf_x(const tdouble x) const;
    virtual const tdouble cdf_x(const tdouble x) const { return cdf_y(x2y(x)); }
    virtual const tdouble mean() const { return exp(lambda+pow2(zeta)/2)+epsilon; }
    virtual const tdouble sd() const { return sqrt(exp(pow2(zeta))-1)*exp(lambda+0.5*pow2(zeta)); }
    virtual const tdouble moment(const int nth);
    virtual void print_para(std::ostream& sout) const;
    virtual const rvtype get_Type() const {return rvLogN;}
    virtual const bool check_value(const tdouble &xval) const;
    
    const tdouble get_CoeffOfVar_withoutEpsilon() { return sd()/(mean()-epsilon); }
};

class FlxRnd_Gumbel : public FlxRnd_base {
  protected:
    const tdouble u;
    const tdouble alpha;
  public:
    FlxRnd_Gumbel(const tdouble u, const tdouble alpha):u(u),alpha(alpha){ };
    FlxRnd_Gumbel(const FlxRnd_Gumbel& Distri);
    virtual const tdouble y2x(const tdouble y) const { return u-(std::log((rv_Phi(y)==1.0)?(rv_Phi(-y)):(-std::log(rv_Phi(y)))))/alpha;}
    virtual const tdouble x2y(const tdouble x) const {return rv_InvPhi(cdf_x(x));}
    virtual const tdouble pdf_x(const tdouble x) const { return alpha*exp(-alpha*(x-u)-exp(-alpha*(x-u))); }
    virtual const tdouble cdf_x(const tdouble x) const { return exp(-exp(-alpha*(x-u))); }
    virtual const tdouble mean() const { return u+GAMMA/alpha; }
    virtual const tdouble sd() const { return PI/(sqrt(6.)*alpha); }
    virtual void print_para(std::ostream& sout) const;
    virtual const rvtype get_Type() const {return rvGumbel;}
};

class FlxRnd_Uniform : public FlxRnd_base {
  protected:
    tdouble a;
    tdouble b;
    tdouble mean_v;
    void reset_bound(const tdouble aNew, const tdouble bNew) { a = aNew; b = bNew; mean_v = (a+b)/2;  }
  public:
    FlxRnd_Uniform(const tdouble a, const tdouble b):a(a),b(b),mean_v((a+b)/2){ }
    FlxRnd_Uniform(const FlxRnd_Uniform& Distri);
    virtual const tdouble y2x(const tdouble y) const { return (y>0)?(b-rv_Phi(-y)*(b-a)):(rv_Phi(y)*(b-a)+a);}
    virtual const tdouble x2y(const tdouble x) const { return (x>mean_v)?(-1*rv_InvPhi((b-x)/(b-a))):(rv_InvPhi((x-a)/(b-a))); };
    virtual const tdouble pdf_x(const tdouble x) const;
    virtual const tdouble cdf_x(const tdouble x) const;
    virtual const tdouble mean() const { return mean_v; }
    virtual const tdouble sd() const { return (b-a)/sqrt(12.); }
    virtual const tdouble moment(const int nth);
    virtual const tdouble moment_central(const int nth);
    virtual void print_para(std::ostream& sout) const;
    virtual const rvtype get_Type() const {return rvUnifrom;}
    virtual const bool check_value(const tdouble &xval) const;
    
    const tdouble Inv_cdf_x(const tdouble p) const;
    
    friend class calc_expectation_numerical_MCI;
    friend class calc_PC_projection_numerical_MCI;
};

class FlxRnd_Exponential : public FlxRnd_base {
  protected:
    const tdouble lambda;
    const tdouble lambda_inv;
  public:
    FlxRnd_Exponential(const tdouble lambda):lambda(lambda),lambda_inv(1./lambda) { };
    FlxRnd_Exponential(const FlxRnd_Exponential& Distri);
    virtual const tdouble y2x(const tdouble y) const { return -1.*std::log((y>0)?(rv_Phi(-y)):(1.-rv_Phi(y)))*lambda_inv; }
    virtual const tdouble x2y(const tdouble x) const {return rv_InvPhi(cdf_x(x));}
    virtual const tdouble pdf_x(const tdouble x) const;
    virtual const tdouble cdf_x(const tdouble x) const;
    virtual const tdouble mean() const { return lambda_inv; }
    virtual const tdouble sd() const { return lambda_inv; }
    virtual void print_para(std::ostream& sout) const;
    virtual const rvtype get_Type() const {return rvExponential;}
    virtual const bool check_value(const tdouble &xval) const;
};

class FlxRnd_Beta : public FlxRnd_base {
  protected:
    const tdouble alpha;
    const tdouble beta;
    const tdouble a;
    const tdouble b;
    const tdouble mu;
    const tdouble lnbetafun;
  public:
    FlxRnd_Beta(const tdouble alpha, const tdouble beta, const tdouble a=0., const tdouble b=1.);
    FlxRnd_Beta(const FlxRnd_Beta& Distri);
    virtual const tdouble y2x(const tdouble y) const;
    virtual const tdouble x2y(const tdouble x) const;
    virtual const tdouble pdf_x(const tdouble x) const;
    virtual const tdouble cdf_x(const tdouble x) const;
    virtual const tdouble mean() const { return (alpha*b+beta*a)/(alpha+beta); }
    virtual const tdouble sd() const { return (b-a)/(alpha+beta)*sqrt(alpha*beta/(alpha+beta+1.)); }
    virtual void print_para(std::ostream& sout) const;
    virtual const rvtype get_Type() const {return rvBeta;}
    virtual const bool check_value(const tdouble &xval) const;

    const tdouble Inv_cdf_x(const tdouble p) const;
    const tdouble E2() const { return pow2((b-a)/(alpha+beta))*(alpha*beta/(alpha+beta+1.)) + pow2(mean()); }
};

//======================= FlxRndCreator ============================================

/**
* @brief A class for managing the generation of random number
*/
class FlxRndCreator {
  private:
    /**
    * @brief random sequence input: reader
    */
    FlxIstream* rndReader;
  public:
    FlxRndCreator() : rndReader(NULL) {};
    ~FlxRndCreator() {}
    /**
    * @brief generate new random samples - standard normal distributed (uncorrelated)
    */
    const tdouble gen_smp();
    /**
    * @brief generate new random samples - uniform distributed on [0;1]
    */
    const tdouble gen_smp_uniform();
    /**
    * @brief generate new random samples - binary RV with Probability p to be true
    */
    const bool gen_smp_binary(const tdouble p) { return (gen_smp_uniform()<=p); }
    void gen_smp(tVec& y);
    /**
    * @brief random sequence input: stop
    */
    void replay_stop(bool doAlert=false);
    /**
    * @brief random sequence input: start
    */
    void replay_start(FlxIstream* rndReaderV);
};


//======================= FlxHLRV ============================================

/**
* @brief The base class of the High Level Random Variables
*/
class FlxHLRVbase : protected FlxBoxBaseR {
  public:
    virtual ~FlxHLRVbase() {}
    /**
    * @brief returns the current realization
    */
    virtual const tdouble get_Realization() = 0;
    virtual const std::string print_info() = 0;
};

/**
* @brief Returns as realization a realization from FlxRndBox.RndCreator
*/
class FlxHLRVstd : public FlxHLRVbase {
  public:
    virtual const tdouble get_Realization();
    virtual const std::string print_info() { return "standard normal distribution."; }
};


//======================= Correlation - Nataf Transformation ============================================

/**
* @brief A class for storing the correlation between random variables
*/
class FlxCorrelation : protected FlxBoxBaseR {
  private:
    FlxMtxSym* rhoCorr;
    /**
    * @brief Cholesky-Decomposition of rhoPrime (calculates the inverse multiplication implicitly)
    */
    FlxMtxSparsLTri* L;
    /**
    * @brief true if a correlation exists
    */
    bool CorrSet;
    bool CorrSolved;	// rnd corr create -> has been executed
    /**
    * @brief calculates the modified correlation needed for the Nataf-transformation.
    */
    const tdouble calc_R_for_rhoPrime(FlxRnd_base* rv1, FlxRnd_base* rv2, const tdouble& rho);
    /**
    * @brief allocates memory for the correlation matrix rhoCorr
    */
    void allocate_rhoCorr(const tuint& nRND);
    void deallocate_L(bool warning=true);
  public:
    FlxCorrelation() : rhoCorr(NULL), L(NULL),CorrSet(false),CorrSolved(false) {}
    ~FlxCorrelation();
    /**
    * @brief Transform from uncorrelated standard normal space y to correlated standard normal space w
    */
    void y2w(const tVec& y, tVec& w) const;
    /**
    * @brief Transform from correlated standard normal space w to uncorrelated standard normal space y
    */
    void w2y(tVec& y, const tVec& w) const;
    /**
    * @brief Set a correlation value
    */
    void set_corr(tuint i, tuint j, tdouble corrV);
    /**
    * @brief Get a correlation value
    */
    const tdouble get_corr(tuint i, tuint j) const;
    /**
    * @brief insert additional random variables
    */
    void insertRV(const tuint& nRND);
    /**
    * @brief creates the matrices needed for the transformation
    * @param onlyFirst if true: create L onlyIF L=NULL
    */
    void create_L(const bool onlyFirst=false);
    /**
    * @brief calculates Jinv=L_ki=d(w_k)/d(y_i)  -> returns a copy of L
    * NOTE: the returned Matrix has to be deallocated!
    */
    const FlxMtxSparsLTri* calc_Jinv() { return L; }
    /**
    * @brief creates a L-matrix for certain random variables
    * NOTE: the returned Matrix has to be deallocated!
    */
    FlxMtxSparsLTri* create_small_L(const iVec &indexV);
};

//======================= FlxRndBox ============================================

class FunBase;

/**
* @brief A class for storing constant values (the result of the evaluated function is stored).
*/
class FlxRndBox {
  private:
    #if FLX_DEBUG
      /**
      * @brief number of instances of this calss - only one class may exist
      */
      static int Cinst;
    #endif
    /**
    * @brief vector with realizations in uncorrelated standard normal space
    */
    tVec* y_smp;
    /**
    * @brief vector with realizations in original space
    */
    tVec* x_smp;
    /**
    * @brief if this is true, no random variables can be created anymore
    */
    int isLocked;   
    /**
    * @brief stores all managed variables
    */
    std::vector<FlxRnd_base*> box;
    /**
    * @brief stores the true indices of the user defined variables
    */
    std::map<tuint, tuint> boxM;
    tuint boxM_size_inclRF;
    /**
    * @brief stores all High Level Random Variables
    */
    std::map<std::string, FlxHLRVbase*> boxH;
    /**
    * @brief stores correlation information of RVs
    */
    FlxCorrelation corr;
    /**
    * @brief if the random box is locked, an error will occur
    */
    void check_lock();
  public:
    /**
    * @brief manages the generation of random values
    */
    FlxRndCreator RndCreator;
    
    FlxRndBox();
    ~FlxRndBox();
    /**
    * @brief insert an automatic random variable
    */
    const tuint insert ( FlxRnd_base* value);
    /**
    * @brief insert an user defined random variable
    */
    const tuint insert (tuint index, FlxRnd_base* value);
    /**
    * @brief insert a High Level Random Variable
    */
    void insert (std::string index, FlxHLRVbase* value); 
    /**
    * @brief insert variables from a random field
    */
    const tuint insert_RFv (std::vector<FlxRnd_base*>& boxRF);
    /**
    * @brief get the i'th managed variable (in order of definition)
    */
    FlxRnd_base* get ( tuint index );
    /**
    * @brief get a pointer of the High Level Random Variable 'index'
    */
    FlxHLRVbase* get ( std::string index );
    /**
    * @brief return the index of a managed variable
    */
    const tuint get ( const FlxRnd_base& rndV );
    /**
    * @brief get the vector of managed random variables (const-vec)
    */
    const std::vector<FlxRnd_base*>& get_vector ( ) { return box; }
    /**
    * @brief get the true index of a user defined random variable
    */
    const tuint get_TrueIndex_from_User(tuint index);
    /**
    * @brief get a user defined random variable by its index
    */
    FlxRnd_base* getRV_UserIndex (tuint index) { return box[get_TrueIndex_from_User(index)]; }
    /**
    * @brief true if user defined random variables exist
    */
    const bool has_userDefined();
    /**
    * @brief it is possible to lock the random box
    */
    void set_lock(bool lockIt);
    /**
    * @brief generate new random samples
    */
    void gen_smp();
    /**
    * @brief number of managed random variables
    */
    const size_t get_NumbRnd() const { return box.size(); }
    /**
    * @brief returns a vector with realizations of the random variables (in original space; managed)
    */
    const tVec& get_x_Vec() const { return *x_smp;}
    void get_x_Vec(tVec& x) const;
    /**
    * @brief returns a vector with realizations of the random variables (in original space; user defined)
    */
    const tVec get_x_Vec_UD() const;
    /**
    * @brief returns a vector with realizations of the random variables (in standard normal space)
    */
    const tVec& get_y_Vec() const { return *y_smp; }
    /**
    * @brief returns a vector with realizations of the random variables (in standard normal space; user defined)
    */
    const tVec get_y_Vec_UD() const;
    /**
    * @brief set the realizations of the random variables (in original space)
    * make sure x is of the correct dimension !!!
    */
    void set_x_Vec(const tVec& x);
    /**
    * @brief set the realizations of the random variables (in standard normal space)
    * make sure y is of the correct dimension !!!
    */
    const void set_y_Vec(const tVec& y);
    /**
    * @brief check if ALL values are in the physical domain
    */
    const bool check_xVec(const tVec &xVec) const;
    /**
    * @brief returns a vector with the mean values of the random variables
    */
    const tVec get_mean_Vec(bool output_just_user=false);    
    /**
    * @brief returns a vector with the standard deviations of the random variables
    */
    const tVec get_sd_Vec(bool output_just_user=false);   
    /**
    * @brief transform a vector from normal space to original space
    */
    void y2x(const tVec& y, tVec& x) const;
    /**
    * @brief transform a vector from original space to normal space
    */
    void x2y(tVec& y, const tVec& x) const;
    /**
    * @brief calculates Jinv=D_ij=d(x_j)/d(y_i)
    * NOTE: the returned Matrix has to be deallocated!
    */
    FlxMtxSparsLTri* calc_Jinv();
    /**
    * @brief Set a correlation between two user defined random variables
    */
    void corr_set(tuint i_udRV, tuint j_udRV, tdouble corrV);
    /**
    * @brief get correlation between to managed random variables
    */
    const tdouble corr_get(const tuint i, const tuint j) const;
    /**
    * @brief Create the matrices needed for the transformation of correlated RVs
    */
    void corr_create(bool onlyFirst=false);
    /**
    * @brief estimates the expectation of the expression numerically
    */
    const tdouble calc_expectation_numerical(FunBase* fun, bool silent=false);
    /**
    * @brief creates a L-matrix for certain random variables
    * NOTE: the returned Matrix has to be deallocated!
    */
    FlxMtxSparsLTri* create_small_L(const iVec &indexV) { return corr.create_small_L(indexV); }
    /**
    * @brief ONLY for random fields which become automatic random fields
    */
    void corret_boxM_size_Rf(const tuint RF_DIM) { boxM_size_inclRF -= RF_DIM; }
    
    static void set_RFbox(FlxRFBox *RFboxV);
    static FlxRFBox* RFbox;
    
    friend class calc_expectation_numerical_Fun;
    friend class calc_expectation_numerical_MCI;
    friend class calc_PC_projection_numerical_MCI;
};




#endif // fesslix_flxrandom_base_H

