/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_pdouble_H
#define fesslix_pdouble_H


#include "flxglobaldef.h"


class pdouble {
  private: 
    tdouble v;
    tdouble c;
  public:
    pdouble() : v(0.0), c(0.0) {}
    pdouble(const tdouble& t) : v(t), c(0.0) {}
    pdouble(const pdouble& p) : v(p.v), c(p.c) {}
    pdouble(const tdouble& v, const tdouble &c) : v(v), c(c) {}
    pdouble& operator=(const tdouble& p) {
      v=p;
      c=0.0;
      return *this;
    }
    pdouble& operator=(const pdouble& p) {
      if (this!=&p) {
	v=p.v;
	c=p.c;
      }
      return *this;
    }
    tdouble cast2double() const { return v; }
  
    pdouble& operator+=(const tdouble& right) {
      tdouble y = right - c;
      tdouble t = v + y;
      c = (t-v)-y;
      v = t;
      return *this;
    }
    pdouble& operator+=(const pdouble& right) { 
      this->operator+=(right.c);
      this->operator+=(right.v);
      return *this;
    }
    pdouble& operator-=(const tdouble& right) {
      tdouble y = right + c;
      tdouble t = v - y;
      c = (t-v)+y;
      v = t;
      return *this;
    }
    pdouble& operator-=(const pdouble& right) { 
      this->operator-=(right.c);
      this->operator-=(right.v);
      return *this;
    }
    pdouble& operator*=(const tdouble& right) {
      v*=right;
      c*=right;
      return *this;
    }
    pdouble& operator*=(const pdouble& right) {
      tdouble v1 = v;
      tdouble c1 = c;
      operator*=(right.v);
      operator+=(c1*right.c);
      operator+=(v1*right.c);
      return *this;
    }
    pdouble& operator/=(const tdouble& right) {
      v/=right;
      c/=right;
      return *this;
    }
    pdouble& operator/=(const pdouble& right) {
      v/=right.v;
      c/=right.v;
      return *this;
    }
    const pdouble operator*(const tdouble& right) const {
      pdouble result(this->v*right,this->c*right);
      return result;
    }
    const pdouble operator*(const pdouble& right) const {
      pdouble result(this->v*right.v);
      result+=this->c*right.c;
      result+=this->v*right.c;
      result+=this->c*right.v;
      return result;
    }
    const pdouble operator/(const tdouble& right) {
      pdouble result(this->v/right,this->c/right);
      return result;
    }
    const pdouble operator/(const pdouble& right) {
      return operator/(right.v);
    }
};

class qdouble {
  private:
    /**
    * @brief if the bin is full, add to this value
    */
    pdouble final_sum;
    /**
    * @brief temporary sum (the bin)
    */
    pdouble t_sum;
    /**
    * @brief number of points per bins
    */
    const unsigned int Np;
    /**
    * @brief current number of points in the bin
    */
    unsigned int cNp;
  public:
    qdouble(const unsigned int Np) : final_sum(0.0), t_sum(0.0), Np(Np), cNp(0) {}
    
    const tdouble cast2double() const { return final_sum.cast2double()+t_sum.cast2double(); }
    const pdouble cast2pdouble() const { pdouble t(final_sum);t+=t_sum;return t; }
    
    qdouble& operator+=(const tdouble& right) {
      t_sum+=right;
      ++cNp;
      if (cNp>=Np) {
	final_sum+=t_sum;
	t_sum = 0.0;
	cNp = 0;
      }
      return *this;
    }
};



#endif // fesslix_pdouble_H