/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxostools.h"

#include <fstream>


#if FLX_BOOST_PO
  #include <boost/program_options.hpp>
  namespace po = boost::program_options;
#endif

/**
* @brief returns the option from the argument array
* @param argc number of arguments in the array
* @param argv array with arguments
* @param pos position of the occurence of CMD
* @param value string with the value of the CMD
* @return false if there was no CMD in the command line
*/
bool arg_CMD(int argc, char** argv, int& pos, std::string& valueS );

#if FLX_BOOST_PO
  void print_help(po::options_description &visible);
#else 
  void print_help();
#endif

void print_version(std::ostream& so, bool pv=false);
void print_author();
void open_GaussFile(std::string gaussFile, ReadStream*& gaussRS);
void print_RETURN_SUCCESS();
void print_RETURN_ERROR();
void create_log(std::string logFile, bool logTrunc);
void delete_FLXcout(bool logOutput, std::ostream *flxcout);


bool arg_CMD(int argc, char** argv, int& pos, std::string& valueS )
{
  valueS = "";
  pos = 1;
  
  while (pos < argc) {
    std::string strV = argv[pos];
    if (strV == "--CMD") {
      for (int i = pos+1; i < argc; ++i) {
	valueS.append(argv[i]);
	valueS+=" ";
      }
      return true;
    }
    pos++;
  }
  return false;
}

#if FLX_BOOST_PO
  void print_help(po::options_description &visible)
  {
    std::string ts;
    #ifndef flxvi_flxvi_H
      ts = "fesslix";
    #else
      ts = "flxvi";
    #endif
    std::cout << "Usage: " << ts << " [options] files [--CMD code]" << std::endl << std::endl
      << "Fesslix - Stochastic Analysis" << std::endl << std::endl
      << visible << std::endl
      ;

  }
#else
  void print_help()
  {
    std::string ts;
    #ifndef flxvi_flxvi_H
      ts = "fesslix";
    #else
      ts = "flxvi";
    #endif
    std::cout << "Usage: " << ts << " files" << std::endl << std::endl
      << "Fesslix - Stochastic Analysis" << std::endl << std::endl
      ;

  }
#endif

void print_version(std::ostream& so, bool pv)
{
  #ifdef flxvi_flxvi_H
  if (pv) {
    so << "FlxVi - Version";
  } else {
    so << "FlxVi";
  }
  so << ":   " << FlxViManager::get_version() << std::endl;
  #endif
  if (pv) {
    so << "Fesslix - Version";
  } else {
    so << "Fesslix";
  }
  so << ": " << FesslixMain::get_version() << std::endl << " ";
  if (!pv) {
    FesslixMain::logInfo(so,true);
  }
}

void print_author()
{
  std::cout << "Fesslix was written by" << std::endl << 
   "    Wolfgang Betz <wolfgang.betz@fesslix.org>" << std::endl <<
   "Please use https://bugs.launchpad.net/fesslix to report bugs." << std::endl
   ;
}

void open_GaussFile(std::string gaussFile, ReadStream*& gaussRS) {
  if ( gaussFile == "{no}" ) return;
  if ( gaussFile == "{default}") {
    gaussFile = GlobalVar.configDir_local;
    gaussFile.append("gausspoints.dat");
    try {
      gaussRS = new ReadStream(gaussFile.c_str());
    } catch (FlxException &e) {
      if (GlobalVar.configDir == GlobalVar.configDir_local) return;
      gaussFile = GlobalVar.configDir;
      gaussFile.append("gausspoints.dat");
      try {
	gaussRS = new ReadStream(gaussFile.c_str());
      } catch (FlxException &e) {
      }
    }
    return;
  }
  gaussRS = new ReadStream(gaussFile.c_str());
}

void print_RETURN_SUCCESS() {
  GlobalVar.slog(4) << "Fesslix: Calculation completed. Fesslix(RETURN_SUCCESS)" << std::endl
		    << "-------------------------------------------------------" << std::endl << std::endl;
}

void print_RETURN_ERROR() {
  print_version(std::cout, true);
  print_version(GlobalVar.slog(3), true);
  GlobalVar.slog(4) << "Fesslix: Termination due to an error. Fesslix(RETURN_ERROR)" << std::endl
		    << "-----------------------------------------------------------" << std::endl << std::endl;
}

void create_log(std::string logFile, bool logTrunc) {
  if (logFile != "{default}") {
    std::ofstream* ofLog;
    if ( logTrunc ) {
      ofLog = new std::ofstream(logFile.c_str());
    } else {
      ofLog = new std::ofstream(logFile.c_str(), std::ios_base::out | std::ios_base::app); 
    }
    if ( ! ofLog->is_open() || ofLog == NULL ) {
      std::ostringstream ssV;
      ssV << "File (" << logFile << ") could not be opened.";
      throw FlxException("create_log_1", ssV.str() );
    } else {
      GlobalVar.set_slog(ofLog);
    }
  }
  print_FLX_GNU(GlobalVar.slog(4));
  print_version(GlobalVar.slog(4), true);
  // print date and time
    time_t Zeitstempel = time(0);
    tm *nun = localtime(&Zeitstempel);
    GlobalVar.slog(4) << "Time: "
      << nun->tm_year+1900 << "-" << nun->tm_mon+1 << "-" << nun->tm_mday 
      << " - " << nun->tm_hour << ':' << nun->tm_min << std::endl;
  
  GlobalVar.slog(4) << "Fesslix: loglevel is " << GlobalVar.logLevel << std::endl;
}

void delete_FLXcout(bool logOutput, std::ostream *flxcout) {
  GlobalVar.set_slog(NULL);
  if (!logOutput || flxcout==NULL) return;
  flxStreamAlloc* fsa = dynamic_cast<flxStreamAlloc*>(flxcout);
  delete fsa;
}